package ExSite::Attribute;
#-----------------------------------------------------------------------
#
#   Copyright 2001-2008 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

# ExSite::Attribute - a general-purpose meta-data tool for databases
#
# Allows you to define attributes (name-value pairs) and link them 
# to any ExSite database record.  All attributes are stored in their
# own table.
#
# This allows the end user to extend the effective sets of values 
# in a record, without modifying the database schema.  A common
# usage is to add customized membership information.


use strict;
use ExSite::Config;
use ExSite::Form;
use ExSite::Misc;
use ExSite::Base;

use vars qw(@ISA);
@ISA = qw(ExSite::Base);

# create an attribute object, and tie it to a record

sub new {
    my ($this,$table,$id) = @_;
    my $obj = {};
    my $class = ref($this) || $this;
    bless $obj, $class;
    $obj->setup($table,$id);
    return $obj;
}

# change record

sub setup {
    my ($this,$table,$id) = @_;
    
    # metadata for this record
    $this->{loaded} = undef;
    $this->{raw} = undef;      # copy of metadata records
    $this->{data} = undef;     # attribute values
    $this->{dirty} = undef;    # flag changed metadata

    # allowed metadata for this table
    # (don't reset this unless we are also changing tables)
    if ($table ne $this->{allow_table}) {
	$this->{allow_table} = undef;
	$this->{allowed} = undef;
    }
    
    $this->{table} = $table;
    $this->{id} = $id;
    return;
}

sub id {
    my ($this,$id) = @_;
    if ($id) {
	if (! $this->{id}) {
	    # current object has acquired an ID
	    $this->{id} = $id;
	}
	elsif ($id ne $this->{id}) {
	    # current object ID has changed; reset the object
	    $this->setup($id);
	}
    }
    return $this->{id};
}

# meta_id returns the id of a named metadata attribute

sub meta_id {
    my ($this,$name) = @_;
    if ($this->ok) {
	$this->load;
	if (exists $this->{raw}{$name}) {
	    return $this->{raw}{$name}{attribute_id};
	}
    }
    return 0;
}

sub ok {
    my ($this) = @_;
    return ($this->{table} && $this->{id});
}

# get the preferred datatype for an attribute

sub get_datatype {
    my ($this,$name) = @_;
    $this->load_allowed();
    return $this->is_allowed($name) ?
	$this->{allowed}{$name}{value} : 
	"string";
}

# load up the attributes

sub load {
    my ($this) = @_;
    if (! $this->{loaded} && $this->ok) {
	my @attr = $this->fetch_match({id=>$this->{id}});
	$this->{raw} = &keywise("name",\@attr);
	foreach my $attr (values %{$this->{raw}}) {
	    $this->{data}{$attr->{name}} = $attr->{value};
	}
	$this->{dirty} = undef;
	$this->{loaded} = 1;
    }
}

sub load_allowed {
    my ($this) = @_;
    if ($this->{allow_table} ne $this->{table}) {
	my @attr = $this->fetch_match({id=>0});
	$this->{allowed} = &keywise("name",\@attr);
	$this->{allow_table} = $this->{table};
    }
}

# get a list of allowed attribute names
# (allowed names are those in the attribute table that have no record ID)
# datatypes are stored in $this->{datatype}

sub get_allowed {
    my ($this) = @_;
    $this->load_allowed();
    return keys %{$this->{allowed}};
}

sub is_allowed {
    my ($this,$name) = @_;
    $this->load_allowed();
    if (scalar keys %{$this->{allowed}} > 0) {
	# we have an explicit allow list for attributes
	return exists $this->{allowed}{$name};
    }
    # there is no allow list;  all attributes are permitted
    return 1;
}

# add an allowed attribute (admin users only)

sub add_allowed {
    my ($this,$name,$datatype) = @_;
    
    if ($share{DB}->is_admin) {
	$this->load_allowed();
	if (exists $this->{allowed}{$name}) {
	    $this->error("ExSite::Attribute::add_allowed: $name already exists");
	}
	else {
	    my $meta = $this->make_record($name,$datatype,0);
	    my $mid = $this->insert($meta);
	    $this->{allowed}{$name} = $meta;
	    return $mid;
	}
    }
    else {
	$this->error("ExSite::Attribute::add_allowed: permission denied");
    }
    return 0;
}

# remove an allowed attribute (admin users only)
# also removes all attributes of that name

sub remove_allowed {
    my ($this,$name) = @_;
    my $db = $share{DB};
    if ($db->is_admin) {
	if ($this->is_allowed($name)) {

	    # remove from database

	    my @attr = $this->fetch_match({name=>$name});
	    $this->trash(@attr);

	    # remove from current object

	    delete $this->{allowed}{$name};
	    delete $this->{raw}{$name};
	    delete $this->{data}{$name};
	    delete $this->{dirty}{$name};

	    return scalar @attr;
	}
	else {
	    $this->error("ExSite::Attribute::remove_allowed: invalid attribute: $name");
	}
    }
    else {
	$this->error("ExSite::Attribute::add_allowed: permission denied");
    }
    return 0;
}

# get all attributes for this record

sub get_all {
    my ($this) = @_;
    $this->load;
    return wantarray ? %{$this->{data}} : $this->{data};
}

# fetch attribute values

sub get {
    my ($this,$name) = @_;
    $this->load;
    if (! $name) { return $this->get_all(); }
    return exists $this->{data}{$name} ?
	$this->{data}{$name} : undef;
}

# set/save an attribute value

sub set {
    my ($this,$name,$value) = @_;
    if ($this->set_nosave($name,$value)) {
	# attribute was successfully set;  now save it
	$this->save($name);
    }
}

sub set_nosave {
    my ($this,$name,$value) = @_;
    
    # set attribute in the object

    if ($this->is_allowed($name)) {
	$this->{data}{$name} = $value;
	$this->{dirty}{$name} = 1;
	return 1;
    }
    else {
	$this->error("ExSite::Attribute::add_allowed: permission denied");
	return 0;
    }
}

sub save { 
    my ($this,$name) = @_;
    return if (! $this->ok());

    my @update;
    if ($name) {
	# save one attribute
	push @update, $name;
    }
    else {
	# save all dirty attributes
	push @update, keys %{$this->{dirty}};
    }

    foreach my $n (@update) {
	if (exists $this->{raw}{$n}) {
	    # update existing attribute
	    $this->update_value($n,$this->{data}{$n});
	}
	else {
	    # new attribute setting
	    my $meta = $this->make_record($n,$this->{data}{$n});
	    my $mid = $this->insert($meta);
	    $this->{raw}{$n} = $meta;
	}
    }
    return;
}

# copy one or all attributes to another record ID

sub copy { 
    my ($this,$id,$name) = @_;
    if ($this->ok) {
	my @keys = $name ? ($name) : keys %{$this->{data}};
	if (@keys > 0 && $id) {
	    my $a = new ExSite::Attribute($this->{table},$id);
	    foreach my $k (@keys) {
		$a->set_nosave($k,$this->get($k));
	    }
	    $a->save();
	}
    }
}

# remove all attribute values

sub clear {
    my ($this,$name) = @_;

    if ($this->ok) {
	$this->load;
	if ($name) {
	    if ($this->{data}{$name}) {
		$this->trash($this->meta_id($name));
		delete $this->{raw}{$name};
		delete $this->{data}{$name};
		delete $this->{dirty}{$name};
		return 1;
	    }
	}
	else {
	    return $this->clear_all();
	}
    }
    
    $this->warn("ExSite::Attribute::clear: no such attribute");
    return 0;
}

sub clear_all {
    my ($this) = @_;

    if ($this->ok) {
	$this->load;
	my @attr = $this->fetch_match({id=>$this->{id}});
	if (@attr > 0) {
	    $this->trash(@attr);
	    $this->{raw} = undef;
	    $this->{data} = undef;
	    $this->{dirty} = undef;
	    return scalar @attr;
	}
    }
    return 0;
}

# generate an input tag to accept attribute values
# NOTE: cannot do foreign key refs in attributes

sub input {
    my ($this,$input_name,$name,$value) = @_;
    if ($this->is_allowed($name)) {
	if ($input_name) {
	    # hard-coded input name;  the form must know what to do with 
	    # this input
	    return $share{DB}->input_exsite(
					    datatype=>($this->get_datatype($name) || "string") ,
					    name=>$input_name,
					    value=>($value || $this->get($name))
					    );
	}
	else {
	    # auto-coded input name; self-processing ExSite forms can figure
	    # this out automatically.
	    my $rec = $this->{raw}{$name} ?
		$this->{raw}{$name}{attribute_id} : undef;
	    my $datatype = $this->get_datatype($name);
	    return $this->input_column(
				       prompt=>$name,
				       record=>$rec,
				       size=>($datatype eq "text") ? 400 : 60,
				       datatype=>($datatype || "string") ,
				       value=>($value || $this->get($name))
				       );
	}
    }
    else {
	return $share{DB}->error("Cannot generate input field for invalid attribute \"$name\"");
    }
}

# DB accessors, for easier overloading
# All low-level DB logic is defined here.

# make_record() - create a datahash suitable for inserting into the DB
# use insert() to actually insert it.

sub make_record {
    my ($this,$name,$value,$id) = @_;
    return { 
	name=>$name, 
	value=>$value, 
	tablename=>$this->{table},
	id=>(defined $id ? $id : $this->{id}), 
	};
}

sub fetch_match {
    my ($this,$match) = @_;
    $match->{tablename} = $this->{table};
    return $share{DB}->fetch_match("attribute",$match);
}

sub insert {
    my ($this,$data) = @_;
    my $id = $share{DB}->insert("attribute",$data);
    $data->{attribute_id} = $id;
    return $id;
}

sub update_value {
    my ($this,$name,$value) = @_;
    if ($this->ok) {
	$this->load();
	my $meta_id = $this->meta_id($name);
	if ($meta_id) {
	    return $share{DB}->update(
				      "attribute",
				      { value=>$value },
				      { attribute_id=>$meta_id }
				      );
	}
    }
    return undef;
}

sub input_column {
    my ($this,%options) = @_;
    $options{table} = "attribute";
    $options{column} = "value";
    return $share{DB}->input_column(%options);
}

sub trash {
    my ($this,@data) = @_;
    my @trash;
    foreach my $attr (@data) {
	push @trash, "attribute", (ref $attr) eq "HASH" ? $attr->{attribute_id} : $attr;
    }
    return $share{DB}->trash(@trash);
}

# deprecated functions

sub fetch {
    my ($this,$name) = @_;
    return $this->get($name);
}

sub rollback {
    my ($this,$name) = @_;
    $this->warn("ExSite::Attribute: rollback no longer supported");
    return 0;
}

sub purge {
    my ($this,$name) = @_;
    $this->warn("ExSite::Attribute: purge no longer supported");
    return 0;
}

1;


