#!/usr/bin/perl
#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::Auth;

use strict;
use ExSite::Config;
use ExSite::Form;
use ExSite::Page;
use ExSite::Base;
use ExSite::Misc;
use ExSite::Util;
use ExSite::Input;
use ExSite::Mail;
use ExSite::Time;
use ExSite::Crypt;
use Digest::MD5 qw(md5_hex);

our @ISA = qw(ExSite::Base);


#========================================================================

=pod

=head1 ExSite::Auth -- authentication and security tools

Authorization is performed before any standard operation on
the database, when the operation is performed by one of ExSite::DB's
descendant classes.  The authorization process has the following
steps:

=over 4

=item 1. identify - determine who the user claims to be

=item 2. lookup - determine whom we understand that username to represent

=item 3. authenticate - determine if the user's claim matches our knowledge

=item 4. authorize - grant a certain level of access to the database

=item 5. approve - approve a specific operation

=back

Each of the steps in the authorization process may be replaced with
a custom handler if the procedure differs from the default ExSite logic.

=head2 identify() - determine who the user claims to be

Determines the user's claimed identity (such as by checking the
identity cookie), and stores the result so that authorization methods
can check it.  This routine does not authenticate the user's identity
claim, and so is usually not called directly.

=head2 lookup() - look up the user to see if we know who they are

Consult our own database to see if we have a record corresponding to
this user's identity (determined from C<identify()>).  Returns the
user record in a hash, if found.  The default logic for looking up a
user record is to look up the login name, which implies that unique
login names are required.  Once this lookup has occurred, the
following identity query functions can be used:

=head2 my_name() - return the login name of the current user

=head2 my_uid() - return the primary key of the current user's record

=head2 my_gid() - return the primary key of the current user's group record

=head2 my_user_record() - return the user's identity hash

=head2 authenticate() - verify the user's identity claim

Returns true (1) if the user's identity matches the information
returned by lookup(), false (0) otherwise.  This serves as a simple
check that the user is logged in, and that their identity is valid.

=head2 authorize() - grant the user a specific level of access

Returns the user's access level.  Access level is a value of 0 or
higher.  0 corresponds to "nobody", or the general public.  Values of
1 or more correspond to authenticated users, with higher values giving
greater permissions to the database.  The maximum value is
C<$config{auth}-E<gt>{root_auth_level}>, which grants unrestricted
"super-user" access to the data.  (This level of access can be
dangerous, as it allows one to mess with sensitive data fields and
referential integrity.)

The default authorize routine assumes that the user's access level
is stored in the user table.  This value is simply returned without
any sanity checks.

=head2 approve() - approve a specific database operation

Usage: C<$db-E<gt>approve( OP, TABLE, RECORD );>

OP is a low-level database operation (one of "read", "update",
"insert", or "delete").  TABLE is the table that this operation is
being performed on. RECORD is the actual record datahash that this
operation is being performed on (only used for "update" and "delete").

The return value is true if the operation is permitted, or false if:

=over 4

=item * the user does not have permission to read or write to the table

=item * the user is referencing records owned by a different user or group.

=back

=cut 

#----------------------------------------------------------------------
# identify : determine whom the user claims to be

sub identify {
    my ($this) = shift;
    my ($name,$pwd,$uid,$gid);
    my $stat = $this->run_handler("identify");
    my $access = 0;
    my $authtype = $config{auth}{type};
    if (defined $stat) {
	# application has its own identify algorithm
	($name,$pwd) = @$stat;
    }
    elsif ($authtype eq "cookie") {
	($name,$pwd) = split /:/, $cookie{$config{site}{identity_cookie}}, 2;
    }
    elsif ($authtype eq "crypt") {
	my $c = new ExSite::Crypt;
	my $id = $c->decrypt($cookie{$config{site}{identity_cookie}});
	($name,$uid,$access) = split /:/,$id;
    }
    elsif ($authtype eq "fixed") {
	$name = $config{auth}->{default_identity}->{name};
    }
    elsif ($authtype eq "session") {
	# User identity claim is implicit in the session ID.
	# Therefore this call is kind of meaningless in the case of sessions.
	# Let's return the session ID, and see what happens.
	my $session_cookie = $config{site}->{id}."_sessionID";
	return $cookie{$session_cookie};
	#return &get_cookie($session_cookie);
    }
    elsif ($authtype eq "none") {
	$name = "nobody";
    }
    $share{identity} = { 
	name => $name,
	uid => $uid,
	access => $access,
	password => $pwd 
	};
    return $name;
}

#----------------------------------------------------------------------
# lookup : look up the user, to see if we know who they are

sub lookup {
    my ($this,$login) = @_;
    if ($share{identity}{lookup} && $login eq $share{identity}{lookup}{login}) {
	return $share{identity}{lookup};
    }
    # check for custom lookup algorithm
    my $stat = $this->run_handler("lookup",$login);
    return $stat if (defined $stat);
    #
    return undef if ($this->{authentication} eq "none");
    my $udata = $share{DB}->fetch_match($config{auth}->{user_table},
			    {$config{auth}->{user_login_column}=>$login});

    if (@$udata > 0) {
	$share{identity}{lookup} = $udata->[0];
	$share{identity}{uid} = $udata->[0]{$config{auth}{user_uid_column}};
	$share{identity}{gid} = $udata->[0]{$config{auth}{user_gid_column}};
	$share{identity}{access} = $udata->[0]{$config{auth}{user_access_column}};
	return $udata->[0];
    }
    return undef;
}

#----------------------------------------------------------------------
# authenticate : verify the user's identity claim
#
# Returns true (1) if the user's identity matches the information
# returned by lookup(), false (0) otherwise.

sub authenticate {
    my ($this) = shift;
    if (exists $this->{authenticated}) {
	return $this->{authenticated};
    }
    # check for custom authentication algorithm
    my $stat = $this->run_handler("authenticate");
    return $stat if (defined $stat);
    
    if ($config{auth}{type} eq "none") {
	$this->{authenticated} = 1;
	return 1;
    }
    elsif ($config{auth}{type} eq "fixed") {
	$this->{authenticated} = 1;
	$share{identity} = $config{auth}{default_identity};
	return 1;
    }
    elsif ($config{auth}{type} eq "session") {
	if (exists $session{identity}) {
	    # we authenticated on a previous visit
	    # copy identity info from session
	    if (! $share{identity}) {
		my $identity = $session{identity};
		$share{identity} = $identity;
		$this->info("restoring identity \"$identity->{name}\" from session");
	    }
	    $this->{authenticated} = 1;
	    return 1;
	}
	else {
	    return 0;
	}
    }
    elsif ($config{auth}{type} eq "crypt") {
	# we believe the claimed identity is secure, since it is encrypted
	if (! exists $share{identity}) { &identify($this); }
	if ($share{identity}{access}) { $this->{authenticated} = 1; }
	return $share{identity}{access};
    }
    else {   # this->{authentication} eq "cookie"

	# check identity
	if (!$share{identity}) { &identify($this); }
	if (!$share{identity}{name}) {
	    # nobody logged in - not necessarily an error
	    return 0;
	}
	# look up the user
	$this->info("authenticating user \"$this->{identity}->{name}\"");
	my $user = &lookup($this,$share{identity}{name});
	if (!$user) {
	    # unknown username
	    $this->warn("no such user \"$this->{identity}->{name}\"");
	    return 0;
	}
	elsif (! $user->{$config{auth}->{user_pwd_column}}) {
	    # password field is blank - deny access
	    $this->warn("blank password for user \"$user->{$config{auth}->{user_login_column}}\"");
	    return 0;
	}

	# compare entered and stored passwords

	my $salt = substr($share{identity}{password},0,2);
	if ($share{identity}{password} eq
	    crypt($user->{$config{auth}{user_pwd_column}},$salt)) {
	    # passwords match
	    $this->{authenticated} = 1;
	    # store extra user info in identity
	    $share{identity}{gid} = $user->{$config{auth}->{user_gid_column}};
	    ####?
	    $share{identity}{access} = $user->{$config{auth}->{user_access_column}};
	    return 1;
	}
	else {
	    # bad password
	    $this->warn("incorrect password for user \"$user->{$config{auth}->{user_login_column}}\" - access denied");
	    $this->{authenticated} = 0;
	    return 0;
	}
    }
}

#----------------------------------------------------------------------
# authorize : grant the user a specific level of access
#
# Access level is a value of 0 or higher.  0 corresponds to "nobody",
# or the general public.  Values of 1 or more correspond to
# authenticated users, with higher values giving greater permissions
# to the database.  The maximum value is $config::root_auth_level,
# which grants unrestricted "super-user" access to the data.  (This
# level of access can be dangerous, as it allows one to mess with
# sensitive data fields and referential integrity.)
#
# The default authorize routine assumes that the user's access level
# is stored in the user table.  This value is simply returned without
# any sanity checks.

sub authorize {
    my ($this) = shift;
    my $stat = $this->run_handler("authorize");
    return $stat if (defined $stat);

    # if not logged in, check for login request
    if (! exists $this->{authenticated}) {	
	# check if credentials were passed in an authtoken
	$this->check_authtoken();
	# authenticate credentials
	&authenticate($this);
    }

    return $this->{authenticated} ? $share{identity}{access} : 0;
}

#----------------------------------------------------------------------
# approve : approve a specific database operation
#
# Arguments:
# op - "read", "update", "insert", "delete"
# table - the table being operated on
# [record] - reference to a datahash containing the record being
#            operated on (used for update, delete only)
#
# Returns true if the operation is permitted.
# Returns false if:
#   - the user does not have permission to read or write to the table
#   - the user is referencing records owned by a different user or group.

sub approve {
    my ($this,$op,$table,$record) = @_;
    my $stat = $this->run_handler("approve",$op,$table,$record);
    return $stat if (defined $stat);
    my $level = $this->authorize();
    my $tablemap = $share{DB}{map}->get_table($table);
    # script okay, now check table permissions
    if (exists $tablemap->{$op}) {
	if ($op =~ /read|insert|search/) {
	    return ($level >= $tablemap->{$op});
	}
	else {
	    # for updates & deletions, must check row ownership
	    if (!$record) {
		if ($level >= $config{auth}->{root_auth_level} - 1) { 
		    return 1; 
		}
		else { return 0; }
	    }
	    elsif ($share{DB}->user_owns($table,$record)) {
		return ($level >= $tablemap->{$op});
	    }
	    else {
		return 0;
	    }
	}
    }
    elsif ($this->{dbmap}) {
	# dbmap kludge - no tablemap or authentication when editing dbmaps
	return 1;
    }
    else {
	# unknown operation
	return 0;
    }
}

# some identity queries, for convenience:

sub my_name {
    my $this = shift;
    if ($this->authenticate) {
	my $stat = $this->run_handler("my_name");
	return $stat if (defined $stat);
	my $udata = $this->my_user_record();
	my @namecol = ref $config{auth}{user_name_column} eq "ARRAY" ?
	    @{$config{auth}{user_name_column}} :
	    ($config{auth}{user_name_column});
	my @name = map { $udata->{$_} } @namecol;
	return join " ",@name;
    }
    return undef;
}

sub user_name {
    my ($this,$uid) = @_;
    my $stat = $this->run_handler("user_name");
    return $stat if (defined $stat);
    my $udata = $this->fetch("member",$uid);
    if (ref $udata eq "HASH") {
	my @namecol = ref $config{auth}{user_name_column} eq "ARRAY" ?
	    @{$config{auth}{user_name_column}} :
	    ($config{auth}{user_name_column});
	my @name = map { $udata->{$_} } @namecol;
	return join " ",@name;
    }
    return undef;
}

sub my_login {
    my $this = shift;
    if ($this->authenticate) {
	my $stat = $this->run_handler("my_login");
	return $stat if (defined $stat);
	# $share{identity}{name} is actually the login ID
	# this should probably be considered a bug.
	return $share{identity}{name};
    }
    return undef;
}

sub my_email {
    my $this = shift;
    if ($this->authenticate) {
	my $stat = $this->run_handler("my_email");
	return $stat if (defined $stat);
	return $share{identity}{lookup}{email};
    }
    return undef;
}

sub my_uid {
    my $this = shift;
    if ($this->authenticate) {
	my $stat = $this->run_handler("my_uid");
	return $stat if (defined $stat);
	return $share{identity}{uid};
    }
    return undef;
}

sub my_gid {
    my $this = shift;
    if ($this->authenticate) {
	my $stat = $this->run_handler("my_gid");
	return $stat if (defined $stat);
	return $share{identity}{gid};
    }
    return undef;
}

sub my_user_record {
    my $this = shift;
    my $u;
    if ($this->authenticate) {
	my $stat = $this->run_handler("my_user_record");
	return $stat if (defined $stat);
	if (exists $share{identity}{lookup}) {
	    $u = $share{identity}{lookup};
	}
	elsif ($this->my_uid) {
	    $u = $share{identity}{lookup} = 
		$share{DB}->fetch($config{auth}{user_table},$this->my_uid);
	}
	else {
	    $u = {};
	}
    }
    else {
	$u = {};
    }
    return (wantarray && defined $u) ? %$u : $u;
}

# user permission tests...

sub is_admin {
    my $this = shift;
    return ($this->authorize >= ($config{auth}{root_auth_level} - 1));
}

sub is_manager {
    my ($this,$site) = @_;
    my $level = $this->authorize;
    # false if insufficient level
    return 0 if ($level < ($config{auth}{root_auth_level} - 2));
    # true if a system admin
    return 1 if ($level >= ($config{auth}{root_auth_level} - 1));
    # otherwise, check for admin key
    return $this->owns($site);
}

sub is_member {
    my ($this,$site) = @_;
    my $stat = $this->run_handler("is_member",$site);
    return $stat if (defined $stat);
    return 0 if (! $this->authorize);
    return 1 if ($this->is_manager($site));
    my $u = $this->my_user_record();
    if ($u->{section_id}) {
	# the user is tied to a particular site
	if (!$site) { $site = $this->this_site; }
	my $sid = ref $site eq "HASH" ? $site->{section_id} : $site;
	return ($sid == $u->{section_id});
    }
    else {
	# the user belongs to all sites
	return 1;
    }
}

# return the effective access level the user is operating at

sub level {
    my $this = shift;
    if (! exists $this->{effective_level}) {
	$this->{effective_level} = $this->real_level;
    }
    return $this->{effective_level};
}

# return the real access level of the user

sub real_level {
    my $this = shift;
    if (! exists $this->{level}) {
	$this->{level} = $this->authorize;
    }
    return $this->{level};
}

# change to a different (lower) effective access level
#
# This is used when publishing, to make pages render as if being viewed
# by the public.

sub set_level {
    my ($this,$level) = @_;
    $this->{effective_level} = ($level <= $this->real_level) ? $level : 0;
    return $this->{effective_level};
}

# return to our real access level.

sub reset_level {
    my $this = shift;
    delete $this->{effective_level};
}

# remove all traces of current user login

sub clear_login {
    my $this = shift;
    delete $this->{authenticated};  # already authenticated flag
    delete $share{identity};        # user identity information
    delete $session{identity};      # preserved identity data in Session
    return;
}

# return a standard login form

sub login_form {
    my ($this,$action,$goto) = @_;
    my $stat = $this->run_handler("login_form",$action,$goto);
    return $stat if (defined $stat);
    my $ml = &get_obj("ML");
    my $out;
    my $in = new ExSite::Input;
    my $data = $in->combine;  # get any input data
    if ($data->{login_action} eq "forgot") {
	return $this->login_reminder_form;
    }
    elsif ($data->{login} && $data->{password} && $this->do_login_form($data)) {
	# successful login
	return undef;
    }
    elsif ($data->{login} || $data->{password}) {
	# failed login
	$out .= $ml->p($msg{"Login failed - please try again."},{class=>"error"});
    }

    # Display site logo
    if ($config{site}{use_logo}){
	$out .= $ml->p($config{site}{logo});
    }
    $out .= $ml->h2($msg{"Login"});

    my %formattr = ( method=>"post" );

    # $action determines who process this form; 
    # default is login.cgi - used if $action is undef
    # if $action is defined but blank/false, no action is used (use this
    # to submit the login form to itself)

    if (defined $action && $action) {
	$formattr{action} = $action;
    }
    elsif (! defined $action) {
	$formattr{action} = "$config{server}{CGIpath}/$config{prog}{login}";
    }

    # $goto determines where the user logs in to; default is login_destination
    $goto = $goto || "home.cgi"; #$config{auth}{login_destination}[$this->authorize()];

    my @login_table = (
		       [ $msg{$config{auth}{login_prompt}}, 
			 $ml->input(undef,{type=>"text",name=>"login"}) ],
		       [ $msg{$config{auth}{password_prompt}}, 
			 $ml->input(undef,{type=>"password",name=>"password"}) ]
		       );
    if ($config{auth}{remember_me}) {
	# add a "remember me" checkbox
	push @login_table, [ ["td", $ml->input(undef,{type=>"checkbox",name=>"remember"}), {align=>"right"}], $msg{$config{auth}{remember_prompt}} ];
    }
    my $formtext = $ml->table(\@login_table, { class=>"LoginForm" } );

    # hidden field for destination
    if ($goto) {
	$formtext .= $ml->input(undef,{type=>"hidden",name=>"goto",value=>$goto});
    }

    # buttons
    (tied %msg)->nohighlight;
    $formtext .= $ml->p(
			$ml->input(undef,{type=>"submit", 
					  value=>" ".$msg{"Log in"}." "}).
			"&nbsp;".
			$ml->input(undef,{type=>"button",
					  value=>" $msg{Cancel} ",
					  onclick=>"history.back()"})
			);
    (tied %msg)->restore;

    $out .= $ml->form($formtext,\%formattr);

    my $forgoturl = &relink(login_action=>"forgot");
    if ($config{auth}{forgot_login}) {
	$out .= $ml->p(
		       $ml->a($msg{$config{auth}{forgot_prompt}},
			      {href=>$forgoturl})
		       );
    }

    return $ml->div($out,{align=>"center"});
}

# process a standard login form - return error message, or redirect

sub do_login_form {
    my ($this,$data) = @_;
    my $goto = $data->{goto};
    if ($this->login($data->{login},$data->{password},$goto,$data->{remember})) {
	# success!
	if (! $goto) {
	    # no place to go;  redirect to default location specified
	    # in config file
	    $goto = $config{auth}{login_destination}[$this->level];
	}
	&redirect($goto);
	return 1;
    }
    return 0;
}

# standard form to send login instructions and password reminders
#
# TODO: assumes cleartext password storage; expand to other cases

sub login_reminder_form {
    my $this = shift;
    if (! $config{auth}{forgot_login}) { return undef; }
    my $stat = $this->run_handler("login_reminder_form");
    return $stat if (defined $stat);
    my $in = new ExSite::Input;
    my $data = $in->combine;  # get any input data
    my $ml = &get_obj("ML");
    if ($data->{login} || $data->{email}) {
	return $this->do_login_reminder($data);
    }
    return $ml->div(
		    $ml->form(
			      $ml->p($msg{"To receive login instructions by e-mail, please provide your $config{auth}{login_prompt} ID if you know it, or your e-mail address if you do not."}).
			      $ml->table([
					  [ $msg{"$config{auth}{login_prompt}:"},
					    $ml->input(undef,{type=>"text",name=>"login"})
					    ],
					  [ $msg{"OR, your email address:"},
					    $ml->input(undef,{type=>"text",name=>"email"})
					    ],
					  ],
					 {cellspacing=>15}
					 ).
			      $ml->p(
				     $ml->input(undef,{type=>"submit",
						       value=>" $msg{Submit} "}).
				     "&nbsp;".
				     $ml->input(undef,{type=>"button",
						       value=>" $msg{Cancel} ",
						       onclick=>"history.back()"})
				     ),
			      { method=>"post" }
			      ),
		    { align=>"center" }
		    );
}

sub do_login_reminder {
    my $this = shift;
    if (! $config{auth}{forgot_login}) { return undef; }
    my $stat = $this->run_handler("do_login_reminder");
    return $stat if (defined $stat);
    my $out;
    my $in = new ExSite::Input;
    my $data = shift || (new ExSite::Input)->combine;  # get any input data
    my $ml = &get_obj("ML");

    # get users and addresses

    my $db = $share{DB} || new ExSite::DB;
    my $admin = $this->owner_email;
    my $user_list;
    if ($data->{login}) {
	$user_list = $db->fetch_match($config{auth}{user_table},
				      {$config{auth}{user_login_column}=>$data->{login}});
    }
    elsif ($data->{email}) {
	$user_list = $db->fetch_match($config{auth}{user_table},
				      {email=>$data->{email}});
    }
	
    # setup template error strings and messages

    my $num_matches = scalar(@$user_list);
    my $url = &relink(login_action=>undef);
    my $forgoturl = &relink(login_action=>"forgot");

    my %template = (
		    nouser => $msg{"Sorry, there is no user named [[login]]."},
		    twousers => $msg{"ERROR: there is more than one user named [[login]]."},
		    pleasecall => $msg{"Please contact the site administrator at [[admin]] for assistance."},
		    missingemail => $msg{"Sorry, there is no email address recorded for the user [[login]]."},
		    email_subject => $msg{"Login instructions for [[site_title]]"},
		    email_body => $msg{"Here are login instructions for [[user]] for the website [[site_title]]"},
		    email_instructions => $msg{"$config{auth}{login_prompt}: [[login]]\n$config{auth}{password_prompt}: [[password]]"},
		    email_error => $msg{"We found an account, but your login information (username, password, 
or access level) is incomplete.  Please contact your site 
administrator for assistance, at [[admin]]"},
		    notmyemail => $msg{"Login instructions have been emailed to [[email]].
If this is not your current e-mail address, then please contact the 
site administrator at [[admin]] for assitance."},
		    );

    if ($num_matches==0) {

	# user not found

	if ($data->{login}) {
	    $out .= $ml->p(&substitute($template{nouser},
				       {login=>$ml->strong($data->{login})}));
	}
	elsif ($data->{email}) {
	    $out .= $ml->p($msg{"Sorry, that email address was not found."});
	}
	$out .= $ml->p($ml->a($msg{"Try again?"},{href=>$forgoturl}));
	$out .= $ml->p($ml->a($msg{"Return to login screen"},{href=>$url}));
    } 
    elsif ($num_matches>1 && $data->{login}) {

	# whoops, too many users found

	$out .= $ml->p(&substitute($template{twousers},
				   {login=>$ml->strong($data->{login})}).
		       $ml->br().
		       &substitute($template{pleasecall},
				   {admin=>$ml->a($admin,{href=>"mailto:$admin"})}));
	$out .= $ml->p($ml->a($msg{"Try again?"},{href=>$forgoturl}));
	$out .= $ml->p($ml->a($msg{"Return to login screen"},{href=>$url}));
    } 
    else {

	# one user found

	my $email = $data->{email} || $user_list->[0]{email};
	if (not $email) {

	    # user does not have an email address

	    $out .= $ml->p(&substitute($template{missingemail},
				       login=>$ml->strong($data->{login})).
			   $ml->br().
			   &substutite($template{pleasecall},
				       admin=>$ml->a($admin,{href=>"mailto:$admin"})));
	    $out .= $ml->p($ml->a($msg{"Try again?"},{href=>$forgoturl}));
	    $out .= $ml->p($ml->a($msg{"Return to login screen"},{href=>$url}));
	}
	else {
	    my $site = $this->this_site();
	    my $site_title = $site ? $site->{title} : $config{site}{name};
	    my $subject = &substitute($template{email_subject},{site_title=>$site_title});
	    my $body = &substitute($template{email_body},
                                   {user=>($data->{login} ? "user $data->{login}" : $email),
                                    site_title=>$site_title});
	    foreach my $u (@$user_list) {
                $body .= "\n";
		if ($u->{login} && $u->{password} && $u->{access}) {
		    $body .= &substitute($template{email_instructions},
                                         { login=>$u->{$config{auth}{user_login_column}},
                                           password=>$u->{$config{auth}{user_pwd_column}} } );
		}
		else {
		    $body .= &substitute($template{email_error},
                                         {admin=>$admin});
		}
                $body .= "\n";
	    }
	    if (&SendMail($email,$admin,$subject,$body)) {
		$out .= $ml->p($msg{"ERROR: unable to send email."}.
			       $ml->br().
			       &substitute($template{pleasecall},
                                           { admin=>$ml->a($admin,{href=>"mailto:$admin"}) })
                               );
		$out .= $ml->p($ml->a($msg{"Try again?"},{href=>$forgoturl}));
		$out .= $ml->p($ml->a($msg{"Return to login screen"},{href=>$url}));
	    } 
	    else {
		$out .= $ml->h1($msg{"Thank you"});
		$out .= $ml->p($msg{"Login instructions have been sent to the e-mail address we have on file."});
		$out .= $ml->p(&substitute($template{nomyemail},
                               {email=>$email,admin=>$ml->a($admin,{href=>"mailto:$admin"})}));
		$out .= $ml->p($ml->a($msg{"Return to login screen"},{href=>$url}));
	    }
	}
    }
    return $out;
}

# validate a login/password pair, and set an identity cookie
#
# If the destination is set, we validate that the user belongs in that
# destination, and allow them in if so.  Otherwise, the login is rejected
# (user attempting to login to one site using their identity from another
# site).
#
# If you want to override this logic, pass a null destination (or leave
# off the destination).
#
# $login = identity that the user has requested to log in under
# $password = the password they have provided as proof of identity
# $destination = the (restricted access) URL they are attempting to visit
# $remember = if true, set an explicit expiry time on cookies

sub login {
    my ($this,$login,$passwd,$destination,$remember) = @_;
    # check for custom login algorithm
    my $stat = $this->run_handler("login",$login,$passwd,$destination);
    return $stat if (defined $stat);

    my $user = $this->lookup($login);

    if (ref $user eq "HASH") {
	# there is such a user
	if ($this->check_password($passwd,$user)) {
	    # and their password matches
	    #if ($share{identity}{lookup}{section_id} && $destination) {
	    if ($config{auth}{scope} eq "local" && 
		$share{identity}{lookup}{section_id} && 
		$destination) {
		# this user is tied to a particular section;
		# is the destination part of that section?
		my $level = $share{identity}{access};
		my $section_id = $this->site_of($destination);
		if ($section_id) {
		    # they are logging in to a particular section
		    #if ($this->is_member($section_id)) {
		    # can't use is_member because it uses the authentication
		    # and identity info, which hasn't been set up yet
		    if ($section_id == $share{identity}{lookup}{section_id} ||
			$level >= ($config{auth}{root_auth_level} - 1)) {
			# the sections match (or this is an admin), proceed
			return $this->do_login($user,$remember);
		    }
		    else {
			# section mis-match, do not proceed
			$this->error("login: you have no access to this site");
			return 0;
		    }
		}
		else {
		    # unknown section; only allow level 2+ users to proceed
		    if ($level > 1) {
			return $this->do_login($user,$remember);
		    }
		    else {
			$this->error("login: permission denied");
			return 0;
		    }
		}
	    }
	    else {
		# user is not tied to a particular section
		return $this->do_login($user,$remember);
	    }
	}
	else {
	    #### track bad login attempts
	    $this->error("login: invalid password");
	    return 0;
	}
    }
    else {
	#### track bad login attempts
	$this->error("login: no such user");
	return 0;
    }
}

# do_login : set login identity, after the user has been authenticated

sub do_login {
    my ($this,$user,$remember) = @_;
    # check for custom login algorithm
    my $stat = $this->run_handler("do_login",$user);
    return $stat if (defined $stat);
    $this->info("login: $user");

    my $cookie;
    my $u = (ref $user eq "HASH" ? $user : $this->lookup($user));

    # set identity for the current request
    # WE NO LONGER GO THROUGH DO_LOGIN_TEMP
    #$this->do_login_temp($u);

    # preserve identity for future requests
    if (ref $u eq "HASH") {
	$share{identity} = {
	    name   => $u->{$config{auth}{user_login_column}},
	    uid    => $u->{$config{auth}{user_uid_column}},
	    gid    => $u->{$config{auth}{user_gid_column}},
	    access => $u->{$config{auth}{user_access_column}},
	    lookup => $u,
	};
	if ($config{auth}{type} eq "session") {
	    my $id = $share{identity};
	    $session{identity} = $id;
	    # NB: $remember ignored for sessions
	}
	elsif ($config{auth}{type} eq "crypt") {
	    my $cname = $config{site}{identity_cookie};
	    my $cval;
	    # see if our encrypted identity has been saved for us
	    if ($share{login_temp_crypt_cval}) {
		$cval = $share{login_temp_crypt_cval};
	    }
	    else {
		# recompute
		my $c = new ExSite::Crypt;
		$cval = $c->encrypt($u->{$config{auth}{user_login_column}}.":".
				    $u->{$config{auth}{user_uid_column}}.":".
				    $u->{$config{auth}{user_access_column}});
	    }
	    $this->info("setting identity cookie: $cname=$cval");
	    if ($remember) {
		(tied %cookie)->store_remember($cname,$cval);
	    }
	    else {
		$cookie{$cname} = $cval;
	    }
	}
	else {	  # if ($config{auth}{type} eq "cookie") {
	    my $cname = $config{site}{identity_cookie};
	    my $cval  = join(":",$u->{$config{auth}{user_login_column}},crypt($u->{$config{auth}{user_pwd_column}},"wd"));
	    $this->info("setting identity cookie: $cname=$cval");
	    if ($remember) {
		(tied %cookie)->store_remember($cname,$cval);
	    }
	    else {
		$cookie{$cname} = $cval;
	    }
	}

	$this->{authenticated} = 1;
	$this->{level} = $this->{effective_level} = $u->{$config{auth}{user_access_column}};

	# optional logging of log-ins
	$this->run_handler("log_login");

	return $this->{level};
    }
    else {
	$this->error("login: no such user: $user");
	return 0;
    }
}	    

# login_temp : temporary login - valid for one page view only.  No persistent
# cookie is set, so the user identity does not persist.  If $temp_cookie is 
# true, then we set a temporary cookie $ENV{HTTP_COOKIE}.  This is needed if
# spawning subprocesses that may not see our cookie jar (eg. publish.cgi).

sub do_login_temp {
    my ($this,$user,$temp_cookie) = @_;
    # check for custom login_temp algorithm
    my $stat = $this->run_handler("login_temp",$user,$temp_cookie);
    return $stat if (defined $stat);

    my $cookie;
    my $u = (ref $user eq "HASH" ? $user : $this->lookup($user));

    if (ref $u eq "HASH") {
	$this->clear_login;
	if ($config{auth}{type} eq "crypt") {
	    my $cname = $config{site}{identity_cookie};
	    my $c = new ExSite::Crypt;
	    my $cval = $c->encrypt($u->{$config{auth}{user_login_column}}.":".
				   $u->{$config{auth}{user_uid_column}}.":".
				   $u->{$config{auth}{user_access_column}});
	    # save so we don't have to recompute
	    $share{login_temp_crypt_cval} = $cval;
	    # set identity for this request
	    (tied %cookie)->store_update($cname,$cval);
	}
	else {	  # if ($config{auth}{type} eq "cookie") {
	    my $cname = $config{site}{identity_cookie};
	    my $cval  = join(":",$u->{$config{auth}{user_login_column}},crypt($u->{$config{auth}{user_pwd_column}},"wd"));
	    # set identity for this request
	    (tied %cookie)->store_update($cname,$cval);
	}
	$share{identity} = {
	    name   => $u->{$config{auth}{user_login_column}},
	    uid    => $u->{$config{auth}{user_uid_column}},
	    gid    => $u->{$config{auth}{user_gid_column}},
	    access => $u->{$config{auth}{user_access_column}},
	};
	$this->{authenticated} = 1;
	$this->{level} = $this->{effective_level} = $u->{$config{auth}{user_access_column}};
	return $this->{level};
    }
    else {
	$this->error("login: no such user: $user");
	return 0;
    }
}

# unset the identity cookie

sub logout {
    my $this = shift;

    $this->clear_login;  # this is sufficient for session logins

    if ($config{auth}{type} =~ /^(cookie|crypt)$/) {
	delete $cookie{$config{site}{identity_cookie}};
    }
    elsif ($config{auth}{type} =~ /^session$/) {
	delete $cookie{$config{site}{id}."_sessionID"};
    }
    return;
}

#----------------------------------------------------------------------
# check password : compare a password with the stored version
#
# Returns true (1) if the passed password matches the user's

sub check_password {
    my ($this,$password,$user) = @_;
    my $udata;
    if (! defined $user) {
        $udata = $this->my_user_record;
    }
    elsif (ref $user eq "HASH") {
	$udata = $user;
    }
    else {
	$udata = $share{DB}->lookup($user);
    }
    if (!$udata) {
	# unknown user
	$this->warn("check_password: unknown user");
	return 0;
    }
    elsif (! $udata->{$config{auth}->{user_pwd_column}}) {
	# password field is blank - deny access
	$this->warn("blank password for user \"$udata->{$config{auth}->{user_login_column}}\"");
	return 0;
    }

    if (! $password) {
	# blank given password - deny
	return 0;
    }
    if (! $udata->{password}) {
	# blank stored password - deny
	return 0;
    }

    my $pass;
    my $passtype = $config{auth}{password_storage_method};
    if ($passtype eq "des") {
	# encrypt the password claim before comparing with stored version
	my $salt = substr($udata->{password},0,2);
	$pass = ($udata->{$config{auth}{user_pwd_column}} eq crypt($password,$salt));
    }
    elsif ($passtype eq "md5") {
	# encrypt the password claim before comparing with stored version
	$pass = ($udata->{$config{auth}{user_pwd_column}} eq &md5_hex($password));
    }
    elsif ($passtype eq "crypt") {
	# decrypt the real password and compare with claim
	require ExSite::Crypt;  # don't load this unless we need it
	my $c = new ExSite::Crypt(mode=>"strong");
	$pass = ($c->decrypt($udata->{$config{auth}{user_pwd_column}}) eq $password);
    }
    else {  # cleartext
	# by default we just compare the two values directly
	$pass = ($udata->{$config{auth}{user_pwd_column}} eq $password);
    }
    return $pass;
}

# change password

sub change_password {
    my ($this,$old,$new,$new2) = @_;
    if ($new) {
	if (! defined $new2 || ($new eq $new2)) {
	    my $re = $share{DB}{map}->regexp('password');
	    if ($new =~ /^$re$/) {
		if (! defined $old || $this->check_password($old)) {
		    my $stat = $this->set_password($new);
		    return $stat if ($stat); # error
		    # update the login cookie
		    my $login = $this->my_login;
		    $this->clear_login;
		    $this->login($login,$new);
		    return undef;
		}
		else {
		    return "password update failed - incorrect password";
		}
	    }
	    else {
		return "insecure password - needs more characters";
	    }
	}
	else {
	    return "new passwords do not match";
	}
    }
    else {
	return "cannot set a blank password";
    }
}

sub set_password {
    my ($this,$password,$uid) = @_;
    if (! $uid) {
	$uid = $this->my_uid;
    }
    if ($uid) {

	# check strength
	my $strength = $this->password_strength($password,$this->my_login());
	my $access = $this->level();
	my $min_strength = $config{auth}{password_strength}{$access} || 
	    $access;
	if ($strength < $min_strength) {
	    return $msg{"Password is too weak.  Strengthen it with more text, numbers, upper-case, or punctuation characters."};
	}
	
	my $pass_store;
	my $passtype = $config{auth}{password_storage_method};
	if ($passtype eq "des") {
	    # des-encrypt the password
	    $pass_store = crypt($password,&randtext(2,'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789'));
	}
	elsif ($passtype eq "md5") {
	    # md5 hash the password
	    $pass_store = &md5_hex($password);
	}
	elsif ($passtype eq "crypt") {
	    # encrypt the password using ExSite encryption
	    require ExSite::Crypt;  # don't load this unless we need it
	    my $c = new ExSite::Crypt(mode=>"strong");
	    $pass_store = $c->encrypt($password);
	}
	else { 
	    # plaintext storage
	    $pass_store = $password;
	}

	return $share{DB}->update($config{auth}{user_table},
			  {$config{auth}{user_pwd_column}=>$pass_store},
			  {$config{auth}{user_uid_column}=>$uid});
    }
    return "no such user";
}

# test password strength
# < 0 is very weak.
# 0-1 is weak
# 1-2 is OK
# 2-3 is fairly strong
# >3 is strong
#
# As a general rule, your password strength should be >= your access level.
#
# For example:
# PASSWORD         STRENGTH
# -------------------------
# password         -3
# password0         0
# passwerd0         0.41
# Passwerd0         1.73
# pAsswerd0         2.23
# pA55w0rd          4.65
# pA55-w0rd!        9.00

sub password_strength {
    my ($this,$password,$login) = @_;
    # login id is optional, but allows for extra tests
    my $strength = $this->run_handler("password_strength",$password,$login);
    return $strength if (defined $strength);

    # starting strength depends on initial password length
    my $len = length $password;
    if ($len < 7) {
	$strength = -1;
    }
    elsif ($len < 10) {
	$strength = 0;
    }
    else {
	$strength = 1;
    }

    my $pass = $password;
    if ($login) {
	# if login is repeated in password, discount it for strength purposes
	$pass =~ s/$login//i;
    }

    # discount any of the polygrams for strength purposes
    # These indicate the use of common words or names in whole or part,
    # and potential vulnerability to dictionary attacks.
    my $foundgram = " ";
    if (-e "conf/polygrams.txt") {
	if (open F, "<conf/polygrams.txt") {
	    my @polygram = <F>;
	    chomp @polygram;
	    foreach my $gram (@polygram) { 
		# if the polygram exists in lower-case, discount completely
		$pass =~ s/$gram//;
		if ($pass =~ /$gram/i && $foundgram !~ /$gram/) {
		    # polygram found, but mixed case; discount slightly
		    $strength -= 1.0;
		    $foundgram .= "$gram ";
		}
	    }
	}
    }

    $len = length $pass;
    # discount repeated characters
    my @pass = split //,$pass;
    my $last = undef;
    foreach my $p (@pass) {
	if ($p eq $last) {
	    $p = undef;
	}
	else {
	    $last = $p;
	}
    }
    $pass = undef;
    foreach my $p (@pass) {
	$pass .= $p if (defined $p);
    }
    $len = length $pass;

    # increase strength by length of discounted password
    $strength += sqrt $len;
    
    # increase strength if password includes digits, decrease if not
    $pass =~ s/\d//g;
    my $newlen = length $pass;
    $strength += (($len - $newlen) || -1);

    # increase strength if password includes uppercase, decrease if not
    $len = $newlen;
    $pass =~ s/[A-Z]//g;
    $newlen = length $pass;
    $strength += (($len - $newlen) || -1);

    # increase strength if password includes non-alphanumeric characters, decrease if not
    $len = $newlen;
    $pass =~ s/[^0-9a-zA-Z]//g;
    $newlen = length $pass;
    $strength += (($len - length($pass)) || -1);

    return $strength;
}

# Authentication Token handling:
# an authtoken is an encrypted string that can be included in the
# query string, which will automatically authenticate the user.

# make_authtoken : generate an authtoken
# $username = the username to auto-validate the user under
# $expiry = number of days in future the authtoken should expire
# $url = the url that the authtoken can be used on

### TODO: support "short" authtokens which use a CRCsum in place of a URL

sub make_authtoken {
    my ($this,$username,$expiry,$url) = @_;
    my $udata = $share{DB}->lookup($username);
    return undef if (! $udata);  # no such user
    if ($expiry !~ /^-?\d+$/) { $expiry = 0; }
    if ($expiry > 0) {
	# we have been passed a number of days in the future;
	# convert this to a unix-timestamp-style day count 
	# (ie. days since Jan 1, 1970)
	my $t = new ExSite::Time;
	$t->add($expiry,"days");
	$expiry = int($t->write("unix_timestamp") / 86400 + 1);
    }
    if ($url) {
	# store complete URL
	# we need to encode the script name, path info, and query args
	# 1) remove the server to save space
	$url =~ s/^$config{server}{server}//;
    }
    $username = &safetext($username);
    my $authtoken_plaintext = "$expiry;$username";
    if ($url) { $authtoken_plaintext .= ";$url"; }
    my $crypt = new ExSite::Crypt;
    return $crypt->encrypt($authtoken_plaintext);
}

# validate_authtoken : return the username, if the authtoken is valid

sub validate_authtoken {
    my ($this,$authtoken) = @_;
    my $crypt = new ExSite::Crypt;
    my $authtoken_plaintext = $crypt->decrypt($authtoken);
    my ($expiry,$username,$url) = split(/;/,$authtoken_plaintext,3);
    $username = &unsafetext($username);

    # validate the expiry time of the authtoken
    $expiry *= 86400;
    if ($expiry > 0 && time > $expiry) {
	$username = undef;
	$this->warn("Sorry, you have tried to connect using an expired authentication token for $username;  access is denied.");
    }

    # validate the user name
    if ($username) {
	my $udata = $share{DB}->lookup($username);
	if (ref $udata ne "HASH") {
	    $this->error("invalid authtoken for $username;  no such user.");
	    $username = undef;
	}
    }

    ### FIXME: use URI for parsing?

    # validate the URL
    if ($username && $url) {
	my ($protocol,$host,$scriptpath,$scriptfile,$pathinfo,$query) = 
	    &parse_url($url);
	if ($share{ENV}{SCRIPT_NAME} =~ /\/$scriptfile$/ &&
	    $share{ENV}{PATH_INFO} =~ /^$pathinfo$/) {
	    my %q1 = &DecodeString($query);
	    my $input = new ExSite::Input;
	    my $q = $input->query;
	    my %q2 = %$q;
	    delete $q2{_auth};
	    foreach my $key (keys %q2) {
		if (exists $q1{$key}) {
		    if ($q1{$key} eq $q2{$key}) {
			delete $q1{$key};
			delete $q2{$key};
		    }
		}
	    }

	    # if q1 and q2 are identical, both should now be empty
	    if (keys %q1 > 0 || keys %q2 > 0) {
		$this->error("invalid authtoken for $username; invalid URL.");
		$username = undef;
		$expiry = 0;
	    }
	}	    
	else {
	    $this->error("invalid authtoken for $username; invalid URL.$share{ENV}{SCRIPT_NAME} != $scriptfile; $share{ENV}{PATH_INFO} != $pathinfo");
	    $username = undef;
	    $expiry = 0;
	}
    }
    return ($username,$expiry);
}

# process_authtoken : login the user, if the authtoken is valid

sub process_authtoken {
    my ($this,$authtoken) = @_;
    my ($login,$expiry) = $this->validate_authtoken($authtoken);
    if ($login) {
	if ($expiry == 0) {
	    # login expires instantly;  do not set a cookie
	    $this->do_login_temp($login);
	}
	else {
	    # login persists
	    $this->do_login($login);
	}
    }
    return;
}

# check_authtoken : find an authtoken in the input and process it

sub check_authtoken {
    my ($this) = shift;
    my $in = new ExSite::Input;
    my $auth = $in->combine()->{_auth};
    if ($auth) {
	$this->process_authtoken($auth);
	# If the authtoken is accepted, it will reset our identity;
	# if not, we will fall back on the previously set identity.
    }
    return;
}

# fetch all admin tickets

sub my_sites {
    my ($this) = @_;
    return $share{DB}->fetch_m2m("section",$config{auth}{user_table},
				 $this->my_uid,"member_site_link");
}

# ditto, but do not count website templates

sub my_real_sites {
    my ($this) = @_;
    my @real_sites;
    foreach my $site ($this->my_sites) {
	if ($site->{type} !~ /template/) {
	    push @real_sites, $site;
	}
    }
    return @real_sites;
}

# return list of sites the user has member-level access to

sub my_access_keys {
    my $this;

    # start with all sites with admin access
    my @sites = $this->my_real_sites;
    
    # add sites with regular member access
    my $user = $share{DB}->my_user_record;
    if ($user->{section_id}) {
	push @sites, $share{DB}->fetch("section",$user->{section_id});
    }
    
    return @sites;
}

# return the current site data (may require some guesswork)

sub this_site {
    my ($this) = @_;
    if ($share{Page}) {
	# we're on a page, so return the page's section
	return $share{Page}->fetch("section",$share{Page}->get("page","section_id"));
    }
    elsif ($share{Section}) {
	# we're on a section, so return it
	return $share{Section}->get();
    }
    elsif ($share{DB}->my_user_record()->{section_id}) {
	# user has a primary site defined
	return $share{DB}->fetch("section",$share{DB}->my_user_record()->{section_id});
    }
    else {
	my @sites = $this->my_real_sites;
	if (@sites > 0) {
	    # ambiguous result if more than one - return first one
	    my @s = $share{DB}->sort_rows("section",\@sites,"section_id");
	    return $s[0];
	}
    }
    return undef;
}

#----------------------------------------------------------------------
# site_of : returns the section of a dyanmic content URL
#
# Only works for page.cgi and content.cgi.  For all other URLs,
# returns "0".

sub site_of {
    my ($this,$url) = @_;
    if ($url =~ /$config{prog}{page}([^?\#]*)\?.*_id=(\d+)/) {
	# specified as id
	my $p = $share{DB}->fetch("page",$2);
	return $p->{section_id};
    }
    elsif ($url =~ /$config{prog}{page}([^?\#]*)\??/) {
	# specified as path
	my $page = new ExSite::Page();
	my ($p,$path) = $page->path_to_page($1);
	return $p->{section_id};
    }
    elsif ($url =~ /$config{prog}{content}\?.*id=(\d+)/) {
	my $c = $share{DB}->fetch("content_data",$1);
	if ($c) { 
	    my $p = $share{DB}->fetch("page",$c->{"page_id"});
	    if ($p) {
		return $p->{section_id};
	    }
	}
    }
    return 0;
}

# test for "ownership" of a site, ie. an admin key

sub owns {
    my ($this,$site) = @_;    # $site can be a section hash, or a section ID
    if (! $site) { $site = $this->this_site(); }
    my $sid = ref $site eq "HASH" ? $site->{section_id} : $site;
    return 0 if (! $sid);
    my @temp = $share{DB}->fetch_match("member_site_link",
				       { member_id => $this->my_uid,
					 section_id => $sid } );
    return (@temp > 0);
}

# test for license to use a service on a site

### DEPRECATED - use ExSite::Module::site_license()

sub license {
    my ($this,$service,$site) = @_;
    if (eval "require Modules::$service") {
	my $s = eval "new Modules::$service";
	if ($s->ioctl("isRestricted")) {
	    # this service requires licenses
	    if (!$site && $share{Page}) {
		$site = $share{Page}->id("section");
	    }
	    if ($site) {
		my $licenses = $share{DB}->fetch_match("service",
						       {section_id => $site,
							name => $service }
						       );
		return (scalar @$licenses > 0);
	    }
	    # hmm, that's strange, we're not on a page
	    $this->warn("$service: unknown website, cannot confirm license");
	    return 0;
	}
	# this modules is not a service - anyone can use it
	return 1;
    }
    # module does not seem to exist
    $this->warn("$service: no such service");
    return 0;
}

# test for license to use a service's control panel
# ie. test that a web app is licensed on any of the user's sites

### DEPRECATED - use ExSite::Module::user_license()

sub license_control_panel {
    my ($this,$service) = @_;
    if (! $share{licenses}) {
	$share{licenses} = {};
	my $db = $share{DB};
	if ($db->{dbtype} =~ /SQL/i) {
	    # use a join query to fetch all licensed services
	    my $uid = $db->my_uid;
	    my @services = $db->custom_query("select service.name from service,section,member_site_link where service.section_id=section.section_id and service.section_id=member_site_link.section_id and member_site_link.member_id=$uid");
	    $share{licenses} = &keywise("name",\@services);
	}
	else {
	    # brute-force loop over all services/sites
	    my @sites = $db->my_real_sites();
	    foreach my $l ($share{DB}->fetch_all("service")) {
		foreach my $s (@sites) {
		    if ($l->{section_id} == $s->{section_id}) {
			$share{licences}{$l->{name}} = 1;
			next;
		    }
		}
	    }
	}
    }
    return (exists $share{licenses}{$service});
}

#
# user management
#

# fetch_admins : fetch all users with unrestricted access

sub fetch_admins {
    my ($this,$site) = @_;
    return $this->fetch_users_by_level($site,$config{auth}{root_auth_level}-1);
}

# fetch all users with manager level access (does not include admins)

sub fetch_managers {
    my ($this,$site) = @_;
    return $this->fetch_users_by_level($site,$config{auth}{root_auth_level}-2);
}

# fetch all users with a certain level of access (does not include admins)

sub fetch_users_by_level {
    my ($this,$site,$level) = @_;
    if (! defined $level) { $level = 1; }
    my @users_by_level;
    if ($site) {
	my $sid;
	$sid = (ref $site eq "HASH") ? $site->{section_id} : $site;

	# fetch everyone with an admin key
	my @users = $share{DB}->fetch_m2m($config{auth}{user_table},"section",$sid,"member_site_link");
	# select only those at our level of access
	foreach my $u (@users) {
	    if ($u->{access} == $level) {
		push @users_by_level, $u;
	    }
	}

	# fetch everyone with this as their home site
	@users = $share{DB}->fetch_match($config{auth}{user_table}, 
					 { access=>$level, section_id=>$sid});
	my @extras;
	# weed out the dupes
      USER:
	foreach my $u (@users) {
	  ADMIN:
	    foreach my $a (@users_by_level) {
		next USER if ($u->{member_id} == $a->{member_id});
	    }
	    # no dupe detected
	    push @extras, $u;
	}
	# add the non-dupes to our admin list
	push @users_by_level, @extras;
    }
    else {
	# no site given, just take everyone
	@users_by_level = $share{DB}->fetch_match($config{auth}{user_table},{access=>$level});
    }
    return @users_by_level;
}

# find most authoritative user on a site 

sub fetch_owner {
    my ($this,$site) = @_;
    if (! $site) { $site = $this->this_site; }
    
    # check for admins/manager who have this as their home site

    my @u = $this->fetch_match($config{auth}{user_table},"section_id=$site->{section_id} and access>=2");
    if (@u > 0) {
	foreach my $u (@u) {
	    if ($u->{access} >= $config{auth}{root_auth_level}-1) {
		return $u;
	    }
	}
	return $u[0];
    }

    # still here;  check for anyone with an admin key

    while ($site) {
	@u = $this->fetch_admins($site);
	if (@u > 0) {
	    return $u[0];
	}
	@u = $this->fetch_managers($site);
	if (@u > 0) {
	    return $u[0];
	}
	# no luck;  try parent site
	if ($site->{parent_id}) { 
	    $site = $this->fetch_owner($share{DB}->fetch("section",$site->{parent_id}));
	}
    }
    return undef;
}

sub owner_email {
    my ($this,$site) = @_;
    if ($site) {
	my $u = $this->fetch_owner($site);
	if (ref $u eq "HASH" && $u->{email}) {
	    return $u->{email};
	}
    }
    # no owner email found; use config administrator email
    return $config{site}{administrator};
}

sub adduser {
    my $this = shift;
    my $udata = shift;  # member datahash
    my @keys = @_;      # admin keys
    my $db = $share{DB};
    my @err;

    # 1) validate the password

    my $stat = $db->validate_column($config{auth}{user_table},"password",$udata->{password});
    if ($stat) {
	push @err, "Invalid/insecure password - please choose another";
    }

    # 2) ensure the login ID is unique

    my @u = $db->fetch_match($config{auth}{user_table}, { login=>$udata->{login} });
    if (@u > 0) {
	push @err, "That $config{auth}{login_prompt} is already in use - please try another";
    }

    # 3) name

    if ($udata->{name} =~ /^\s*$/) {
	push @err, "Cannot insert member - no name provided";
    }

    # 4) validate the email

    $stat = $db->validate_column($config{auth}{user_table},"email",$udata->{email});
    if ($stat) {
	push @err, "Invalid email address";
    }

    # 5) home site

    if ($udata->{section_id}) {
	$stat = $db->validate_column($config{auth}{user_table},"section_id",$udata->{section_id});
	if ($stat) { 
	    push @err, $stat;
	}
    }

    # abort if there are errors
    if (@err > 0) {
	$this->error(@err);
	return 0;
    }
    
    # insert the user
    if ($udata->{member_id}) { 
	$this->warn("removing member ID of new user");
	delete $udata->{member_id}; 
    }
    my $uid = $db->insert("member",$udata);
    if (! $uid) {
	$this->error("failed to add user $udata->{login}");
	return 0;
    }

    # add admin keys

    foreach my $key (@_) {
	next if (ref $key ne "HASH");
	$key->{member_id} = $uid;
	my @err = $db->validate_record("member_site_link",$key);
	if (@err) {
	    $this->error(@err);
	}
	else {
	    $db->insert("member_site_link",$key);
	}
    }

    return $uid
}

sub deluser {
    my $this = shift;
    my $user = shift;
    my $uid = ref $user eq "HASH" ? $user->{member_id} : $user;
    $share{DB}->trash_r($config{auth}{user_table},$uid);
}

# referer check - is there a referer?
# This check is used mostly to reduce crawling or snooping of the database.

sub referer_check {
    my $this = shift;
    # check can be disabled across the board
    return 1 if (! $config{auth}{referer_check});
    # check can be disabled for logged-in users
    if (! $config{auth}{member_referer_check}) {
	my $level = $this->authorize();
	return 1 if ($level);
    }
    return defined $ENV{HTTP_REFERER};
}

1;

