#!/usr/bin/perl
#-----------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

=pod

=head1 Content.pm : manage content objects

A content object is an element of a web page, such as text, html, 
stylesheets, scripts, images, or downloadable files.  This module
provides methods to act on generic content objects.

This is part of the ExSite content stack, so it inherits from the
ContentData (and then ContentBase) class, and is in turn inherited by
Page.

=cut

package ExSite::Content;

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::ContentData;
use MIME::Base64;

use vars qw(@ISA);
@ISA = qw(ExSite::ContentData);

=pod

=head1 Configuration & Setup

=head2 new()

(This method is actually inherited from ExSite::ContentBase.)

=head2 setup()

=head2 setupContent()

Instantiate and/or configure a content object.  C<%option> includes the 
the configuration parameters, described below.

Usage:

    my $content = new ExSite::Content(%option);
    $content->setup(%option);
    $content->setupContent(%option);

C<setup> and C<setupContent> are synonymous;  they are used to setup or
reconfigure an existing Content object.  C<new()> instantiates a new content
object (this method is actually inherited from C<ExSite::ContentBase>).

C<%option> is used to pass the content description information that
allows the object to be configured.  It can include the following content
description parameters:

=over 4

=item id

The numeric content ID will be used to fetch the content description from 
the database.

=item page

The content record can be passed as a hash reference, if it has already been
fetched.

=item page, name

Given a page ID and content object name, the content can be looked up
in the database.

=item (null)

If no content description is provided, the object will
configure an empty content object that will have limited
functionality.

=back

The setup methods are normally invoked automatically by the class
contructor method C<new()>, which receives the same C<%option>
parameters.

=cut

# setup info for Content class

sub setup {
    &setupContent(@_);
}

sub setupContent {
    my ($this,%option) = @_;

    # fetch the object in question
    ###### CHECK PERMISSIONS.
    my $content;
    if ($option{id}) {
	# direct lookup
	$content = $this->fetch("content",$option{id});
    }
    elsif ($option{content}) {
	# record already fetched
	$content = $option{content};
    }
    elsif ($option{name} && $option{page}) {
	########################
	### search templates for matching content for general display of page,
	### but don't search in CMS management mode

	my @content = $this->fetch_match("content",
					 {name=>$option{name},
					  page_id=>$option{page}},
					 "content_id");
	# should be only 1, but in case there are more, we'll take the
	# most recently added one
	$content = $content[-1];
    }

    if (ref $content eq "HASH") {
	if ($content->{type} eq "alias") {
	    # points to another content record
	    $content = $this->fetch("content",$content->{description});
	    if (ref $content ne "HASH") {
		$this->info("setting up null Content");
		return;
	    }
	}

	# found something
	$this->info("setting up Content $content->{content_id}");
	$this->{data} = $content;
	if ($option{revision}) {
	    # this will load a revision, allowing the content to be rendered
	    $this->set_context("child",$option{revision});
	}
	else {
	    # this is faster for simple content queries
	    $this->set_context("none");
	}
	
	#$this->{name} = $content->{name};
#	my $context = "none";
#	if (! defined $option{context} || $option{context}) {
#	    $context = $option{context} ? $option{context} : "all";
#	}
#	$this->set_context($context,$content);
    }
    else {
	$this->info("setting up null Content");
    }
}

=pod

=head1 Authorization

=head2 authtype()

Returns the content/authorization type, ie. "editorial" or "design", 
which determines who is allowed to update the content.

=head2 authorize()

Returns TRUE if the current user is permitted to update the content.
This depends on the the user's access level, and their admin key for
the current section, which determines their access privileges.

=cut

# authtype = editorial|design

sub authtype {
    my $this = shift;
    return $this->get("content","type");
}

# allow the current user to modify the content

sub authorize {
    my $this = shift;
    $this->set_context();
    my $section_id = $this->get("section","section_id");
    if (! $section_id) {
	my $p = $share{DB}->fetch("page",$this->get("content","page_id"));
	$section_id = $p->{section_id};
    }
    return 0 if (! $section_id);
    
    my @key = $share{DB}->fetch_match("member_site_link",
				      {section_id=>$section_id,
				       member_id=>$share{DB}->my_uid});
    my $type = $this->authtype;
    foreach my $key (@key) {
	return 1 if ($type eq "editorial" && $key->{priveleges} =~ /editor/);
	return 1 if ($type eq "design" && $key->{priveleges} =~ /designer/);
    }
    # no matching priveleges found
    return 0;
}

=pod

=head1 Content Management

=head2 update()

This creates a new revision of the content; it is synonymous with
C<ExSite::ContentData::make_content_data()>.

=head2 rollback()

This removes the most recent revision of the content.

=head2 count()

Returns the number of revisions on file for this content.

=cut


# update the content - synonym for make_content_data

sub update {
    return &ExSite::ContentData::make_content_data(@_);
}

# roll back to the previous revision of the content

sub rollback {
    my $this = shift;
    $this->set_context("child");
    my $cdata = new ExSite::ContentData(content_data=>$this->{context}{content_data});
    $cdata->delete;
    if ($cdata->errorcheck()) {
	# copy the errors into this object, so we can see them later
	$this->error($cdata->fetch_diagnostics("error"));
    }

    ### this will have unintended consequences if the revision in use
    ### is not the latest revision!

    # reconfigured the context with the previous revision
    $this->{context}{"content_data"} = $this->find_content_data();
}

# count the number of revisions on file

sub count {
    my $this = shift;
    return $share{DB}->count("content_data",
			     {content_id=>$this->id("content")});
}

=pod

=head1 Creating/Modifying Content Objects

NB: modifying means changing the content attributes, not updating the 
content.  The latter operation is done by calling the C<update()>
method.

=head2 validate()

    my @errors = $content->validate($data);
    if (@errors == 0) { # content is OK }

Returns TRUE if the values in the the C<$data> hashref comprise a
valid and legal content description.  Returns an array of error
messages, which will be empty if the content validates.

=head2 make()

=head2 make_content()

    $content->make($data,$revision,$opt);

Creates a new content object, with the attributes given in the
C<$data> hashref.  If the user is permitted to create new content
objects and the attributes validate, then the content record will be
created.

If C<$revision> is defined, then a ContentData object will be created
as well, using the content_data description in C<$revision>, and any
extra options in $opt.  These are passed to
C<ExSite::ContentData::make_content_data()>.

=head2 delete()

Removes the content object from the system.  Attempts to unpublish its
files, as well.

=cut

# validate 

# validate - check integrity of datahash

sub validate {
    my ($this,$data) = @_;
    my $content;
    if ($data) {
	$content = new ExSite::Content(content=>$data);
    }
    else {
	$content = $this;
	$this->set_context("child");
	$data = $this->get("content");
    }
    my @err = $share{DB}->validate_record("content",$data);
    if (! $data->{name}) {
	push @err, "content has no name";
    }
    else {
	# simplify name, in case it contains special characters
	$data->{name} = &clean_filename(&latin1_to_ascii($data->{name}));
	
	my $p = $share{DB}->fetch("page",$data->{page_id});
	if (! ($data->{name} eq "body" && $p->{type} eq "library")) {
	    # cannot have two content objs of the same name in the same
	    # page, unless it's a library of bodies
	    my @c = $share{DB}->fetch_match("content",{page_id=>$data->{page_id},name=>$data->{name}});
	    if (@c > 0) {
		push @err, "content of that name is already installed";
	    }
	}
    }
    if ($data->{page_id}) {
	if (! $content->owns()) {
	    push @err, "ownership mismatch";
	}
    }
    return @err;
}


# make_content : insert a new content record
#
# can optionally pass a content data record to seed it with

sub make {
    return &make_content(@_);
}

sub make_content {
    my ($this,$content,$cdata,$opt) = @_;
    if ($content->{content_id}) {
	# invalid content
	$this->error("invalid attempt to insert an already-defined content record");
	return undef;
    }
    if (! $this->allow_insert("content",$content)) {
	# permission denied
	$this->error("make_content: permission denied");
	return undef;
    }

    if (! $content->{type}) {
	$content->{type} = "editorial";
    }
    my @err = ExSite::Content::validate($this,$content);
    if (@err > 0) {
	$this->error(@err);
	return undef;
    }
    # remove any residual timestamps
    delete $content->{mtime};
    delete $content->{ctime};
    $share{DB}->run_handler("pre_make_content",$content);
    my $cid = $share{DB}->insert("content",$content);
    if (! $cid) {
	$this->error("failed to install new content $content->{name}");
    }
    else {
	$this->info("installed new content $content->{name}");
	# update the prefetch cache
	$content->{content_id} = $cid;
	$this->prefetch_update("content",$content);
	if (ref $cdata eq "HASH") {
	    $cdata->{content_id} = $cid;
	    my $cdid = $this->make_content_data($cdata,$opt);
	    # ignore return code; make_content_data will report its own errs
	}
	$content->{content_id} = $cid;
	$share{DB}->run_handler("post_make_content",$content);
    }
    return $cid;
}

# everything else is inherited :-)

### FIXME:
# - is_published() - determine whether content needs to be published
# - publish_date() - determine when content was last published

# delete the content object

sub delete {
    my $this = shift;
    my $id = $this->id;
    if ($id && $this->allow_delete("content",$this->{data})) {
	$this->unpublish();
	$share{DB}->trash_r("content",$id);
    }
}

1;
