#!/usr/bin/perl
#-----------------------------------------------------------------------
#
#   Copyright 2001-2006 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::ContentBase;

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::ML;
use ExSite::Util;
use ExSite::Base;
use ExSite::Input;
use ExSite::Cache;

use vars qw(@ISA @class_structure %class_index %revision);
@ISA = qw(ExSite::Base);
@class_structure = ("section", "page", "content", "content_data");
%class_index = ("ExSite::Section"=>0, 
		"ExSite::Page"=>1, 
		"ExSite::Content"=>2, 
		"ExSite::ContentData"=>3, 
		);
%revision = ( current => "active",
	      active => "active",
	      published => "active",
	      newest => "newest" );

#------------------------------------------------------------------------

=pod

=head1 ContentBase.pm : generic content-management methods

This is the base class for the ExSite core content management
framework.  (Core content pertains to basic web page content
such as HTML, images, stylesheets, etc.)  The base class contains
tools that can be used/inherited by all CMS data types (ie. sites,
pages, content definitions, and content data), as well as tools
that deal with the entire content stack at once.

NOTE: CMS objects can exist in two forms:  (1) as data objects (a hash
of attribute keys/values, equivalent to a DB record), or (2) as code
objects.  The names of the data objects correspond to the names of 
the DB tables used to store their attributes.  The names of the code
objects correspond to the Perl modules/classes that define them. 
These names are similar, but do not match exactly:

    +------------------------------+-----------------------------+
    |  DATA OBJECT                 |  CODE OBJECT                |
    +------------------------------+-----------------------------+
    |  section                     |  Section                    |
    |  page                        |  Page                       |
    |  content                     |  Content                    |
    |  content_data                |  ContentData                |
    +------------------------------+-----------------------------+

It can be considered a bug that we use both conventions when referring to
object types, not always consistently.

ContentData objects are often called "revisions".

=head2 new()

This is a general-purpose constructor that is used by all CMS
classes.  As such, it is generally called via one of the other
CMS classes (ie. C<ContentData>, C<Content>, C<Page>, or C<Section>).
Usage is typically:

    my $page = new ExSite::Page(id=>$page_id);      # numeric ID

    my $page = new ExSite::Page(page=>$page_data);  # page datahash

where "C<Page>" can be replaced with the name of the code object, and
"C<page>" can be replaced with the name of the data object.  For
certain classes, other parameters are also allowed.  See the
documentation for those classes for details.

=cut

#------------------------------------------------------------------------


# options
# name => content object name
# page => page this content is assigned to
# id => direct lookup of content object
# revision => the revision to select (current|active or newest)
# DB => database object (optional)

sub new ($%) {
    my($this, %option) = @_;

    # objectify object
    my $cobj = {};
    my $class = ref($this) || $this;
    bless $cobj, $class;

    # setup content cache - in share, so that multiple objects can share it

#    if (! exists $share{ccache}) {
#	$share{ccache} = new ExSite::Cache;
#    }

    # setup Base class features

    $cobj->initialize_object;

    # register CMS-specific handlers

#    $cobj->{DB}->handler("select_foreign_key",&CMS_select_foreign_key);

    # setup context hash

    $cobj->{context} = {};

    # set preferred revision

    if ($option{revision}) { 
	$cobj->set_revision($option{revision});
    }

    # setup metadata

    $cobj->{metadata} = {};

    # setup prefetch cache for faster content lookups

    if (! exists $share{prefetch}) { $share{prefetch} = {}; }

    # end generic constructor; now run local class setup

    $cobj->setup(%option);
    return $cobj;
}

######################### CONTEXT METHODS ############################

=pod

=head1 Context Handling

The "context" consists of the related elements in the content "stack".
For instance, an image object will belong to a certain page, which will
belong to a certain section;  it will also have a current revision among 
the known revisions.

=head2 set_context()

    $this->set_context($mode,$revision);

The rules for identifying parents in the stack is simply to use the
foreign key reference in the child record.  The rules for identifying
children in the stack are more complicated:

=over 4

=item * the child of a site is the index page for the site

=item * the child of a page is the "page" content object for the page

=item * the child of a content object is the current revision of the content

=back

Setting the context consists of finding and loading the other objects 
in this stack.  It is assumed that the current data object is already 
loaded.

C<$mode> is C<none|all|parent|child>, depending on whether one needs
no context, full context, parent context, or child context only.
"all" is most versatile; other modes may save a bit of work and be
more efficient for particular cases.

C<$revision> is "current" or "active" to select the latest approved/published
revision of the content, or "newest" to select the latest revision even if
not approved.

Many of the context-testing functions below will not return useful
results if the context has not been set.  However, we don't want to 
automatically set the context every time, because it is costly and 
many basic operations do not require it.

=head2 clear_context()

Removes any context settings that have already been made.

=head2 reset_context()

    $this->reset_context($mode,$revision);

C<set_context()> saves its results so that it will not execute again
unnecessarily.  C<reset_context()> forces the context to be set over
again. Parameters are the same as C<set_context()>

=head2 set_revision()

    $this->set_revision($revtype);

Sets the preferred revision, eg. "active" or "newest".  
("active" is the default)

=head2 get_revision()

Returns the preferred revision, eg. "active" or "newest".
("active" is the default)

=cut

sub set_context {
    my ($this,$mode,$revision) = @_;
    return if $this->{context_set};
    $mode = $mode || "all";
    if ($mode eq "all") { $this->{context_set} = 1; }
    # start with current object
    my $class = ref $this;
    my $index = $class_index{$class};
    $this->{context}{$class_structure[$index]} = $this->{data};
    return if ($mode eq "none");
    if ($mode =~ /all|parent/) {
	# move up thru the content stack by following foreign key refs
	for (my $i = $index; $i > 0; $i--) {
	    my $table = $class_structure[$i];
	    my $ptable = $class_structure[$i-1];
	    my $pkey = $ptable."_id";
	    my $id = $this->{context}{$table}{$pkey};
	    if ($id) {
		$this->{context}{$ptable} = $this->fetch($ptable,$id);
	    }
	    else {
		$this->warn("no $table found");
		last;
	    }
	}
    }
    if ($mode =~ /all|child/) {
	# move down thru the content stack
	my @cstack = @class_structure;
	while (scalar @cstack && $cstack[0] ne 
	       $class_structure[$class_index{$class}]) { 
	    shift @cstack; 
	};
	foreach my $ctype (@cstack) {
	    next if (ref $this->{context}{$ctype} eq "HASH"); # already loaded
	    if ($ctype eq "page") {
		$this->{context}{$ctype} = $this->find_page("index.%");
	    }
	    elsif ($ctype eq "content") {
		$this->{context}{$ctype} = $this->find("page","data");
	    }
	    elsif ($ctype eq "content_data") {
		$this->{context}{$ctype} = $this->find_content_data($revision);
	    }
	}
    }
}

# clear the context

sub clear_context {
    my $this = shift;
    $this->{context_set} = 0;
    $this->{context} = {};
    # $this->{data} = {};   # not the context
    return;
}

# force new context, even if context was already set

sub reset_context {
    my $this = shift;
    $this->clear_context;
    $this->set_context(@_);
}

# set the preferred revision

sub set_revision {
    my $this = shift;
    my $rev = shift;
    $this->{revision} = $revision{$rev} || "active";
}

sub get_revision {
    my $this = shift;
    return $this->{revision} || "active";
}

#------------------------------------------------------------------------

=pod

=head2 whatami()

Returns the type of the current object, one of "Page", "Section",
"Content", or "ContentData".

=head2 is()

Tests if the current object is of a certain type.  C<$type> can be one
of "Page", "Section", "Content", or "ContentData".

Example - given a content object C<$x>:
    
    if ($x->is("Page")) { # we are a Page object

=head2 has()

    if ($this->has($type)) { ...

Test if the context contains a particular type of CMS object.
C<$type> can be one of "Page", "Section", "Content", or "ContentData".
Returns TRUE if the given type is found in the current object's
context.  If FALSE, the given type may be found if you call
C<set_context()>.

Example - given a content object C<$x>:
    
    if ($x->has("Section")) { # we have our section data loaded into the context

=head2 has_content()

Returns TRUE if the object contains actual web content (ie. a revision).
Returns FALSE if not, or if the context has not been set.

=head2 get()

    my $value = $this->get($type,$attribute);

Returns the value of the given attribute, for the given object
C<$type> in the context.  C<$type> can be one of "Page", "Section",
"Content", or "ContentData".  C<$attribute> can be any attribute of
the C<$type>.  If C<$attribute> is not given, all attributes are
returned in a datahash.

Example - given a content object called C<$img>, get the filename of 
the page it is located in:
    
    my $filename = $img->get("page","filename");

Or, get all attriubtes of the page it is located in:
    
    my $page = $img->get("page");

If C<$type> is not a context item, then we will try to return an
attribute of the object that is named C<$type>.

=head2 id()

    my $id = $this->id($type);

Returns the numeric ID of the requested C<$type> in the context.

C<$type> can be one of "Page", "Section", "Content", or "ContentData",
which must be loaded into the context.  If C<$type> is not given, 
returns the numeric ID of the current object.

Example - given a content object called C<$img>:

    my $id = $img->id();            # ID of the image
    my $page_id = $img->id("page"); # ID of the image's page

=cut

#------------------------------------------------------------------------

sub whatami {
    my ($this,$datafmt) = @_;
    my $iam = ref $this;
    $iam =~ s/^.+::(.+)$/$1/;
    if ($datafmt) {
	# underscore kludge because class names do not match data names
	$iam = lc($iam);
	return $iam eq "contentdata" ? "content_data" : $iam;
    }
    return $1;
}

sub is {
    my ($this,$type) = @_;
    # return true if this object is a $type
    return (ref $this =~ /^$type$/i);
}

sub has {
    my ($this,$type) = @_;
    # return true if this object has a $type loaded in its context
    return defined $this->{context}{$type};
}

sub has_content {
    my ($this) = shift;
    # return true if this has content data defined 
    if ($this->has("content_data")) {
	return ($this->{context}{content_data}{data} || 
		$this->{context}{content_data}{fdata} || 
		$this->{context}{content_data}{file} || 
		$this->{context}{content_data}{url});
    }
    return undef;
}

sub get { 
    my ($this,$type,$attribute) = @_;
    if ($type) {
	# default to ::Base behaviour if $type is not a context item
	return $this->{$type} if (ref $this->{context}{$type} ne "HASH");
	# if $type is a context item, return data for it
	return $attribute ? 
	    $this->{context}{$type}{$attribute} : 
	    $this->{context}{$type};
    }
    else {
	# no args - return this object's data
	return $this->{context}{$class_structure[$class_index{ref $this}]};
    }
}

sub get_my { 
    my ($this,$attribute) = @_;
    return (ref $this->{data} eq "HASH" && exists $this->{data}{$attribute}) ?
	$this->{data}{$attribute} : 
	undef;
}

sub id {
    my ($this,$type) = @_;
    my $type = $type ? $type : $class_structure[$class_index{ref $this}];
    return $this->{context}{$type}{$type."_id"};
}

#------------------------------------------------------------------------

=pod

=head1 CMS Access Control

Access control in the ExSite CMS is normally handled on a section-by-section
basis.  Each section has its own set of managers, who can each have the
following types of access to the section:

=over 4

=item editorial

The manager can create, update, and delete editorial objects.

=item design

The manager can create, update, and delete design objects and templates.

=item administrator

The manager can create, delete, and publish pages.

=back

Who has manager access to a section, and what types of access they have, 
is determined by their administrator keys to the section.  An 
administrator key is a hash with the following keys/values:

=over 4

=item member_id

The UID of the user whom this key belongs to.

=item section_id

The section ID the key is for.

=item priveleges

A plain text list of the types of access (above) that are granted by
this key.  (Yes, it is spelled wrong.)

=back 

If you have a group of items that require different access
privileges from the rest of the section, consider placing these items
into their own section and issuing a new set of keys for them.

=head2 get_keys()

Returns the key(s) that the current user has for the current section.
They are returned in an array (or arrayref in scalar context), even
if there is only one.

=head2 allow_update()

    if ($this->allow_update($type,$data)) { ...

Returns TRUE if the user can modify the C<$type>, which can be one of
"page", "section", "content", or "content_data".  C<$data> is the
attribute datahash for the C<$type>, if not using the one in the
context.

=head2 allow_insert()

    if ($this->allow_insert($type,$data)) { ...

Returns TRUE if the user can create a new C<$type>, which can be one of
"page", "section", "content", or "content_data".  C<$data> is the
attribute datahash for the C<$type>.

=head2 allow_delete()

    if ($this->allow_delete($type,$data)) {...

Returns TRUE if the user can delete the C<$type>, which can be one of
"page", "section", "content", or "content_data".  C<$data> is the
attribute datahash for the C<$type>, if not using the one in the
context.

=head2 owns()

Returns TRUE if the user owns the object, that is, if the user has an
administrator key for the section that the object belongs to.

=cut

#------------------------------------------------------------------------

sub get_keys {
    my $this = shift;
    $this->set_context();  # need section ID to do authorization
    my $db = $share{DB};
    return $db->fetch_match("member_site_link",
			    {member_id=>$db->my_uid(),
			     section_id=>$this->id("section")});
}

sub allow_update {
    my ($this,$type,$data) = @_;
    if (! $type) { $type = $this->whatami(1); }
    if (! $data) { $this->set_context(); $data = $this->{context}{$type}; }
    foreach my $key ($this->get_keys()) {
	if ($key->{priveleges} =~ /administrator/) { return 1; }
    }
    $this->error("Configure $type: permission denied - you need administrator privileges on this site.");
    return 0;
}

sub allow_insert {
    my ($this,$type,$data) = @_;
    my $db = $share{DB};
    if (! $type) { $type = $this->whatami(1); }

    # with inserts, we often have no preexisting object to set a context
    # on, so we must manually figure out what section the insert is 
    # going into.
    my $section_id;
    if ($type eq "section") {
	return 1 if ($share{DB}->is_admin);
    }
    elsif ($type eq "page") {
	$section_id = $data->{section_id};
    }
    elsif ($type =~ /content/) {
	my $content = $type eq "content" ? $data : $this->fetch("content",$data->{content_id});
	my $page_id = $content->{page_id};
	if ($page_id) {
	    my $page = $this->fetch("page",$page_id);
	    if ($page) { $section_id = $page->{section_id}; }
	}
    }
    my @keys;
    if ($section_id) {
	@keys = $this->fetch_match("member_site_link",
				   {member_id=>$db->my_uid(),
				    section_id=>$section_id});
    }

    foreach my $key (@keys) {
	$_ = $key->{priveleges};
	if ($type =~ /^section$/) {
	    return 1 if ($share{DB}->is_admin);
	}
	elsif ($type =~ /^page$/) {
	    return 1 if (/administrator/);
	    return 1 if (/designer/ && $data->{type} =~ /template|library/);
	    # editors cannot create pages
	}
	elsif ($type =~ /^content$/) {
	    return 1 if (/administrator|designer/);
	    return 1 if (/editor/ && $data->{type} eq "editorial");
	}
	elsif ($type =~ /^content_data$/) {
	    my $content = $this->fetch("content",$data->{content_id});
	    my $ctype = $content->{type};
	    return 1 if (/designer/ && $ctype eq "design");
	    return 1 if (/editor/ && $ctype eq "editorial");
	}
    }
    $this->error("Insert new $type: permission denied - your administrator key does not allow you to do this.");
    return 0;
}

sub allow_delete {
    my ($this,$type,$data) = @_;
    if (! $type) { $type = $this->whatami(1); }
    if (! $data) { $this->set_context(); $data = $this->{context}{$type}; }
    foreach my $key ($this->get_keys()) {
	$_ = $key->{priveleges};
	if ($type =~ /^section$/) {
	    return 1 if (/administrator/ && $share{DB}->is_admin);
	}
	elsif ($type =~ /^page$/) {
	    return 1 if (/administrator/);
	    return 1 if (/designer/ && $data->{type} =~ /template/);
	    # editors cannot remove pages
	}
	elsif ($type =~ /^content$/) {
	    return 1 if (/administrator|designer/);
	    return 1 if (/editor/ && $data->{type} eq "editorial");
	}
	elsif ($type =~ /^content_data$/) {
	    return 1 if (/administrator/);
	}
    }
    $this->error("Delete $type: permission denied - your administrator key does not allow you to do this.");
    return 0;
}

sub owns {
    my ($this) = @_;
    return ($this->get_keys() > 0);
}

### DEPECATED:  see ExSite::Auth for access control tools
sub access {
    my ($this,$method,$class,$id) = @_;
    my $db = $share{DB};
    my $uid = $db->my_uid;
    my @sites = $db->fetch_m2m("section","member",$uid,"member_site_link");
    $this->set_context;

    # crude access control for now - simply check for a section match

    foreach my $site (@sites) {
	if ($site->{section_id} == $this->id("section")) {
	    return 1;
	}
    }
    return 0;
}

#------------------------------------------------------------------------

=pod

=head1 Content Searches

When building a page, ExSite encounters various tags that reference
content objects such as images, HTML, etc.  ExSite must search for
matching objects to resolve these references, within the parameters of
the current context.  (For example, there could be many C<index.html>
pages in the system; which one is meant depends on where one begins
looking for it.)

=head2 find()

    $this->find($name,$mode);

Locates a named content object, starting from the current context
(which will be set automatically if not already set).  The object is
sought in the following locations:

    - in current page's content objects
    - in templates' content objects, including inherited templates
    - in content libraries for this section
    - in content libraries for ancestral sections

The first eligible match satisfies the algorithm, and searching will stop.

Returns the found object's attribute hash if C<$mode> is "data".  If
C<$mode> is "object" (default), C<find()> will return a new content object.

=head2 find_prefetch()

    $this->find_prefetch($name,$mode);

Performs the actual content search for C<find()> by looking through
the C<prefetch()> cache.  This is efficient, but it will perform a
more costly search directly in the database if the prefetch cache
appears not to contain anything useful.

=head2 find_db()

    $this->find_db($name,$mode);

Performs the actual content search for C<find()> by looking through
the database.  This may require many queries, and is substantially
slower than C<find_prefetch()>.  However, it is guaranteed to find the
content, whereas the prefetch cache could in principle be out of sync
with the database.

=head2 find_db_nonlocal()

    $this->find_db_nonlocal($name,$mode);

Content searches that exhaust the current section may have parent
sections that share content, or public templates/libraries that can
also be searched.  This method searches external sections for missing
content.

=head2 find_page()

    my $page = $this->find_page($name,$section_id);

Returns the attribute hash for a page with filename C<$name>.  The
C<$section_id> may optionally be specified, if looking outside the
current context.

=head2 find_content()

    my $content = $this->find_content($name,$page_id);

Returns the attribute hash for a content object named C<$name>.  The
C<$page_id> may optionally be specified, if looking outside the
current context.  This is used by the more powerful C<find()> method
to search individual pages, templates, and libraries for matching
content.

=head2 find_content_data()

    my $cdata = $this->find_content_data($type,$content_id);

Returns the attribute hash for the revision in the content object 
indicated by C<$content_id>.  C<$content_id> is optional, and only needed
in the case where the content object is different from that in the 
context.

=cut

#------------------------------------------------------------------------

sub find {
    my ($this,$cname,$mode) = @_;

    my $stat = $this->run_handler("cms_find",$cname,$mode);
    return $stat if (defined $stat);

    # this method attempts to find the content in the prefetch caches
    # before hitting the database
    my $c = $this->find_prefetch($cname,$mode);
    if (!$c) {
	# in case the browser has borked our spaces/escape codes, try again
	if ($cname =~ / /)      { $cname =~ s/ /%20/g; }
	elsif ($cname =~ /%20/) { $cname =~ s/%20/ /g; }
	$c = $this->find_prefetch($cname,$mode);
    }

    # this method (deprecated) goes straight to the database to find 
    # the content (much slower).
    #my $c = $this->find_db($cname);

    return $c if ($mode eq "data");
    require ExSite::Content;
    return new ExSite::Content(content=>$c,revision=>$this->get_revision());

    ### FIXME:  revision?!?!?!
}

# the find algorithm, using the prefetch caches

sub find_prefetch {
    my ($this,$cname,$mode) = @_;

    my ($c,$sid,$pid,$page,$version);

    if (! $this->{context_set}) {
	$this->set_context;
    }
    $sid = $this->id("section");
    if (! $this->is_prefetched($sid)) {
	return $this->find_db($cname,$mode);

	## OR, do we just preload that section too?
    }
    $pid = $this->id("page");
    $page = $this->get("page");
    return undef if (! $page);
    my $version = $this->get("page","version");
    
    # search the current page
    if (exists $share{prefetch}{$sid}{$pid}{$cname}) {
	return $share{prefetch}{$sid}{$pid}{$cname}{_};
    }

    # search the master page
    if ($page->{type} eq "alternate") {
	if (exists $share{prefetch}{$sid}{$page->{parent_id}}{$cname}) {
	    return $share{prefetch}{$sid}{$page->{parent_id}}{$cname}{_};
	}
    }

    # search the template stack
    my $sid_orig = $sid;
    my %visited;  # detect infinite loops
    while ($page->{template_id} && ! $visited{$page->{template_id}}) {
	$visited{$page->{template_id}} = 1;
	if (exists $share{prefetch}{$sid}{$page->{template_id}}) {
	    if (exists $share{prefetch}{$sid}{$page->{template_id}}{$cname}) {
		return $share{prefetch}{$sid}{$page->{template_id}}{$cname}{_};
	    }
	    $page = $share{prefetch}{$sid}{$page->{template_id}}{_};
	}
	else {
	    # the template is not part of the current site
	    my $template = $this->fetch("page",$page->{template_id}) || {};
	    if ($template->{section_id} != $sid) {
		$this->prefetch($template->{section_id});
		$sid = $template->{section_id};
		$visited{$page->{template_id}} = 0;  # allow a 2nd pass
	    }
	    else {
		# bad template reference :-(
		$page = {};
	    }
	}
    }

    # search the libraries, if this is not the body
    $sid = $sid_orig;
    my $section = $this->get("section");
    my $bestguess;
    if ($cname ne "body") { 
	%visited = ();
	while ($sid && ! $visited{$sid}) {
	    $bestguess = undef;
	    $visited{$sid} = 1;
	    foreach my $pid (keys %{$share{prefetch}{$sid}}) {
		next if ($pid eq "_");
		$page = $share{prefetch}{$sid}{$pid}{_};
		next if ($page->{type} ne "library");
		next if ($page->{section_id} != $sid_orig &&
			 $page->{access} eq "administrators"); # private lib
		next if (! exists $share{prefetch}{$sid}{$pid}{$cname});
		# we have a matching content object in this page
		if ($version eq $page->{version}) {
		    # perfect match
		    return $share{prefetch}{$sid}{$pid}{$cname}{_};
		}
		elsif (! $page->{version}) {
		    # imperfect match (default version)
		    $bestguess = $share{prefetch}{$sid}{$pid}{$cname}{_};
		}
	    }
	    return $bestguess if ($bestguess);

	    # goto previous section
	    $sid = $section->{parent_id};
	    if ($sid) {
		$section = $this->fetch("section",$sid);
		$this->prefetch($sid);
	    }
	}
    }

    # entire search path is exhausted - the content does not exist
    return undef;
}

# find_db:  the find algorithm, using direct DB lookups.
# Slower than using prefetch caches.
# This method searches the same locations that are covered by the
# prefetch caches.  Use find_db_nonlocal() to search outside the scope
# of the prefetch caches.

sub find_db {
    my ($this,$cname,$mode) = @_;

    if (! $this->{context_set}) {
    #if (! $this->{context}{page}) {
        # missing our context - can't just check context_set flag, because
        # set_context calls us before the context is fully set.
        $this->set_context;
    }

    my $page = $this->{context}{page};
    return undef if (! $page);

    # search the current page
    my $c = $this->find_content($cname);

    # search the master page
    if (! $c && $page->{type} eq "alternate") {
	$c = $this->find_content($cname,$page->{parent_id});
    }

    if (! $c) {
	$c = $this->find_db_nonlocal($cname,$mode);
    }
    return $c;
}

# find_db in locations that are (potentially) outside the current site

sub find_db_nonlocal {
    my ($this,$cname,$mode) = @_;
    my $c;
    my $page = $this->{context}{page};
    return undef if (! $page);

    # search the template stack
    while (! $c && $page->{template_id}) {
	$page = $this->fetch("page",$page->{template_id});
	$c = $this->find_content($cname,$page->{page_id});
    }

    # search the libraries, if this is not the body
    if (! $c && $cname ne "body") { 
	# get current section
	my $section = $this->{context}{section};
	######## set the parent context first, to be sure it is defined?
	return undef if (! $section);
	my $sid = $section->{section_id};
	while ($sid) {
	    # get all libraries in this section
	    my @libs = $this->fetch_match("page",{type=>"library",
						  section_id=>$sid});
	    if (@libs > 0) {
		# search each library for a matching content object
		my $lib = shift @libs;
		while ($lib && ! $c) {
		    $c = $this->find_content($cname,$lib->{page_id});
		    $lib = shift @libs;
		}
		last if ($c);
	    }
	    # goto previous section
	    $sid = $section->{parent_id};
	    if ($sid) {
		$section = $this->fetch("section",$sid);
	    }
	}
    }
    return $c;
}

# more find tools

sub find_page {
    my ($this,$name,$section) = @_;
    my $sid = ref $section eq "HASH" ? $section->{section_id} : $section;
    if (! $sid) {
	return undef if (! $this->has("section"));
	$sid = $this->id("section");
    }
    $this->prefetch($sid) if (! $this->is_prefetched($sid));
    if (exists $share{prefetch}{$sid}{$name}) {
	return $share{prefetch}{$sid}{$name}{_};
    }
    return undef;
}

sub find_content {
    my ($this,$name,$page) = @_;
    my $pid = ref $page eq "HASH" ? $page->{page_id} : $page;
    if (! $pid) {
	return undef if (! $this->has("page"));
	$pid = $this->id("page");
    }
    my $pdata = ref $page eq "HASH" ? $page : $this->fetch("page",$pid);
    my $sid = $pdata->{section_id};
    $this->prefetch($sid) if (! $this->is_prefetched($sid));
    if (exists $share{prefetch}{$sid}{$pid}{$name}) {
	return $share{prefetch}{$sid}{$pid}{$name}{_};
    }
    return undef;
}

sub find_content_data {
    my ($this,$revision,$content) = @_;
    my $cid = ref $content eq "HASH" ? $content->{content_id} : $content;
    if (! $cid) {
	return undef if (! $this->has("content"));
	$cid = $this->id("content");
    }
    my $cdata = ref $content eq "HASH" ? $content : $this->fetch("content",$cid);
    my $pid = $cdata->{page_id};
    my $pdata = $this->fetch("page",$pid);
    my $sid = $pdata->{section_id};
    $this->prefetch($sid) if (! $this->is_prefetched($sid));
    if (exists $share{prefetch}{$sid}{$pid}{$cid}) {
	my $rev = $revision || $this->get_revision();
	return $share{prefetch}{$sid}{$pid}{$cid}{$rev};
    }
    return undef;
}

#------------------------------------------------------------------------

=pod

=head1 Metadata

CMS objects can carry around various metadata.  These data are largely
descriptive, and have no functional purpose within the CMS.  However,
they can be included in HTML output, and may be significant for other
purposes such as search engine indexing.  Metadata are stored in a 
hash table, indexed under various keys.

To include metadata in HTML output, you can include a tag like:

    <!--$metadata_key-->

which will be replaced with the metadata value for that key.  For example, 
to insert the title into the HTML, use:

    <!--$title-->

=head2 fetch_metadata()

Resets/loads the default metadata for a CMS object.  The metadata for
any CMS object is initialized to the contents of the current page
record.  The contents of the section record are added to this.  All
section keys are prepended with "site_" so they will not collide with
the page keys.  For instance, use the key C<title> to get the page
title, or C<site_title> to get the section title.

=head2 set_metadata()

    $this->set_metadata($key,$value);

You can add to the metadata, or change default metadata values with
this method.  If you re-use a key that is already set by
C<fetch_metadata()>, then you replace the default value.  This is
typically done by plug-ins that want to adjust page titles and
descriptions to reflect specialized data objects that they are
displaying, rather than rely on the generic page title.  For example:

    $share{Page}->set_metadata("title","Foo Bar");

=cut

sub fetch_metadata {
    my $this = shift;

    my $meta = $this->run_handler("fetch_metadata");
    return $meta if (defined $meta);

    # initialize to page datahash
    $meta = $this->{context}{page};
    # add the section datahash
    foreach my $key (keys %{$this->{context}{section}}) {
	$meta->{"site_".$key} = $this->{context}{section}{$key};
    }

    # add any explicit metadata
    foreach my $key (keys %{$this->{metadata}}) {
	$meta->{$key} = $this->{metadata}{$key};
    }

    return wantarray ? %$meta : $meta;
}

sub set_metadata {
    my ($this,$key,$val) = @_;
    $this->{metadata}{$key} = $val;
}

#------------------------------------------------------------------------

=pod

=head1 Content Expansion

=head2 get_start_html()

    my $html = $this->get_start_html($mode);

Initalizes the HTML representation of an object.

Page content objects are typically built starting from an HTML 
framework that comes out of a template.  This template might be precompiled,
which means that it resides on disk as a partially-processed HTML file.  
If not precompiled, then it resides in the database as raw HTML.  This
method identifies and returns the appropriate block of initial HTML to
begin building the page from.

For non-page content objects, this simply returns an HTML representation
of the object.

=head2 expand()

    my $html = $this->expand(%options);

Expands the content in the current context, by replacing CMS tags with
the URLs or the HTML content that they represent.  Returns the
expanded for the content.

Options:

=over 4

=item html

Use this as the starting HTML, instead of using the context.

=item content

a hash of content names and values to use preferentially over 
those in the database

=item method

a combination of B<data, url, content, page, module,> depending on 
which expansions you want to apply to this content.  All are 
selected, by default.  You can also specify B<dummy-content> or
B<dummy-module> to substitute placeholder graphics instead of full 
content.

=item expand

if set to "template", we only expand the template/design content.
Remaining tags are left in place.  DCD tags are left unexpanded.

=item cms

if true, include CMS tools in the output, for interactive page
editing.

=item mode

"dynamic" (fetch everything from DB source) or "static" (default;
fetches published and precompiled versions where possible)

=back

=cut

#------------------------------------------------------------------------

sub get_start_html {
    my ($this,$mode) = @_;

    my $stat = $this->run_handler("get_start_html",$mode);
    return $stat if (defined $stat);

    # for "page" objects, check to see if the template is precompiled,
    # as that will save us a lot of content lookups later.

    if ($mode ne "dynamic" &&
	$this->{context}{page} && 
	$this->{context}{content}) {
	# If we have a full page loaded,
	if ($this->{context}{page}{type} eq "page") {
	    # and it really is a page (not a template),
	    if ($this->{context}{content}{name} eq "page") {
		# and the content object is the base HTML for the page,
		if ($config{page}{prepublish_dynamic} && 
		    $this->{context}{page}{publish_method} eq "dynamic" && 
		    $this->{revision} eq "active" && 
		    exists $share{Page} && 
		    ! $share{Page}->is_publishing()) {
		    # and this is a precompiled dynamic page,
		    # ...then we can use the precompiled page
		    my $pagefile = $share{Page}->diskpath();
		    if (-e $pagefile) {
			# should we cache this?
			return &get_file($pagefile);
		    }
		}
		elsif ($this->{context}{content}{page_id} != 
		    $this->{context}{page}{page_id}) {
		    # and this content object comes from an external template,
		    # ...then we can use a precompiled template, if available.
		    my $template_id = $this->get("page","template_id");
		    my $t = $this->fetch("page",$template_id);
		    if ($t->{publish_method} eq "static") {
			my $template = new ExSite::Page(page=>$t);
			my $path = $template->diskpath;
			# check for a precompiled template
			if (-e "$path/index.html") {
			    # should we cache this?
			    return &get_file("$path/index.html");
			}
		    }
		}
	    }
	}
    }
    # no precompiled template found;  convert the content_data object to HTML
    return $this->get_html($mode);
}

sub substitute {
    return &expand(@_);
}

sub expand { 
    my ($this,%option) = @_;
    return $this->{html} if ($this->{expanded} || $this->{noexpand});
    $this->info("expanding html");

    # monitor the content objects that we have expanded, to avoid
    # infinite loops
    $this->{expanded} = {};

    # fetch raw template data
    #$this->{html} = $this->get_html($option{mode});
    # fetch prepublished template data, if possible; otherwise fetch raw data
    $this->{html} = $option{html} || $this->get_start_html($option{mode});

    if (! exists $option{content}) { $option{content} = {}; }
    my $method = $option{method} || "data,url,content,page,module"; 
    my $cmsmenu = 0;

    do {

	# clear substitution counter
	$this->{htmlsub} = 0;

	# 1) content-URL substitutions, eg. SRC=[[content-object-name]]

	if ($method =~ /url/) {
	    # strict
	    #$this->{html} =~ s/\[\[([\w\.\ \%-]+)\]\]/$this->get_content_url($1,\%option)/ges;
	    # liberal
	    $this->{html} =~ s/\[\[(.+?)\]\]/$this->get_content_url($1,\%option)/ges;
	}

	# 2) page-URL substitutions, eg. HREF={{page-name/id}}
	# (we skip this on templates, since public templates can be used on 
	# other sites with different pages)

	if ($method =~ /page/ && ($option{expand} ne "template")) { 
	    $this->{html} =~ s/\{\{([\w\.-]*)\}\}/$this->get_page_url($1,\%option)/ges;
	}

	# 3) HTML content substitutions, eg. <!--content(content-obj-name)-->

	if ($method =~ /content/) { 
	    # repeat until all substitutions are done
	    while ($this->{html} =~ /<\!--\&?content\((.+?)\)-->/) {
		$this->{html} =~ s/<\!--\&?content\((.+?)\)-->/$this->get_content($1,\%option)/ges;
	    }
	}

	# 4) Dynamic Content Driver substitutions, eg. <!--&module.method(args)-->
	if ($method =~ /module/) {
	    # add a CMS menu DCD tag in advanced CMS mode
	    # NB: editmode selects a non-default way of handling inline CMS 
	    #     funcs; only add the menu if this is not set
	    if ($option{cms} && (! $config{page}{editmode}) && ! $cmsmenu ) {
		$this->{html} =~ s/(<body[^>]*>)/$1\n<!--\&CMS\(cmscmd=pagemenu\)-->/i;
		# ensure that cmsmenu is only inserted once
		$cmsmenu = 1;
	    }
	    # insert DCD content
	    do {
		$this->{expand_redo} = 0;
		my $html = $this->{html};
		# direct substitution of dynamic content (same server request)
		$html =~ s/<\!--\&(\w+)\((.*?)\)-->/$this->get_dynamic_content($1,$2,\%option)/ges;
		# indirect substitution of dynamic content (separate request)
		$html =~ s/<\!--\&\&(\&)?(\w+)\((.*?)\)-->/$this->get_dynamic_content_indirect($2,$3,\%option,$1)/ges;
		if (! $this->{expand_redo} && ! $this->{noexpand}) {
		    $this->{html} = $html;
		}
	    } until (! $this->{expand_redo} || $this->{noexpand});

	    # Normally there is only one pass through above do...until loop.
	    # However, some DCDs may set flags to change the logic:
	    #    expand_redo - start over
            #    noexpand - abort

	}
    } until ($this->{htmlsub} == 0);

    # build a hash of meta data
    my $meta = $this->fetch_metadata();
    # perform meta data substitutions, eg. <!--$name-->
    # (we skip this on templates, since metadata is page-specific)
    # (we also do this last, in case plug-ins want to change our metadata)
    if ($method =~ /data/ && ($option{expand} ne "template") ) { 
	$this->{html} =~ s/<!--\$(\w+)-->/$meta->{$1}/gs;
    }

    if ($option{expand}) {
	# partial expansion was requested, which means we should now have
	# one or more masked CMS tags in our HTML.  Now we have to unmask them.
	$this->{html} = &restore_cms_tags($this->{html});
    }

    # flag expansion as done; also frees the hash we used to keep track
    # of expansions
    $this->{expanded} = 1;  
    return $this->{html};
}

=pod

=head2 get_page_url()

    my $url = $this->get_page_url($id);

Return replacement text for a C<{{$id}}> CMS tag.  C<$id> refers to a
page--tags such as C<{{123}}> or C<{{index.html}}> will be replaced
with the correct URL to the indicated page (referenced by a numeric
page ID, or the page file name).

=head2 get_content_url()

    my $url = $this->get_content_url($name,$option);

Return replacement text for a C<[[$name]]> CMS tag.  C<$name> is the
name of a content object--tags such as C<[[image.gif]]> will be
replaced with the correct URL to the indicated content object.
C<$option> is a reference to an option hash that accepts the following
parameters:

=over 4

=item mode

C<static|dynamic> - force the URL to fetch the published or database revisions
of the content, respectively.  This is optional;  the system defaults to
static, if it is available, dynamic otherwise.

=item expand

C<template> - expand only the template elements; leave page-specific 
content unexpanded.

=back

=head2 get_content()

    my $html = $this->get_content($name,$option);

Return replacement HTML for a C<< <!--content(...)--> >> CMS tag

This is similar to C<get_content_url()>, except the full HTML is substituted
instead of just the URL.  Accepts the same options as C<expand()>, and 
acts on the following:

=over 4

=item content

This can point to a hash of predefined content definitions, which you can
use to override the content database.

=item method

Same as the method item passed to C<expand()>.  The "B<dummy-content>" 
method is the only one used in this method.

=item expand

"B<template>" if we are precompiling a template, and wish to leave
page-specific data unexpanded.

=item cms

Inserts interactive page editing tools, if true.

=back

=head2 placeholder_image()

    my $imgtag = $this->placeholder_image($content_id);

This generates an HTML image tag that can be used as a placeholder
image for unexpanded content objects.  This is used by the HTML editor
to represent external content objects that you cannot edit.  However,
the content object as a whole can be deleted, by removing the
placeholder image.

=cut

sub get_page_url {
    my ($this,$id) = @_;
    my ($p,$Page,$query,$url,$static);
    if ($id =~ /\?/) {
	($id,$query) = split /\?/, $id;
    }

    # get the base URL for the page

    if (! $id) {
	# no id means a self-link
	$Page = $share{Page};
	$static = defined $query;
    }
    elsif ($id eq "!") {
	# ! means self-link to dynamic version
	$Page = $share{Page};
	$static = 0;
    }
    else {
	$p = $this->fetch("page",$id);
	if ($p) {
	    $Page = new ExSite::Page(page=>$p,noload=>1);
	}
	elsif ($id =~ /^\d+$/) {
	    $Page = new ExSite::Page(id=>$id,noload=>1);
	}
	else {
	    $Page = new ExSite::Page(filename=>$id,
				     section_id=>$this->id("section"));
	}
	$static = 1;
    }

    # change version automatically
    my $new_version = $Page->get_my("version");
    my $old_version = $share{Page} ? $share{Page}->get_my("version") : undef;
    if ($old_version && $new_version ne $old_version) {
	$Page->set_version($old_version);
    }

    # modify the base URL with extra query params

    if ($query) {
	my $uri = $Page->get_uri_dynamic();
	my %qdata = &DecodeString($query);
	$uri->parameters(%qdata);
	$url = $uri->write();
    }
    else {
	$url = $static ? $Page->get_url() : $Page->get_url_dynamic();
    }
    return $url;
}

# get_content_url : return replacement text for a [[...]] CMS tag

sub get_content_url {
    my ($this,$id,$option) = @_;
    # content reference
    my ($subobj, $vfile);
    $subobj = $this->find($id);
    $subobj->set_context("child");  # load revisions
    if ((! $subobj || ! $subobj->defined) && $id =~/\%/) {
	# no matching content - try unescaping any escape chars 
	# embedded in the id
	$id =~  s/%(..)/pack("c",hex($1))/ge;
	$subobj = $this->find($id);
	$subobj->set_context("child");  # load revisions
    }
    if (! $subobj || ! $subobj->defined) {
	# no matching content - escape the tag chars in case this
	# is not a CMS tag
	$vfile = &html_escape("[[$id]]","[]");
    }
    elsif ($option->{expand} eq "template" && 
	   ($subobj->get("content","type") eq "editorial")) {
	# defer substitution of local page content by hiding the tags 
	# temporarily.  We must undo this in &ExSite::Util::restore_cms_tags
	# which should be called automatically by expand().
	$vfile = "[|[$id]|]";
    }
    else {
	$vfile = $subobj->get_url($option->{mode});
    }
    return $vfile;
}

# get_content : return replacement text for a <!--&content(...)--> CMS tag

sub get_content {
    my ($this,$id,$option) = @_;

    $this->{htmlsub}++;
    my ($start,$end,$cobj,$cobjname,$csub,$newtext,@cobjname);
    @cobjname = split(/,/,$id);
    # first content object in list is substituted directly
    $cobjname = shift @cobjname;
    $this->{expanded}{$cobjname} = 1;

    if ($option->{method} =~ /dummy-content/) {
	# we're not really substituting content, 
	# so use a placeholder graphic instead
	$newtext = $this->placeholder_image($id);
    }
    elsif ($option->{content}->{$cobjname}) {
	# predefined content object
	$newtext = $option->{content}->{$cobjname};
    }
    else {
	# look for matching content object in DB
	$cobj = $this->find($cobjname);
	#$cobj->set_context();  # load full context
	if ($option->{expand} eq "template" && 
	    $cobj->get("content_data","page_id") == $cobj->get("page","page_id") ) {
	    # defer substitution of local page content
	    # (we are using the placeholder image to "escape" the CMS tag)
	    $newtext = $this->placeholder_image($id);
	}
	elsif ($option->{expand} eq "design" && 
	    $cobj->get("content","type") eq "editorial") {
	    # defer substitution of editorial content 
	    $newtext = $this->placeholder_image($id);
	}
	else {
	    # substitute now
	    my $in = new ExSite::Input;
	    my $q = $in->query;

	    # FIXME:
	    # first two clauses that follow are experiments in inline
	    # content editing for a simplified CMS interface.  This should
	    # be cleaned up when it becomes clear which method is preferred.

	    if ($id eq "body" && $config{page}{editmode} eq "easy" && 
		$option->{cms} && $q->{cmscmd} =~ /update/) {
		# inline edit - do not output content
		$newtext = "<!-- body -->\n";
	    }
	    elsif ($id eq "body" && $config{page}{editmode} eq "MySite" && 
		$option->{cms} && $q->{ecmd} eq "update") {
		# inline edit - do not output content
		$newtext = "<!-- body -->\n";
	    }
	    else {
		$newtext = $cobj->get_html;
	    }
	    if ($option->{cms} && 
		$cobj->{context}{content}{page_id} == 
		$this->{context}{page}{page_id}) {
		###### should allow hotlinks to undefined content
		###### objects that are being inherited from template
		# can only manage content tied directly to this page
		# to manage templates, view the template in cms mode
		
		# inline editing links
		if ($share{DB}->owns() && 
		    (($config{page}{editmode} ne "MySite") || $id eq "body")) {
		    $newtext .= $this->edit_content_link($cobj);
		    # highlight editable content sections
		    my $ml = new ExSite::ML;
		    if ($option->{cms} > 1 && $config{page}{editmode} eq "MySite") {
			# for the noobs
			$newtext .= 
			    $ml->br().
			    $ml->a("Return to website control panel",{href=>"javascript:window.close()"});
		    }
		    $newtext = $ml->div($newtext,{style=>"border-width:1px; border-color:red; border-style:dashed; padding:0px;"});
		}
	    }
	}
    }
    
    # remaining content objects in list are content objects to 
    # insert into the first content object
    
    my $icobj = 1;
    foreach $cobjname (@cobjname) {
	if ($cobjname =~ /^\".+\"$/) {
	    # quoted string - plaintext substitution
	    # simple strings only - (no commas!)
	    $cobjname =~ s/^\"//;
	    $cobjname =~ s/\"$//;
	    $newtext =~ s/<!--\&?content\($icobj\)-->/$cobjname/g;
	}
	else {
	    # unquoted string - treat as a content object reference
	    $newtext =~ s/<!--\&?content\($icobj\)-->/<!--content($cobjname)-->/g;
	}
	$icobj++;
    }
    return $newtext;
}

sub placeholder_image {
    my ($this,$id) = @_;
    ### FIXME: use html_escape();
    $id =~ s/\"/%22/g;
    $id =~ s/</%3C/g;
    $id =~ s/>/%3E/g;
    my $ml = new ExSite::ML;
    return $ml->img(undef,{alt=>"content($id)", 
			   src=>"$config{server}{HTMLpath}/_ExSite/images/checker.png",
			   title=>$id,
			   name=>$id,
			   height=>50,
			   width=>100,});
}

=pod

=head2 get_dynamic_content()

    my $html = $this->get_dynamic_content($module,$args,$option);

Return replacement text for a C<< <!--&Module(...)--> >> CMS tag

Loads the C<$module> plug-in, and calls its C<write()> method, passing
the C<$args> to it.  Returns the output of the module.  If the module 
fails to load/compile, a Perl error string will be returned.

C<$option> is the same as is passed to C<expand()>.  If the "B<expand>" option
or "B<dummy-module>" methods are selected, a placeholder image will be
substutited instead of the module content.

If the "B<static-module>" method is selected, each module will be
queried to see if its output is static for this page; only if static
will the content will be substituted.  (This allows for precompiling 
certain modules' output, eg. menus, while rendering others at page 
view time.)

=head2 get_dynamic_content_indirect()

    my $html = $this->C<get_dynamic_content_indirect($module,$args,$option);

This is an AJAX version of C<get_dynamic_content()>.

This method returns replacement text for CMS tags of the form:

    <!--&&Module(...)-->    # indirect substitution, direct re-links
    <!--&&&Module(...)-->   # indirect substitution, indirect re-links

This fetches the DCD content using a separate server request, instead 
of inlining it directly.

The main advantage is that you can publish the main page to a static 
HTML file, yet keep some page elements dynamic.  This is especially 
useful for index.html pages, which must be static, but may contain 
dynamic elements (eg. recent news, upcoming events, current specials,
etc.).  The solution is either continuous republishing of the index
page (which may still be a better solution for heavily loaded sites)
or using an indirect dynamic content fetch.

The disadvantage is that the full page is slower (although the base
page may be much faster), and that JavaScript must be enabled to 
perform the secondary content fetches.

The '&&' variant does direct re-links, ie. links from the dynamic content
point to full URLs that generate a new page.  The '&&&' variant does 
indirect re-links, ie. links back to the same module in the same page
only fetch the DCD content and inline it dynamically into the current
page without generating a whole new page.

=cut

sub get_dynamic_content {
    my ($this,$module,$modargs,$option) = @_;

    my ($endname,$modname,$modmethod,$newtext);
    # save dcd objects for re-use
    if (! $this->{dcd}) { $this->{dcd} = {}; }

    # flag a substitution as having occured, so we can expand these results
    $this->{htmlsub}++;

    ($modname,$modmethod) = split(/\./,$module,2);
    if (!$modmethod) { $modmethod = "write"; }

    my $ml = new ExSite::ML;
    
    # run the DCD
    
    if ($option->{method} =~ /dummy-module/ || $option->{expand}) {
	my $icon = "$config{server}{HTMLpath}/_Modules/$modname/icon.gif";
	if (! -e "$config{server}{HTMLroot}$icon") {
	    $icon =~ s/\.gif$/\.png/;
	}
	if (! -e "$config{server}{HTMLroot}$icon") {
	    # no icon; use default ExSite app icon
	    $icon = "$config{server}{HTMLpath}/_Modules/icon.gif";
	}
	$newtext = $ml->img(undef,{alt=>"$module($modargs)",
				   src=>$icon,
				   title=>"$module($modargs)",
				   name=>"$module($modargs)",
				   border=>3,
				   style=>"border-style:outset;"});
    }
    elsif ($modmethod =~ /^(read|write|ioctl)$/) {
	# find the module
	my $dcd = $this->{dcd}{$modname} || $this->get_dcd($modname);
	if ($dcd) {
	    $this->{dcd}{$modname} = $dcd;
	    if ($option->{method} =~ /static-module/) {
		# only expand modules whose output is constant on a given page
		if ($dcd->ioctl('isStatic')) {
		    # execute the DCD method
		    $newtext = eval {
			# try to trap fatal errors
			$this->{dcd}{$modname}->$modmethod($modargs);
		    };
		    if ($@) {
			$newtext = $this->error($@);
		    }
		}
		else {
		    # pass the tag through - mask the tag by adding '*'; 
		    # we'll unmask it after expansion is complete
		    $newtext = "<!--*&${module}(${modargs})-->";
		}
	    }
	    else {
		# execute the DCD method
		$newtext = eval {
		    # try to trap fatal errors
		    $this->{dcd}{$modname}->$modmethod($modargs);
		};
		if ($@) {
		    $newtext = $this->error($@);
		}
	    }
	    # disable substitution for read and ioctl methods
	    if ($modmethod ne "write") {
		$newtext = $ml->Comment("$modname.$modmethod executed");
	    }
	}
	else {
	    $newtext = $ml->Comment("$modname failed");
	    $newtext .= $this->show_diagnostics("error","html");
	}
    }
    else {
	$newtext = $ml->Comment("Error: invalid dynamic content driver method: $modmethod");
	$this->error("invalid DCD method: $modmethod");
    }
    return $newtext;
}

# AJAX version
#
# FIXME: in principle, we don't have to display a warning message in
# <noscript>...</noscript>.  We could actually display the the real
# content, as in a direct substitution.  Unfortunately, that would involve
# running the DCD twice, once for the <noscript> direct sub, and again for
# the JS indirect sub.  If we cached this data (in the %session?) we could
# spare ourselves that extra hit, and give ourselves some robustness to
# non-JS clients (eg. GoogleBot).
#
# Actually, there's no need to put it in the <noscript> section;  you could
# put it in the regular DIV, in which case, all clients see it and use it.
# Javascript-enabled clients then overwrite it with the most current 
# dynamic version, if there is one.  The double-hit is a reasonable price
# to pay if we are publishing, since the 1st hit only occurs once for the
# lifetime of the file.

sub get_dynamic_content_indirect {
    my ($this,$module,$modargs,$option,$indlink) = @_;
    my $newtext;
    my $ml = new ExSite::ML;
    if ($option->{method} =~ /dummy-module/ || $option->{expand}) {
	my $icon = "$config{server}{HTMLpath}/_Modules/$module/icon.gif";
	if (! -e "$config{server}{HTMLroot}$icon") {
	    $icon =~ s/\.gif$/\.png/;
	}
	if (! -e "$config{server}{HTMLroot}$icon") {
	    # no icon; use default ExSite app icon
	    $icon = "$config{server}{HTMLpath}/_Modules/icon.gif";
	}
	$newtext = $ml->img(undef,{alt=>"&${indlink}$module($modargs)",
				   src=>$icon,
				   title=>"&${indlink}$module($modargs)",
				   name=>"&${indlink}$module($modargs)",
				   border=>3,
				   style=>"border-style:outset;"});
    }
    else {
	if ((ref $this) !~ /Page/) {
	    # we're probably in some special mode such as a preview or
	    # search index routine, in which case AJAX is not appropriate
	    return $ml->Comment("AJAX call to $module ignored");
	}
	my $ref;
	if ($this->is_publishing) {
	    # referrer isn't really publish.cgi; should probably be page.cgi
	    my $pid = $this->id("page");
	    # note escaped ?
	    $ref = "$config{server}{CGIpath}/$config{prog}{page}%3f_id=$pid";
	}
	else {
	    $ref = &url_escape($ENV{REQUEST_URI});
	}
	my $il = $indlink ? "_ajax=$module&" : ""; 
	$newtext = $ml->script(undef,{type=>"text/javascript",
				      src=>"$config{server}{HTMLpath}/_ExSite/js/httprequest.js"});
	$newtext .= $ml->div(
			     $ml->script("<!--//
document.write(getRequestData('$config{server}{CGIpath}/$config{prog}{dcd}/$module?_opt=$modargs&${il}_ref=$ref'));
//-->", { type=>"text/javascript" }).
			     $ml->noscript("(You must have javascript enabled to view $module.)"),
			     { id=>$module }
			     );
    }
    return $newtext;
}

=pod

=head1 CMS Data Fetching Methods

These are wrappers for the DB data fetching methods.  The primary
difference from the generic DB data fetching methods is that we use the
prefetch caches as much as possible to try and avoid hitting the
database for records we have already fetched.

The following methods take the same parameters as the equivalent
methods in C<ExSite::DB>.

=head2 fetch()

Attempts to locate page, content, and content_data records in the prefetch
caches before resorting to a database query.

=head2 fetch_child()

Attempts to locate content, and content_data records in the prefetch
caches before resorting to a database query.

=head2 fetch_match()

At present, this is a simple wrapper to a normal database query.  
(Ie. the prefetch caches are not used.)

=cut


# we provide alternatives to the generic DB fetching routines here, so that
# we can take advantage of some CMS caches and optimizations.

sub fetch {
    my ($this,$table,$id,@other) = @_;
    return undef if (! $id);
    my $data;
    if ($this->is_prefetched()) {
	if ($table eq "section") {
	    if (exists $share{prefetch}{$id}) {
		$data = exists $share{prefetch}{$id}{_} ? 
		    $share{prefetch}{$id}{_} : undef;
	    }
	}
	elsif ($table =~ /^(page|content)$/) {
	    # which sections should we search?
	    my @section;
	    if ($id =~ /^\d+$/) {
		# we've specified a specfic item, okay to search all
		@section = keys %{$share{prefetch}};
	    }
	    elsif ($table eq "page" && $this->id("section")) {
		# search for pages only in the current section
		@section = ( $this->id("section") );
	    }
	    elsif ($table eq "page" && $this->id("page")) {
		# search for pages only in the current section
		@section = ( $this->get("page","section_id") );
	    }
	    else {
		# not sure; search all sections
		@section = keys %{$share{prefetch}};
	    }

	  SECTION:
	    foreach my $sid (@section) {
		if ($table eq "page" && $this->is_prefetched($sid)) {
		    if (exists $share{prefetch}{$sid}{$id}) {
			$data = exists $share{prefetch}{$sid}{$id}{_} ? 
			    $share{prefetch}{$sid}{$id}{_} : undef;
			last SECTION if ($data);
		    }
		}
		else {  # content

		  PAGE:
		    foreach my $pid (keys %{$share{prefetch}{$sid}}) {
			next PAGE if ($pid eq "_");
			if ($table eq "content" && exists $share{prefetch}{$sid}{$pid}{$id}) {
			    $data = exists $share{prefetch}{$sid}{$pid}{$id}{_} ? 
				$share{prefetch}{$sid}{$pid}{$id}{_} : undef;
			    last SECTION if ($data);
			}
		    }
		}
	    }
	}
    }
 
    if ($data) {
	return wantarray ? ($data) : $data;
    }
    return $share{DB}->fetch($table,$id,@other);
}

sub fetch_child {
    my ($this,$table,$parent_table,$id,$sort,$limit) = @_;
    my @data;
    if ($this->is_prefetched()) {
	if ($table eq "page" && $parent_table eq "section") {
	    $this->prefetch($id) if (!$this->is_prefetched($id));
	    foreach my $key (keys %{$share{prefetch}{$id}}) {
		next if ($key eq "_");
		next if ($key !~ /^\d+$/);
		push @data, $share{prefetch}{$id}{$key};
	    }
	}
	if ($table =~ /^content$/ && 
	    $parent_table =~ /^(page|library|template)$/) {
	    my $parent = $this->fetch("page",$id);
	    my $sid = $parent->{section_id};
	    $this->prefetch($sid) if (!$this->is_prefetched($sid));
	    foreach my $key (keys %{$share{prefetch}{$sid}{$id}}) {
		next if ($key eq "_");
		next if ($key !~ /^\d+$/);
		push @data, $share{prefetch}{$sid}{$id}{$key}{_};
	    }
	}
    }
    if (@data) {
	if ($sort) {
	    my @s = ref $sort eq "ARRAY" ? @$sort : ($sort);
	    @data = $share{DB}->sort_rows($table,\@data,@s);
	}
	if ($limit && @data > $limit) {
	    @data = @data[0..($limit-1)];
	}
	return wantarray ? @data : \@data;
    }
    return $share{DB}->fetch_child($table,$parent_table,$id,$sort,$limit);
}

sub fetch_match {
    my ($this,$table,$match,$sort) = @_;
    # no prefetch optimizations... yet.
    return $share{DB}->fetch_match($table,$match,$sort);
}

=pod

=head1 Prefetch Cache

Content prefetching can greatly reduce the number of database queries
that get executed in the course of building a page.  They do this by
prefetching content related to a page or site, and then trying to
resolve content requests by using these prefetch caches rather than
database queries.

Prefetching must be done by the caller to take full advantage of this.
It is done by page.cgi, so normal page views will benefit from this
performance enhancement.

The structure of the prefetch cache is as follows:

    $share{prefetch}{SECTION}{PAGE}{CONTENT}{REVISION}

"SECTION" is a section ID.  "PAGE" is a page ID or filename.
"CONTENT" is a content ID or a content object name.  "REVISION" is a
content data ID, or a revision name.  That means any object is
addressible by ID or name, and the following styles of lookup (or a
mix of both) work:

    $share{prefetch}{6}{89}{121}{887}
    $share{prefetch}{6}{"index.html"}{"body"}{"active"}

These examples fetch a revision record from the prefetch cache.  To
fetch a section, page, or content record, use the key "_", eg.

    my $pagedata = $share{prefetch}{6}{"index.html"}{_};

=head2 is_prefetched()

    $this->is_prefetched($section_id);

Returns TRUE if we have loaded prefetch data for the given section.
If no section ID is given, returns TRUE if we have loaded prefetch
data for any section.

=head2 prefetch()

    $this->prefetch($section,$allrevs);

Preloads all useable content records (pages, content objects, and
revisions) for the given section, and stores them in a data structure
under C<$share{prefetch}>.  C<$section> can be a section ID or a
section datahash.  If not provided, we use the section from the
current context.  Note that inactive pages are not fetched.

By default we only store only the newest and active revisions of
content data in the prefetch cache.  All other revisions are
discarded.  You can override this by setting $allrevs to a TRUE value.

=head2 prefetch_update()

    $this->prefetch_update($type,$data);

Updates the prefetch cache with the datahash C<$data> which represents
a C<$type> (which can be one of "section", "page", "content", or
"content_data".  This ensures that page builds or previews following a
content update will render using the new content, rather than the
information that was in the prefetch cache before the update occurred.

=cut

sub is_prefetched {
    my ($this,$section_id) = @_;
    if ($section_id) {
	if (exists $share{prefetch}{$section_id}) {
	    return (scalar keys %{$share{prefetch}{$section_id}} > 0);
	}
	else {
	    return 0;
	}
    }
    else {
	return (scalar keys %{$share{prefetch}} > 0);
    }
}

sub prefetch {
    my ($this,$section,$allrevs) = @_;
    my ($sid,$cid,$pid,$dat,
	@pages,@content,@cdata,
	@cid,%pid,$published,$query);

    if (! $section) {
	$this->set_context("parent");
	$section = $this->get("section");
	$sid = $this->id("section");
    }
    elsif (ref $section eq "HASH") {
	$sid = $section->{section_id};
    }
    else {
	$sid = $section;
	$section = $share{DB}->fetch("section",$sid);
    }

    # sections

    return 0 if (! $sid);
    return 1 if (ref $share{prefetch}{$sid} eq "HASH");
    $share{prefetch}{$sid} = {};
    $share{prefetch}{$sid}{_} = $section;

    # pages 

    @pages = $this->fetch_match("page",{section_id=>$sid,
					status=>"active"});
    foreach my $p (@pages) {
	# store it under the page filename and page id
	$share{prefetch}{$sid}{$p->{page_id}} = {};
	$share{prefetch}{$sid}{$p->{page_id}}{_} = $p;
	# also index under filename, if a normal page
	if ($p->{type} =~ /page|alternate/) {
	    $share{prefetch}{$sid}{$p->{filename}} = 
		$share{prefetch}{$sid}{$p->{page_id}};
	}
    }

    # content

    $query = $config{mysql_timestamp_hack} ?
	"prefetch content with mysql_timestamp_hack" :
	"prefetch content";
    @content = $share{DB}->get_query($query,$sid);

    foreach my $c (@content) {
	$pid = $c->{page_id};
	$cid = $c->{content_id};
	$share{prefetch}{$sid}{$pid}{$cid} = {};
	$share{prefetch}{$sid}{$pid}{$cid}{_} = $c;
	$share{prefetch}{$sid}{$pid}{$c->{name}} = 
	    $share{prefetch}{$sid}{$pid}{$cid};
	push @cid,$cid;
	$pid{$cid} = $pid;
    }

    # revisions

    $query = $config{mysql_timestamp_hack} ?
	"prefetch content_data with _fsize and mysql_timestamp_hack" :
	"prefetch content_data with _fsize";
    @cdata = $share{DB}->get_query($query,$sid);

# legacy code for use with queries that fetch entire file, not _fsize;
# these queries are deprecated because the file data can potentially inflate
# the query results beyond MySQL's query size limit.
#
#    if (! exists $cdata[0]{_fsize}) {
#	# trim the data down to size
#	foreach my $rev (@cdata) {
#	    if ($rev->{fdata}) {
#		# drop the file contents from the data, to keep our cache
#		# size reasonably low
#		my ($file,$filedata) = split(/\#/,$rev->{fdata},2);
#		delete $rev->{fdata};
#		$rev->{fdata} = "$file#";
#		$rev->{_fsize} = length $filedata;
#	    }
#	    if ($rev->{thumb}) {
#		# again, for the thumb
#		my ($file,$filedata) = split(/\#/,$rev->{thumb},2);
#		delete $rev->{thumb};
#		$rev->{thumb} = "$file#";
#		$rev->{_tsize} = length $filedata;
#	    }
#	}
#    }

    # loop backwards over the data, taking the first useable version for each 
    # content_id.  If the first version is unpublished, then we also take the
    # first published version as well.

    while (@cdata > 0) {
	$dat = pop @cdata;
	$cid = $dat->{content_id};
	$pid = $pid{$cid};
	$published = ( $dat->{ptime} !~ /^0+/ );
	if ($allrevs) {
	    # we remember every revision under its rev id
	    $share{prefetch}{$sid}{$pid}{$cid}{$dat->{content_data_id}} = $dat;
	}
	if (! exists $share{prefetch}{$sid}{$pid}{$cid}{newest}) {
	    $share{prefetch}{$sid}{$pid}{$cid}{newest} = $dat;
	    # also set active version, in case no published version is found.
	    $share{prefetch}{$sid}{$pid}{$cid}{active} = $dat;
	}
	elsif ($published) {
	    if ($share{prefetch}{$sid}{$pid}{$cid}{active}{ptime} =~ /^0+/) {
		# our active version is just a fallback in case no published
		# version was found;  replace it with this published version.
		$share{prefetch}{$sid}{$pid}{$cid}{active} = $dat;
	    }
	}
    }

    return 1;
}

# if new data is inserted to the DB, we can update the prefetch
# cache so the data can be found on the same request.

sub prefetch_update {
    my ($this,$what,$data) = @_;
    my ($sid,$pid,$p,$cid,$c,$cdid);
    if ($what =~ /page/) { 
	$sid = $data->{section_id};
	return if (! exists $share{prefetch}{$sid});
	# store it under the page filename and page id
	$share{prefetch}{$sid}{$data->{page_id}} = {};
	$share{prefetch}{$sid}{$data->{page_id}}{_} = $data;
	# also index under filename, if a normal page
	if ($data->{type} =~ /page|alternate/) {
	    $share{prefetch}{$sid}{$data->{filename}} = 
		$share{prefetch}{$sid}{$data->{page_id}};
	}
    }
    elsif ($what =~ /^content$/) { 
	$pid = $data->{page_id};
	$p = $this->fetch("page",$pid);
	$sid = $p->{section_id};
	return if (! exists $share{prefetch}{$sid});
	$cid = $data->{content_id};
	$share{prefetch}{$sid}{$pid}{$cid} = {};
	$share{prefetch}{$sid}{$pid}{$cid}{_} = $data;
	$share{prefetch}{$sid}{$pid}{$data->{name}} = 
	    $share{prefetch}{$sid}{$pid}{$cid};
    }
    elsif ($what =~ /^content_data$/) { 
	$cid = $data->{content_id};
	$c = $this->fetch("content",$cid);
	$pid = $c->{page_id};
	$p = $this->fetch("page",$pid);
	$sid = $p->{section_id};
	return if (! exists $share{prefetch}{$sid});
	$cdid = $data->{content_data_id};
	$share{prefetch}{$sid}{$pid}{$cid}{newest} = $data;
	if ($data->{ptime} !~ /^0+$/) {
	    $share{prefetch}{$sid}{$pid}{$cid}{active} = $data;
	}
    }
}

### The following DCD-fetching methods are deprecated.
### Use ExSite::Module instead.

# get_all_dcd : return an array of installed DCD names 

sub get_all_dcd {
    my $this = shift;
    my @dcd;
    opendir MODDIR, "$config{server}{CGIroot}/Modules";
    @dcd = readdir MODDIR;
    closedir MODDIR;
    @dcd = grep /\.pm$/, @dcd; # only accept Perl modules
    foreach (@dcd) { s/\.pm$//; }
    return sort @dcd;
}

# get_my_dcd : return an array of DCD names available to the current site

sub get_my_dcd {
    my $this = shift;
    my $section = shift;
    if (!$section) {
	$section = $this->id("section") || 
	    $share{DB}->this_site()->{section_id};
    }
    my @dcd = $this->get_all_dcd;
    my @dcd_allow;
    foreach my $dcd (@dcd) {
	my ($mod,undef) = split /\./,$dcd;
	if (eval "require Modules::$mod") {
	    my $s = eval "new Modules::$mod";
	    if ($s) {
		if ($s->ioctl("isRestricted")) {
		    # this service requires licenses
		    my $match = $section ? 
		       { section_id => $section, name => $mod } : 
		       { name => $mod };
		    my @licenses = $share{DB}->fetch_match("service",$match);
		    if (@licenses > 0) {
			push @dcd_allow, $dcd;
		    }
		}
		else {
		    push @dcd_allow, $dcd;
		}
	    }
	    else {
		$this->error($@);
	    }
	}
    }
    return @dcd_allow;
}

# get_dcd : return a running DCD object

sub get_dcd {
    my ($this,$modname) = @_;
    my $dcd;
    my $stat = eval "require Modules::$modname";
    if ($stat) {
	$dcd = eval "new Modules::$modname";
	if ($dcd) { 
	    # check licensing
	    if ($dcd->ioctl("isRestricted")) {
		# this service requires licenses
		$this->set_context;
		my @licenses = $share{DB}->fetch_match("service",
						       {section_id => $this->id("section"),
							name => $modname }
						       );
		if (@licenses == 0) {
		    $this->error("$modname: permission denied (no license)");
		    $dcd = undef;
		}
	    }
	}
	else {
	    # $@ is $EVAL_ERROR
	    $this->error("$modname: $@");
	    $dcd = undef;
	}
    }
    else {
	# $@ is $EVAL_ERROR
	$this->error("$modname: $@");
	$dcd = undef;
    }
    return $dcd;
}

=pod

=head1 Publishing

=head2 safe_path()

    my $safepath = $this->safe_path($unsafepath,$prefix);

Makes a path safe for publishing to.  If the path has already been
untainted, then it is returned as-is.  If not, then it must conform to
the following rules:

    - no ".." elements allowed
    - all directories must begin with a word character (\w)
    - the path must begin with the system HTML area directory
      (you can override this using $prefix)

If the path does not conform, then nothing is returned.  Otherwise, 
the untainted path is returned.

=cut


#------------------------------------------------------------------------
# publishing utilities
#
# safe_path - test if a diskpath can be safely written to
#           - no ".." elements allowed
#           - all directories must begin with a word char (\w)
#           - $safe_prefix lets you specify alternate safe areas to allow
#           - returns the safe path, or nothing if not safe

sub safe_path {
    my ($this,$path,$safe_prefix) = @_;
    if (&is_tainted($path)) {
	# path is unsafe, so let's try to clean it
	$path =~ /^($safe_prefix(\/\w[\w\-\.]+)+)$/;
	return $1;
    }
    else {
	return $path;
    }
}

#------------------------------------------------------------------------
#
# placeholder methods : these are dummies that will be overloaded in
# higher-level CMS classes.  They are provided here to allow ContentBase
# to run stand-alone for testing.

sub get_raw {
    return;
}

sub get_html {
    return;
}

sub get_url {
    return;
}

sub setup {
    return;
}

sub edit_content_link {
    return;
}
    

1;
