#!/usr/bin/perl
#-----------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

=pod

=head1 ContentData.pm : manage content revisions.

A ContentData object is a particular revision of some piece of web content.
It contains the actual, viewable web content (eg. HTML, image, other).

The terms "ContentData", "content_data", and "revision" are often used
synonymously.

ContentData is part of the ExSite content stack, so it inherits from
the ContentBase class, and is in turn inherited by the Content class.
Each Content object contains a history list of content_data (revision)
records.  Each update to the content creates a new revision that is
appended to this list.

Any number of revisions may be retained in the history, but there are
two important revisions in this list:

=over 4

=item current (or active)

The current revision is the most recently published revision.  It is
the revision that is approved for public view.

=item newest

The newest revision is the most recently added revision.  It is a
working copy that is not displayed to the public until it is
published.

=back

These may, of course, refer to the same revision.

=cut

package ExSite::ContentData;

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::ContentBase;
#use ExSite::Cache;
use ExSite::Image;
use ExSite::ML;
use MIME::Base64;

use vars qw(@ISA);
@ISA = qw(ExSite::ContentBase);

#------------------------------------------------------------------------

=pod

=head1 Configuration & Setup

=head2 new()

(This method is actually inherited from ExSite::ContentBase.)

=head2 setup()

=head2 setupContentData()

Instantiate and/or configure a ContentData object.  C<%option> includes the 
the configuration parameters, described below.

Usage:

    my $rev = new ExSite::ContentData(%option);
    $rev->setup(%option);
    $rev->setupContentData(%option);

C<setup> and C<setupContentData> are synonymous; they are used to
setup or reconfigure an existing revision.  C<new()>
instantiates a new ContentData object (this method is actually
inherited from C<ExSite::ContentBase>).

C<%option> is used to pass the revision description information that
allows the object to be configured.  It can include the following revision
description parameters:

=over 4

=item id

The numeric content_data ID will be used to fetch the revision
description from the database.

=item page

The content_data record can be passed as a hash reference, if it has
already been fetched.

=item (null)

If no content_data description is provided, the object will
configure an empty content_data object that will have limited
functionality.

=back

The setup methods are normally invoked automatically by the class
contructor method C<new()>, which receives the same C<%option>
parameters.

=cut

#------------------------------------------------------------------------

# setup info for ContentData class

sub setup {
    &setupContentData(@_);
}

# fetch the content_data record

sub setupContentData {
    my ($this,%option) = @_;
    my $cdata;
    if ($option{id}) {
	$cdata = $this->fetch("content_data",$option{id});
    }
    elsif ($option{content_data}) {
	$cdata = $option{content_data};
    }
    if (ref $cdata eq "HASH") {
	$this->info("setting up ContentData $cdata->{content_data_id}");
	$this->{data} = $cdata;
	$this->set_context("none");
    }
    else {
	$this->warn("setting up null ContentData");
    }
}

=pod

=head1 Revision Queries

=head2 defined()

Returns TRUE if the revision references actual web data.  The data can
be in one of the following forms:

=over 4

=item raw data

Typically text data, stored in verbatim form.

=item file data

Encoded file upload, typically used for images.  No check is performed
that the file data decodes to a meaningful file format.

=item file

A path to a file stored outside the CMS.  No check is performed that
this file actually exists and is readable.

=item URL

A URL to a resource stored outside the system.  No check is performed
that this resource is accessible.

=back

=head2 content_type()

Returns the MIME type of the revision.  This is taken directly from
the content_data description, and not inferred from the data itself.
Therefore, if the data has been incorrectly described, that incorrect
description will be returned.

=head2 timestamp()

Returns the creation time of the revision, in database (eg. MySQL
timestamp) format.

=head2 is_file()

Returns TRUE if the revision data is an encoded file upload.

=head2 is_image()

Returns TRUE if the MIME type of the revision indicates that it is an
image.

=head2 is_text()

Returns TRUE if the MIME type of the revision indicates that it is
text (ie. HTML, CSS, or plain text).  Text data formats such as CSV
or tab-delimited values will return FALSE.

=head2 is_publishable()

Returns TRUE if the revision can be published to its own file.  This
is the case for all non-HTML MIME types.

=head2 get_fileinfo()

Returns an array of the filename, size (in kilobytes), and MIME-type
of the revision's encoded file.  Returns undef if the revision is not
an encoded file.

=cut

# is this content defined?

sub defined {
    my $this = shift;
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	return ($data->{data} || $data->{fdata} || 
		$data->{file} || $data->{url});
    }
    return 0;
}

# content queries

# return the content-type

sub content_type {
    my $this = shift;
    return $this->get("content_data","mime_type");
}

# return the timestamp

sub timestamp {
    my $this = shift;
    return $this->get("content_data","ctime");
}

# determine whether content is in a virtual file

sub is_file {
    my $this = shift;
    my $dat = $this->get("content_data");
    return ( ! $dat->{data} && $dat->{fdata} );
}

# determine whether content is an image

sub is_image {
    my $this = shift;
    return $this->get("content_data","mime_type") =~ /^image\//;
}

# determine whether content is text

sub is_text {
    my $this = shift;
    return $this->get("content_data","mime_type") =~ /^text\//;
}

# determine whether content is publishable to a file
# -> essentially true for all non-HTML mime types in static pages/libraries

sub is_publishable {
    my $this = shift;
    my $mimetype = $this->get("content_data","mime_type");
    if ($mimetype && $mimetype !~ /html/) {
	# the content is publishable, but what about its page?
	$this->set_context("parent");
	my $p = new ExSite::Page(page=>$this->get("page"));
	return $p->is_publishable();
    }
    return 0;
}

# deprecated
sub publishable {
    return &is_publishable(@_);
}

# in the event that the content data is an upload, return the file info
# (filename, size (K), mime-type)

sub get_fileinfo {
    my $this = shift;
    return undef if (! $this->defined);
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if ($data->{fdata}) {
	    my ($file,$fdata) = split /$config{form}{sepchar}/,$data->{fdata};
	    my $mime = &MimeType($file);
	    my $size = $data->{_fsize} ? $data->{_fsize} : length $fdata;
	    $size = int($size*6/8/100)/10;
	    return ($file,$size,$mime);
	}
	elsif ($data->{file}) {
	    my @path = split(/\//,$data->{file});
	    my $file = $path[-1];
	    if (-e $data->{file}) {
		my $mime = &MimeType($file);
		my $size = -s $data->{file};
		$size = int($size*6/8/100)/10;
		return ($file,$size,$mime);
	    }
	    else {
		return ($file, 0, "Warning: ile not found");
	    }
	}
	else {
	    return undef;
	}
    }
    return undef;
}

=pod

=head1 Retrieving/Displaying Revisions

=head2 get_raw()

Returns the raw data for the revision as a text or binary string.
This can be printed to stdout to send it directly to the client, or
manipulated by other code.

=head2 get_html()

    my $html = $rev->get_html($attributes,$url);

Returns an HTML string to render the revision.  For HTML revisions,
this is equivalent to C<get_raw()>.  Other MIME types will be
re-encoded to make them HTML-ready.  Text types will be wrapped in
appropriate HTML tags to make them render appropriately.  File types
will be converted to HTML tags that will allow the revision to be
downloaded in a separate request (eg. an IMG tag).

You can optionally specify some custom attributes, or a URL to the 
content data.  The system will specify suitable defaults for these 
values.

=head2 get_html_dynamic()

Same as previous, but forces the URL of content data to a dynamic 
representation.

=head2 get_url()

    $url = $rev->get_url($mode);

Returns a URL to fetch the revision in a stand-alone HTTP request.
This is used to set up the SRC and HREF attributes in IMG, LINK, and A
tags.

The returned URL may be static, if the revision has been published to
its own file, or dynamic if it must be read from the database.
The choice can be forced by passing the optional C<$mode> parameter.
Dynamic URLs use the C<content.cgi> program to fetch the data.

=head2 get_static_url()

Returns a URL to fetch the revision in a stand-alone HTTP request.
The URL is always to the static, published file, or to a dummy checker
graphic, if no such URL can be constructed.

=head2 get_dynamic_url()

Returns a URL to fetch the revision in a stand-alone HTTP request.
The URL is always to the dynamic version of the content, or to a dummy
checker graphic, if no such URL can be constructed.

=head2 get_thumb_html()

Same as C<get_html()> except that it returns an HTML string to render
a thumbnail of the content.

The revision data may include an explicit thumbnail, in which case
this is used.  Otherwise, if the revision is an image, we simply
return the image with reduced dimensions.  If the revision is not an
image, and no explicit thumbnail is provided, we return C<get_html()>.

=head2 get_thumb_url()

Same as C<get_url()> except that it returns a URL to the thumbnail.
If no explicit thumbnail exists, this returns the same result as
C<get_url()>.

=head2 get_thumb_raw()

Returns the raw data for the thumbnail as binary string.  Only works
if an explicit thumbnail is defined, otherwise C<undef> is returned.

=head2 preview()

    my $html = $rev->preview($mode);

Returns HTML that comprises an administrator preview of the revision.

Similar to C<get_html()>, except that it allows for expansion of CMS
tags to show other embedded content objects, and for the viewing of 
raw HTML (C<$mode> = "raw").

=cut

# return the raw content

##### check for published file, file pointer, DB file, and raw content

sub get_raw {
    my $this = shift;
    return undef if (! $this->defined);
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if ($data->{data}) {
	    return $data->{data};
	}
	elsif ($data->{fdata}) {
	    my $cdat = $data;
	    if ($data->{_fsize}) {
		# the file contents have been dropped from the data to 
		# conserve memory;  we need to reload it
		$cdat = $this->fetch("content_data",$cdat->{content_data_id});
	    }
	    my ($file,$fdata) = split /$config{form}{sepchar}/,$cdat->{fdata};
	    return decode_base64($fdata);
	}
	elsif ($data->{file}) {
	    return &get_file($data->{file});
	}
    }
    $this->warn("ContentData: no raw content defined");
    return undef;
}

# return the HTML tags needed to inline the content into a page

sub get_html {
    my ($this, $extra_attributes, $url) = @_;
    return undef if (! $this->defined);
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if (! $data->{content_data_id}) {
	    # nothing to get!
	    return "";
	}
	my $attributes = {};
	if ($data->{attributes}) {
	    $attributes = &DecodeAttributes($data->{attributes});
	}
	if (ref $extra_attributes eq "HASH") {
	    $attributes =  { %$attributes, %$extra_attributes };
	}
	my $type = $this->content_type;
	my $ml = new ExSite::ML;
	if ($type =~ /^text\/html$/) {
	    # html can be returned straight-up
	    return $this->get_raw;
	}
	elsif ($type =~ /^text\/css$/) {
	    if ($config{content}{inline_css}) {
		my $data = $this->get_raw;
		$attributes->{type} = "text/css";
		return $ml->style($data,$attributes);
	    }
	    else {
		if (! $url) { $url = $this->get_url(); }
		$attributes->{rel} = "stylesheet";
		$attributes->{type} = "text/css";
		$attributes->{href} = $url;
		return $ml->link(undef,$attributes);
	    }
	}
	elsif ($type =~ /^text/) {
	    if ($config{content}{preformatted_plaintext}) {
		return $ml->pre($this->get_raw);
	    }
	    else {
		return $this->get_raw;
	    }
	}
	elsif ($type =~ /^image/) {
	    # images are rendered using IMG tags
	    if (! $url) { $url = $this->get_url(); }
	    if (! exists $attributes->{alt}) {
		# add an alt to make W3C happy
		$attributes->{alt} = $this->get("content","name");
	    }
	    if (! exists $attributes->{title} && $config{content}{img_title}) { 
		# tooltip (using title attribute)
		my $title = $this->get("content",$config{content}{img_title}); 
		if ($title) { $attributes->{title} = $title; }
	    }
	    $attributes->{src} = $url;
	    return $ml->img(undef, $attributes);
	}
	elsif ($type =~ /^application\/x-shockwave-flash$/) {
	    # flash objects require some embedded object cruft
	    if (! $url) { $url = $this->get_url(); }
	    my $filename = $this->filename;
	    # attributes typically has dimensions for the movie
	    my $objattr = $attributes;
	    $objattr->{type} = "application/x-shockwave-flash";
	    $objattr->{data} = $url;
	    # return a flash object containing appropriate params
	    # followed by an IE Flash hack script
	    my $objcontents = 
		$ml->param(undef,{name=>"allowScriptAccess",value=>"sameDomain"}).
		$ml->param(undef,{name=>"quality",value=>"best"}).
		$ml->param(undef,{name=>"movie",value=>$url});
	    if ($attributes->{img}) {
		# optional img attribute specifies an image that will be used
		# if flash is not enabled
		$objcontents .= $ml->img(undef,$attributes->{img});
		delete $objattr->{img};
	    }
	    return $ml->object($objcontents,$objattr).
		$ml->script(undef,{ src=>"$config{server}{HTMLpath}/_ExSite/js/flashinsert.js" });
	}
	else {
	    # miscellaneous types are shown as hotlinks
	    if (! $url) { $url = $this->get_url(); }
	    my $filename = $this->filename || "download $type";
	    $attributes->{href} = $url;
	    return $ml->a($filename,$attributes);
	}
    }
    $this->warn("ContentData: no html content defined");
    return undef;
}

# return the HTML tags needed to inline the content into a page

sub get_html_dynamic {
    my ($this, $extra_attributes) = @_;
    my $url = $this->get_dynamic_url();
    return $this->get_html($extra_attributes,$url);
}

# return a URL that fetches the content as a separate HTTP request
# mode = static|dynamic (defaults to static, if found, dynamic otherwise)

sub get_url {
    my $this = shift;
    my $mode = shift;
    if ($this->{publish}) { $mode = undef; }
    return undef if (! $this->defined);
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if (! $data->{content_data_id}) {
	    # nothing to get!
	    return "";
	}
	# if there is a published revision, use it
	if ($mode ne "dynamic" && ($data->{data} || $data->{fdata})) {
	    my $file = $this->filename;
	    my $diskpath = $this->diskpath;
	    if (-e $diskpath) {
		$this->set_context;
		my $data = $this->get("page");
		require ExSite::Page;
		my $p = new ExSite::Page(page=>$data);
		my $url;
		if ($p->is_page()) {
		    # page data is written directly to the section directory
		    require ExSite::Section;
		    my $sec = new ExSite::Section(section=>$this->get("section"));
		    $url = $sec->get_url;
		}
		else {
		    # page files are written to the library/template subdirectory
		    $url = $p->get_url_static;
		}

		## make sure we know our section
		#$this->set_context; 
		#require ExSite::Section;
		#my $sec = new ExSite::Section(section=>$this->get("section"));
		## content is saved in the section, presumably
		#if (-e "$config{server}{HTMLroot}$config{server}{HTMLpath}$url/$file") {
		### fails if $url contains server info
		    return $url."/".$file;
		#}
	    }
	}
	if ($mode ne "static") {
	    # use dynamic revision
	    my $fname = $this->filename;
	    my $url = "$config{server}{server}$config{server}{CGIpath}/$config{prog}{content}/${fname}?id=".$this->id("content_data");
	    # $this->set_context;  # needed to get name, but do we really need that?
	    my $cobj = $this->get("content");
	    my $name = &url_escape($cobj->{name});
	    $url .= "&name=$name"; 
	    return $url;
	}
    }
    # nothing suitable found
    $this->warn("ContentData: no content URL defined");
    return "$config{server}->{HTMLpath}/_ExSite/images/checker.png";
}

sub get_static_url {
    my $this = shift;
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if ($data->{content_data_id}) {
	    my $file = $this->filename;
	    my $diskpath = $this->diskpath;
	    if (-e $diskpath) {
		$this->set_context;
		my $data = $this->get("page");
		require ExSite::Page;
		my $p = new ExSite::Page(page=>$data);
		my $url;
		if ($p->is_page()) {
		    # page data is written directly to the section directory
		    require ExSite::Section;
		    my $sec = new ExSite::Section(section=>$this->get("section"));
		    $url = $sec->get_url;
		}
		else {
		    # page files are written to the library/template subdirectory
		    $url = $p->get_url;
		}
		return $url."/".$file;
	    }
	}
    }
    # nothing suitable found
    $this->warn("ContentData: no content URL defined");
    return "$config{server}->{HTMLpath}/_ExSite/images/checker.png";
}

sub get_dynamic_url {
    my $this = shift;
    return undef if (! $this->defined);
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if ($data->{content_data_id}) {
	    my $url = "$config{server}{server}$config{server}{CGIpath}/$config{prog}{content}?id=".$this->id("content_data");
	    my $cobj = $this->get("content");
	    $url .= "&name=$cobj->{name}"; 
	    return $url;
	}
    }
    # nothing suitable found
    $this->warn("ContentData: no content URL defined");
    return "$config{server}{server}$config{server}{HTMLpath}/_ExSite/images/checker.png";
}

# return the HTML tags needed to inline a thumbnail of the content into a page

sub get_thumb_html {
    my $this = shift;
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if (! $data->{content_data_id}) {
	    # nothing to get!
	    return "";
	}
	my $ml = new ExSite::ML;
	if ($data->{thumb}) {
	    # yay, a real thumb!
	    my $url = $this->get_thumb_url;
	    my $thattr = $data->{thumb_attributes} ? 
		&DecodeAttributes($data->{thumb_attributes}) : {};
	    if (! exists $thattr->{alt}) { 
		$thattr->{alt} = $this->get("content","name"); 
	    }
	    if (! exists $thattr->{title} && $config{content}{img_title}) { 
		my $title = $this->get("content",$config{content}{img_title}); 
		if ($title) { $thattr->{title} = $title; }
	    }
	    return $ml->img(undef,{src=>$url, class=>"thumbnail", border=>0, %$thattr});
	}
	elsif ($this->content_type =~ /^image/) {
	    # fake a thumb by shrinking the image
	    my $url = $this->get_url;
	    my $attr = $data->{attributes} ?
		&DecodeAttributes($data->{attributes}) : {};
	    my $width = $attr->{width};
	    my $height =  $attr->{height};
	    if (! $width || ! $height) {
		# missing height, width attributes - compute from scratch
		my $img = $this->get_image();
		if ($img) {
		    ($width,$height) = $img->dim;
		}
	    }
	    $attr->{src} = $url;
	    $attr->{class} = "thumbnail";
	    $attr->{border} = 0;
	    if (! exists $attr->{alt}) { 
		$attr->{alt} = $this->get("content","name"); 
	    }
	    if (! exists $attr->{title} && $config{content}{img_title}) { 
		my $title = $this->get("content",$config{content}{img_title}); 
		if ($title) { $attr->{title} = $title; }
	    }
	    if ($width =~ /^\d+$/ && $height =~ /^\d+$/) {
		if ($width <= 100 && $height <= 100) {
		    $attr->{width} = $width;
		    $attr->{height} = $height;
		}
		elsif ($width > $height) {
		    $attr->{width} = 100;
		    $attr->{height} = int($height / $width * 100);
		}
		else {
		    $attr->{width} = int($width / $height * 100);
		    $attr->{height} = 100;
		}
	    }
	    return $ml->img(undef,$attr);
	}
	else {
	    return $this->get_html();
	}
    }
    $this->warn("ContentData: no content defined");
    return undef;
}

# return the URL to a thumbnail

sub get_thumb_url {
    my $this = shift;
    return undef if (! $this->defined);
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if (! $data->{content_data_id}) {
	    # nothing to get!
	    return "";
	}
	if ($data->{thumb}) {
	    my $diskpath = $this->diskpath;
	    $diskpath =~ s/\.([^\.]*)$/_T.$1/;
	    if (-e $diskpath) {
		$this->set_context;
		my $data = $this->get("page");
		require ExSite::Page;
		my $p = new ExSite::Page(page=>$data);
		my $url;
		if ($p->is_page) {
		    # page data is written directly to the section directory
		    require ExSite::Section;
		    my $sec = new ExSite::Section(section=>$this->get("section"));
		    $url = $sec->get_url;
		}
		else {
		    # page files are written to the library/template subdirectory
		    $url = $p->get_url_static();
		}
		my $file = $this->filename;
		$file =~ s/\.([^\.]*)$/_T.$1/;
		return $url."/".$file;
	    }
	    # not published - return dynamic url
	    # getdata.cgi deprecated for CMS work
	    #return "$config{server}{CGIpath}/$config{prog}{getdata}?tab=content_data&col=thumb&id=$data->{content_data_id}";
	    my $id = $this->id("content_data");
	    return "$config{server}{CGIpath}/$config{prog}{content}?id=${id}&thumb=1";
	}
	else {
	    return $this->get_url;
	}
    }
    $this->warn("ContentData: no content defined");
    return undef;
}

sub get_thumb_raw {
    my $this = shift;
    my $img = $this->get_thumb();
    if ($img) {
	return $img->get();
    }
    $this->warn("ContentData: no content defined");
    return undef;
}

sub get_thumb {
    my $this = shift;
    return $this->get_image("thumb");
}

sub get_image {
    my $this = shift;
    my $itype = shift || "fdata";
    return undef if (! $this->defined);
    my $data = $this->get("content_data");
    return undef if (ref $data ne "HASH") || (! $data->{$itype});
    my $cdat = $data;
    my $filename = $cdat->{$itype};
    $filename =~ s/\#.*//;
    my $mimetype = &MimeType($filename);
    return undef if ($mimetype !~ /^image/);
    my $size = ($itype eq "thumb") ? "_tsize" : "_fsize";
    if ($data->{$size}) {
	# the file contents were dropped from the data to save memory
	# we need to reload it
	$cdat = $this->fetch("content_data",$cdat->{content_data_id});
    }
    return new ExSite::Image($cdat->{$itype});
}

# preview: display the content for an administrator

sub preview {
    my $this = shift;
    my $mode = shift;
    my $type = $this->content_type;
    if ($type eq "text/html" && $mode ne "raw") {
	# expand the content
	# do not use published elements (ie. always show latest revision)
	# do not expand dynamic modules
	return $this->expand(mode=>"dynamic",method=>"data,url,content,page");
    }
    elsif ($type =~ /^text/) {
	# display raw content for inspection, rather than convert it to HTML
	my $raw = $this->get_raw;
	my $ml = new ExSite::ML(safe_content=>0);
	return $ml->pre($raw);
    }
    else {
	# convert to HTML - force dynamic preview mode for images
	return $this->get_html_dynamic();
    }
}

=pod

=head1 Addressing Revisions

=head2 filename()

Returns a file name for the revision.  If the revision is not already
a file, then a filename is composed from the content ID and MIME type.

Note that content objects are often named for the original file name
in their first revision.  However, as the content gets updated, new
files may be uploaded with different names.  The content object name
remains the same, but the revision filename will change.  For example,
a content object named "logo.gif" may eventually resolve to an actual
file named "logo-64px-BW.jpg".  This can be confusing when first
encountered.

=head2 httppath()

Returns the URL path to the file (ie. the part of the static URL after
the host name), not including the filename itself.  This is identical
to the httppath to the page contianing the content.

=head2 diskpath()

Returns the full disk path to the file, including the filename.  

If the content belongs to a regular page, the file will be published
into the general section directory.  That could result in filename
conflicts, which would cause different content objects under different
pages, to publish to the same file.  In this case, one will overwrite
the other, and you will get unexpected results.

=cut

### FIXME: do we need a get method that returns an administrative link to 
### the content data (or content object)?

# determine the filename of a content datum "virtual file"

sub filename {
    my $this = shift;
    my $file;
    my $data = $this->get("content_data");
    if (ref $data eq "HASH") {
	if ($data->{data}) {
	    my $suffix = &MimeToFile($data->{mime_type});
	    $file = "$data->{content_id}.$suffix";
	}
	elsif ($data->{fdata}) {
	    ($file, undef) = split /$config{form}{sepchar}/,$data->{fdata};
	    # strip inconvenient characters from filenames
	    $file = &clean_filename($file);
	}
	elsif ($data->{file}) {
	    # strip path info
	    my @path = split /[:\\\/]/, $data->{file};
	    $file = $path[-1];
	}
	else {
	    # strip path info
	    my @path = split /\//, $data->{url};
	    $file = $path[-1];
	}
	return $file;
    }
    return undef;
}

# determine the path to the file, if published

sub httppath {
    my $this = shift;
    $this->set_context;
    require ExSite::Page;
    my $page = new ExSite::Page(page=>$this->get("page"));
    # the path to the content data is the same as the path to its page
    return $page->httppath;
}

# determine the path to the file, if published

sub diskpath {
    my $this = shift;
    my $dir = shift;  # return directory only if true
    $this->set_context;
    #return undef if (! $this->{cdata});
    #return $config{server}->{HTMLroot}.$this->httppath;
    my $cdata = $this->get("content_data");
    if ($cdata->{data} || $cdata->{fdata}) {
	my $data = $this->get("page");
	require ExSite::Page;
	my $p = new ExSite::Page(page=>$data);
	my $path;
	if ($p->is_page) {
	    # page files are written directly to the section directory
	    require ExSite::Section;
	    my $sec = new ExSite::Section(section=>$this->get("section"));
	    $path = $sec->diskpath;
	}
	else {
	    # page files are written to the library/template subdirectory
	    $path = $p->diskpath;
	}
	return $path if ($dir);
	$path .= "/".$this->filename;
	return $path;
    }
    elsif ($cdata->{file}) {
	if ($dir) {
	    my $path = $cdata->{file};
	    my @path = split /[:\\\/]/, $cdata->{file};
	    pop @path;
	    return join "/",@path;   # *nix only
	}
	return $cdata->{file};
    }
    elsif ($cdata->{url}) {
	return undef;
    }
}

=pod

=head1 Publishing

WARNING: you need write access to the publication directories to
publish files.  The webserver normally does not have this access.
Negotiating this access is normally done by the publish program
(ie. publish.cgi), which has special permissions.  If you do not have
these permissions, then you have to ask publish.cgi to do this task on
your behalf, using C<ExSite::Util::run_publisher()>.  The publish
program does not have access to specific revisions; you can only ask
it to publish a Content object, which will publish the newest revision
of the content.

=head2 set_publish_target()

    my $target = $rev->set_publish_target($path,$filename);

Determines where a revision's file will be published.  The path and
filename can be passed explicitly, or it will be inferred from the
C<diskpath()> and C<filename()>.  The full path to be published to
will be returned, but the target will also be stored in the revision
object under the following attributes:

A revision in isolation has no context to determine where it ought to
be published, so this method must be called before you can publish.
This is done automatically by C<ExSite::Page::publish()>, which can
inform the revisions which directory the page's files will publish to.

=over 4

=item publish_target_filename

The filename to publish to.

=item publish_target

The complete path, including filename, to publish to.

=back

=head2 is_published()

Returns the timestamp from the last time the content was published.
If never published, returns 0.  This is suitable for a simple logical
test for whether the revision has been published.

=head2 publish()

    my $status_message = $rev->publish($path,$filename);

Publishes the revision, and returns a text string reporting the status
of the operation.

The publication C<$path> and C<$filename> can be passed explicitly,
but this is optional if we have a full context to determine which page
this content revision belongs to.

Publication of a revision accomplishes two things:

1) It writes out the revision's file(s) to disk.  If the revision
includes a thumbnail, it will also be written.

2) It marks the revision as published and therefore viewable by the
public.

There is one special case worth noting: if the revision refers to an
external file, then this file is assumed to already exist on disk and
in theory does not need to be published.  However, the CMS falls back
on this method of revision storage for revisions that are too large
for storage in the database.  In this case, the file may not yet exist
on disk because it has just been uploaded.  So we check in the POST
data for possible replacements for this file, under the following
parameters:

=over 4

=item from

This parameter contains a path to a temporary file that should be
copied to the publication location.

=item fdata

This parameter contains an ExSite-encoded binary file that should be
decoded and written to the publication location.

=back

=head2 unpublish()

Removes the revision's file and thumbnail from disk.  It will fail if
the file is located outside the system HTMLroot.  (This is not
considered a bug, because those files are technically outside the CMS
and should not be deleted by the CMS.)

=head2 mkdir()

    my $status_message = $rev->mkdir($path);

Creates the directories in path, if they do not exist.  It will only
create directories inside the system's HTMLroot for security reasons.
If you need to create directories outside this area, you must untaint
the C<$path> yourself before passing it to C<mkdir()>.

=head2 rmdir()

    my $status_message = $rev->rmdir($path);

Deletes the directory in C<$path>, if it exists.  It will only
delete directories inside the system's HTMLroot for security reasons.
If you need to remove directories outside this area, you must untaint
the C<$path> yourself before passing it to C<rmdir()>.

=head2 rm()

    my $status_message = $rev->rm($path);

Deletes the file specified by C<$path>, if it exists.  It will only
delete directories inside the system's HTMLroot for security reasons.
If you need to remove files outside this area, you must untaint
the C<$path> yourself before passing it to C<rm()>.

=cut

sub set_publish_target {
    my ($this,$path,$filename) = @_;
    my $out;
    my $data = $this->get("content_data");
    
    if ($this->is_publishable()) {
	if (! $path) {
	    $path = $this->diskpath(1);
	    if ($path) { 
		$out .= $this->mkdir($path); 
	    }
	}
	if (! $filename) {
	    $filename = $this->filename();
	}
	$this->{publish_target_filename} = "$filename";
	$this->{publish_target} = "$path/$filename";
    }
    else {
	$this->{publish_target_filename} = undef;
	$this->{publish_target} = undef;
    }
    return $this->{publish_target};
}

# return the timestamp from the last time the content was published

sub is_published {
    my ($this,$path,$filename) = @_;

    my $ptime = $this->get("content_data","ptime");
#    if ($ptime !~ /^0+$/) {     # FIME: fails on MySQL 4.1, use mysql_timestamp_hack on prefetch() to fix timestamps
    if ($ptime !~ /^0+/) {
	# the database has a publish timestamp
	if ($this->{publish_target}) {
	    # we have a specific filename to publish to
	    return (-e $this->{publish_target});
	}
	return $ptime;
    }
    return 0;
}

sub publish {
    my ($this,$path,$filename) = @_;
    if (! $this->allow_update()) {
	# publishing has same permission rules as updates
	print "Permission denied - administrator privileges are required to publish.";
	return;
    }

    my $out;
    
    # fetch the raw content data
    my $data = $this->get("content_data");
    if (! $data->{content_data_id}) {
	# abort if we have no content data defined
	return " ";
    }

    # figure out where to publish to
    $this->set_publish_target($path,$filename);

    # check if the content has already been published
    if ($this->is_publishable() && $data->{ptime} && $data->{ptime} !~ /^0+$/) {
	# timestamp indicates that it was published at one time
	if ($this->is_published($path,$filename)) {
	    # there appears to be a matching file already on disk
	    return " ";
	}
    }

    # set the publication timestamp
    # this indicates, among other things, that the content is approved
    # for public viewing (even if the content object does not get
    # written to its own file).

    my $ctime = "ctime";  # so we can set ctime to itself
    $share{DB}->update("content_data",
		       {content_data_id=>$data->{content_data_id},
			ctime=>\$ctime,
			ptime=>undef});

    # return if this content does not go to its own file
    if (! $this->is_publishable()) {
	return " ";
    }
 
    my $ml = new ExSite::ML;
    $out .= $ml->p("Publishing $this->{publish_target_filename}...");
    if ($this->{publish_target}) {
	my $data = $this->get("content_data");
	my $cobjname = $this->get("content","name");
	if (ref $data eq "HASH") {
	    my ($file,$fdata);
	    $file = $this->{publish_target_filename};
	    my $in = new ExSite::Input;
	    my $updata = $in->post;

	    if ($data->{file} && keys %$updata > 0) {
		# raw data is stored in a disk file...
		# but that means one of 2 things:
		# - it is alreay published
		# - user has uploaded a file too large for revision control
		# In either case, check POST data for a newly uploaded
		# revision of the file, and republish it in that case.
		if ($updata->{from}) {
		    # raw data is stored in a temp file, which needs to be
		    # copied to the destination
		    $this->info("Loading oversize file: $updata->{from}");
		    my $safe_tmpfile = $updata->{from};
		    $safe_tmpfile =~ /^(\/tmp\/ExSite[0-9]+)$/;
		    $safe_tmpfile = $1;
		    if ($safe_tmpfile) {
			$fdata = &get_file($updata->{from});
			if ($fdata =~ /^[^\#]{1,40}\.[\w]{2,4}\#[A-Za-z0-9+\/=]+$/) {
			    # exsite-encoded file
			    my (undef,$mimedata) = split /\#/,$fdata,2;
			    $fdata = &decode_base64($mimedata);
			}
			# remove the temp file
			$this->info("Removing temp file $updata->{from}");
			unlink $safe_tmpfile or 
			    $out .= $ml->p("Remove temp file: $!");
		    }
		    else {
			$this->error("permission denied: $updata->{from}");
			return undef;
		    }
		}
		elsif ($updata->{fdata}) {
		    if ($updata->{fdata} =~ /^[^\#]{1,40}\.[\w]{2,4}\#[A-Za-z0-9+\/=]+$/) {
		    #if ($updata->{fdata} =~ /^[^\#]{1,40}\.[\w]{2,4}\#/) {
			# looks like an exsite-encoded file
			my (undef,$mimedata) = split /\#/,$updata->{fdata},2;
			$fdata = &decode_base64($mimedata);
		    }
		    else {
			# could be a multipart-encoded file
			$fdata = $in->fetch_file("fdata","raw");
		    }
		    $out .= $ml->p( (length $fdata)." bytes in file" );
		    if (! $fdata) {
			$this->warn("ContentData::publish: No file data found");
			return undef;
		    }
		}
		else {
		    # nothing to publish
		    if (! -e $data->{file}) {
			# the file pointer goes nowhere...
			return $this->warn("ContentData::publish: $data->{file} does not exist");
		    }
		    return undef;
		}
	    }
	    elsif ($data->{data} || $data->{fdata}) {
		# raw data is stored in database
		$fdata = $this->get_raw;
	    }
	    else {
		# nothing to publish
		return undef;
	    }
	    #my $outputpath = $path ? "$path/$file" : $file;
	    my $outputpath = $this->{publish_target};
	    #### allow periods in path
	    # eg. /home/www.domain.com/htdocs/...
	    #$outputpath =~ /^(((\/[\w-]+)+)\/[\w][\w\-\.]+)$/;
	    #my $secure_outputpath = $1;
	    my $secure_outputpath = $this->safe_path($outputpath);
	    if (! $secure_outputpath) {
		$this->error("publish content data: invalid filename: $this->{publish_target}");
	    }
	    elsif (open OUTFILE, ">$secure_outputpath") {
		$this->info("publish content data $data->{content_data_id} to $secure_outputpath");
		print OUTFILE $fdata;
		close OUTFILE;
		$out .= "Successfully published $cobjname to $file.<br>\n";
		if ($data->{thumb}) {
		    # add _T to filename for thumnail
		    $secure_outputpath =~ s/\.([^\.]*)$/_T.$1/;
		    $file =~ s/\.([^\.]*)$/_T.$1/;
		    $fdata = $this->get_thumb_raw;
		    if (open OUTFILE, ">$secure_outputpath") {
			print OUTFILE $fdata;
			close OUTFILE;
			$out .= "Successfully published thumbnail of $cobjname to $file.<br>\n";
		    }
		    else {
			$out .= "Failed to publish thumbnail of $cobjname to $file: $!<br>\n";
		    }
		}
		return $out;
	    }
	    else {
		$this->error("Failed to publish $cobjname to $file: $!");
		my $x = $this->get("content_data");
		return $ml->p("Failed to publish $cobjname to $file: $!",{class=>"error"});
	    }
	}
    }
    $this->warn("ContentData::publish: No publication directory specified");
    return undef;
}

# remove the file associated with the content data

sub unpublish {
    my ($this,$path) = @_;
    # return if this content does not go to its own file
    return " " if (! $this->is_publishable());
    my $out;
    if (! $path) { $path = $this->diskpath(1); }
    if ($path) {
	my $ml = new ExSite::ML;
	my $br = $ml->br();
	my $data = $this->get("content_data");
	my $cobjname = $this->get("content","name");
	if (ref $data eq "HASH") {
	    my ($file,$fdata);
	    $file = $this->filename;
	    my $outputpath = $path ? "$path/$file" : $file;
	    $outputpath =~ /^($config{server}{HTMLroot}$config{server}{HTMLpath}((\/[\w-]+)*)\/[\w][\w\-\.]+)$/;
	    my $secure_outputpath = $1;
	    if (-e $secure_outputpath) {
		$this->info("removing file $secure_outputpath (content data $data->{content_data_id})");
		if (unlink $secure_outputpath) {
		    $out .= "Removed $file.$br";
		}
		else {
		    $out .= $this->error("Failed to remove $file: $!");
		}
	    }
	    else {
		return $this->error("ContentData::unpublish: $outputpath: permission denied");
	    }
	    
	    if ($data->{thumb}) {
		# add _T to filename for thumnail
		$secure_outputpath =~ s/\.([^\.]*)$/_T.$1/;
		$file =~ s/\.([^\.]*)$/_T.$1/;
		if (-e $secure_outputpath) {
		    $this->info("removing file $secure_outputpath (thumbnail for content data $data->{content_data_id})");
		    if (unlink $secure_outputpath) {
			$out .= "Removed $file.$br";
		    }
		    else {
			$out .= $this->error("Failed to remove $file: $!");
		    }
		}
	    }

	    return $out;
	}
    }
    $this->warn("ContentData::unpublish: cannot determine content location");
    return undef;
}

# utility for creating publication directories

sub mkdir {
    my ($this,$dir) = @_;
    my $out = " ";  # dummy return value

    {
	# new block to limit scope of $1
	$dir =~ /(\/\w[\w_\-\.\/]+)/;
	if (! $1) {
	    $this->error("mkdir: malformed directory: $dir");
	    return undef;
	}
    }

    my @path = split(/\//,$dir);
    shift @path; # first one is a dummy
    my ($path,$secure_dir);
    foreach my $dir (@path) {
	$path .= "/$dir";
	next if (-e "$path");

	{
	    # new block to limit scope of $1
	    # new directories must be in the system docs path
	    #$path =~ /^($config{server}{HTMLroot}$config{server}{HTMLpath}(\/[\w-\/]+)?)$/;
	    #$secure_dir = $1;
	    $secure_dir = $this->safe_path($path,"$config{server}{HTMLroot}$config{server}{HTMLpath}");
	    if (! $secure_dir) { 
		# if you get this error, you are attempting to publish
		# outside the system HTdocs area.  You cannot create 
		# directories outside your HTdocs area for security reasons,
		# so if you want to publish there, you must create the 
		# directories manually, or clean the directory path using
		# safe_path()
		$this->error("mkdir: permission denied: directory '$path' is insecure");
		return undef;
	    }
	}

	$out .= "Creating site directory $dir... ";
	if (! mkdir "$secure_dir") {
	    $this->error("mkdir $dir: failed: $!");
	    return undef;
	}
	else {
	    $this->info("created site directory $dir");
	    $out .= "succeeded.";
	}
    }
    if ($out =~ /\S/) {
	my $ml = new ExSite::ML;
	return $ml->p($out);
    }
    return $out;
}

# utility for removing publication directories

sub rmdir {
    my ($this,$dir) = @_;
    my $out = " ";  # dummy return value
    my $ml = new ExSite::ML;
    my $br = $ml->br();
    # can only remove directories in the HTMLpath
    #$dir =~ /^($config{server}{HTMLroot}$config{server}{HTMLpath}\/[\w-\/]+)$/;
    #my $secure_dir = $1;
    my $secure_dir = $this->safe_path($dir,"$config{server}{HTMLroot}$config{server}{HTMLpath}");
    if (! $secure_dir) { 
	$this->error("rmdir: invalid directory: $dir");
	return undef;
    }
    elsif (-e $secure_dir) {
	opendir D,$secure_dir;
	my @files = readdir D;
	closedir D;
	foreach my $f (@files) {
	    next if ($f =~ /^\.\.?$/);
	    $f =~ /^(.+)$/;
	    my $secure_file = $secure_dir."/".$1;
	    if (-d $f) {
		# subdir; better remove it, too
		$out .= $this->rmdir("$dir/$f");
	    }
	    else {
		# must move unlink to its own if statement, or it causes
		# an insecure dependency error 
		if (unlink $secure_file) {
		    $out .= "Removed $f.$br\n";
		}
		else {
		    $this->error("rmdir: failed to remove $f: $! (you may be trying to remove directories that are not managed by ExSite)");
		    return undef;
		}
	    }
	}
	if (rmdir $secure_dir) {
	    $out .= "Removed $secure_dir.$br\n";
	}
	else {
	    $this->error("rmdir: failed to remove $secure_dir: $!");
	    return undef;
	}
    }
    return $out;
}

# utility for removing published files

sub rm {
    my ($this,$path) = @_;
    my $out = " ";  # dummy return value
    my $ml = new ExSite::ML;
    my $br = $ml->br();
    # can only remove files in the HTMLpath
    if ($path =~ /\.\./) {
	$this->error("rm: not removing suspicious path: $path");
	return undef;
    }
    #$path =~ /^($config{server}{HTMLroot}$config{server}{HTMLpath}\/.+)$/;
    #my $secure_path = $1;
    my $secure_path = $this->safe_path($path,"$config{server}{HTMLroot}$config{server}{HTMLpath}");

    if (! $secure_path) { 
	$this->error("rm: invalid path: $path");
	return undef;
    }
    elsif (-e $secure_path && ! -d $secure_path) {
	if (unlink $secure_path) {
	    $out .= "Removed $secure_path.$br\n";
	}
	else {
	    $this->error("rm: failed to remove $secure_path: $!");
	    return undef;
	}
    }
    return $out;
}

=pod

=head1 Revision Management

=head2 validate()

    my @errors = $rev->validate($revdata);
    if (@errors == 0) { # revision is OK }

Returns TRUE if the values in the the C<$revdata> hashref comprise
a valid and legal revision description.  Returns an array of error
messages, which will be empty if the revision validates.

=head2 make()

=head2 make_content_data()

   $section->make($revdata,$options);

Creates a new revision using the attributes in the C<$revdata> hashref.

The information in C<$revdata> is validated, and authorization checks
are performed to ensure the user has update privileges.  The data may
be further modified as follows:

=over 4

=item * file uploads that are too large to store in the database may
be switched to file pointers

=item * guess a MIME-type, if one was not specified

=item * add height and width attributes to images

=item * generate image thumbnail

=item * clean the filename to make it friendly to *nix servers

=item * add a comment, if not was not specified

=back

=head2 write_to_file()

    $rev->write_to_file($content_id,$data);

Write a file straight to disk.  This is used when an oversized file
was passed to C<make()> which could not be stored in the database.
Instead, we repost the file to C<publish.cgi>, which will take care of
writing it direct to disk.  C<$content_id> tells the publish program
where this file belongs, and C<$data> is the ExSite-encoded file data.

=head2 restore()

Creates a new revision that is identical to this revision.  This is
used to "restore" an old revision to use, without rolling back through
all the intermediate revisions.

=head2 delete()

Deletes the revision from the database.  Does not attempt to unpublish
it.  This is functionally equivalent to "rolling back" to the previous
revision (which will now be the newest revision).

=head2 purge()

Deletes all but this revision from the database.

=cut

# validate - check integrity of datahash

sub validate {
    my ($this,$data) = @_;
    my $cdata;
    if ($data) {
	$cdata = new ExSite::ContentData(content_data=>$data);
    }
    else {
	$cdata = $this;
	$data = $this->get("content_data");
    }
    my @err = $share{DB}->validate_record("content_data",$data);
    if ($cdata->{content_id}) {
	if (! $cdata->owns()) {
	    push @err, "permission denied - ownership mismatch";
	}
    }
    if (! $data->{data} && 
	! $data->{fdata} && 
	! $data->{file} && 
	! $data->{url}) {
	push @err, "content revision is empty";
    }
    return @err;
}

# make_content_data : insert a new content_data record
#
# this provides some generic algorithms for:
# - guessing the mime type (automatic)
# - determining image attributes (on request)
# - thumbnailing the image (on request)

sub make {
    return &make_content_data(@_);
}

sub make_content_data {
    my ($this,$cdata,$option) = @_;
    if ($cdata->{content_data_id} || ! $cdata->{content_id}) {
	# invalid content
	$this->error("failed to make improperly defined content_data");
	return undef;
    }
    if (! $this->allow_insert("content_data",$cdata)) {
	# permission denied
	$this->error("make_content_data: permission denied");
	return undef;
    }

    # remove any residual timestamps
    delete $cdata->{ctime};

    my ($file,$filedata,$length,$filepath);

    # data size

    if ($cdata->{data}) {
	delete $cdata->{fdata};   # clear ambiguous definition
	$length = length $cdata->{data};
	if ($length > $config{form}{bigfile_maxsize}) {
	    $this->error("data too large");
	    return undef;
	}
    }
    elsif ($cdata->{fdata}) {
	($file,$filedata) = split(/$config{form}{sepchar}/,$cdata->{"fdata"});
	$length = (length $filedata) / 4 * 3;
	if ($length > $config{max_upload_size}) {
	    $this->error("file too large");
	    return undef;
	}
	elsif ($length > $config{form}{bigfile_maxsize}) {
	    # too big for database, but not to big to write out to disk
	    delete $cdata->{fdata};
	    $file = &clean_filename($file);
	    # determine the future path to the published file
	    my $cid = $cdata->{content_id};
	    my $content = $share{DB}->fetch("content",$cid);
	    if ($content) {
		require ExSite::Page;
		my $page = new ExSite::Page(id=>$content->{page_id});
		$filepath = $page->diskpath(1);
		if ($filepath) {
		    $cdata->{file} = "$filepath/$file";
		}
		else {
		    $this->error("could not determine path to $file");
		}
	    }
	    else {
		$this->error("could not determine context for publishing $file");
	    }
	}
    }

    # set the mime-type

    if (! $cdata->{mime_type}) {
	if ($cdata->{data}) {
	    # look for suggestive patterns
	    if ($cdata->{data} =~ /<(!--|[\/a-zA-Z])/) {
		$cdata->{mime_type} = "text/html";
	    }
	    elsif ($cdata->{data} =~ /\/\*|[\w-]+:\s*[\w\#]+\;|^\w*\.\w+\s*\{/) {
		$cdata->{mime_type} = "text/css";
	    }
	    else {
		$cdata->{mime_type} = "text/plain";
	    }
	}
        else {
	    # look at file type
	    my $file;
	    if ($cdata->{fdata}) {
		($file,$filedata) = split(/$config{form}{sepchar}/,$cdata->{"fdata"});
	    }
	    else {
		$file = $cdata->{"file"} || $cdata->{"url"};
	    }
	    $cdata->{mime_type} = &MimeType($file);
	}
    }

    # comment

    if (! $cdata->{note}) {
	$cdata->{note} = "inserted by ExSite::ContentData::make()";
    }

    # image handling - optional

    if ($cdata->{fdata} && $cdata->{mime_type} =~ /^image\/(gif|jpeg|png)$/) {
	my $img = new ExSite::Image($cdata->{fdata});
	if ($option =~ /attribute/ && ! $cdata->{attributes}) {
	    my ($x,$y) = $img->dim;
	    if ($x && $y) {
		$cdata->{attributes} = "width=\"$x\" height=\"$y\"";
	    }
	}
	if ($option =~ /thumb/ && ! $cdata->{thumb}) {
	    my $stat = ($option =~ /square-thumb/) ? 
		$img->square_thumb() : 
		$img->thumb();
	    if ($stat) {
		$cdata->{thumb} = "th_".$img->encode;
		my ($x,$y) = $img->dim;
		if ($x && $y) {
		    $cdata->{thumb_attributes} = "width=\"$x\" height=\"$y\" border=\"0\"";
		}
	    }
	}
    }
    elsif ($cdata->{fdata}) {
	($file,$filedata) = split(/$config{form}{sepchar}/,$cdata->{"fdata"});
	$length = $filedata / 4 * 3;
	if ($length > $config{max_upload_size}) {
	    $this->error("file too large");
	    return undef;
	}
	elsif ($length > $config{form}{max_upload_size}) {
	    # too big for database, but not to big to write out to disk
	    $file = &clean_filename($file);
	}
    }

    # brute force the validate call into this class
    # $this->validate(...) fails for content updates, because $this is a 
    # Content object and we end up using Content::validate() instead.
    my @err = &ExSite::ContentData::validate($this,$cdata);
    if (@err > 0) {
	$this->error(@err);
	return undef;
    }

    my $cdid = $share{DB}->insert("content_data",$cdata);
    if (! $cdid) {
	$this->error("failed to install new revision of $file");
    }
    else {
	$cdata->{content_data_id} = $cdid;
	$this->info("installed new revision $cdid $file as $cdata->{mime_type}");
	if ($filepath && ! $cdata->{fdata}) {
	    # looks like fdata was removed from the record, which means
	    # we're expected to publish the file directly
	    $this->write_to_file($cdata->{content_id},"$file#$filedata");
	}

	# update the prefetch cache
	$this->prefetch_update("content_data",$cdata);
    }
    return $cdid;
}

# write a content file straight to disk
# -> outsource job to publish.cgi program to get necessary permissions
# cobj = content ID
# path = file to publish to
# data = exsite-encoded file data 

sub write_to_file {
    my ($this,$cobj,$data) = @_;

    # make a temporary copy of the file

    my $tmpfile = "$config{server}{temppath}/ExSite$$";
    open TMP,">$tmpfile";
    print TMP $data;
    close TMP;
    my $postdata = "from=$tmpfile";

    my $postdata = "fdata=$data";
    $postdata =~ s/\+/%2B/g;          # mask MIME chars from http

    # save current environment
    my $qstr = $ENV{QUERY_STRING};
    my $clen = $ENV{CONTENT_LENGTH};
    my $reqm = $ENV{REQUEST_METHOD};
    my $ctyp = $ENV{CONTENT_TYPE};
    # overload environment
    $ENV{QUERY_STRING} = "content=$cobj&silent=1";
    $ENV{CONTENT_LENGTH} = length $postdata;
    $ENV{REQUEST_METHOD} = "POST";
    $ENV{CONTENT_TYPE} = "application/x-www-form-urlencoded";

    # fire up the publishing program, and post data to it
    open F, "|$config{server}{CGIroot}/$config{prog}{publish}" or print "$?\n";
    print F $postdata,"\n";
    close F;

    # restore the environment of the current request, just to be safe
    $ENV{QUERY_STRING} = $qstr;
    $ENV{CONTENT_LENGTH} = $clen;
    $ENV{REQUEST_METHOD} = $reqm;
    $ENV{CONTENT_TYPE} = $ctyp;

    return;
}

# restore a previous revision

sub restore {
    my $this = shift;
    my $cdata = shift || $this->get("content_data");
    return 0 if (! $cdata);
    my %data;
    if (ref $cdata eq "HASH") {
	%data = %$cdata;
    }
    else {
	%data = $share{DB}->fetch("content_data",$cdata);
    }
    return 0 if (scalar keys %data == 0);
    delete $data{content_data_id};
    if ($this->allow_insert("content_data",\%data)) {
	delete $data{ctime};
	$data{note} .= " (RESTORED)";
	my $cdid = $share{DB}->insert("content_data",\%data);
	if (! $cdid) {
	    $this->error("content_data::restore: failed to restore old revision");
	}
	else {
	    $this->info("content_data::restore: restored old revision $cdid");
	}
	return $cdid;
    }
    return 0;
}

# delete the revision

sub delete {
    my $this = shift;
    my $id = $this->id;
    if ($id && $this->allow_delete("content_data",$this->{data})) {
	$share{DB}->trash_key("content_data",$id);
    }
}

# purge older revisions

sub purge {
    my $this = shift;
    my $data = $this->get("content_data");
    if ($data && $this->allow_delete("content_data",$data)) {
	my $db = $share{DB};
	my @cdata = $db->fetch_child("content_data","content",$data->{content_id},"content_data_id");
	foreach my $cdat (@cdata) {
	    last if ($cdat->{content_data_id} >= $data->{content_data_id});
	    $db->trash_key("content_data",$cdat->{content_data_id});
	}
    }
}

1;
