#!/usr/bin/perl
#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

=pod

BUGS

- long tab labels cause formatting problems

=cut

package ExSite::HTML;

use strict;
use ExSite::Config;

use vars qw($list_count %idcount %script_loaded $js_output $active_tab);


#========================================================================

=pod

=head1 ExSite::HTML

=head2 HTML methods and templates

This package contains routines and functions to generate HTML snippets
in various predefined formats.

Most of their appearance and behaviours are controlled by CSS, so you 
need to be using a stylesheet that is sensitive to these classes.
See the ExSite stylesheet for examples.  You can output some generic
compatible CSS using the CSS subroutine.

=head2 Boxes 

These are framed units of content.

=over 4

=item BasicBox 

a framed box with titlebar and content pane.

=item ScrollBox

a BasicBox with fixed height, and a scrollbar if the pane overflows.

=item CropBox 

a BasicBox with fixed height but no scrollbar.  If the pane overflows, 
it will be cropped.

=item FrameBox 

a pane that is framed by a rule.  The frame title is overlaid on the
top left corner of the frame rule.

=item MenuBox

a left-side vertical menu, with a content pane on the right.

=item AccordionBox 

A multi-paned box, in which each pane is collapsible/expandible,
like an accordion.

=item TabBox 

a tab bar, with a content pane beneath it.

=item DynTabBox 

a TabBox, with all tabs/panes preloaded and switchable using DHTML.

=back

=head2 Popups

=over 4

=item Popup

Pops up a box when a link is clicked.  The box covers content underneath it.

=item HelpPopup

Pops up a box when a link is clicked.  The box is inlined into the HTML, so
it does not cover content underneath it.

=item ToolTip

Works like Popup(), but works on a mouseover rather than a click.

=item Overlay

Works like Popup(), but overlays the whole page with a semi-opaque screen
so that the popup must be closed before you can do anything else.

=back

=head2 Other HTML Constructions

=over 4

=item DynList

a nested list, where the sub-lists are normally hidden.  There are
expand/hide icons beside the list items, which can be used to expand
the sub-lists, or hide them.

=item Button

a regular hyperlink, but styled to look like a button.

=item Icon

a graphic icon and label.

=back

=cut

#========================================================================

BEGIN {
    use Exporter   ();
    use vars       qw($VERSION @ISA @EXPORT @EXPORT_OK);
    $VERSION     = 1.00;
    @ISA         = qw(Exporter);
    @EXPORT = qw(   );
    @EXPORT_OK = qw(&BasicBox &ScrollBox &CropBox &FrameBox
		    &FrameBox &MenuBox &TabBox &DynTabBox &AccordionBox
		    &DynList &Icon &Button &Alert &Confirm
		    &HelpPopup &Popup &Overlay
		    &MailTo
		    &CSS
		    );
}

use vars qw($tag);

sub tag {
    return $tag if $tag;
    $tag = &get_obj("ML");
    return $tag;
}

#------------------------------------------------------------------------
# BasicBox - output a framed box containing content, with a titlebar
#
# title=>      titlebar text
# pane=>       main panel text
# width=>      width of the box
# height=>     height of the box
# padding=>    pane padding
# paneID=>     id of the pane, in case you want to use it with DHTML
# float=>      left|right, or 1 (left)|0 (none)
#
# WARNING: if your height is insufficient, the pane contents will run
# outside your box.
#------------------------------------------------------------------------

sub BasicBox {
    my %opt = @_;
    my %boxopt = ( 
		   head => $opt{title}, 
		   pane => $opt{pane},
		   paneClass=>$opt{paneClass},
		   paneID=>$opt{paneID},
		   );
    if ($opt{height}) { $boxopt{paneStyle} = "height:$opt{height};"; }
    if ($opt{width}) { 
	$boxopt{wrapperStyle} = "width:$opt{width};"; 
	#$boxopt{paneStyle} .= "width:$opt{width}";  
    }
    if ($opt{float}) {
	$boxopt{wrapperStyle} .= ($opt{float} =~ /left|right/i) ?
	    "float:$opt{float};" : "float:left;";
    }
    if ($opt{padding}) { $boxopt{paneStyle} .= "padding:$opt{padding}"; }
    return &box(%boxopt);
}

#------------------------------------------------------------------------
# ScrollBox - BasicBox with a preset height.  Overflow content
# is viewable using a scrollbar.
#
# Same parameters as BasicBox.
#------------------------------------------------------------------------

sub ScrollBox {
    return &BasicBox(@_,(paneClass => "scrollpane"));
}

#------------------------------------------------------------------------
# CropBox - BasicBox with a preset height, but no scrollbar.
# Overflow content is cropped.
#
# Same parameters as BasicBox.
#------------------------------------------------------------------------

sub CropBox {
    return &BasicBox(@_,(paneClass => "croppane"));
}

#------------------------------------------------------------------------
# FrameBox - framed box, with a title set into the frame
# NB.: not based on BasicBox.
#
# title=>      frame title text
# pane=>       main panel text
# bgcolor=>    you need to specific the general background color, if not white
# color=>      frame color (#ccc for white bgcolor, else defaults to bgcolor)
# style=>      solid,inset,outset,groove
# width=>      width of overall box
# thickness=>  width of the frame (in pixels)
# float=>      left|right, or 1 (left)|0 (none)
#------------------------------------------------------------------------

sub FrameBox {
    my %opt = @_;
    my (%boxopt, %headopt, %paneopt);

    my $title = $opt{title} || "Info";
    $headopt{"background-color"} = $opt{bgcolor} || $config{bgcolor} || "white";
    $paneopt{"border-color"} = $opt{color} || 
	($headopt{"background-color"} eq "white" ? "#ccc" : "black");
    $paneopt{"border-style"} = $opt{style} || "solid";
    if ($opt{width}) {
	$boxopt{width} = $opt{width};
    }
    $paneopt{"border-width"} = ($opt{style} eq "groove") ? 2 : ($opt{thickness} || 1);
    if ($opt{float}) {
	$boxopt{float} = ($opt{float} =~ /left|right/i) ?
	    "float:$opt{float};" : "float:left;";
    }

    my @styles = map { "$_:$boxopt{$_}" } keys %boxopt;
    my $boxstyle = join("; ",@styles);
    @styles = map { "$_:$headopt{$_}" } keys %headopt;
    my $headstyle = join("; ",@styles);
    @styles = map { "$_:$paneopt{$_}" } keys %paneopt;
    my $panestyle = join("; ",@styles);

    my $head = &tag->div(&tag->span($title,{class=>"framehead",style=>$headstyle}),
			 {class=>"framehead"});
    my $pane = &tag->div($opt{pane},{class=>"framepane",style=>$panestyle});

    my $out = &tag->Comment("begin FrameBox");
    $out .= &tag->div( $head.$pane, { class=>"framebox", style=>$boxstyle } );
    $out .= &tag->Comment("end FrameBox");
    return $out;
}

#------------------------------------------------------------------------
# MenuBox - static menu box;  left hand side is a menu of options,
#           right-hand side is the box pane.
#
# This element must be regenerated in its entirety when another menu
# item is selected.
#
# tabs=>        arrayref of menu labels, or menu hashrefs {label=>,url=>}
# menu=>        same as above
# thistab=>     index of the active menu item (starting from 0)
# pane=>        HTML contents of tab pane
# pane_id=>     ID of the pane DIV
# width=>       width of the pane
# height=>      height of the pane (auto scrolls, if necessary)
#
# WARNING: if you pass complete links, instead of menu hashrefs, you will
# not get highlighting of thistab.
#------------------------------------------------------------------------

sub MenuBox {
    my %opt = @_;
    if (! exists $opt{tabs}) { $opt{tabs} = $opt{menu}; }
    return &Entity("p",{class=>"error"},"Invalid MenuBox") 
	if (ref $opt{tabs} ne "ARRAY");

    # Shares a lot of logic with TabBox, below, which is why you'll 
    # references to tabs in the code.  A "tab" in this case is a menu 
    # item.

    my ($tabbar,$tabpane);

    my $ntab = scalar @{$opt{tabs}};

    # build the menu bar

    for (my $itab=0; $itab < $ntab; $itab++) {
	my $ttab = ++$idcount{tab};
	my $id = (defined $opt{thistab} && ($itab == $opt{thistab})) ? 
	    "thistab" : "tab$ttab";
	my $tablabel = ref $opt{tabs}[$itab] eq "HASH" ?
	    &tag->a($opt{tabs}[$itab]{label},
		    { href=>$opt{tabs}[$itab]{url}, id=>$id }) :
#	    "<a href=\"$opt{tabs}[$itab]{url}\" id=\"$id\">$opt{tabs}[$itab]{label}</a>" :
	    $opt{tabs}[$itab];
	$tabbar .= "$tablabel\n";
    }

    # build the tab pane

    if ($opt{pane}) {
	my $tpane = ++$idcount{tabpane};
	my %paneopt = (class=>"boxpane");
	$paneopt{id} = $opt{pane_id} ? $opt{pane_id} : "tabpane$tpane";
	if ($opt{height}) {
	    $paneopt{style} = "height:$opt{height}";
	}
	$tabpane = &tag->div($opt{pane},\%paneopt);
    }

    my %menuboxopt = (class=>"menubox");
    if ($opt{width}) { $menuboxopt{style} = "width:$opt{width};"; }

    my $vmenu = &tag->td($tabbar,{valign=>"top",class=>"menubox_vmenu"});
    my $pane = &tag->td($tabpane,{valign=>"top"});
    my $menutab = &tag->table(
			      &tag->tr($vmenu.$pane),
			      { class=>"menubox", 
				cellspacing=>0, 
				cellpadding=>0, 
				border=>0 }
			      );


    my $out = &tag->Comment("begin MenuBox");
    $out .= &tag->div($menutab,\%menuboxopt);
    $out .= &tag->Comment("end MenuBox");
    return $out;
}

#------------------------------------------------------------------------
# TabBox - static tab box
#
# This element must be regenerated in its entirety when another tab is
# selected.
#
# tabs=>        arrayref of tab labels, or tab hashrefs {label=>,url=>}
# thistab=>     index of the active tab (starting from 0)
# pane=>        HTML contents of tab pane
# pane_id=>     ID of the pane DIV
# width=>       width of the pane
# tabwidth=>    width of the tabs (default is 125px)
# tabfill=>     a string to right-fill the tabbar with
# height=>      height of the pane (auto scrolls, if necessary)
#
# WARNING: if you explicitly encode your menu lables, you will not get
# highlighting of the current tab.
#------------------------------------------------------------------------

sub TabBox {
    my %opt = @_;
    return &tag->p("Invalid TabBox",{class=>"error"}) 
	if (ref $opt{tabs} ne "ARRAY");

    my ($tabbar,$tabpane,$ttab,$id,$class,$tablabel,$tabopt);

    my $ntab = scalar @{$opt{tabs}};

    # build the tab bar

    for (my $itab=0; $itab < $ntab; $itab++) {
	$ttab = ++$idcount{tab};
	$id = "tab$ttab";
	if ($itab == $opt{thistab}) {
	    $class = "thistab";
	    $active_tab = $id;
	}
	else {
	    $class = "tab";
	}
	$class = ($itab == $opt{thistab}) ? "thistab" : "tab";
	$tablabel = ref $opt{tabs}[$itab] eq "HASH" ?
	    &tag->a($opt{tabs}[$itab]{label},{href=>$opt{tabs}[$itab]{url}}) :
	    $opt{tabs}[$itab];
	$tabopt = {class=>$class,id=>$id};
	if ($opt{tabwidth}) { $tabopt->{style} = "width:$opt{tabwidth}";}
	$tabbar .= &tag->td($tablabel,$tabopt);
	if ($itab != $ntab-1) {
	    # add a small gap between the tabs
	    $tabbar .= &tag->td(undef,{class=>"tabgap"});
	}
    }
    $tabbar .= &tag->td($opt{tabfill} || "&nbsp;",{class=>"notab"});
    $tabbar = &tag->tr($tabbar,{class=>"tabbar"});

    # build the tab pane

    if ($opt{pane}) {
	my $tpane = ++$idcount{tabpane};
	my $id = $opt{pane_id} ? $opt{pane_id} : "tabpane$tpane";
	my $span = $ntab * 2;
	
	my %panedivopt = ( class=>"tabpane", id=>$id );
	if ($opt{height}) { $panedivopt{style} = "height:$opt{height}"; }
	my $panediv = &tag->div($opt{pane},\%panedivopt);

	$tabpane = &tag->tr(
			    &tag->td($panediv,{class=>"tabpane",colspan=>$span}),
			    {class=>"tabpane"}
			    );
    }

    my %wrapopt = ( class=>"tabbox" );
    if ($opt{width}) { $wrapopt{style} = "width:$opt{width};"; }

    my $out = &tag->Comment("begin TabBox");
    $out .= &tag->div(
		      &tag->table($tabbar.$tabpane,
				  {class=>"tabbox",cellspacing=>0,
				   cellpadding=>0,border=>0}),
		      \%wrapopt);
    $out .= &tag->Comment("end TabBox");
    return $out;
}

#------------------------------------------------------------------------
# DynTabBox (DHTML version of TabBox)
#
# This version of the tabbox uses DHTML to do the tab switching
# and pane loading automatically.
#
# tabs=>        array of tab labels
# panes=>       array of pane contents
# thistab=>     index of the active tab (starting from 0)
# pane_id=>     ID of the pane DIV
# width=>       width of the pane
# height=>      height of the pane (auto scrolls, if necessary)
#------------------------------------------------------------------------

sub DynTabBox {
    my %opt = @_;
    return &tag->p("Invalid TabBox",{class=>"error"}) 
	if (ref $opt{tabs} ne "ARRAY");
    my (@tabs,@panes,$style,$selected_tab,$ipane);
    $ipane = 0;
    my $tid = $idcount{tab};
    foreach my $tablabel (@{$opt{tabs}}) {
	$tid++;
	my $pane = shift @{$opt{panes}};
	push @tabs, &tag->div($tablabel,{class=>"tablabel",onclick=>"loadtab($tid,this)"});

	# We use overflow rather than display to show/hide the panes, because
	# undisplayed panes can result in obscure window.onload errors in some
	# browsers.  It seems like undisplayed content is not a 1st class 
	# citizen of the DOM, and may not be found or initialized in some
	# Javascript apps.  This approach works slightly better, although not 
	# a whole lot...

	if ($ipane == $opt{thistab}) {
#	    $display = "block";
	    $style = "overflow:visible;height:auto;";
	    $selected_tab = $tid;
	}
	else {
	    $style = "overflow:hidden;height:0px;";
#	    $display = "none";
	}
	push @panes, &tag->div($pane,{id=>"dyntabpane$tid",style=>$style});
	$ipane++;
    }

    # basic TabBox

    my $out = &TabBox(
		      tabs=>\@tabs,
		      pane=> join("\n",@panes),
		      thistab=>$opt{thistab},
		      width=>$opt{width},
		      height=>$opt{height},
		      tabwidth=>$opt{tabwidth},
		      tabfill=>$opt{tabfill},
		      );

    # tab switching code

    if (! $script_loaded{DynTabBox}) { 
	$out .= &tag->script("
var current_tab = $selected_tab;
function loadtab(pane,tab) {
    // hide old pane
    var loc = document.getElementById(\"dyntabpane\"+current_tab);
    loc.style.overflow = \"hidden\";
    loc.style.height = \"0px\";
    // show new pane
    var newtabpane = \"dyntabpane\" + pane;
    var loc = document.getElementById(newtabpane);
    loc.style.overflow = \"visible\";
    loc.style.height = \"auto\";
    // shade old tab
    var loc = document.getElementById(\"tab\"+current_tab);
    loc.className = \"tab\";
    // highlight new tab
    tab.parentNode.className = \"thistab\";
    // save new current tab
    current_tab = pane;
}
");
	$script_loaded{DynTabBox} = 1;
    }

    return $out;
}

#------------------------------------------------------------------------
# AccordionBox - exploding list of boxes (uses DHTML)
#
# An AccordionBox is a sequence of boxes arranged in a stack.  Each box
# can be expanded/shrunk.
#   titles=>[array of box titles]
#   panes=>[array of pane contents]
#   states=>[array of booleans, indicating whether the given pane is open]
#------------------------------------------------------------------------

sub AccordionBox {
    my %opt = @_;

    my $out;
    if (! $script_loaded{AccordionBox}) { 
	$out .= &tag->script("function accordion_expand(id) {
    var controlid = \"accordioncontrol\" + id;
    var loc = document.getElementById(controlid);
    loc.innerHTML = \"<a class='accordioncontrol' href='javascript:accordion_contract(\"+id+\")'><img src=\\\"$config{server}{HTMLpath}/_ExSite/images/minus.gif\\\"></a>\";
    var paneid = \"accordionpane\" + id;
    var loc=document.getElementById(paneid);
    loc.style.display=\"block\";
}
function accordion_contract(id) {
    var controlid = \"accordioncontrol\" + id;
    var loc = document.getElementById(controlid);
    loc.innerHTML = \"<a class='accordioncontrol' href='javascript:accordion_expand(\"+id+\")'><img src=\\\"$config{server}{HTMLpath}/_ExSite/images/plus.gif\\\"></a>\";
    var paneid = \"accordionpane\" + id;
    var loc=document.getElementById(paneid);
    loc.style.display=\"none\";
}\n",
			     {language=>"javascript"});
	$script_loaded{AccordionBox} = 1;
    }

    ### ML markup needed

    my %boxopt = ( class=>"accordionbox" );
    if ($opt{width}) {
	$boxopt{style} = "width:$opt{width}";
    }

    foreach my $title (@{$opt{titles}}) {
	my $paneid = ++$idcount{accordionpane};
	my $pane = shift @{$opt{panes}};
	my $state = 1;
	if (ref $opt{states} eq "ARRAY") {
	    $state = shift @{$opt{states}};
	}
	my $img = $state ? "minus.gif" : "plus.gif";
	my %paneopt = ( class=>"accordionpane",id=>"accordionpane$paneid" );
	if (! $state) { $paneopt{style} = "display:none"; }
	my $action = $state ? "contract" : "expand";

	
	$out .= &tag->tr(
			 &tag->th(
				  &tag->span(
					     &tag->a(
						     &tag->img(undef,{src=>"$config{server}{HTMLpath}/_ExSite/images/$img"}),
						     {class=>"accordioncontrol",href=>"javascript:accordion_$action($paneid)"}
						     ),
					     {class=>"accordioncontrol",id=>"accordioncontrol$paneid"}
					     )."&nbsp;$title"
				  )
			 );

	$out .= &tag->tr( &tag->td( &tag->div($pane,\%paneopt) ) );
    }

    return &tag->Comment("begin AccordionBox").
	&tag->table($out,\%boxopt).
	&tag->Comment("end AccordionBox");
}

#------------------------------------------------------------------------
# DynList - exploding list (uses DHTML)
#
# This supports nested sublists, to any depth.
#
# A list is a sequence of hashes, each describing a list element:
#   text=>full text of the list element
#   label=>anchor text of the list element, or text string if no url
#   url=>destination of list element link
#   target=>where list element link opens
#   open=>1|0 determines whether this list element is opened
#   sublist=>referece to another list, nested under this element
#   class=>class of the list item wrapper DIV
#   prefix=>HTML that should be positioned before the list control
#------------------------------------------------------------------------

sub DynList {

    my $out;
    if (! $script_loaded{DynList}) { 
	$out .= &tag->script("function list_expand(id) {
    var controlid = \"listcontrol_\" + id;
    var sublist = \"list_\" + id;
    var loc = document.getElementById(controlid);
    loc.innerHTML = \"<a class='listcontrol' href='javascript:list_contract(\"+id+\")'><img src=\\\"$config{server}{HTMLpath}/_ExSite/images/darr.gif\\\"></a>\";
    var loc=document.getElementById(sublist);
    loc.style.display=\"block\";
}
function list_contract(id) {
    var controlid = \"listcontrol_\" + id;
    var sublist = \"list_\" + id;
    var loc = document.getElementById(controlid);
    loc.innerHTML = \"<a class='listcontrol' href='javascript:list_expand(\"+id+\")'><img src=\\\"$config{server}{HTMLpath}/_ExSite/images/rarr.gif\\\"></a>\";
    var loc=document.getElementById(sublist);
    loc.style.display=\"none\";
}\n",
			     {language=>"javascript"});
	$script_loaded{DynList} = 1;
    }

    foreach my $item (@_) {
	$out .= &DynListItem($item);
    }
    my $tag = &tag;
    return $tag->Comment("begin DynList").
	$tag->div($out,{class=>"DynList"}).
	$tag->Comment("end DynList");
}

sub DynListItem {
    my $opt = shift;
    my $tag = &tag;
    return $tag->p("Invalid list item",{class=>"error"}) 
	if (ref $opt ne "HASH");
    my %opt = %$opt;
    my ($next, $script, $itemtext, $img, $target, $sublist_style);
    $list_count++;  # global
    $next = $list_count + 1;
    my $class = $opt->{class} || "listitem";

    my $out;
    if ($opt{open}) {
	$script = "list_contract";
	$sublist_style = {};
    }
    else {
	$script = "list_expand";
	$sublist_style = {style=>"display:none;"};
    }

    my %aopt = ( href=>$opt{url}, class=>"listlink" );
    if ($opt{target}) { $aopt{target} = $opt{target}; }
    $itemtext = $opt{url} ? $tag->a($opt{label},\%aopt) : $opt{label};

    my $dimg = $tag->img(undef,{height=>9,width=>9,border=>0,src=>"$config{server}{HTMLpath}/_ExSite/images/darr.gif"});
    my $rimg = $tag->img(undef,{height=>9,width=>9,border=>0,src=>"$config{server}{HTMLpath}/_ExSite/images/rarr.gif"});

    if ($opt{sublist} || $opt{text}) {

	# output list control elements
	$out .= 
	    $opt{prefix}.
	    $tag->span(
		       $tag->a(
			       $opt{open} ? $dimg : $rimg,
			       {class=>"listcontrol",href=>"javascript:$script($list_count)"}
			       ),
		       { id=>"listcontrol_$list_count" }
		       );

	# now the item text
	$out .= $itemtext;

	# now the sublist
	my $subout;
	my $count = $list_count; 
	if ($opt{sublist}) {
	    foreach my $item (@{$opt{sublist}}) {
		$subout .= &DynListItem($item);
	    }
	}
	$out .= $tag->div( 
			   $opt{text}.$subout,
			   { class=>"list", id=>"list_$count", %$sublist_style }
			   );
    }
    else {
	$out .= 
	    $opt{prefix}.
	    $tag->span(
		       $tag->img(undef,{height=>9,width=>9,border=>0,src=>"$config{server}{HTMLpath}/_ExSite/images/null.gif"}),
		       { id=>"listcontrol_$list_count" }
		       );
	$out .= $itemtext;
    }

    return $tag->div($out,{class=>$class});
}

#------------------------------------------------------------------------
# Popup
#
# Displays a CSS popup.  Uses absolute positioning, so it overlays
# content beneath it.  May have problems in forms, or mixed with 
# embedded objects such as Flash.
#
# pane=>      the contents of the popup
# label=>     optional anchor for the popup link/button
# width=>     optional width for the popup
# height=>    optional height for the popup
# top=>       offset of top of popup
# right=>     offset of right of popup
# left=>      offset of left of popup
#
# Offsets are relative to the anchor label.
#------------------------------------------------------------------------

sub Popup {
    my %opt = @_;
    my $out;
    my $tag = &tag();
    my $tpop = ++$idcount{Popup};
    my $pop_id = "popup$tpop";
    my %popopt = (class=>"popup", id=>$pop_id);
    if ($opt{width}) { $popopt{style} = "width:$opt{width}px;"; }
    if ($opt{height}) { $popopt{style} .= "height:$opt{height}px;"; }
    if ($opt{top}) { $popopt{style} .= "top:$opt{top}px;"; }
    if ($opt{right}) { $popopt{style} .= "right:$opt{right}px;"; }
    elsif ($opt{left}) { $popopt{style} .= "left:$opt{left}px;"; }
    if (! $script_loaded{Popup}) { 
	$out .= $tag->script("function showpopup(popid) {
    var loc = document.getElementById(popid);
    loc.style.display=\"block\";
}
function hidepopup(popid) {
    var loc = document.getElementById(popid);
    loc.style.display=\"none\";
}");
	$script_loaded{Popup} = 1;
    }
    my $close = $tag->a("close",
			{href=>"javascript:hidepopup('$pop_id')",
			 class=>"popup_close"}
			);

    my $label = $opt{label} || "Info";
    $out .= $tag->span(
		       $tag->a($label,{class=>"popup",href=>"javascript:showpopup('$pop_id')"}).
		       $tag->div($close.$opt{pane}, \%popopt),
		       { class=>"popup_p" }
		       );
    return $out;
}

#------------------------------------------------------------------------
# ToolTip
#
# Displays a rollover CSS popup.  Otherwise, works like Popup()
#------------------------------------------------------------------------

sub ToolTip {
    my %opt = @_;
    my $out;
    my $tag = &tag();
    my $tpop = ++$idcount{Popup};
    my $pop_id = "popup$tpop";
    my %popopt = (class=>"popup", id=>$pop_id);
    if ($opt{width}) { $popopt{style} = "width:$opt{width}px;"; }
    if ($opt{height}) { $popopt{style} .= "height:$opt{height}px;"; }
    $opt{top} or $opt{top} = 2;
    $opt{right} or $opt{right} = 5;
    $popopt{style} .= "top:$opt{top}px;";
    $popopt{style} .= "right:$opt{right}px;";
    if (! $script_loaded{Popup}) { 
	$out .= $tag->script("function showpopup(popid) {
    var loc = document.getElementById(popid);
    loc.style.display=\"block\";
}
function hidepopup(popid) {
    var loc = document.getElementById(popid);
    loc.style.display=\"none\";
}");
	$script_loaded{Popup} = 1;
    }
    my $label = $opt{label} || "Info";
    $out .=
	$tag->span(
		   $tag->span($label,{class=>"tooltip_hotspot",onmouseover=>"javascript:showpopup('$pop_id')",onmouseout=>"javascript:hidepopup('$pop_id')"}).
		   $tag->div($opt{pane}, \%popopt),
		   { class=>"popup_p" }
		   );
    return $out;
}

#------------------------------------------------------------------------
# Overlay
#
# Displays a CSS overlay - like a popup, but it takes over the whole
# screen, and fades out the background page.  (Doesn't work in IE6 due
# poor implementation of transparent PNGs and fixed positioning.)
#
# pane=>      the contents of the popup
# label=>     optional anchor for the popup link/button
# width=>     optional width for the popup
# height=>    optional height for the popup
#------------------------------------------------------------------------

sub Overlay {
    my %opt = @_;
    my $out;
    my $tag = &tag();
    my $tpop = ++$idcount{Popup};
    my $pop_id = "popup$tpop";
    my %olopt = (class=>"overlay", id=>$pop_id);
    my %popopt = ();
    if ($opt{width}) { $popopt{style} = "width:$opt{width}px;"; }
    if ($opt{height}) { $popopt{style} .= "height:$opt{height}px;"; }
    if (! $script_loaded{Popup}) { 
	$out .= $tag->script("function showpopup(popid) {
    var loc = document.getElementById(popid);
    loc.style.display=\"block\";
}
function hidepopup(popid) {
    var loc = document.getElementById(popid);
    loc.style.display=\"none\";
}");
	$script_loaded{Popup} = 1;
    }
    my $close = $tag->a($msg{close},
			{href=>"javascript:hidepopup('$pop_id')",
			 class=>"popup_close"}
			);

    my $label = $opt{label} || $msg{Info};
    $out .= $tag->a($label,{href=>"javascript:showpopup('$pop_id')"});
    $out .= $tag->div(
		      $tag->div($close.$opt{pane},\%popopt),
		      \%olopt
		       );
    return $out;
}

#------------------------------------------------------------------------
# HelpPopup
#
# Displays a CSS popup with some help text when you click a help link.
# Uses relative positioning, so it will push other elements aside, 
# which is useful when mixing with unfriendly elements like some form
# controls, Flash objects, etc.
#
# message=>   the help message
# label=>     optional anchor for the help link
# width=>     optional width for the popup
#------------------------------------------------------------------------

sub HelpPopup {
    my %opt = @_;
    my $out;
    my $tag = &tag();
    my $thelp = ++$idcount{HelpPopup};
    my $help_id = "help$thelp";
    my %helpopt = (class=>"help_popup", id=>$help_id);
    if ($opt{width}) { $helpopt{style} = "width:$opt{width}px;"; }
    if (! $script_loaded{HelpPopup}) { 
	$out .= &tag->script("function showhelp(helpid) {
    var loc = document.getElementById(helpid);
    loc.style.display=\"block\";
}
function hidehelp(helpid) {
    var loc = document.getElementById(helpid);
    loc.style.display=\"none\";
}");
	$script_loaded{HelpPopup} = 1;
    }
    my $br = &tag->br();
    $opt{message} =~ s/\n\n/$br$br\n/g;
    my $label = $opt{label} || "?";
    my $close = $tag->a("$msg{close}&nbsp;&nbsp;&times;",
			{href=>"javascript:hidehelp('$help_id')",
			 class=>"popup_close"}
			);
    $out .= &tag->span($label,{class=>"help_hotspot", onclick=>"showhelp('$help_id')"});
    $out .= &tag->div($close.$opt{message},
		      \%helpopt);
    return $out;
}

#------------------------------------------------------------------------
# Icon
#
# label=>     icon text
# img=>       image url
# img_x=>     image height
# img_y=>     image width
# img_size=>  image height & width
# url=>       url to link to
# target=>    target window/frame
#------------------------------------------------------------------------

sub Icon {
    my %opt = @_;
    my %imgopt = ( src=>$opt{img}, border=>0 );
    my %aopt = ( href=>$opt{url} );
    if ($opt{target}) { $aopt{target} = $opt{target}; }
    my $w = $opt{img_x} || $opt{img_size};
    my $h = $opt{img_y} || $opt{img_size};
    if ($h && $w) {
	$imgopt{width} = $w;
	$imgopt{height} = $h;
    }
    return &tag->div(
		     &tag->a(
			     &tag->img(undef,\%imgopt).&tag->br().$opt{label},
			     \%aopt
			     ),
		     {class=>"icon"});
}

#------------------------------------------------------------------------
# Button
#
# label=>    button text
# url=>      url to link to
# type=>     link|button (link is default)
# target=>   target window/frame
# class=>    additional class to apply to the button
#------------------------------------------------------------------------

sub Button {
    my %opt = @_;
    my $class = "button";
    if ($opt{class}) { $class .= " $opt{class}"; }
    if ($opt{type} eq "button") {
	my %aopt = ( class=>$class, type=>"button", onClick=>"javascript:window.location.href='$opt{url}'" );
	return &tag->button($opt{label},\%aopt);
    }
    else {
	my %aopt = ( href=>$opt{url}, class=>"button" );
	if ($opt{target}) { $aopt{target} = $opt{target}; }
	return &tag->a($opt{label},\%aopt);
    }
}

#------------------------------------------------------------------------
# Alert - creates a link that generates a Javascript alert box 
# containing arbitrary text.
#
# label=>    button text
# button=>   1, if link should be presented as a button
# message=>  text of alert message
#------------------------------------------------------------------------

sub Alert {
    my %opt = @_;
    # escape quotes
    $opt{message} =~ s/\"/\\\"/g;
    $opt{message} =~ s/\'/\\\'/g;
    # insert linebreaks, if not already done
    if ($opt{message} !~ /\\n/ && $opt{message} =~ /\n/) {
	$opt{message} =~ s/\r?\n/\\n/g;
    }
    if ($opt{button}) {
	return &Button(label=>$opt{label},
		       url=>"javascript:alert('$opt{message}')");
    }
    else {
	return &tag->a($opt{label},{ href=> "javascript:alert('$opt{message}')" });
    }
}

#------------------------------------------------------------------------
# Confirm - creates a link that requires a confirmation before visiting it
#
# label=>    button text
# button=>   1, if link should be presented as a button
# message=>  text of alert message
# url =>     url to link to
#------------------------------------------------------------------------

sub Confirm {
    my %opt = @_;
    my $out;
    # escape quotes
    $opt{message} =~ s/\"/\\\"/g;
    $opt{message} =~ s/\'/\\\'/g;
    # insert linebreaks, if not already done
    if ($opt{message} !~ /\\n/ && $opt{message} =~ /\n/) {
	$opt{message} =~ s/\r?\n/\\n/g;
    }
    if (! $script_loaded{Confirm}) {
	$out = $tag->script("function confirm_custom(message, url) {
    if(confirm(message)) {
	window.location=url;
    }
    return; // false;
}");
    }

    if ($opt{button}) {
	$out .= &Button(label=>$opt{label},
			url=>"javascript:confirm_custom('$opt{message}','$opt{url}')");
    }
    else {
	$out .= &tag->a($opt{label},{ href=> "javascript:confirm_custom('$opt{message}','$opt{url}')" });
    }
    return $out;
}

#------------------------------------------------------------------------
# MailTo - builds mailto: links that are spambot-resistant
#
# label => anchor text
# email => target email address
#------------------------------------------------------------------------

sub MailTo {
    my ($email,$label) = @_;
    my ($user, $domain) = split(/\@/,$email,2);
    my $anchor = $label ? "'$label'" : "addr";
    my $out = &tag->script("
<!-- //
user = \"$user\";
domain = \"$domain\";
addr = user + '\@' + domain;
document.write('<a href=\"mai' + 'lto:' + addr + '\">' + $anchor + '</a>');
// -->
",
			   { language=>"JavaScript" }
			   );
    $out .= &tag->noscript("$user AT $domain");
}

#------------------------------------------------------------------------
# CSS
#
# This can be used to auto-insert the necessary CSS if your stylesheet
# does not have it.
#------------------------------------------------------------------------

sub CSS {
    return &tag->link(undef,
		      {rel=>"stylesheet", type=>"text/css",
		       href=>"$config{server}{HTMLpath}/_ExSite/css/HTML.css"});
}

#------------------------------------------------------------------------
# Internal box-building function
#
# wrapperClass
# wrapperStyle
# headClass
# headStyle
# head
# paneClass
# paneStyle
# pane
#------------------------------------------------------------------------

sub box {
    my %opt = @_;

    my ($wrapperClass, $wrapperStyle, $headClass, $headStyle,$paneClass,
	$paneStyle, $paneID, $head, $pane);

    my %headopt = ( class => $opt{headClass} || "boxhead" );
    my %paneopt = ( class => $opt{paneClass} || "boxpane" );
    my %wrapopt = ( class => $opt{wrapperClass} || "box" );

    if ($opt{wrapperStyle}) { $wrapopt{style} = $opt{wrapperStyle}; }
    if ($opt{headStyle})    { $headopt{style} = $opt{headStyle}; }
    if ($opt{paneStyle})    { $paneopt{style} = $opt{paneStyle}; }
    if ($opt{paneID})       { $paneopt{id}    = $opt{paneID}; }
    $head = $opt{head} ? &tag->div($opt{head},\%headopt) : "";
    $pane = &tag->div($opt{pane},\%paneopt);
    return &tag->Comment("begin box").
	&tag->div($head.$pane,\%wrapopt).
	&tag->Comment("end box");
}

#------------------------------------------------------------------------
# Internal javascript output function
#------------------------------------------------------------------------

### is this still used?

sub js {
    return undef if ($js_output);
    $js_output = 1;
    my $out = &tag->script("imgpath = '$config{server}{HTMLpath}/_ExSite/images'\n",
			   { type=>'text/javascript', src=>"$config{server}{HTMLpath}/_ExSite/js/misc.js"});
    return $out;
}

1;
