#!/usr/bin/perl
#========================================================================
# ExSite::Mail  --  miscellaneous E-Mail functions
# Author: Alan Douglas, Exware Solutions, April 2003
#========================================================================
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::Mail;

use strict;
use ExSite::Config;
use ExSite::Misc;
use Net::SMTP;

BEGIN {
    use Exporter   ();
    use vars       qw($VERSION @ISA @EXPORT);
    $VERSION     = 1.00;
    @ISA         = qw(Exporter);
    @EXPORT = qw(   &SendMail &SendMail_Multipart
		    );
}

# ExSite::Mail::send - multi-purpose email sending tool
# Options:
# to => scalar or array of email addresses
# from => email address
# subject => subject string
# body => email message, plain text or HTML
# attach => array of files to attach

sub send {
    my (%opt) = @_;
    if ($opt{attach}) {
	# email with attachments
	return &SendMail_Multipart(
				   $opt{to},
				   $opt{from},
				   $opt{subject},
				   $opt{body},
				   (ref $opt{attach} eq "ARRAY") ? @{$opt{attach}} : $opt{attach},
				   );
    }
    elsif ($opt{body} =~ /[^[:ascii:]]/ or $opt{subject} =~ /[^[:ascii:]]/) {
	# non-ascii email - use multipart for better encoding
	return &SendMail_Multipart(
				   $opt{to},
				   $opt{from},
				   $opt{subject},
				   $opt{body},
				   );
    }
    else {
	# simple ascii email
	return &SendMail(
			 $opt{to},
			 $opt{from},
			 $opt{subject},
			 $opt{body},
			 );
    }
}

sub SendMail {
    if ($config{mail}{method} eq "sendmail") {
	return SendMail_2(@_);
    }
    else {
	return SendMail_1(@_);
    }
}

sub SendMail_1{
   my ($to,$from,$subject,$body) = @_;
   my @lines;
   return 2 if ($to !~ /\w/);
 
   my $smtp = new Net::SMTP('localhost');
   $smtp->mail($from);
   # 20040416 MB - allow for multiple recipients
   my @to = split(/,/,$to);
   if ($config{mail}{bcc}) {
       push @to, $config{mail}{bcc};
   }
   $smtp->to(@to);
   $smtp->data;
   $smtp->datasend("To: $to\nFrom: $from\nSubject: $subject\n\n");
   @lines=split(/\n/,$body);
   foreach (@lines) {
      $_ =~ s/"/\\"/g;
      $_ =~ s/\n//g;
      $_ =~ s/\r//g;
      $smtp->datasend("$_\n");
   }
   $smtp->dataend;
   $smtp->quit;
   return 0;
}

sub SendMail_2 {
#----------------------------------------------------------------------
# SimpleSendMail: a simplified version of the SendMail function.
#                 It has also been adjusted to allow sending even from
#                 badly misconfigured mail servers
#
# Arguments:
#   to      - recipient email address
#   from    - sender email address
#   subject - email subject header
#   message - message body
#
# Returns 0 if it worked 
#         1 if not

    my ($to,$from,$subject,$message) = @_;
	
    if (!open (SENDMAIL, "|$config{mail}{sendmailpath} -f $from $to")) {
        return 1;
    }
    else {
        print SENDMAIL "To: $to\nFrom: $from\nSubject: $subject\n\n$message";
        close SENDMAIL;
        return 0;
    }
}

# Multipart email
sub SendMail_Multipart {
    # $tolist: comma delimited email address
    # $from: from address
    # $subject: subject
    # $body: email body
    # @files: attachments(an array of hash reference)
    #	each file attachment has 2 hash keys:
    #		filename => "myfilename.txt", # string
    #		data => <attachment>, # binary data
    my ($tolist,$from,$subject,$body,@files) = @_;
    return 2 if ($tolist !~ /\w/);
    return 2 if ($from !~ /^[\w](\.?[\w\-]+)*@([\w\-]+\.)+[A-Za-z]{2,4}$/);
    my @to = split /,/,$tolist;
    if ($config{mail}{bcc}) {
	push @to, $config{mail}{bcc};
    }
    require MIME::Entity;
    foreach my $to(@to){
	if ($to !~ /^[\w](\.?[\w\-]+)*@([\w\-]+\.)+[A-Za-z]{2,4}$/){
	    # Invalid email address
	    next;
	}
	# Build email
	if ($subject =~ /[[:^ascii:]]/) {
	    require MIME::Words;
	    $subject = &MIME::Words::encode_mimewords($subject,Encoding=>"Q",Charset=>$config{charset});
	}
	my $mime = MIME::Entity->build(
				       Type=>"multipart/mixed",
				       From=>$from,
				       To=>$to,
				       Subject=>$subject,
				       );
	my $part1 = MIME::Entity->build(
					Type=>"multipart/alternative",
					);
	my $html_message;
	my $text_message;
	# HTML tag checks
	if ($body !~ /<[^>]*>/gi){
	    # Send out plain text email
	    $text_message = $body;
	    $part1->attach(Type=>"text/plain; charset=\"$config{charset}\"",Data=>"$text_message");
	}
	else {
	    # Send out multipart email
	    $html_message = $body;
	    $text_message = &html_to_plaintext($html_message);
	    $part1->attach(
			   Type=>"text/plain; charset=\"$config{charset}\"",
			   Data=>$text_message,
			   );
	    $part1->attach(
			   Type=>"text/html; charset=\"$config{charset}\"",
			   Data=>$html_message,
			   );
	}
	$mime->add_part($part1);
	my $part2 = MIME::Entity->build(
					Type=>"multipart/mixed",
					);
	# Add file attachments
	my $include_attachment = 0;
	foreach my $file(@files){
	    next if !$file;
	    next if ref($file) !~ /hash/i;
	    next if (keys(%$file) < 1);
	    $include_attachment = 1;
	    $part2->attach(Filename=>$file->{filename},Type=>&MimeType($file->{filename}),Data=>$file->{data});
	}
	if ($include_attachment){
	    $mime->add_part($part2);
	}
	my $smtp = new Net::SMTP('localhost');
	$smtp->mail($from);
	$smtp->to($to);
	$smtp->data;
	$smtp->datasend($mime->stringify);
	$smtp->dataend;
	$smtp->quit;
    }
    return 0;
}
1;
