#!/usr/bin/perl
#-----------------------------------------------------------------------
#
#   Copyright 2001-2007 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::Object;

use strict;
use ExSite::Config;
use ExSite::Base;
use ExSite::Form;
use ExSite::Misc;
use ExSite::ObjectList;

use vars qw(@ISA);
@ISA = qw(ExSite::Base);

#=== OBJECT CREATION

sub new ($%) {
    my ($this,%opt) = @_;
    my $obj = {};
    my $class = ref($this) || $this;
    bless $obj, $class;
    $obj->initialize_object;
    $obj->setup(%opt);
    return $obj;
}

# set/return the database that stores the object
# you can use $obj->set("DB",$some_other_db) to use another DB

sub DB {
    my $this = shift;
    return $this->{DB} || $share{DB};
}

#=== IDENTIFICATION

# setup(%): standard protocol for data object setup is:
#
# type => table name (defaults to my_type())
# id => record ID
# data => datahash
#
# However, we support an older calling convention:
#
# setup($$), where the parameters can be:
# (type,id)
# (undef,id)
# (id)

sub setup {
    my $this = shift;
    if (scalar @_ <= 2 && scalar @_ > 0 && 
	$_[0] ne "id" && $_[0] =~ /^\w*$/ && 
	(! $_[1] || $_[1] =~ /^\d+$/)) {
	# old style args
	return $this->setup_old(@_);
    }
    my %opt = @_;
    $this->{type} = $opt{type} || $this->my_type;
    $this->{id} = $opt{id};
    $this->{data} = $opt{data};
    if ($this->{data} && ! $this->{id}) {
	$this->{id} = $this->{data}{$this->key()};
    }
    # flag whether we need to save data - assumes data is a clean record,
    # otherwise you should manually set the dirty flag
    $this->{dirty} = 0;
    return;
}

sub setup_old {
    my $this = shift;
    my $arg = shift;
    if ($arg =~ /[a-z]/) {
	# setup($type,$id)
	$this->{type} = $arg;
	$this->{id} = shift;
    }
    elsif (! defined $arg) {
	# setup(undef,$id)
	$this->{id} = shift;
    }
    elsif ($arg =~ /^\d+/) {
	# setup($id)
	$this->{id} = $arg;
    }
    $this->{data} = undef;
    # flag whether we need to save data
    $this->{dirty} = 0;
    return;
}

sub my_type { return; }    # generic object has no preset type

sub type {
    my $this = shift;
    return $this->{type} || $this->my_type();
}

sub id {
    my $this = shift;
    # fetch the id
    if ($this->{id}) { 
	return $this->{id};
    }
    elsif ($this->{data}) {
	$this->{id} = $this->{data}{$this->key()};
	return $this->{id};
    }
    return undef;
}

sub name {
    my $this = shift;
    if ($this->defined()) {
	return $this->DB()->title($this->{type},$this->{id},$this->{data});
    }
    return "n/a";
}

sub key {
    my $this = shift;
    my $type = $this->type();
    if ($type) {
	return $this->DB()->{map}->get_table_attr($type,"key");
    }
    return undef;
}

# ok : returns true if object is fully specified

sub ok {
    my $this = shift;
    $this->{status} = undef;
    # return true if object has an identity
    my $type = $this->type();
    if ($type) {
	$this->{status} = "Undefined $type object." if (! $this->id());
    }
    else {
	$this->{status} = "Undefined object type.";
    }
    return (! $this->{status});
}

# exists : returns true if the object as specified is found in the DB

sub exists {
    my $this = shift;
    if ($this->ok) {
	# completely defined record - load it if necessary
	if (! defined $this->{data}) { $this->load(@_); }
	return (ref $this->{data}) eq "HASH";
    }
    return undef;
}

# defined : returns true if the object's data is present in the object

sub defined {
    my $this = shift;
    return $this->exists() || (ref $this->{data}) eq "HASH";
}

sub loaded { return &defined(@_); } # deprecated

#=== DATA

sub get {
    my ($this,$key) = @_;
    return defined $key ? $this->SUPER::get($key) : $this->{data};
}

sub getdata {
    my ($this,$key) = @_;
    if ($this->defined) {
	return $this->{data}{$key};
    }
    return undef;
}

sub setdata {
    my ($this,$key,$val) = @_;
    # NB: if not defined, this will define the object
    $this->{data}{$key} = $val;
    $this->{dirty} = 1;
}

sub showdata {
    my ($this,$key) = @_;
    if ($this->defined) {
	my $auth = $this->showauth($key);
	my $show = $auth ? "show_data_noauth" : "show_data";
	return $this->DB()->$show($this->{type},
				  $key,
				  $this->getdata($key),
				  $this->{id});
    }
    return undef;
}

sub showauth { 
    my ($this,$key) = @_;
    # default behaviour is to use the dbmap rules
    return 0;
}

sub dirty {
    my $this = shift;
    return $this->{dirty};
}

#=== DISPLAY

# show : display the object for an end-viewer
#
# There is no default show method, show we simply use the report.

sub show {
    my ($this,%opt) = @_;
    return $this->report(%opt);
}

# report : technical or administrator view of the object

sub report {
    my ($this,%opt) = @_;
    if ($this->ok) {
	if ($this->defined()) {
	    return $this->DB()->report({ table=>$this->{type},
					 data=>[$this->{data}], 
					 columns=>"full|brief",
					 type=>"full", 
					 title=>$this->name,
					 %opt });
	}
	else {
	    my $name = ucfirst $this->name;
	    return $this->error("$name does not exist.");
	}
    }
    return $this->error($this->{status});
}

#=== DB

# load record : can pass an ID or a datahash

sub load {
    my ($this,$data) = @_;
    my $id;
    if (ref $data ne "HASH") { $id = $data; $data = undef; }
    $this->{data} = $data;
    my $type = $this->type;
    if (! $type) { $this->error("Undefined object type."); return 0; }
    if (! $id) { $id = $this->id(); }
    if ($id) { 
	$this->{data} = $this->DB()->fetch($type,$id); 
    }
    else { 
	$id = $this->id(); 
    }
    if (! $id && $this->{data}) { 
	$this->error("Mis-configured $type object."); 
	return undef;
    }
    elsif (! $id || ! $this->{data}) { 
	$this->error("Undefined $type object."); 
	return undef;
    }
    return $this->{data};
}

# modify database directly

sub insert {
    my $this = shift;
    my $type = $this->type;
    my @err;
    if (! $type) {
	push @err, "Undefined object type.";
    }
    my $data = scalar @_ > 0 ? {@_} : $this->{data};

    if ($data->{$this->key()}) {
	# hm... object already seems to have an identity
	push @err, "cannot insert object that already has a record";
	return undef;
    }
    push @err, $this->validate();
    if (@err > 0) {
	$this->error(@err);
	return undef;
    }

    $this->{id} = $this->DB()->insert($type,$data);
    if ($this->{id}) {
	# successful insert; update our data with the id
	$this->{data}{$this->key()} = $this->{id};
	$this->{dirty} = 0;
    }
    return $this->{id};
}

sub update {
    my $this = shift;
    my $type = $this->type;
    my @err;
    if (! $type) {
	push @err, "Undefined object type.";
	return undef;
    }
    my $data = scalar @_ > 0 ? {@_} : $this->{data};
    push @err, $this->validate();
    if (@err > 0) {
	$this->error(@err);
	return undef;  # invalid update
    }

    # update record
    if ($this->dirty()) {
	$this->DB()->update($type,$data);
	$this->{dirty} = 0;
	return 1;   # record changed
    }
    return 0;   # valid, but no changes
}

sub action {
    my $this = shift;
    return $this->id() ? "update" : "insert";
}

sub save {
    my $this = shift;
    return $this->error("No data to save.") if (ref $this->{data} ne "HASH"); 
    my $action = $this->action;
    return $this->$action();
}

# misc DB ops

sub validate {
    my $this = shift;
    return ($msg{"no data"}) if (ref $this->{data} ne "HASH");
    my $action = $this->action;

    if ($action eq "update") {
	return $this->DB->validate_record_update($this->{type},$this->{data});
    }
    else {
	return $this->DB->validate_record_insert($this->{type},$this->{data});
    }
}

# prepare_image : multipurpose function to prep (scale) images before insertion
# $key = source column name
# $key2 = destination column name (useful for thumbnail fields)
# $dim = max width/height
# $size = max filesize

sub prepare_image {
    my ($this,$key,$dim,$size,$key2) = @_;
    return undef if (! $this->{data}{$key});
    $key2 or $key2 = $key;
    my ($fname,$fdata) = split /$config{form}{sepchar}/,$this->{data}{$key};
    if (&MimeType($fname) =~ /^image/) {
	my $img = new ExSite::Image($this->{data}{$key});
	$img->shrink(dim=>$dim||500,size=>$size||64);
	$this->{data}{$key2} = $img->encode;
	return 1;
    }
    return 0;
}

sub delete {
    my $this = shift;
    if ($this->ok) {
	$this->DB()->trash_r($this->{type},$this->{id});
	return 1;
    }
    return undef;
}

sub copy {
    my $this = shift;
    if ($this->exists) {
	my $data = $this->{data};
	my $key = $this->DB()->get_key($this->{type});
	delete $data->{$key};
	return $this->DB()->insert($this->{type},$this->{data});
    }
    $this->error("cannot copy a non-existent $this->{type}");
    return undef;
}

#=== FORMS

# make: create a basic record

sub make {
    my $this = shift;
    my $type = $this->type;
    return $this->error("Undefined object type.") if (! $type);
    return $this->DB()->make(table=>$type,@_);
}

# edit: edit basic record

sub edit {
    my $this = shift;
    if ($this->ok) {
	return $this->DB()->make(table=>$this->{type},record=>$this->{id},@_);
    }
    return $this->show_diagnostics("error","html");
}

# process forms

sub do_make {
    my $this = shift;
    return $this->DB()->do();
}

sub do_edit {
    my $this = shift;
    return $this->DB()->do();
}

# get child lists

sub get_children {
    my $this = shift;
    my %child;
    if ($this->ok) {
	my @child = $this->DB()->{map}->get_children($this->{type});
	foreach my $ch (@child) {
	    my $list = new ExSite::ObjectList(type=>$ch);
	    my @ref = $this->DB()->{map}->get_reference_column($ch,$this->{type});
	    my $ref = shift @ref;
	    my $chkey = $this->DB()->{map}->get_table_attr($ch,"key");
	    $list->load({$ref=>$this->{id}},$chkey);
	    if (@ref > 0) {
		# looks like record has multiple reference columns
		while ($ref = shift @ref) {
		    $list->loadmore({$ref=>$this->{id}});
		}
		$list->sort($chkey);
	    }
	    $child{$ch} = $list;
	}
    }
    return wantarray ? %child : \%child;
}

1;

=pod

=head1 ExSite::Object

Generic, multi-purpse CMS objects are data objects that an end-user can be
expected to create, edit, delete, and display in reports of one kind or
another.  This class defines a useful set of core behaviours for objects 
that fit this description.

This class can be inherited by specialized objects that can override
the core beheviours to specify custom forms, reports, data handling, etc.

=head2 Attributes

Each Object has two important attributes, B<type> and B<id>.  In the
generic Object, these are taken to be a table name and record ID, and
are used as an "address" to look up the data, and also as raw material for
composing a name (label) for the object.  The type can be treated as a
pseudo-class, since two objects of different types will have slightly
different behaviours, as specified by the lower-level database record
handling tools.

Each Object has a B<data> attribute, consisting of the actual database
record that the Object represents.  This data is only loaded when needed.
You can pass an actual datahash in when instantiating the record, if you
have already loaded it another way.

=head1 Methods

=head2 new(%opt)

Instantiate a new object; the type and id can be passed in the C<%opt> hash.

=head2 setup($type,$id)

Transform the object so that it represents a different record.

=head2 type()

Returns the object's type (by default, it's table).

=head2 id($id)

Returns the object's ID.  An object normally only has an ID if it was
fetched from the database.  An object that has data, but no ID, is a
working object that must be saved for it to be asigned an ID.

=head2 name()

Returns a human-friendly description of the object.

=head2 ok()

Returns true if the object's type and ID are defined.

=head2 key()

Returns the name of the key in the objects datahash that looks up the
objects ID.

=head2 get($attr)

Returns the value of the named attribute (eg. C<type>, C<id>, C<data>,
...); if no attribute name is given, returns the Object's
datahash.

=head2 getdata($key)

Returns the value of the named key in the object's datahash.

=head2 setdata($key,$val)

Sets the value of the named key in the object's datahash.  Only sets
the value in the object; to make this change permanent, the object
must be saved.

=head2 showdata($key)

Formats the value of the named key in the object's datahash as HTML.
Also performs access-control checks to ensure the user is permitted to
view the data.

=head2 dirty()

Returns true (1) if the object's data has been modified and is out of
sync with the saved data.  Returns false (0) if the object matches the 
saved data.

=head2 show()

Formats the entire Object as an HTML report.

=head2 load($data)

Fetches the Object's data.  If C<$data> is passed, it is used as the
object's datahash;  otherwise, the data is fetched from the database.

=head2 insert(%data)

Creates a new record, using C<%data>.  If no data is passed, the
Object's datahash is used.

=head2 update(%data)

Updates an existing record, using C<%data>.  If no data is passed, the
Object's datahash is used.

=head2 save()

Writes the objects datahash out to the database, using either the insert()
or update() method, as appropriate.

=head2 delete()

Moves the record represented by the Object to the trash.

=head2 copy()

Creates a new record that is a copy of the record represented
by the Object (except for the record ID).

=head2 make()

Generates an HTML form that allows the user to enter data for a new
object/record.

=head2 edit()

Generates an HTML form that allows the user to edit the data of the
current Object.

=head2 do_make()

Process the form generated by C<make()>.

=head2 do_edit()

Process the form generated by C<edit()>.

=cut

