#!/usr/bin/perl
#----------------------------------------------------------------------------
#
#   Copyright 2001-2006 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

=pod

=head1 ExSite::Page - templated page generation and management

The Page class handles the storage, retrieval, templating, assembly,
and dispatch of web pages.  It is part of the ExSite content stack, so
it inherits from C<ExSite::Content>, and is in turn inherited by
C<ExSite::Section>.  It provides a more sophisticated page-building
model than the older C<ExSite::PageBuf> class.

The Page class also handles templates and libraries, which are special
cases of the general page object.  In all of these cases, a page is a
"container" that holds content, so there is a lot of common
functionality, even though conceptually they may seem quite different.

C<ExSite::Page> inherits from:

=over 4

=item ExSite::Content

=item ExSite::ContentData

=item ExSite::ContentBase

=back

Many important methods are actually drawn from these classes,
especially content-handling and assembly methods, which are inherited
from C<ExSite::ContentBase> and used by all classes in the content
stack.

=cut


package ExSite::Page;

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::ML;
use ExSite::Util;
use ExSite::URI;
use ExSite::Section;
use ExSite::Content;
use MIME::Base64;

use vars qw(@ISA);
@ISA = qw(ExSite::Content);

# new() constructor is inherited from ContentBase class.

=pod

=head1 Configuration & Setup

=head2 new()

(This method is actually inherited from ExSite::ContentBase.)

=head2 setup()

=head2 setupPage()

Instantiate and/or configure a page object.  C<%option> includes the 
the configuration parameters, described below.

Usage:

    my $page = new ExSite::Page(%option);
    $page->setup(%option);
    $page->setupPage(%option);

C<setup> and C<setupPage> are synonymous;  they are used to setup or
reconfigure an existing Page object.  C<new()> instantiates a new page
object (this method is actually inherited from C<ExSite::ContentBase>).

C<%option> is used to pass the page description information that
allows the object to be configured.  It can include the following page
description parameters:

=over 4

=item id

The numeric page ID will be used to fetch the page description from 
the database.

=item page

The page record can be passed as a hash reference, if it has already been
fetched.

=item section_id, filename

Given a section_id and filename, the page can be looked up in the database.

=item path

The section path and filename can be specified as a text string, eg.
"C</section/subsection/filename.html>".  With this information, ExSite
can locate the page in the database, although the lookup is more costly
than other methods, above.

=item (null)

If no page description is provided, the object will reuse any previously
loaded page info, if present.  Otherwise it will configure an empty page
object that will have limited functionality.

=back

The following additional options are also accepted:

=over 4

=item version

An alternate version of the page may be specified (see C<set_version()>,
below).

=item cms

If this parameter is true and we have CMS permissions to manage this
page, we will select the latest working revisions of content for
display.  Otherwise we select the current approved revisions.

=item publish

Set this flag to 1 if the page is publishing.  This is used to distinguish 
between dynamic and static page views.

=back

If the page is not set up with an actual page description, the object
will have limited functionality.  The setup methods are normally invoked
automatically by the class contructor method C<new()>, which receives the
same C<%option> parameters.

=head2 C<set_version()>

Usage:

    $page->set_version($version);

Switches to an alternate version of the page.  Versions are
alternative presentations of the same content, and are typically used
to provide a page in alternate languages.  C<$version> is a version
name (eg. "Chinese") used to find the alternate version.  If there is
no such alternate page tied to the current page, the default version
of the page is used.  For instance, if we request the Chinese version
of the home page in a certain site, but no Chinese version of that
page exists, then we use the default home page (for instance, the
English C<index.html>) instead.

=cut

sub setup {
    # so that we can do setup using a non-overloaded method
    &setupPage(@_);
}

# options:
# id = id of page to look up
# page = prefetched page record
# noload = skip content loading phase (much faster if not rendering page)

sub setupPage { 
    my ($this,%option) = @_;

    my $page;
    if ($option{id}) {
	# lookup page record
	$page = $this->fetch("page",$option{id});
	if (! $page) {
	    $this->error("page $option{id} not found");
	}
    }
    elsif ($option{page}) {
	# page record was passed
	if (ref $option{page} eq "HASH") {
	    $page = $option{page};
	}
	else {
	    $this->error("invalid page parameters passed to setupPage");
	}
    }
    elsif ($option{filename} && $option{section_id}) {
	# lookup page by name & section
	my @pages = $this->fetch_match("page",{filename=>$option{filename},
					   section_id=>$option{section_id}});
	$page = @pages[-1];
	if (! $page) {
	    $this->error("\"$option{filename}\" not found");
	}
    }
#    elsif (! $this->{data} && $ENV{PATH_INFO}) {
    elsif ($option{path}) {
	# get the path
	my $path = $option{path};
	my $in = new ExSite::Input;
	if ($path == 1) { $path = $in->path_info(); }
	# find the page specified by this path
	my $part_path;
	($page,$part_path) = $this->path_to_page($path);
	# indicate the part of the path that we have used
	$in->path("CMS",$part_path);
#	$page = $this->setup_path(%option);
    }
    elsif (ref $this->{data} eq "HASH") {
	# looks like it's already loaded
	$page = $this->{data};
    }

    if ($page) {
	$this->info("setting up Page $page->{page_id}");
	$this->{data} = $page;
	# preload content
	if ($option{preload}) {
	    $this->prefetch();
	    $this->set_context();
	}
	else {
	    $this->set_context("none");
	}
    }
    else {
	$this->info("setting up null Page");
	$this->{data} = {};
    }

    # set the revisions we wish to see
    if (defined $option{cms} && $this->owns()) {
	$this->set_revision("newest");
    }
    else {
	$this->set_revision("active");
    }

    # see if an alternate version has been requested
    if ($option{version}) {
	$this->set_version($option{version});
    }

    # publish flag
    if (exists $option{publish}) { 
	$this->{publish} = $option{publish}; 
    }

}

# given a URL, setup a page object for the page it represents
# FIXME: extend to work with static URLs

sub setup_url { 
    my ($this,$url) = @_;
    my $uri = new ExSite::URI(uri=>$url);
    if ($uri->get("script_name") !~ /$config{prog}{page}/) {
	# unrecognized url
	return $this->setup();
    }
    my $id = $uri->get("query_data")->{"_id"};
    if ($id) {
	return $this->setup(id=>$id);
    }
    my $path = $uri->get("path_info");
    if ($path) {
	# find the page specified by this path
	my ($page,$part_path) = $this->path_to_page($path);
	# don't setup input if that's not where the data came from
	#$in->path("CMS",$part_path);
	return $this->setup(page=>$page);
    }
    # unknown page
    return $this->setup();
}

# given a path string, return
#    1) the page record that the path points to
#    2) the partial-path that was used to find this page
#       (in case the path has more than just page data in it)

sub path_to_page {
    my ($this,$path) = @_;

    my $page;
    my @path = split(/\//,$path);
    my @pagepath;
    push @pagepath, shift @path;
    # build lookup table of sections
    #my @section = $this->fetch_match("section",{type=>"native website"});
    #my %section = &keywise("publish_to",\@section);
    my %section;
    foreach my $s ($share{DB}->fetch_all("section")) {
	if ($s->{type} =~ /native/) {
	    $section{$s->{publish_to}} = $s;
	}
	else {
	    $section{$s->{section_id}} = $s;
	}
    }
    my ($section,$sdata);
    
    # search through path until we can't find a matching section
    my ($section,$section_id);
    if ($path[0] && $section{$path[0]}) {
	while ($path[0] && $section{$path[0]}) {
	    $sdata = $section{$path[0]};
	    $section_id = $sdata->{section_id};
	    push @pagepath, shift @path;
	}
	$section = new ExSite::Section(section=>$sdata);
    }
    else {
	# no matching sections - set the section to the base native section
	$section = new ExSite::Section(section=>$section{""});
	$section_id = $section->id;
    }
    $this->{context}{section} = $section->get();
    
    # no more sections - see if next path element is a page
    if (@path) {
	my @page = $this->fetch_match("page",{filename=>$path[0],
					      section_id=>$section_id}
				      );
	if (@page > 0) {
	    push @pagepath, shift @path;
	    $page = $page[0];
	}
    }
    
    if (! $page) {
	# no page found - use default/index page for section
	$section->set_context;
	$page = $section->get("page");	
    }

    my $part_path = join("/",@pagepath);
    return ($page,$part_path);
}

# fix_path()
#
# If we are supposed to be using path URLs, but have come in via
# a query URL, then some URL-generation functions may give mixed
# results further down the line.  To avoid this situation, we patch
# the input buffers to make it look like we used a path URL all along.
# This is only needed in cases where you are generating entire pages.

sub fix_path {
    my $this = shift;
    if ($this->is_page() && $config{page}{dyn_url_type} eq "path") {
	my $input = new ExSite::Input;
	$input->path("CMS",$this->cmspath());
	$input->delete("_id");
    }
}

sub set_version {
    my ($this,$version) = @_;
    if ($version ne $this->{page}{version}) {
	my $master = $this->{page}{version} ? 
	    $this->{data}{parent_id} : $this->{data}{page_id};
	
	my @pages = $this->fetch_match("page",{
	    section_id=>$this->{data}{section_id},
	    parent_id=>$master,
	    version=>$version});
	if (@pages > 0) {
	    $this->{data} = $pages[-1];
	    $this->{context_set} = 0;
	    $this->set_context("none");
	    $this->info("setting up $version version of Page $this->{page}{page_id}");
	}
	else {
	    $this->warn("$version version of Page $this->{page}{page_id} not found.");
	}
    }
}

### basic queries

=pod

=head1 Page Attribute Queries

=head2 title()

Returns the page title as a simple string.

=head2 description()

Returns the page description as a simple string.

=head2 keywords()

Returns the page keywords as a simple string.

=head2 is_active()

Returns TRUE if the page can be served, ie. it has not been disabled
in the CMS.

=head2 is_template()

Returns TRUE if the page is a template.

=head2 is_library()

Returns TRUE if the page is a library.

=head2 is_page()

Returns TRUE if the page is a normal web page, ie. a destination URL
for website visitors.  This is true for pages of type "page" and type
"alternate".

=head2 is_public()

Returns TRUE if the page is viewable by unauthenticated members of the
public.

=head2 is_homepage()

Returns TRUE if the page is the default homepage for the section, ie.
it is an index page with an appropriate suffix.

=head2 is_publishable()

Returns TRUE if the page can be published to publicly-viewable files.

=head2 is_published()

Returns TRUE is the page actually has been published, ie. it has no
content that is younger than the page filename's mtime flag.

=head2 is_publishing()

Returns TRUE if the page is currently publishing, ie. writing itself
out to publicly-viewable files.  This method is usually used to detect
when the page should assume no special access permissions, even though
the current user is an administrator who has full access.

=head2 timestamp()

Returns a timestamp that gives the age of the page.  The timestamp is
defined as the most recent of the timestamps of the page's template,
the page record itself, and the content objects in the page.  The
timestamp is a database timestamp, eg. a MySQL timestamp.

=cut

sub title { 
    my $this = shift;
    return $this->{data}{title};
}

sub description { 
    my $this = shift;
    return $this->{data}{description};
}

sub keywords { 
    my $this = shift;
    return $this->{data}{keywords};
}

sub is_active { 
    my $this = shift;
    return ($this->{data}{status} ne "disabled");
}

sub is_template { 
    my $this = shift;
    return ($this->{data}{type} eq "template");
}

sub is_library { 
    my $this = shift;
    return ($this->{data}{type} eq "library");
}

sub is_page { 
    my $this = shift;
    return ($this->{data}{type} =~ /page|alternate/);
}

sub is_public { 
    my $this = shift;
    return ($this->{data}{access} eq "public" && $this->is_active());
}

sub is_homepage { 
    my $this = shift;
    return ($this->{data}{filename} =~ /^index$config{page}{suffix}$/);
}

# does this page ever get published?

sub is_publishable { 
    my $this = shift;
    if ($this->is_template) {
	return ((!$this->{data}{status} || $this->{data}{status} eq "active") && 
		(!$this->{data}{publish_method} || $this->{data}{publish_method} eq "static"));
    }
    else {
	return ((!$this->{data}{access} || $this->{data}{access} eq "public") && 
		(!$this->{data}{status} || $this->{data}{status} eq "active") && 
		(!$this->{data}{publish_method} || $this->{data}{publish_method} eq "static"));
    }
}

# has the latest revision of this page's content been published?

### warning: this only gives accurate results for normal pages
### warning: as a check to see if we need to publish, this is probably 
###          costlier than just publishing

sub is_published { 
    my $this = shift;

    if ($this->is_publishable) {
    
	# the page timestamp is a MySQL timestamp
	my $time = $this->timestamp;
	$time =~ /^(....)(..)(..)(..)(..)(..)$/;
	my ($y1,$M1,$d1,$h1,$m1,$s1) = ($1,$2,$3,$4,$5,$6);

	# the published file has a Unix timestamp
	my @stat = stat($this->diskpath);
	my ($s2,$m2,$h2,$d2,$M2,$y2,undef,undef,undef) = 
	    localtime $stat[9]; # ie. mtime
	$M2 += 1;
	$y2 += 1900;
    
	# compare the two timestamps;  return TRUE if the published
	# file is older than the page data

	if    ($y2 < $y1) { return 0; }
	elsif ($y2 > $y1) { return 1; }

	elsif ($M2 < $M1) { return 0; }
	elsif ($M2 > $M1) { return 1; }

	elsif ($d2 < $d1) { return 0; }
	elsif ($d2 > $d1) { return 1; }

	elsif ($h2 < $h1) { return 0; }
	elsif ($h2 > $h1) { return 1; }

	elsif ($m2 < $m1) { return 0; }
	elsif ($m2 > $m1) { return 1; }

	elsif ($s2 < $s1) { return 0; }
	elsif ($s2 > $s1) { return 1; }

    }

    return 0;

}

# is the page currently being published?

sub is_publishing { 
    my $this = shift;
    return $this->{publish};
}

sub publishing { return &is_publishing(@_); } # deprecated

# return the timestamp of the page
# defined as the most recent of the timestamps of:
# - the template
# - the page
# - the content objects in the page

sub timestamp {
    my $this = shift;
    my @times;
    # timestamp of current page
    push @times, $this->get("page","mtime");
    # timestamp of all content objects in this page
    my @cobjs = $this->fetch_all_content_objs(revision=>"newest");
    foreach my $cobj (@cobjs) {
	push @times, $cobj->timestamp;
    }
    # plus the same, for this page's template
    my $template_id = $this->get("page","template_id");
    if ($template_id) {
	my $template = new ExSite::Page(id=>$template_id);
	push @times, $template->timestamp;
    }
    my @sorted_times = sort @times;
    return pop @sorted_times;
}

=pod

=head1 Access Controls

=head2 is_valid()

Returns TRUE if the page belongs to the site the URL claims.
Valid pages are those that belong to the site they are invoked from.
Ie. if site x.com owns page 98 and site y.org owns page 99, then the
url http://x.com/cgi/page.cgi?_id=99 is not valid.

However, if the user owns the page/site then we allow them to inspect 
it from anywhere.  (This allows them to preview their pages from
a webservice CMS, for instance.)  System administrators can view any
pages from anywhere.

=head2 authorize()

Returns TRUE if the viewer is permitted to see the page contents in a
dynamic page view.  FALSE will normally be returned if the user is not
a site member but is attempting to view a members-only page, or if the
user is not an administrator and is attempting to view an
administrators-only page.

=cut

sub is_valid {
    my $this = shift;
    return 1 if ($share{DB}->is_admin);
    my $domain;
    if ($config{server}{auto}) {
	# config's domain info should be accurate
	$domain = $config{server}{domain};
    }
    else {
	# parse the domain right out of the HTTP_HOST info
	$ENV{HTTP_HOST} =~ /^((.+?)\.)?(.*?\..*?)(:\d+)?$/;
	$domain = $3;
    }
    my $s = new ExSite::Section(id=>$this->id("section"));
    my $url = $s->get_host;
    if ($url =~ /^(\w+\.)?$domain$/) {
	# the section URL matches the calling host
	return 1;
    }
    return 0;
}

# is the current user permitted to view this page dynamically?

sub authorize {
    my $this = shift;
    my $stat = $this->run_handler("page_authorize");
    return $stat if (defined $stat);

    # skip authcheck if we are publishing, since the page is public by nature

    return 1 if ($this->publishing);

    # sysadmins can view anything from anywhere

    my $db = $share{DB};
    return 1 if ($db->is_admin);

    # regular admins can view pages from their own sites from any URL

    return 1 if ($db->is_manager($this->get("section")));

    # non-admins can only view active pages

    return 0 if (! $this->is_active());

    # check for a valid page URL

    return 0 if ($config{cms}{valid_page_check} && ! $this->is_valid);

    # check access permissions for regular members

    my $pdata = $this->get();
    return 1 if ($pdata->{access} eq "public");
    return 1 if ($pdata->{access} =~ /member/ && $db->is_member());

    # non-public, non-member page, and we are not an admin

    return 0;
}

=pod

=head1 Displaying Pages

=head2 show()

Dispatches the page to the viewer.  First it sends any headers that
have been accumulated by the page, then it performs all CMS content
expansions and substitutions, and finally it dumps the page contents
to stdout.

=head2 cancel()

Aborts the page, so that no contents will be displayed to the viewer.
This is typically used if the page is superceded by a redirection.

=head2 header()

Usage:

    $page->header($hdr);

Adds the HTTP header line in $hdr to the list of header directives
that should be sent with this page.  No validation of C<$hdr> is performed.

=head2 show_headers()

Dispatches the accumulated headers for the page to the viewer.  If
there is no content-type header, one is added.

=head2 redirect()

Same as...

=head2 location()

Usage:

    $page->redirect($url);

Cancels the current page, and redirects the viewer to the specified URL.
The redirection is performed by issuing a Location: header.

=head2 errorpage()

Usage:

    $page->errorpage($html);

Cancels the current page, and generates an error page in its place.
C<$html> contains the error message.

This will attempt to use a special _ERROR template to format the page
if the section has such a template defined.  If not, the error message
will be issued "bare", ie. without any wrapper template.

This method is called if there is any problem generating the page as
requested.  Most commonly, this occurs when a non-member attempts to 
view a members-only page.  The error message in this case is the 
login prompt.

=cut

### some page display tools

sub show {
    my $this = shift;

    if (! $this->{done}) {
	# dump headers

	$this->show_headers();

	# dump page

	if (! $this->{redirect}) {
	    print $this->expand;
	}

	$this->{done} = 1;
    }
    return;
}

sub cancel {
    my $this = shift;
    # do NOT show this page
    $this->{done} = 1;
    return;
}

sub halt {
    my $this = shift;
    # abort current pass of content generation, 
    # but leave content in its current state
    $this->{noexpand} = 1;
    return;
}

sub redo {
    my $this = shift;
    # abort current pass of content generation and start over
    $this->{expand_redo} = 1;
    return;
}

# some HTTP header functions

sub header {
    my $this = shift;
    my $hdr = shift;
    if ($hdr !~ /\n$/) { $hdr .= "\n"; }
    $this->{http_header} .= $hdr;
    return;
}

sub show_headers {
    my $this = shift;
    my $content_type = $this->content_type || "text/html";
    print $this->{http_header};
    if ($this->{http_header} !~ /content-type:/i) {
	print "Content-type: $content_type\n";
    }
    print "\n";
    return;
}

# redirect using a location: header (address bar does not change)

sub redirect {
    return &location(@_);
}

sub location {
    my ($this,$url) = @_;
    $this->header("Location: $url");
    $this->{redirect} = 1;
    return;
}

### DEPRECATED - use %cookie
sub cookie {
    my ($this,$name,$value,$path,$domain,$expiry) = @_;
    if (! $path) { $path = $config{server}{CGIpath}; }
    if (! $domain) { $domain = $config{server}{domain}; }
    my $cookie = "Set-cookie: $name=$value; path=$path; domain=$domain";
    if ($expiry) { $cookie .= "; Expires=$expiry"; }
    $this->header($cookie);
    return;
}

sub errorpage {
    my ($this,$html) = @_;

    # see if the section has defined a special template for exceptions

    ### FIXME: this only finds error templates in the current site;
    ### Would be nice to inherit error templates from parent sites.
    ### Also: does this fail to identify the section, if the URL has our
    ### domain, but no page identifier info?
    my @t = $share{DB}->fetch_match("page",{type=>"template",
					    section_id=>$this->id("section"),
					    filename=>"_ERROR"});

    if (@t > 0) {
	# reset the page template
	$this->{data}{template_id} = $t[-1]{page_id};
	$this->reset_context;
	$this->{expanded} = 0;
	$this->expand(content=>{body=>$html});
	return;
    }
    else {
	# no error-handler page;
	# just output the bare html
	$this->{html} = $html;
	$this->{noexpand} = 1;
	return;
    }
}

# exception_page : change the template to handle irregular occurences
#
# Special templates can be defined to handle exceptional conditions
# (errors, permission denied, login screens, etc.).  When the code calls
# exception_page($html,"ABC"), then the system will attempt to replace
# the correct template for the page with a template named '_ABC' (always
# upper case), and then substitute $html into the body of this template.
#
# ExSite knows about and will try to use the following exception templates:
#
# _LOGIN : authentication failures, requiring login to proceed
# _ERROR : errors in building or displaying the page
#
# Note that '_ERROR' is the default exception template, and will be used
# if no explicit exception template is named, or in cases where another 
# exception template was requested but not found.  For example, if the
# "LOGIN" template is requested, but not defined, but an "ERROR" template
# _is_ defined, the _ERROR template will be used instead.
#
# Any other exception templates may be defined, and invoked by plug-in
# modules or other custom code.

sub exception_page {
    my ($this,$html,$exception) = @_;

    $exception = $exception ? uc($exception) : "ERROR";

    # see if the section has defined a special template for exceptions

    ### FIXME: this only finds error templates in the current site;
    ### Would be nice to inherit error templates from parent sites.
    ### Also: does this fail to identify the section, if the URL has our
    ### domain, but no page identifier info?
    my @t = $share{DB}->fetch_match("page",{type=>"template",
					    section_id=>$this->id("section"),
					    filename=>"_$exception"});
    if (@t == 0 && $exception ne "ERROR") { 
	# requested exception page not found;
	# try _ERROR instead.
	@t = $share{DB}->fetch_match("page",{type=>"template",
					     section_id=>$this->id("section"),
					     filename=>"_ERROR"});
    }

    if (@t > 0) {
	# reset the page template
	$this->{data}{template_id} = $t[-1]{page_id};
	$this->reset_context;
	$this->{expanded} = 0;
	$this->expand(content=>{body=>$html});
	return;
    }
    else {
	# no error-handler page;
	# just output the bare html
	$this->{html} = $html;
	$this->{noexpand} = 1;
	return;
    }
}

=pod

=head1 Page Location

=head2 filename()

Returns the publication name of the page.  For regular pages, this is
the filename; for templates and libraries, it is the directory name.
The bare name is returned; see C<diskpath()> if a full path is needed.
The filename is processed by C<ExSite::Misc::clean_filename()> to
removed sensitive shell characters.  If the page does not have a
filename (should not normally occur), a filename based on the page ID
will be generated.

Note that the return value may not refer to an actual file if the page is 
always rendered dynamically.

=head2 httppath()

Returns the web path to the static page file.  This is the part of the
static URL after the domain name.  For templates and libraries
this returns the folder containing the template and library files.
For normal pages, this returns the folder for the current section.

Note that the return value may not refer to an actual file if the page is 
always rendered dynamically.

=head2 diskpath()

    my $path = $page->diskpath($dir_only);

Returns the server disk path to the static page file.  This returns
the path to the page's file if it is a normal page and $dir_only is
not TRUE.  Otherwise it returns the path to the directory that the
files are written to (ie. the section directory, template directory,
or library directory, as appropriate).

Note that the return value may not refer to an actual file if the page is 
always rendered dynamically.

=head2 get_url()

Return the URL to the page.  If the page is static, it checks to make
sure the page has actually been published before returning a static
URL.  If not, then a dynamic URL is returned instead, to ensure that
the URL will actually generate a meaningful result.

=head2 get_url_canonical()

Returns the correct public URL to the page.  This simply means that if
the page is supposed to be static, then the static URL is always
returned, even if the page has not yet been published.

=head2 get_url_static()

Returns the static URL to a page.  This may not be a valid URL if the
page is configured to only render dynamically.

=head2 get_url_dynamic()

    my $url = $page->get_url_dynamic($path_method);

Returns the dynamic URL to a page.  If C<$path_method> is TRUE, then
the URL is of the form:

    /cgi/page.cgi/section/page.html

Otherwise, the URL is of the form:

    /cgi/page.cgi?_id=123

=head2 link()

    my $url = $page->link(%args);

Generates a URL to a dynamic view of a page, including any optional
query parameters that are given in %args.  It is thus analagous to 
C<ExSite::BaseDCD::link()> in usage.

=cut

# name of page file - overloaded variation on ContentData method

sub filename {
    my $this = shift;
    my $page = $this->get;
    if ($page) {
	my $filename = $page->{filename};
	if ($this->is_page()) {
	    $filename = &clean_filename($filename);
	}
	else {
	    # "." not allowed in dir names
	    $filename =~ s/[^\w]/_/g;
	    $filename =~ s/_{2,}/_/g;
	}

	if ($page->{type} eq "page") {
	    return $filename || "page_".$page->{page_id}.".html";
	}
	else {
	    # directory name for templates and libraries
	    return $filename || $page->{page_id};
	}
    }
    return undef;
}

# cmspath - returns the CMS path to the page (section path + page filename)

sub cmspath {
    my $this = shift;
    $this->set_context("parent");
    my $s = new ExSite::Section(section=>$this->get("section"));
    my $path = $s->sectionpath_url;
    my $filename = $this->filename;
    if ($this->is_page) {
	$path .= "/$filename";
    }
    elsif ($this->is_template) {
	$path .= "/_Template/$filename";
    }
    elsif ($this->is_library) {
	$path .= "/_Library/$filename";
    }
    return $path;
}

# httppath: the URL path to the page

sub httppath {
    my $this = shift;
    my $page = $this->get;
    if ($page) {
	my $s_httppath;
	if ($page->{section_id}) {
	    my $s = new ExSite::Section(id=>$page->{section_id});
	    $s_httppath = $s->httppath;
	}
	else {
	    $s_httppath = $config{server}{HTMLpath};
	}
	$_ = $page->{type};
	if (/template/) {
	    return "$s_httppath/_Template/$page->{filename}";
	}
	elsif (/library/) {
	    return "$s_httppath/_Library/$page->{filename}";
	}
	else {
	    return $s_httppath;
	}
    }
    return undef;
}

# disk path to file

sub diskpath {
    my $this = shift;
    my $dir = shift;  # return directory only if true
    my $page = $this->get("page");
    if ($page) {
	my $s_diskpath;
	if ($page->{section_id}) {
	    my $s = new ExSite::Section(id=>$page->{section_id});
	    $s_diskpath = $s->diskpath;
	}
	else {
	    $s_diskpath = $config{server}{HTMLroot};
	}
	if ($s_diskpath) {
	    if (! $dir || ! $this->is_page) { 
		if ($page->{type} eq "template") {
		    $s_diskpath .= "/_Template";
		}
		elsif ($page->{type} eq "library") { 
		    $s_diskpath .= "/_Library";
		}
#		elsif (! $this->is_publishable && ! $this->is_homepage) {
#		    # EXPERIMENTAL: precompile dynamic page templates
#		    # non-publishing pages can have their precompiled 
#		    # templates published to the _Template folder
#		    $s_diskpath .= "/_Template";
#		}
	    }
	    return $s_diskpath if ($dir && $page->{type} =~ /page|alternate/);
	    return "$s_diskpath/".$this->filename;
	}
    }
    return undef;
}

# overloaded get_url: return a URL for the page

sub get_url {
    my $this = shift;
    my $stat = $this->run_handler("page_get_url",@_);
    return $stat if (defined $stat);

    # first parameter can be used to force this method to return the
    # static url;  otherwise it returns the static URL only if it
    # exists.
    my $static_req = shift; # static url required?
    if ($static_req &&
        ($this->is_publishable || $config{cms}{static_urls})) {
        return $this->get_url_static;
    }
    my $static = defined $static_req ? $static_req : 1;  # default to static

    if ($this->is_publishable) {
	if (! $this->is_page) {
	    return $this->get_url_static;
	}
	elsif ($cookie{ExSiteCMS}) {
	    return $this->get_url_dynamic();
	}

	# if we get here, we should return the static URL, if it exists
	my $page = $this->get;
	return undef if (! $page);
	if ($page->{section_id}) {
	    my $file = $this->filename;
	    my $s = new ExSite::Section(id=>$page->{section_id});
	    my $s_url = $s->get_url;
	    my $s_diskpath = $s->diskpath;
	    my ($path_to_page,$url_to_page);
	    if ($page->{type} eq "template") { 
		$path_to_page = $s_diskpath."/_Template/".$file;
		$url_to_page = $s_url."/_Template/".$file;
	    }
	    elsif ($page->{type} eq "library") { 
		$path_to_page = $s_diskpath."/_Library/".$file;
		$url_to_page = $s_url."/_Library/".$file;
	    }
	    else {
		$path_to_page = $s_diskpath."/".$file;
		$url_to_page = $s_url."/".$file;
	    }
	    return $url_to_page if (-e $path_to_page);

	    # cannot find a published page - fall back on dynamic
	    return $this->get_url_dynamic();
	}
	# shouldn't get here (sectionless page)
	return undef;
    }
    else {
	if ($config{page}{dyn_url_type} eq "redirect") {
	    # static redirect to dynamic page
	    return $this->get_url_static();
	}
	else {
	    return $this->get_url_dynamic();
	}
    }
}

# get_url_canonical : return the static url for static pages, even if
# the static page has not been published.

sub get_url_canonical {
    my $this = shift;

    if ($this->is_publishable || ! $this->is_page) {
	# static page
	return $this->get_url_static;
    }
    elsif ($config{page}{dyn_url_type} eq "redirect") {
	# static redirect to dynamic page
	return $this->get_url_static();
    }
    else {
	# straight to dynamic page
	return $this->get_url_dynamic();
    }
}

# get_url_static : return the static url for the page, even if it is not
# actually published

sub get_url_static {
    my $this = shift;
    my $page = $this->get;
    return undef if (! $page);
    if ($page->{section_id}) {
	my $file = $this->filename;
	my $s = new ExSite::Section(id=>$page->{section_id});
	my $s_url = $s->get_url;
	if ($page->{type} eq "template") { 
	    return $s_url."/_Template/".$file;
	}
	elsif ($page->{type} eq "library") { 
	    return $s_url."/_Library/".$file;
	}
	else {
	    return $s_url."/".$file;
	}
    }
    return undef;
}

# get_url_dynamic()
#
# This is a crufty method that needs to be refactored.  Use one of the
# more specific methods that follow, if you can.
#
# there are two options:
# full
#   if true:  http://mydomain.com/cgi/page.cgi...
#   if false: /cgi/page.cgi...
# path
#   if true:  /cgi/page.cgi/section/page.html
#   if false: /cgi/page.cgi?id=XXX

sub get_url_dynamic {
    my $this = shift;
    my $full = shift;  # tells us to use fully-qualified URL
    my $use_path = shift;  # use path_info method, if true
    if (! defined $use_path) { 
	$use_path = ($config{page}{dyn_url_type} eq "path"); 
    }
    #return undef if (! $this->is_page);
    my ($url,$s);
    if ($full) {
	$url = $config{server}{server};
    }
    if ($use_path) {
	$url = "$config{server}{CGIpath}/$config{prog}{page}";
	$this->set_context("parent");
	$s = new ExSite::Section(id=>$this->id("section"));
	$url .= $s->sectionpath_url;
	$url .= "/".$this->filename;
    }
    else {
	my $id = $this->id();
	$url .= "$config{server}{CGIpath}/$config{prog}{page}?_id=$id";
    }
    ### fixme: this is really convoluted logic...
    return $url if ($full || $config{page}{relative_dynamic_url});

    # make url fully-qualified, if necessary
    if ((ref $s) !~ /Section/) { 
	$s = new ExSite::Section(id=>$this->id("section"));
    }
    $s->get_url() =~ /^(https?:\/\/[^\/]*)/;
    return "$1$url";
}

# return full url

sub get_url_dynamic_full {
    my $this = shift;
    my $url = $this->get_url_dynamic_relative();
    my $s = new ExSite::Section(id=>$this->id("section"));
    $s->get_url() =~ /^(https?:\/\/[^\/]*)/;
    return "$1$url";
}

# return relative url (type is based on config prefs)

sub get_url_dynamic_relative {
    my $this = shift;
    return ($config{page}{dyn_url_type} eq "path") ?
	$this->get_url_dynamic_path() :
	$this->get_url_dynamic_query();
}

# return relative url using path notation; ie. page.cgi/filename.html

sub get_url_dynamic_path {
    my $this = shift;
    my ($url,$s);
    $url = "$config{server}{CGIpath}/$config{prog}{page}";
    $this->set_context("parent");
    $s = new ExSite::Section(id=>$this->id("section"));
    $url .= $s->sectionpath_url;
    $url .= "/".$this->filename;
    return $url;
}

# return relative url using query notation; ie. page.cgi?_id=XXX

sub get_url_dynamic_query {
    my $this = shift;
    my $id = $this->id();
    return "$config{server}{CGIpath}/$config{prog}{page}?_id=$id";
}

# return a URI object representing the dynamic page address

sub get_uri_dynamic {
    my $this = shift;
    my $url = $this->get_url_dynamic(@_);
    my $uri = new ExSite::URI();
    $uri->setup($url);
    return $uri;
}

# link - like get_url, but always returns a dynamic link, and you can add 
# your own parameters to it.

sub link {
    my ($this,%args) = @_;
    my $id = $this->id("page");
    my $url = new ExSite::URI;
    if ($config{page}{dyn_url_type} eq "path") {
	my $path = $this->cmspath();
	$url->setup("$config{server}{server}$config{server}{CGIpath}/$config{prog}{page}$path");
    }
    else {
	$url->setup("$config{server}{server}$config{server}{CGIpath}/$config{prog}{page}?_id=$id");
    }
    $url->query(%args);
    return $url->write;
}

=pod

=head1 Content Management

=head2 fetch_all_content_objs()

    my @content = $page->fetch_all_content_objs(%options);

Returns an array of content objects for each content item defined
within the current page.  The C<%options> are passed to the C<new()>
method when instantiating each content object.

=head2 edit_content_link()

Generates a hotlink to allow the administrator to edit the contents
of the current page.  This is inserted into the page content so that
the content can be edited directly from a page preview.

The result of following this link depends on the setting of
C<$config{page}{editmode}>.  For instance, it may invoke the Website
Manager or MySite CMS applications.

=head2 set_template()

    $page->set_template($template);

Changes the template of the current page to C<$template> (which may be
a template ID or a template record hashref), if permitted.

=head2 set_body()

    $page->set_body($body);

Changes the body of the current page to C<$body> (which may be a
content ID or a content record hashref), if permitted.  This is used
to copy content from a model body into the current page.  To simply
update the page body, use C<ExSite::Content::update()> on the body
object.

=cut

# fetch_all_content_objs - returns array of content objects for this page
# only returns content belonging to this page.

sub fetch_all_content_objs {
    my ($this,%options) = @_;
    my @contents = $this->fetch_child("content","page",$this->id(),"content_id");
    my @cobjs;
    foreach my $content (@contents) {
	push @cobjs, new ExSite::Content(content=>$content,%options);
    }
    return @cobjs;
}

# overload ContentBase::find, for page prefetching optimizations
#
#sub find {
#    my ($this,$cname,$mode) = @_;
#
#    # setup
#    if (! $this->{context_set}) {
#	$this->set_context;
#    }
#
#    my $c = $this->fetch_content($cname);
#    if ($c) {
#	return $c if ($mode eq "data");
#
#	# fixme: do not use current if admin preview ??
#
#	my $content = new ExSite::Content(content=>$c,revision=>"current");
#	#$c->set_context("child","current");
#	return $content;
#    }
#
#    # fallback on base method if there is nothing in the cache
#    return ExSite::ContentBase::find($this,$cname,$mode);
#}

# create a hotlink to edit content data

sub edit_content_link {
    my ($this,$cobj) = @_;
    return if (! $cobj);

    my $path = "$this->{context}{page}{section_id}/$this->{context}{page}{page_id}/$cobj->{context}{content}{content_id}";

    my $name = "$cobj->{context}{content}{name}";

    my $out;

    if ($config{page}{editmode} eq "easy") {
	$out .= "<!--&CMS(mode=inline&cmscmd=edit_content_link&name=${name}&path=/$path)-->";
    }
    elsif ($config{page}{editmode} eq "MySite") {
	$out .= "<!--&MySite(ecmd=edit_content_link)-->";
    }
    else {
	my $ml = new ExSite::ML;

	my $url = "$config{server}->{CGIpath}/$config{prog}{ctrlpanel}/CMS?path=/$path";

	# update content, if user has editorial/design priviledges

	$out .= $ml->a(
		       $ml->img(undef, { src=>"$config{server}->{HTMLpath}/_ExSite/images/plus16.gif", height=>16, width=>16, border=>0, title=>"update $name"}),
		       { href=>"${url}&cmscmd=update" }
		       );

	# manage content, if user has editorial/design priviledges

	$out .= $ml->a(
		       $ml->img(undef,{src=>"$config{server}->{HTMLpath}/_ExSite/images/box16.gif", height=>16, width=>16, border=>0, title=>"manage $name"}),
		       { href=>$url }
		       );
    }

    return $out;
}

# change template for this page

sub set_template {
    my ($this,$template) = @_;
    my $tid = ref $template eq "HASH" ? $template->{page_id} : $template;
    $this->set_context();

    # sanity checks

    if (! $tid) {
	return $this->error("set_template: no template given");
    }
    if (! $this->owns()) {
	return $this->error("set_template: you do not have sufficient priveleges to modify this page");
    }
    my $s = new ExSite::Section(section=>$this->get("section"));
    my @allowed_templates = $s->get_templates(0);
    my $t;
    foreach $t (@allowed_templates) {
	if ($t->{page_id} == $tid) {
	    $share{DB}->update("page",{template_id=>$tid},{page_id=>$this->id()});
	    return undef;
	}
    }
    return $this->error("set_template: you do not have permission to use this template");
}

# change body for this page

sub set_body {
    my ($this,$body) = @_;
    my ($bid,$content);
    if (ref $body eq "HASH") {
	$bid = $body->{content_id};
	$content = new ExSite::Content(content=>$body);
    }
    else {
	$bid = $body;
	$content = new ExSite::Content(id=>$body);
    }
    $content->set_context();

    # sanity checks

    if (! $bid) {
	return $this->error("set_body: no page body");
    }
    if ($content->get("content","name") ne "body") {
	return $this->error("set_body: not a page body");
    }
    if (! $content->owns()) {
	return $this->error("set_body: you do not have sufficient priveleges to work with this page body");
    }

    my $s = new ExSite::Section(section=>$content->get("section"));
    my @allowed_bodies = $s->get_bodies;
    my $b;
    foreach $b (@allowed_bodies) {
	if ($b->{content_id} == $bid) {
	    # get actual body
	    my $db = $share{DB};
	    my @pbody = $db->fetch_match("content",{page_id=>$this->id(),
						    name=>"body"});
	    my $bodydata = $content->get("content_data");
	    delete $bodydata->{content_data_id};
	    $bodydata->{content_id} = $pbody[0]{content_id};
	    if (@pbody == 0) {
		# no body - make one!
		$this->make_content({name=>"body",
				     page_id=>$this->id(),
				     type=>"editorial"},
				    $bodydata);
	    }
	    # update the existing body
	    my $c = new ExSite::Content(content=>$pbody[0]);
	    $c->update($bodydata);
	    return undef;
	}
    }
    return $this->error("set_body: you do not have sufficient priveleges to work with this page body");
}

# update body for this page

sub update_body {
    my ($this,$body) = @_;
    # $body is assumed to be HTML
    my $b = $this->find_content("body");
    my $cdata = { content_id=>$b->{content_id},
		  data=>$body };
    my $c = new ExSite::Content(content=>$b);
    $c->update($cdata);
    return;
}

=pod

=head1 Publishing

=head2 publish()

Write a page and its files out to disk, and mark all of its content as
published.

Publishing is equivalent to "approving for public view", so pages and
content that do not get written to static files can still be
published.

In some cases, dynamic pages and templates can be published to static
files.  The purpose in this case is to precompile the page/template so
that less assembly is required at the actual time of page view.
Dynamic page generation will begin from the precompiled file, rather
than from scratch.  Although the precompiled pages can in principle be
viewed using static URLs, the pages will include unresolved CMS tags,
and ExSite will not create links to those files.

If C<$config{search}{reindex_on_publish}> is TRUE, the search index
for the page may also get rebuilt at this time.

WARNING: you need write access to the page's C<diskpath()> to publish
files there.  The webserver normally does not have this access.
Negotiating this access is normally done by the publish program
(ie. publish.cgi), which has special permissions.  If you do not have
these permissions, then you have to ask publish.cgi to do this task on
your behalf, using C<ExSite::Util::run_publisher()>.

=head2 unpublish()

Removes a page's files from disk.  Also deindexes the page from the
search index, if C<$config{search}{reindex_on_publish}> is TRUE.  Does
not, however, unapprove the page content for public viewing.  See
C<publish()> for other warnings.

=cut

# write the page to a file

sub publish {
    my $this = shift;
    if (! $this->allow_update()) {
	# publishing has same permission rules as updates
	print "Permission denied - administrator privilges required to publish.";
	return;
    }
    my $out;
    $this->{publish} = 1;  # flag this page as going to disk
    my $diskpath = $this->diskpath(1); # directory only
    my $page = $this->get;
    my $ml = new ExSite::ML;
    if (! $page) { 
	$this->error("Page::publish: no page to publish");
	return "No page to publish!";
    }
    else {
	return "" if ($page->{status} eq "disabled");
	$this->info("Page::publish: publishing page $page->{page_id}");
	$share{DB}->set_level(0);  # build pages as an unauthenticated user
	$this->set_context;
	my $mkdir = $this->mkdir($diskpath);
	if (! $mkdir) {
	    $share{DB}->reset_level();
	    return $this->show_diagnostics("error","html");
	}
	$out .= $mkdir;
	# First, publish virtual files and external content objects.
	# This allows the page to find the published files and link to them
	# instead of the dynamic revisions.
	#### Might be better to save all content objects actually 
	#### inserted into the page during substitute(), and use those.
	my @contents = $this->fetch_all_content_objs(revision=>"newest");
	foreach my $content (@contents) {
	    $out .= $content->publish($diskpath);
            if ($content->errorcheck()) {
		$out .= $content->show_diagnostics("error","html");
            }
	}
        if ($this->is_page) {
	    # set the version/language
	    (tied %msg)->version($this->get_my("version"));
	    # rebuild environment as if we were viewing this page
	    $this->update_env();
	    if ($this->is_publishable()) {
		# publish normal static page
		$out .= $this->write_page_file($this->filename(),
					       $diskpath,
					       $this->expand(),
					       "page",$ml);
	    }
	    elsif ($this->is_active && $config{page}{prepublish_dynamic}) {
		# publishes a dynamic page (containing unresolved CMS tags)
		### EXPERIMENTAL
		my $pagedata = $this->expand(method=>"data,url,content,page,static-module");
		# unmask the delayed expansion tags
		$pagedata =~ s/<\!--\*\&(\w+)\((.*?)\)-->/<!--&$1($2)-->/g;
		$out .= $this->write_page_file($this->filename(),
					       $diskpath,
					       $pagedata,
					       "templated page",$ml);
	    }
	    elsif ($this->is_active && $config{page}{dyn_url_type} eq "redirect") {
		# publishes a pseudo-page containing a meta-refresh to 
		# redirect to the dynamic page.
		### EXPERIMENTAL - SimpleMenu does not get the urls right
		### because its get_url doesn't use get_url_canonical
		my $url = $this->get_url_dynamic_relative();
		$ml->Doctype('<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" "http://www.w3.org/TR/html4/loose.dtd">');
		$ml->_html(
			   $ml->head(
				     $ml->meta(undef,{'http-equiv'=>"Refresh",content=>"0;URL=$url"})
				     ).
			   $ml->body("")
			   );
		
		# publish page
		$out .= $this->write_page_file($this->filename(),
					       $diskpath,
					       $ml->Write(),
					       "redirect page",$ml);
	    }

	    if ($config{search}{reindex_on_publish} && $share{Search}) {
		# update the search index for this page
		$out .= $share{Search}->index_page($this);
	    }
	}
	elsif ($this->is_template && $page->{publish_method} eq "static" ) {
	    # publish precompiled template
	    $out .= $this->write_page_file("index.html",
					   $diskpath,
					   $this->expand(expand=>"template"),
					   "template",$ml);
	}
	$share{DB}->reset_level();  # redundant?
    }
    return $out;
}

# actually write the page file to disk
sub write_page_file {
    my ($this,$filename,$diskpath,$pagedata,$type,$ml) = @_;
    $type or $type = "page";
    $ml or $ml = new ExSite::ML;
    $filename =~ /^([\w-]+$config{page}{suffix})$/;
    my $secure_filename = $1;
    my $secure_diskpath = $this->safe_path($diskpath);
    my $out;
    if ($secure_filename) {
	my $url = $this->get_url(1); # get static url
	my $pubmsg = "Saving $type $this->{page_id} to ".
	    $ml->a($secure_filename,{href=>$url}).
	    "... ";
	if (open OUTFILE, ">$secure_diskpath/$secure_filename") {
	    print OUTFILE $pagedata;
	    close OUTFILE;
	    $pubmsg .= "succeeded.";
	    # remove our URI object, so it doesn't contaminate caller
	    delete $share{URI};
	    $out .= $ml->p($pubmsg);
	}
	else {
	    $this->error("Page::publish: publish $secure_filename failed: $!");
	    $pubmsg .= $ml->span("FAILED: $!",{class=>"error"});
	    $share{DB}->reset_level();
	    $out .= $ml->p($pubmsg);
	}
    }
    else {
	$this->error("Page::publish: invalid filename to publish to: $filename");
	$out .= $ml->p("Invalid filename to publish to: $filename.",
		       {class=>"error"});
    }
    return $out;
}

# remove the page file(s)

sub unpublish {
    my $this = shift;
    my $out;
    my $diskpath = $this->diskpath();
    my $page = $this->get;
    if (! $page) { 
	return $this->error("Page::unpublish: no page to remove");
    }
    else {
	$this->info("Page::unpublish: removing $page->{filename}");
	$this->set_context;
        if ($page->{type} eq "library" || $page->{type} eq "template") {
	    # remove the library/template directory
	    $this->rmdir($diskpath);
	}
        else {
	    # remove the page file
	    $this->rm($diskpath);
	    if ($this->is_page()) {
		if ($config{search}{reindex_on_publish} && $share{Search}) {
		    # clear the search index for this page
		    $out .= $share{Search}->clear_url($this->get_url());
		}
	    }
	}
    }
    return $out;
}

# update the environment to resemble the current page object; 
# ie. make it look like Apache is serving a request for this page.
# We do this when generating complete pages for view, so that their 
# input bufs, self-referential links, etc. are appropriate to the page.

sub update_env {
    my $this = shift;

    # doctor the URI to spoof normal page generation
    my $uri = &get_obj("URI");
    $uri->setup($this->get_url_dynamic_full());
    $uri->use_input();
    
    # WARNING: tampering with environment to spoof normal 
    # page generation context;  this should make the input
    # buffers look like a normal page call.
    $ENV{SCRIPT_NAME} = $uri->get("script_name");
    $ENV{QUERY_STRING} = $uri->get("query");
    $ENV{PATH_INFO} = $uri->get("path_info");

    # clear and regenerate input buffers to force a 
    # reload from our spoofed environment
    delete $share{input};
    my $in = new ExSite::Input;
    $in->query;
    $in->path_info;
    $this->fix_path();

    return;
}


=pod

=head1 Page Relationships

=head2 get_child_pages()

    my @child_pages = $page->get_child_pages($version);

Returns an array of page objects representing the children of this
page (ie. all pages in a submenu of this page).  If C<$version> is
specified, then the fetched pages will be of that version
(eg. language), if they exist, or the default version otherwise.

=head2 parent()

Returns the parent page as a page object.  This is the page whose
submenu the current page belongs to.

=head2 master()

Returns the master version (ie. the default version) of the current 
page.  This will be the page itself, unless the page type is "alternate".

=cut

# get child pages, in appropriate order

sub get_child_pages {
    my ($this,$version) = @_;

    # use the current version by default
    if (! $version) { $version = $this->{page}{version}; }

    # determine the parent page to find children of
    my $parent = $this->{page}{type} eq "alternate" ?
	$this->{page}{parent_id} : 
	$this->id;
    my @pages = $this->fetch_match("page",{parent_id=>$parent},"rank");

    # create an array of Page objects
    my @Page;
    foreach my $page (@pages) {
	if ($version) {
	    # alternate version requested
	    my @alt = $this->fetch_match("page",{parent_id=>$page->{page_id},
						 type=>"alternate",
						 version=>$version,
						 },"page_id");
	    if (@alt > 0) {
		# alternate version of this page was found
		# swap it in place
		$page = $alt[-1];
	    }
	}
	push @Page, new ExSite::Page(page=>$page);
    }
    return @Page;
}

# get master page, as a Page object

sub master {
    my ($this) = @_;
    if ($this->{data}{type} eq "alternate") {
	return new ExSite::Page(id=>$this->{data}{parent_id});
    }
    return $this;
}

# get alternate version, as a Page object

sub alternate {
    my ($this,$version) = @_;
    return new ExSite::Page(id=>$this->id,version=>$version);
}

# get parent page, as a Page object

sub parent {
    my ($this) = @_;
    my $page = $this->get();
    my $version = $page->{version};
    # get master
    if ($page->{type} eq "alternate") {
	$page = $this->fetch("page",$page->{parent_id});
    }
    # take the parent of the master, in the appropriate version
    return new ExSite::Page(id=>$page->{parent_id},version=>$version);
}

=pod

=head1 Creating/Modifying Pages

NB: modifying means changing the page attributes, not updating the page
content.  The latter operation is done by calling the C<update()>
method of the particular content object being updated.

=head2 validate()

    my @errors = $page->validate($pagedata);
    if (@errors == 0) { # page is OK }

Returns TRUE if the values in the the C<$pagedata> hashref comprise a
valid and legal page description.  Returns an array of error messages,
which will be empty if the page data validates.

=head2 is_dupe()

    if ($page->is_dupe($pagedata)) { # similar page already exists }

Returns TRUE if the pagedata has identical publication characteristics
(ie. same filename, page type, and section) to an existing page in the
system.  Called automatically by C<validate()>.

=head2 make()

=head2 make_page()

    $page->make($pagedata,$body,$opt);

Creates a new page, with the attributes given in the C<$pagedata>
hashref.  If the user is permitted to make pages and the attributes
validate, then the page record will be created.

If C<$body> is defined, then a "body" content object will be created
in the new page, and initialized to the content in C<$body> (presumed
to be HTML).  C<$opt> is an options hashref that is passed to
C<ExSite::Content::make()> when the body is created.

=head2 delete()

Removes the current page from the system.  Attempts to unpublish its
files, as well.

=cut

# validate a page record
#
# FIXME: validate list: datatypes

sub validate {
    my ($this,$data) = @_;
    my ($page,$type,@err);
    if (! $data) { $data = $this->get("page"); }

    # generic validation (using DBmap validation rules)
    push @err, $share{DB}->validate_record("page",$data);

    # special validation
    $type = $data->{type};
    if ($type eq "page" || $type eq "alternate") {
	if (! $data->{template_id}) {
	    push @err, "page is missing its template";
	}
	if (! $data->{filename} || $data->{filename} !~ /$config{page}{suffix}$/) {
	    push @err, "page has an invalid filename";
	}
	if (! $data->{label}) {
	    push @err, "page is missing a label";
	}
	if (! $data->{access}) {
	    push @err, "page has no access control defined";
	}
	if (! $data->{visibility}) {
	    push @err, "page has no visibility setting";
	}
	if (! $data->{publish_method}) {
	    push @err, "page has no publish method defined";
	}
    } 
    elsif ($type eq "template") {
	if (! $data->{filename} || $data->{filename} !~ /^[\w_\-]+$/) {
	    push @err, "template has an invalid publication name";
	}
	if (! $data->{access}) {
	    push @err, "template has no access control defined";
	}
    }
    elsif ($type eq "library") {
	if (! $data->{filename} || $data->{filename} !~ /^[\w_\-]+$/) {
	    push @err, "library has an invalid publication name";
	}
	if (! $data->{access}) {
	    push @err, "library has no access control defined";
	}
    }
    if (@err == 0) {
	if ($this->is_dupe($data)) {
	    push @err, "a $type with that publish name already exists";
	}
    }
    return @err;
}

# is_dupe : returns TRUE if a page with matching publication characteristics
# exists in the system

sub is_dupe {
    my ($this,$data) = @_;
    return 0 if (! $data->{section_id} ||
		 ! $data->{type} ||
		 ! $data->{filename});

    # check for dupe pages in this site

    my @p = $this->fetch_match("page",{section_id=>$data->{section_id},
				       filename=>$data->{filename}});
    my @page;
    if (@p > 1) {
	my @p2;
	if ($data->{type} =~ /page|alternate/) {
	    foreach my $p (@p) {
		if ($p->{type} =~ /page|alternate/) {
		    push @p2, $p;
		}
	    }
	}
	else {
	    foreach my $p (@p) {
		if ($p->{type} eq $data->{type}) {
		    push @p2, $p;
		}
	    }
	}
	if (@p2 > 1) {
	    # more than one already installed
	    $this->warn("duplicate $data->{type} $data->{filename} detected in section $data->{section_id}");
	    return 1;
	}
	@p = @p2;
    }
    if (@p == 1) {
	# only one installed;  it's a dupe if it's ID is different from data
	if ($data->{page_id}) {
	    return (($data->{page_id} != $p[0]{page_id}) && 
		    (($data->{type} =~ /page|alternate/ && $p[0]{type} =~ /page|alternate/) ||
		     ($data->{type} !~ /page|alternate/ && $data->{type} eq $p[0]{type})));
	}
	else {
	    return 1;
	}
    }
    return 0;
}

# make_page : insert a new page record
#
# can optionally pass a body (content_data) record to seed it with
#
# FIXME:  should be able to pass a body template ID, too

sub make {
    return &make_page(@_);
}

sub make_page {
    my ($this,$page,$body,$opt) = @_;
    if ($page->{page_id}) {
	# invalid page
	$this->error("invalid attempt to insert an already-defined page record");
	return undef;
    }
    if (! $this->allow_insert("page",$page)) {
	# permission denied
	$this->error("make_page: permission denied");
	return undef;
    }
    my @err = $this->validate($page);
    if (@err > 0) {
	$this->error(@err);
	return undef;
    }

    # remove any residual timestamps
    delete $page->{mtime};
    delete $page->{ctime};
    $share{DB}->run_handler("pre_make_page",$page);
    my $pid = $share{DB}->insert("page",$page);
    if (! $pid) {
	$this->error("failed to install new content $page->{filename}");
    }
    else {
	$this->info("installed new page $page->{filename}");
	my $cid;
	if (ref $body eq "HASH") {
	    $cid = $this->make_content({name=>"body",
					type=>"editorial",
					page_id=>$pid},
				       $body,$opt);
	    # ignore return code; make_content_data will report its own errs
	}
	$body->{content_id} = $cid;
	$share{DB}->run_handler("post_make_page",$page);
    }
    return $pid;
}

# delete the page

sub delete {
    my $this = shift;
    my $id = $this->id;
    if ($id && $this->allow_delete("page",$this->{data})) {
	$this->unpublish();
	$share{DB}->trash_r("page",$id);
	return 1;
    }
    return 0;
}

1;


