package ExSite::Search;

#----------------------------------------------------------------------------
#
#   Copyright 2001-2007 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#-----------------------------------------------------------------------

use strict;
use ExSite::Config;
use ExSite::Auth;
use ExSite::Section;
use ExSite::Module qw(&get_module);
use ExSite::Page;
use ExSite::ML;
use ExSite::Base;

use vars qw(@ISA);
@ISA = qw(ExSite::Base);

sub new {
    my $this = shift;
    my $obj = {};
    $obj->{section_id} = shift;
    my $class = ref($this) || $this;
    bless $obj, $class;
    $obj->initialize_object;

    $obj->{weight} = $config{search}{weight};

    $obj->reset;

    return $obj;
}

sub get_skipwords {
    my $this = shift;
    if (ref $config{search}{skipwords} eq "ARRAY") {
	# skipwords can be written into the configuration
	return $config{search}{skipwords};
    }
    elsif (! ref $config{search}{skipwords} &&
	   -e "conf/$config{search}{skipwords}") {
	# alternatively, they can be kept in a separate file
	if (open F, "<conf/$config{search}{skipwords}") {
	    my @skipwords = <F>;
	    chomp @skipwords;
	    return \@skipwords;
	}
    }
    return [];
}

sub reset {
    my $this = shift;
    $this->{title} = {};
    $this->{descr} = {};
    $this->{site} = {};
    $this->{privacy} = {};
    $this->{module} = {};
    $this->{clear_section} = [];
    $this->{clear_url} = [];
    return;
}

sub search_form {
    my ($this,$term,$title,$width) = @_;
    my $url = $this->link(search=>"1");
    if (! $width) { $width = 20; }
    my $out;
    my $ml = &get_obj("ML");
    if ($title) { $out .= $ml->h1($title); }
    $out .= $ml->form(
		      $ml->input(undef,{type=>"text",size=>$width,name=>"searchterm",value=>$term}).
		      $ml->input(undef,{type=>"submit",value=>"Search"}),
		      {method=>"POST",action=>$url,class=>"search",id=>"searchform"}
		      );
    return $out;
}

sub do_search {
    my ($this,$term) = @_;
    my $ml = &get_obj("ML");
    my $out .= $ml->h1("Search Results for \"$term\"");
    $out .= $this->display_results($this->search($term));
    $out .= $this->search_form($term,"Search");
    return $out;
}

# index_site : indexes all content in a site

sub index_site {
    my ($this,$section) = @_;
    my $site = ref $section ?
	new ExSite::Section(section=>$section) :
	new ExSite::Section(id=>$section);

    # first, index all of the site's pages
    my $out .= $this->index_pages($site->get_all_pages(status=>"active"));

    # second, see if any of the website services wish to add to the index
    $out .= $this->index_plugins($site);

    # remember to clear old index for this site before updating
    push @{$this->{clear_section}},$site->id();

    return $out;
}

# index_plugins : indexes special content for all plugins in a site

sub index_plugins {
    my ($this,$section) = @_;
    my ($site,$out);
    if ((ref $section) =~ /Section/) {
	$site = $section;
    } 
    else {
	if (ref $section) { $site = new ExSite::Section(section=>$section); }
	else { $site = new ExSite::Section(id=>$section); }
    }
    my @service = $site->get_services;
    foreach my $service (@service) {
	if ($service->{page_id}) {
	    my $p = new ExSite::Page(id=>$service->{page_id});
	    my $modname = $service->{name};
	    my $dcd = &get_module($modname,0,$section);
	    #my $dcd = $site->get_dcd($modname);
	    if ($dcd) {
		$out .= $this->index_plugin($dcd,$site,$p);
	    }
	}
    }
    return $out;
}

# index_plugin : indexes special content for a specific plugin

sub index_plugin {
    my ($this,$dcd,$section,$page) = @_;
# this fails for reasons unknown
#    my $search = eval "$dcd->ioctl('Search')";
#    if ($@) {
#	return $this->error("Search::index_plugin: $@");
#    }
    my $search = $dcd->ioctl('Search');
    if ($search) {
	# this service has its own search indexing tools
	return $dcd->$search($this,$section,$page);
    }
}

sub index_pages {
    my ($this,@pages) = @_;
    my $out;
    foreach my $p (@pages) {
	$out .= $this->index_page($p);
	if ($config{search}{scanpause}) { sleep($config{search}{scanpause}); }
    }
    return $out;
}

# indexpage : index the metadata and content of a page

sub index_page {
    my ($this,$page) = @_;
    if (ref $page !~ /Page/) {
	if (ref $page) { $page = new ExSite::Page(page=>$page); }
	else { $page = new ExSite::Page(id=>$page); }
    }
    return undef if (! $page->is_page);   # do not index libs, templates
    #return undef if (! $page->is_public); # do not index private pages
    $share{Page} = $page;
    $page->set_context;
    my $url = $page->get_url;
    return $this->error("$url: Permission denied") 
	if (! $share{DB}->owns($page->get_my("section_id")));
    my $filename = $page->filename;
    my $title = $page->title;
    my $descr = $page->description;
    my $privacy = $page->is_public ? "public" : "members-only";
    $this->add_url($url,$page->id("section"),$title,$descr,$privacy);
    
    my @content = $page->fetch_all_content_objs();
    foreach my $c (@content) {
	$c->set_context();
	my $ctype = $c->content_type;
	next if ($ctype !~ /^text/);
	my $text = $c->get_raw;
	if ($ctype =~ /html/) {
	    $c->expand(html=>$text);
	    $text = $c->{html};
	    # remove stuff outside the body
	    if ($text =~ /<body>(.+)<\/body>/si) { $text = $1; }
	    # remove CSS
	    $text =~ s/<style>.+?<\/style>/ /gsi;
	    # remove Javascript
	    $text =~ s/<script.*?>.+?<\/script>/ /gsi;
	    # remove HTML tags
	    $text =~ s/<\/?[^>]+>/ /g;
	}
	$this->index($url,$this->{weight}{body},$text);
    }

    $this->index($url,$this->{weight}{title},$title);
    $this->index($url,$this->{weight}{keywords},$page->keywords);
    $this->index($url,$this->{weight}{description},$descr);
#    $this->index($url,$this->{weight}{body},$body);
    $this->index($url,$this->{weight}{filename},$filename);

    my $ml = &get_obj("ML");
    my $br = $ml->br;
    return "indexing $url...$br\n";
}

# index : indexes the words in a block of plaintext

sub index {
    my ($this,$url,$weight,$text) = @_;
    return if (!$text);
    return if (! exists $this->{index}{$url});

    # strip endlines
    $text =~ s/\r?\n/ /g;
    # strip all non-word characters
    $text =~ s/[^\w]/ /g;
    # convert to lower case
    $text = lc($text);
    # convert to an array of words
    my @word = split /\s+/,$text;

    # index the words found
    foreach my $word (@word) {
	$this->index_term($url,$word,$weight);
    }
}

# index_term : add a term to the index

sub index_term {
    my ($this,$url,$term,$weight) = @_;
    return 0 if (! $url);
    return 0 if (! $term);
    # don't add the term if it is in the skip list
    if (! exists $this->{skipwords}) { 
	$this->{skipwords} = $this->get_skipwords; 
    }
    return 0 if (grep /^$term$/,@{$this->{skipwords}});
    $weight |= 1;
    $this->{index}{$url}{$term} += $weight;
    # should return current weight for that term
}

# add_url : add a url to the index

sub add_url {
    my ($this,$url,$section,$title,$descr,$privacy,$module) = @_;
    if (! exists $this->{index}{$url}) {
	$this->{index}{$url} = {};
	# remember to clear old index for this URL before updating
	push @{$this->{clear_url}},$url;
    }
    $this->{title}{$url} = $title;
    $this->{descr}{$url} = $descr;
    $this->{site}{$url} = $section;
    $this->{privacy}{$url} = $privacy || "public";
    if ($module && $share{DB}{map}->is_mapped("searchurl","module")) {
	# this URL is specific to a module
	$this->{module}{$url} = $module;
    }
    return;
}

# clear : clear old search data for a particular site

sub clear_site {
    my ($this,$section_id) = @_;
    my $db = $share{DB};
    if ($section_id) {
	my @oldurl = $db->fetch_match("searchurl",{section_id=>$section_id});
	foreach my $oldurl (@oldurl) {
	    # delete links and urls
	    $this->clear_url($oldurl->{searchurl_id});
	}
    }
}

# clear search data for a single URL
# $url can be a searchurl_id or a full URL

sub clear_url {
    my ($this,$url) = @_;
    my $db = $share{DB};
    my @urlid;
    if ($url =~ /^\d+$/) { 
	$db->delete("searchterm",{searchurl_id=>$url});
	$db->delete("searchurl",{searchurl_id=>$url});
    }
    else {
	my @oldurl = $db->fetch_match("searchurl",{url=>$url});
	foreach my $u (@oldurl) {
	    if ($db->owns($u->{section_id})) {
		$db->delete("searchterm",{searchurl_id=>$u->{searchurl_id}});
		$db->delete("searchurl",{searchurl_id=>$u->{searchurl_id}});
	    }
	}
    }
}

# update : update the DB index

sub update {
    my ($this,$threshold) = @_;
    if (!$threshold) { $threshold = 1; }
    my $db = $share{DB};

    # first, remove older indexes
    foreach my $site (@{$this->{clear_section}}) {
	$this->clear_site($site);
    }
    foreach my $url (@{$this->{clear_url}}) {
	$this->clear_url($url);
    }

    # then, add the new indexing data
    foreach my $url (keys %{$this->{index}}) {

	# clear old search data
	my @oldurl = $db->fetch_match("searchurl",{url=>$url});
	foreach my $oldurl (@oldurl) {
	    $this->clear_url($oldurl->{searchurl_id});
	}
	
	# add new data

	my %surl = (
		    url=>$url,
		    title=>$this->{title}{$url},
		    description=>$this->{descr}{$url},
		    section_id=>$this->{site}{$url},
		    privacy=>$this->{privacy}{$url},
		    );
	if ($share{DB}{map}->is_mapped("searchurl","module")) {
	    # include module, if supported
	    $surl{module} = $this->{module}{$url};
	}

	my $url_id = $db->insert("searchurl",\%surl);
	foreach my $word (keys %{$this->{index}{$url}}) {
	    next if (! $word);
	    $db->insert("searchterm",{searchurl_id=>$url_id,
				      term=>$word,
				      weight=>$this->{index}{$url}{$word}});
	}
    }
}

# search the search table

sub search {
    my ($this,$terms,$module) = @_;

    my $db = $share{DB};
    #$terms =~ s/[^\w]/ /g;  # eliminate punctuation
    $terms =~ s/(?<=\S)[+-]|[^\w+-]/ /g;  # eliminate punctuation
    my @term = split /\s+/,lc($terms);
    my $nterm = 0;

    # "+foo" means "foo" is a required term in the results
    # "foo" means "foo" is an optional term in the results
    # "-foo" means "foo" is a forbidden term in the results

    my (@required,@optional,@blocked,$resultset,$required_terms);
    my $section = $share{Page} ? $share{Page}->id("section") : 0;
    my $col_searchurl_id = "searchurl.searchurl_id";
    foreach my $term (@term) {
	if ($term =~ /^\+/) {
	    # required term
	    $resultset = \@required;
	    $required_terms = 1;
	    $term =~ s/^\+//;
	    $nterm++;
	}
	elsif ($term =~ /^\-/) {
	    # forbidden term
	    $resultset = \@blocked;
	    $term =~ s/^\-//;
	}
	else {
	    # optional term
	    $resultset = \@optional;
	    $nterm++;
	}
	my $match = {
	    term=>$term,
	    "searchterm.searchurl_id"=>\$col_searchurl_id
	    };
	if ($config{search}{scope} ne "global") {
	    # restrict to local section, unless global searching is enabled
	    $match->{section_id} = $section;
	}
	if ($module && $share{DB}{map}->is_mapped("searchurl","module")) {
	    # restrict search to a particular plug-in
	    $match->{module} = $module;
	}
	push @$resultset, $db->select(["searchterm","searchurl"],undef,$match);
    }

    # combine the results into a single result hash

    my ($termresult,%result);
    
    $this->{search_results} = {};
    foreach $termresult (@required) {
	$result{$termresult->{searchurl_id}} += ($termresult->{weight} / $nterm);
	$this->{search_results}{$termresult->{searchurl_id}} = $termresult;
    }
    foreach $termresult (@optional) {
	# optional results count if there are no required terms, or 
	# if the result URL was already selected by the required terms
	if (! $required_terms || exists $result{$termresult->{searchurl_id}}) {
	    $result{$termresult->{searchurl_id}} += ($termresult->{weight} / $nterm);
	    $this->{search_results}{$termresult->{searchurl_id}} = $termresult;
	}
    }
    foreach $termresult (@blocked) {
	# remove blocked URLs from the results
	delete $result{$termresult->{searchurl_id}};
	delete $this->{search_results}{$termresult->{searchurl_id}};
    }
    
    return %result;
}

# format the search results 

sub display_results {
    my ($this,%results) = @_;
    my @url_id = sort { $results{$b} <=> $results{$a}; } keys %results;
    my $ml = &get_obj("ML");
    my $out;
    my $db = $share{DB};
    my $nresult = 0;
    foreach my $url_id (@url_id) {
	my $url = $this->{search_results}{$url_id} || 
	    $db->fetch("searchurl",$url_id);
	next if ($url->{privacy} && 
		 $url->{privacy} ne "public" && 
		 ! $db->level); # members only
	$nresult++;
	my $score = int($results{$url_id}*10)/10;
	if ($url->{url} =~ /^$config{server}{CGIpath}/) {
	    # show complete URLs
	    $url->{url} = $config{server}{server}.$url->{url};
	}
	$out .= $ml->dt( $ml->a($url->{title},{href=>$url->{url}}) );
	my $descr = $url->{description} . 
	    ($url->{description} ? $ml->br : "") .
	    $ml->span("[$url->{url}, score: $score]",{class=>"searchdetails"});
	$out .= $ml->dd($descr);
	last if ($nresult == $config{search}{maxresult});
    }
    if ($nresult == 0) {
	$out .= $ml->dt($config{search}{no_results});
    }
    return $ml->dl($out);
}

=pod

=head1 Search.pm

The search system is based on a simple search index that uses 2 tables:

=over 4

=item * searchurl

every searchable URL is represented once here

=item * searchterm

every indexed word at each url is represented once here,
with a weight that is used for calculating relevance.  The weight
is based on the location of the word at the URL, and the number of
occurences.

=back

The search index is constructed by finding all content that is unique
to a URL (eg. the content objects that are page-specific), stripping
out all mark-up, and counting the incidence of each remaining word in
the text.  For web pages, we also index words in filenames, titles,
descriptions, and keywords.

Indexing of a web site consists of:

=over 4

=item 1 indexing the regular content on each page

The search system finds this content and indexes it automatically.  It
ignores content that is not page-specific, such as text that comes
from templates, menus, and so forth.

=item 2 indexing the content in each plug-in

Individual plug-ins can devise their own content indexing logic.  To
make a plug-in search indexing tool available to the system, the
plugin must be defined as a site service, and must reply with a code
reference to the the "Search" ioctl command.  This code reference is
the plugin's search indexer.  It will be invoked with three
parameters:

=over 4

=item an ExSite::Search object

Using this, the plug-in can add terms into the search index.

=item an ExSite::Section object

Using this, the plug-in can constrain which section's content gets indexed.

=item an ExSite::Page object

This indicates the page that should be used to deliver the plugin's
content in saerch results.  In other words, the plug-in should index
URLs that generate this page, albeit with alternate query string
parameters.

=back

=back

=head2 Usage

Make a search object:

    my $s = new ExSite::Search;

You must first index your site(s) before you can perform any searches:

    $s->index_site($section);

C<$section> can be a section ID or a section datahash.

To generate a search form:

    my $form_html = $s->search_form($term,$title,$width);

The parameters are all optional.  C<$term> is a term to prepopulate
the search field with.  C<$title> is a title/heading.  C<$width> is
the size of the search field (in characters).

To perform a search on the terms in a search string:

    my $results_html = $s->do_search($searchstring);

To get just the list of search hits:

    my $results_html = $s->display_results( $s->search($searchstring) );

The Search plug-in provides a simple interface to these functions.

=head2 Search Term Rules

The search system breaks each block of content down to a stream of
plain text.  All tags and non-text content (such as scripts and CSS)
are removed, to leave just the human-readable words and text on the
page.  Then we strip out all punctuation and other non-word characters
to leave just alphanumeric text and whitespace.  We convert the text
to lower case, and break it out into individual terms, splitting on
whitespace.  This has a few consequences that may be important for the
developer to understand, such as:

=over 4

=item * hyphenated words such as "over-easy" will be broken into two terms,
"over" and "easy".

=item * contractions such as "haven't" will be broken into two terms,
"haven" and "t".

=item * it only works on pages/sites that use languages that delimit
their words by whitespace and punctuation.  Other languages, such as
Chinese, will not work.

=back

Each term is then counted, and the count is multiplied by a weight
factor for that content block.  The resulting score determines how
significant a hit on that term is for that URL.

=head3 Advanced Searching Options

Search terms can optionally be prefixed with a + or - character, which
changes the search rules:

=over 4

=item term

The term is desired, but optional, in the search results.  Since at
least one term must produce a hit, if only one optional term is given,
then it is effectively a required term.  If more than one optional
term is given, at least one of them is required.

=item +term

The term is B<required> in the search results.  Results that do not
contain this term will not be reported.

=item -term

The term is B<forbidden> in the search results.  Results that 
contain this term will B<not> be reported.

=back

You can combine these for some extra logical control over your
searches.  For example:

=over 4

=item foo bar

Search for "foo" B<or> "bar".  (But pages that have both terms will tend to
be more relevant.)

=item +foo +bar

Search for "foo" B<and> "bar".

=item +foo bar

Search for "foo" and optionally "bar".  (Ie. search for "foo", but
if "bar" is also found, it will increase the relevance of the hit.)

=item foo -bar

Searches for pages containing "foo", but excludes pages containing
"bar" from the results.

=back

=head3 Skipwords

Certain terms can be ignored entirely by the search index.  These
skipwords are simply not inserted into the index, no matter how often
or where they appear.  They are ignored in search queries, and
attempts to search for just these terms will find nothing.

There are two ways to define the list of skipwords.  Method 1 is to
simply list them in the configuration parameter
C<$config{search}{skipwords}>.  You can add to this list using the
configuration file notation:

    search.skipwords += foo
    search.skipwords += bar

If the C<search.skipwords> parameter is not an array of works, but is
just a scalar string, that string is understood to be a file
containing the skipwords, one per line.  For example:

    search.skipwords = skipwords.txt

This file will be sought in the C<conf> subdirectory of C<cgi-bin>.
A fairly comprehensive sample file is included with ExSite, containing
over 500 words that by themselves carry little meaning and therefore
do not help to distinguish one search topic from another.  This file
may be edited or replaced as needed.

=head2 Limitations

You cannot search for partial words.  For example "surf" does not
match "surfing".

Quotes are ignored, and any words in a quoted phrase are searched for
individually.

Searches for negative numbers, eg. "-99" will be understood to mean
"exclude '99' from the search results".

It does not index alt tags on images.

It does not index any plug-ins that have not been configured as a service.

Only English skipwords are provided.

=cut

1;
