#!/usr/bin/perl
#----------------------------------------------------------------------------
#
#   Copyright 2001-2006 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

=pod

=head1 ExSite::Section - section and website management

The Section class handles configuration, publication, and other management
of website sections.

Formally, a section is a group of pages with similar publication and
access control attributes.  All static pages in a section publish into
the same directory, and all member-only pages are accessible by the
same group of members.

Many small websites are configured as a single section; therefore the
terms "section", "website", and "site" are sometimes used
interchangeably.

C<ExSite::Section> is part of the ExSite content stack, so it inherits
from C<ExSite::Page>, which in turn inherits from C<ExSite::Content>,
C<ExSite::ContentData>, and C<ExSite::ContentBase>.

Note that sections have somewhat different behaviours, depending on their
type:

=over 4

=item stand-alone website

"Stand-alone" means the section must explicitly specify its
publication folder (as a complete directory path), and its domain
name.

=item native website

"Native" means the section implicitly infers its publication path and
domain name from the parent section.  If there is no parent, then these
are inferred from the ExSite system configuration.  The publication
directory is simply a directory name.

=item website template

A website template is a model section, used to initialize new sections
with thesame pages that the website template has.  Website templates
do not publish.

=back

=cut

package ExSite::Section;

########### needed:
# get_webmasters
# is_webmaster
# section_path
# get_pages ! this section only
# get_pages_from_path ! all sections in path

use strict;
use ExSite::Config;
use ExSite::ML;
use ExSite::Misc;
use ExSite::Page;
use ExSite::Module;
use MIME::Base64;

use vars qw(@ISA);
@ISA = qw(ExSite::Page);

=pod

=head1 Configuration & Setup

=head2 new()

(This method is actually inherited from ExSite::ContentBase.)

=head2 setup()

=head2 setupSection()

Instantiate and/or configure a section object.  C<%option> includes the 
the configuration parameters, described below.

Usage:

    my $section = new ExSite::Section(%option);
    $section->setup(%option);
    $section->setupSection(%option);

C<setup> and C<setupSection> are synonymous;  they are used to setup or
reconfigure an existing Section object.  C<new()> instantiates a new section
object (this method is actually inherited from C<ExSite::ContentBase>).

C<%option> is used to pass the section description information that
allows the object to be configured.  It can include the following section
description parameters:

=over 4

=item id

The numeric section ID will be used to fetch the section description from 
the database.

=item section

The section record can be passed as a hash reference, if it has already been
fetched.

=back

If the section is not set up with an actual section description, the object
will have limited functionality.  The setup methods are normally invoked
automatically by the class contructor method C<new()>, which receives the
same C<%option> parameters.

=cut

# new() constructor is inherited from ContentData class.

sub setup {
    # so that we can do setup using a non-overloaded method
    &setupSection(@_);
}

sub setupSection { 
    my ($this,%option) = @_;

    my $section;
    if ($option{id}) {
	# lookup section record
	$section = $this->fetch("section",$option{id});
    }
    elsif (ref $option{section} eq "HASH") {
	# section record was passed
	if (ref $option{section} eq "HASH") {
	    $section = $option{section};
	}
	else {
	    $this->error("invalid section parameters passed to setupSection");
	}
    }
    if ($section) {
	$this->info("setting up section $section->{section_id}");
	$this->{data} = $section;
	$this->set_context("none");
    }
    elsif ($option{id}) {
	$this->error("section $option{id} not found");
    }
    else {
	$this->warn("setting up null section");
    }
}

# previous - change the section object to next one up in path

sub previous {
    my $this = shift;
    my $section = $this->get;
    if ($section->{parent_id}) {
	$this->setupSection(id=>$section->{parent_id});
	return $section->{parent_id};
    }
    else {
	return undef;
    }
}

=pod

=head1 Section Address

=head2 filename()

Returns the publication name of the section.  This is only meaningful
for native websites.

=head2 diskpath()

Returns the full disk path where the section's files will be published.
The disk path depends on the section's type and relationship to its 
parent section.

=head2 httppath()

(DEPRECATED) Returns the path relative to the webroot where the
section's files will be found.  

=head2 get_url()

Returns the URL to the section.  The URL references the section
directory, which normally defaults to the C<index.html> or equivalent
file published in that directory.  The URL includes the hostname
and protocol information.

=head2 get_host()

Returns just the machine name part of the URL, eg. "www.mydomain.com".

=head2 sectionpath()

Sections can be nested to any depth.  This method returns an array of
section objects representing all sections in this path, from the root
section to the current section.

=head2 sectionpath_url()

Returns the C<sectionpath()> as a simple string, of the format
"/section1/section2/...".  Native sections in the path will be named
for their publication location; stand-alone sections will be named
using their section ID.  For instance:

    /native1/2/native3

This might represent a native website "native3" that is a child of a
stand-alone website (section ID 2), which is a child of a native
website "native1".  Note that this in turn may be a child of a native
website with a blank publication directory.  (The root section in a
system is typically a native website with no publication directory,
since it publishes to the default directory specified by the
webserver.)

=cut

# name of section directory - overloaded variation on ContentData method

sub filename {
    my $this = shift;
    my $section = $this->get;
    if ($section) {
	if ($section->{type} =~ /native/) {
	    return $section->{publish_to};
	}
    }
    # filename is not defined for other website types
    return undef;
}

# disk path to section

sub diskpath {
    my $this = shift;
    my $section = $this->get;
    if ($section) {

	# no disk path for templates and remote sites
	return undef if ($section->{type} =~ /template/);

	# clean the path so it can be published to
	my $path;
	{
	    # new block to limit scope of $1
	    $section->{publish_to} =~ /^(([\/\w][\w\-\.]+)+)$/;
	    $path = $1;
	}

	if ($section->{publish_to} && ! $path) {
	    # path failed validation check
	    $this->error("Invalid section publish directory");
	}
	elsif ($section->{type} =~ /stand-alone/) {
	    # stand-alone sites hold full path
	    return $path;
	}
	else {
	    if ($section->{parent_id}) {
		# this section nests within the parent section
		my $psec = new ExSite::Section(id=>$section->{parent_id});
		my $parent_path = $psec->diskpath;
		return "$parent_path/$path" if ($parent_path);
	    }
	    elsif ($section->{publish_to}) {
		return "$config{server}{HTMLroot}$config{server}{HTMLpath}/$path";
	    }
	    else {
		# root section of the native site
		return "$config{server}{HTMLroot}$config{server}{HTMLpath}";
	    }
	}
    }
    return undef;
}

# web path to section

sub httppath {
    my $this = shift;
    my $section = $this->get;
    if ($section) {
	# can only give a partially-qualified URL for native sites
	if ($section->{type} =~ /native/) {
	    if ($section->{parent_id}) {
		# this section nests within the parent section
		my $psec = new ExSite::Section(id=>$section->{parent_id},
					       context=>"none");
		my $parent_path = $psec->httppath;
		# if parent is undefined, that's because we are nested 
		# under a non-native website
		return undef if ! defined $parent_path;
		return "$parent_path/$section->{publish_to}";
	    }
	    elsif ($section->{publish_to}) {
		return "$config{server}{HTMLpath}/$section->{publish_to}";
	    }
	    else {
		# root section of the native site - return base HTML path
		return $config{server}{HTMLpath};
	    }
	}
	elsif ($section->{type} !~ /template/) {
	    return $section->{url};
	}
    }
    return undef;
}

# overloaded get_url: return a URL for the section

sub get_url {
    my $this = shift;
    my $url;
    my $section = $this->get;
    return undef if ($section->{type} =~ /template/);
    while ($section) {
	if ($section->{type} =~ /native/) {
	    if ($section->{publish_to}) {
		$url = "/".$section->{publish_to}.$url;
	    }
	}
	else {
	    $url = $section->{url}.$url;
	    last;
	}
	if ($section->{parent_id}) {
	    $section = $this->fetch("section",$section->{parent_id});
	}
	else { 
	    $section = undef;
	}
    }
    if (! $url || $url =~ /^\/\w/) {
	#$url = $config{server}{HTMLpath}.$url;
	# DANGER: admin server may not be the same as public server
	$url = $config{server}{HTMLpath}.$url;
	return $url if ($config{page}{relative_static_url});
	$url = $config{server}{server}.$url;
    }
    elsif ($url =~ /^\w+\./) {
	$url = "http://$url";
    }
    return $url;
}

# get_host : return only the machine name part of the section URL

sub get_host {
    my $this = shift;
    my $url = $this->get_url;
    $url =~ /^https?:\/\/([^\/]*)/;
    return $1;
}

# sectionpath : return an array of the sites/sections traversed to get to 
#               this site

sub sectionpath {
    my $this = shift;
    my $section = $this->get();
    if ($section) {
	my @parent_path;
	if ($section->{parent_id}) {
	    my $psec = new ExSite::Section(id=>$section->{parent_id});
	    @parent_path = $psec->sectionpath;
	}
	return (@parent_path,$section);
    }
}

# sectionpath_url : returns the section path as a URL-friendly string

sub sectionpath_url {
    my $this = shift;
    my @path;
    my @s = $this->sectionpath;
    while (@s) {
	my $s = pop @s;
	unshift @path, $s->{type} =~ /native/ ? 
	    $s->{publish_to} : 
	    $s->{section_id};
    }
    if ($path[0]) { unshift @path, ""; }
    return join("/",@path);
}

sub service_page {
    my ($this,$module) = @_;
    return &ExSite::Module::service_page($module,$this);
}

=pod

=head1 Content Management

=head2 get_all_pages()

Returns an array of page objects for all of the active pages in the
section.

=head2 get_mapped_pages()

Returns an C<ExSite::Tree> object containing all of the section's
pages organized into a site map.  The site map ignores the page
visibility attribute that determines whether to include the pages in
displayed site maps and menus, so code that constructs such menus
should consider the page's visibility when deciding whether to show
each page.

=head2 fetch_libraries()

Returns an array of library records (not objects) that are accessible
to the current section.  This includes libraries in the current
section, shared libraries in the the sectionpath, and public libraries
anywhere in the system.

=head2 fetch_library_content()

Returns an array of content objects for all of the content in all of
the libraries returned by the previous method.

=head2 get_templates()

Returns all templates available to the section, as an array of 
page records.  This may include templates from other sections, depending
on their privacy settings and relationship to the current section.

    my @templates = $section->get_templates(%option)

C<%option> can specify:

=over 4

=item section_id

This is the section to search under, if not the current section.

=item mode

This is an integer that determines which templates will be selected
for inclusion:

=over 4

=item "-1"

select all templates on the system (administrators only, otherwise
treated as 0)

=item "0"

select templates useable by this site (default) (use when changing
template of a site)

=item "1"

select templates useable by children of this site
(use when creating a child site)

=item "2"

select public templates only
(use when creating a new independent site)

=back

=back

=head2 set_template()

    $section->set_template($template)

Changes the graphic design template for the section to C<$template> (which
may be a page ID, or a page hashref).  If allowed, the section's default
template is changed, and all of the regular pages are changed to use the
template.  Nothing is returned.

If not allowed, an error message is returned, which can be issued
directly to the browser.

=head2 get_bodies()

Returns an array of content records, representing model page bodies that 
can be used by the current section.  These can be used to initialize the
body of any page in the current section.

=head2 get_services()

Returns an array of services running on this section.  A service is 
a special record connecting a section to a DCD (plug-in).  It is used
for authorizing access to restricted-access DCDs, and finding special
service pages that should be used to handle all requests to a DCD.

=head2 find_dcd()

Finds pages using a particular DCD (plug-in).

    my @pages = $section->find_dcd($module);

Returns an array of page records representing those pages that contain
a tag invoking the DCD C<$module> in the newest revision of their content.

=cut

# get home page (as a page object)

sub homepage {
    my ($this,$version) = @_;
    $this->set_context();
    my $p = new ExSite::Page(page=>$this->get("page"));
    $p->set_version($version) if ($version);
    return $p;
}

# get all pages for the section - returns array of Page objects
#
# %opt is a match hash for pages

sub get_all_pages {
    my ($this) = @_;
    my ($sid,@Page);
    $sid = $this->id() || 0;
    if ($this->is_prefetched($sid)) {
	my %found;
	foreach my $p (keys %{$share{prefetch}{$sid}}) {
	    # skip the section record in the prefetch cache
	    next if ($p eq "_");
	    # pages are in the prefetch cache twice - once under the ID
	    # and again under the filename;  skip if we already have one
	    # of them.
	    next if ($found{$share{prefetch}{$sid}{$p}{_}{page_id}});
	    push @Page, new ExSite::Page(page=>$share{prefetch}{$sid}{$p}{_});
	    $found{$share{prefetch}{$sid}{$p}{_}{page_id}} = 1;
	}
    }
    else {
	my @pages = $this->fetch_match("page",{section_id=>$sid,status=>"active"});
	foreach my $page (@pages) {
	    push @Page, new ExSite::Page(page=>$page);
	}
    }
    return @Page;
}

# get all pages for the section, organized into a map (ExSite::Tree)
#
# options are:
# version => alternate version/language
#
# Note: the map includes pages with reduced visibility.  Applications
# that honour page visibility (eg. menu builders) should test the 
# visibility attribute themselves.

sub get_mapped_pages {
    my ($this,$parent_id,$opt) = @_;
    my $match = {section_id=>$this->id(),
		 status=>"active",
		 type=>"page",
		 };
    my @pages = $this->fetch_match("page",$match,["rank","page_id"]);
    

    # make a basic Tree/map from the regular pages
    require ExSite::Tree;
    my $map = new ExSite::Tree("page_id","parent_id",@pages);

    # do we need to consider alterate pages?
    my $version = $opt->{version};
    if (! $version && $parent_id) {
	# no version explicitly specified, but we should check the version
	# of the parent page, just to be sure

	# see if the parent is already loaded in the map
	my $p = $map->getnode_data($parent_id);
	# if not, look it up in the DB
	if (!$p) { $p = $this->fetch("page",$parent_id); }
	if ($p->{version}) { $version = $p->{version}; }
    }
    
    # overlay the alternate pages
    if ($version) {
	$match->{type} = "alternate";
	@pages = $this->fetch_match("page",$match,["rank","page_id"]);
	# clone the list, because we modify it in replacenode, and don't want
	# to modify the version in cache
	my $clonedpages = &clone(\@pages);
	foreach my $alt (@$clonedpages) {
	    # the alternate's parent points to the page it replaces
	    $map->replacenode($alt->{parent_id},$alt);
	}
    }
    return $map;
}

sub get_mapped_subsections {
    my ($this,$opt) = @_;
    my @site = $this->fetch_all("section","section_id");

    # make a global Tree/map
    require ExSite::Tree;
    my $allmap = new ExSite::Tree("section_id","parent_id",@site);

    # extract the subtree corresponding to us
    my $submap = $allmap->subtree($this->get_my("section_id"));

    # drop ourselves from the subtree, so we only get our children
    $submap->splice($this->get_my("section_id"));

    return $submap;
}

# fetch_libraries : fetches all libraries in the site path.

sub fetch_libraries {
    my $this = shift;
    my @sections = $this->sectionpath;
    my @Library;
    my $first = 1;
    while (my $s = pop @sections) {
	my @libs = $this->fetch_match("page",{type=>"library",section_id=>$s->{section_id}});
	foreach my $lib (@libs) {
	    # only the last one in the path is accessible by administrators
	    next if ($lib->{access} eq "administrators" && ! $first);
	    push @Library,$lib;
	    $first = 0;
	}
    }
    return @Library;
}

# fetch_library_content : grabs all content objects in all libraries in the
#                         site path.  This could be useful for inserting 
#                         shared content into pages, or it could be a pig.
#
# mime_re is a regexp to filter fetched content with, eg. to select 
# all pdf files, or all jpgs, or all non-text items.

sub fetch_library_content {
    my ($this,$mime_re) = @_;
    my @sections = $this->sectionpath;
    my @libs = $this->fetch_libraries;
    my @Content;
    foreach my $l (@libs) {
	my @cont = $this->fetch_match("content",{page_id=>$l->{page_id}});
	foreach my $c (@cont) {
	    my $cobj = new ExSite::Content(content=>$c);
	    $cobj->set_context("child");
	    if (! $mime_re || $cobj->content_type() =~ /$mime_re/) {
		push @Content, $cobj;
	    }
	}
    }
    return @Content;
}

# get all templates available to the section
# - includes public and semi-private templates from other sections
#
# opt:
#   section_id - the section to search under, if not the current section
#   mode - the template selection method:
#      -1 => select all templates on the system
#           (administrators only, otherwise treated as 0)
#      0 => select templates useable by this site (default)
#           (use when changing template of a site)            
#      1 => select templates useable by children of this site
#           (use when creating a child site)
#      2 => select public templates only
#           (use when creating a new independent site)

sub get_templates {
    my ($this,%opt) = @_;
    my ($sid,$db,@templates);
    $sid = $opt{section_id} || $this->id;
    $db = $share{DB};

    if ($opt{mode} < 0 && $db->is_admin()) {
	# fetch all templates in the system
	@templates = $db->fetch_match("page",{type=>"template",
					      status=>"active"});
    }
    else {
	# fetch public templates
	@templates = $db->fetch_match("page",{type=>"template",
					      access=>"public",
					      status=>"active"});
	
	if ($opt{mode} < 2) {
	    # ... plus the member-only templates from the current section
	    push @templates, $db->fetch_match("page",{type=>"template",
						      access=>"members",
						      status=>"active",
						      section_id=>$sid});
	    if (! $opt{mode}) {
		my $sec = $db->fetch("section",$sid);
		if ($sec && $sec->{parent_id}) {
		    # ... plus the member-only templates from the parent section
		    push @templates, $db->fetch_match("page",{type=>"template",
							      access=>"members",
							      status=>"active",
							      section_id=>$sec->{parent_id}});
		}
		# ...plus the private templates from this section
		push @templates, $db->fetch_match("page",{type=>"template",
							  access=>"administrators",
							  status=>"active",
							  section_id=>$sid});
	    }
	}
    }
    return @templates;
}

# change template for all pages

sub set_template {
    my ($this,$template) = @_;
    my $tid = ref $template eq "HASH" ? $template->{page_id} : $template;
    my $sid = $this->id();

    # sanity checks

    if (! $tid) {
	return $this->error("set_template: no template given");
    }
    if (! $sid) {
	return $this->error("set_template: no section given");
    }
    if (! $this->owns()) {
	return $this->error("set_template: you do not have sufficient priveleges to modify this website");
    }
    my @allowed_templates = $this->get_templates(0);
    my $t;
    foreach $t (@allowed_templates) {
	if ($t->{page_id} == $tid) {
	    my $db = $share{DB};
	    # update site
	    $db->update("section",{template_id=>$tid},{section_id=>$sid});
	    # update pages
	    $db->update("page",{template_id=>$tid},{section_id=>$sid,type=>"page"});
	    $db->update("page",{template_id=>$tid},{section_id=>$sid,type=>"alternate"});
	    return;
	}
    }
    return $this->error("set_template: you do not have permission to use this template");

}

# get all bodies available to the section (and its parent sections)

sub get_bodies {
    my ($this) = @_;
    my ($sid,$db,@liblist,@libs,@body_templates,$sql);
    $sid = $this->id;
    $db = $share{DB};

    # find all the libraries that we can take bodies from

    @libs = $db->fetch_match("page",{type=>"library",section_id=>$sid});
    my $section = $this->get("section");
    while ($section->{parent_id}) {
	$sid = $section->{section_id};
	push @libs, $db->fetch_match("page",{type=>"library",section_id=>$sid});
	$section = $this->fetch("section",$sid);
    }

    # get all the bodies in our list of libraries
    foreach my $l (@libs) { 
	push @liblist, $l->{page_id};
    }
    # warning: not portable to non-SQL DB drivers
    $sql = "select * from content where name='body' and page_id in (".join(',',@liblist).")";
    @body_templates = $db->custom_query($sql);

    return wantarray ? @body_templates : \@body_templates;
}

# get_services : fetches all web application services running on this site

sub get_services {
    my $this = shift;
    my @service = $share{DB}->fetch_child("service","section",$this->id);
    return wantarray ? @service : \@service;
}

# find_dcd : find all active oysters for a particular module

sub find_dcd {
    my ($this,$module) = @_;
    my $sid = $this->id();
    my @moddata;
    if ($sid && $module =~ /^\w+$/) {
    
	### WARNING: elaborate custom query - may be mysql-specific

	# We must search for all page-specific content (ie. not template 
	# content) that contains an oyster for the given module.
	#
	# Note: this query selects the "newest" revision, not the active 
	# revision.

	my $regexp = '<!--&+'.$module.'\\\(';
	@moddata = $share{DB}->get_query("oysters",$sid,$regexp);

#	my $query = "select page.*,content.content_id from page,content,content_data where page.page_id=content.page_id and content.content_id=content_data.content_id and page.section_id=$sid and content_data.mime_type=\"text/html\" and page.type=\"page\" and content_data.data rlike \"$regexp\" order by content_data.content_data_id desc;";
#	@moddata = $share{DB}->custom_query($query);
    }

    my (@page,%found);
    foreach my $dat (@moddata) {
	next if ($found{$dat->{content_id}});
	$found{$dat->{content_id}} = 1;
	delete $dat->{content_id};
	push @page, $dat;
    }
    return wantarray ? @page : \@page;
}

# sitemap tool
#
# DEPRECATED - use get_mapped_pages() instead

sub map {
    my ($this,$options,$pageoptions) = @_;
    my $out;
    my $db = $share{DB};
    my $ml = new ExSite::ML;
    my %match = (section_id=>$this->id,
		 type=>"page",
		 status=>"active",
		 parent_id=>0,);
    if (! $options->{show_hidden}) {
	$match{visibility} = "visible";
    }
    my @pages = $db->fetch_match("page",\%match,"rank");
    my $map;
    foreach my $page (@pages) {
	$map .= $this->map_r($page,$options,$pageoptions);
    }
    $out = $ml->p($ml->ul($map,{class=>"sitemap"}));
    if (! $options->{show_hidden} && $options->{footnote}) {
	$out .= $ml->p("NOTE: hidden pages do not appear in the site map.");
    }
    return $out;
}

sub map_r {
    my ($this,$page,$options,$pageoptions) = @_;
    # recursive mapping routine
    my $ml = new ExSite::ML;
    my $db = $share{DB};
    my $pid = $page->{page_id};
    my $url;

    if ($options->{link_to} eq "cms") {
	$url = &relink(path=>"/$page->{section_id}/$pid",cmscmd=>undef);
    }
    elsif ($options->{link_to} eq "dynamic") {
	my $query;
	if (ref $pageoptions eq "HASH") {
	    $pageoptions->{_id} = $pid;
	    $query = &EncodeHash(%$pageoptions);
	}
	else {
	    $query = "_id=$pid";
	}
	#$url = "$config{server}{CGIpath}/page.cgi?$query";
	$url = "$config{server}{CGIpath}/$config{prog}{page}?$query";
    }
    else {
	my $p = new ExSite::Page(page=>$page);
	$url = $p->get_url;
    }
    
    my $linkopt = { href=>$url };
    if ($options->{target}) { $linkopt->{target} = $options->{target}; }
    my $pageitem = $ml->a($page->{label} || "(no label)", $linkopt);

    my @attr;
    if ($options->{show_filename}) { 
	$pageitem .= " ($page->{filename})"; 
    }
    if ($page->{access} ne "public") { 
	push @attr, "RESTRICTED ACCESS"; 
    }
    if ($page->{visibility} eq "hidden") { 
	push @attr, "HIDDEN";
    }
    if (@attr > 0) {
	$pageitem .= " [".join(", ",@attr)."]";
    }
    my $out = $ml->li($pageitem);
    my $subpage = $db->fetch_match("page",{parent_id=>$pid,
					   type=>"page",
					   visibility=>"visible"},
				   "rank");
    if (@$subpage > 0) {
	my $submap;
	foreach my $page (@$subpage) { 
	    $submap .= $this->map_r($page,$options,$pageoptions); 
	}
	$out .= $ml->ul($submap, { class=>"sitemap" });
    }
    return $out;
}

=pod

=head1 Publishing

WARNING: you need write access to the section's publication area to
publish files there.  The webserver normally does not have this
access.  Negotiating this access is normally done by the publish
program (ie. publish.cgi), which has special permissions.  If you do
not have these permissions, then you have to ask publish.cgi to do
this task on your behalf, using C<ExSite::Util::run_publisher()>.

=head2 publish()

    $section->publish($opt);

Creates the section directories on disk, and publishes all publishable
pages and content to these directories.  If C<$opt> is set to
"nofiles", then only the section's base directory is created, and
nothing is published to it.  Otherwise, pages are recursively
published by instantiating page objects for each, and invoking the
C<publish()> method on each.

If C<$config{search}{reindex_on_publish}> is TRUE, then the section's
search index will also be rebuilt at this time.

The output of the publish method is printed direct to stdout, so that
progress can be monitored in real time.  Otherwise for large sections,
long delays can be experienced as the lengthly publish output is
buffered until completion.  This is why publish progress screens are
not templated.

=head2 unpublish()

Removes a section's files from disk.  Recursively unpublishes each
page by callings its C<unpublish()> method, and recursively
unpublishes child sections as well.

=cut

# write the section to disk

sub publish {
    my ($this,$opt) = @_;
    my $section = $this->get;
    if (! $section) { 
	$this->error("Section::publish: no section to publish");
	print "No section to publish!";
	return;
    }
    elsif (! $section->{type} =~ /template/) { 
	$this->error("Section::publish: cannot publish website templates");
	print "Cannot publish website templates!";
	return;
    }
    else { # if native website
	if (! $this->allow_update()) {
	    # publishing has same permission rules as updates
	    print "Permission denied - administrator privileges are required to publish.";
	    return;
	}
	my $diskpath = $this->diskpath;
	if (! $diskpath) { 
	    print $this->show_diagnostics("error","html");
	    return;
	}
	$this->info("Section::publish: publishing section $section->{section_id}");
	# 1) create the section directories
	my $mkdir = $this->mkdir($diskpath);
	if (! $mkdir) {
	    print $this->show_diagnostics("error","html");
	    return;
	}
	print $mkdir;
	return if ($opt eq "nofiles");

	# 2) prefetch site content for better speed
	$this->prefetch($section);
	my $sid = $this->id;
	my @Page = $this->get_all_pages;
	foreach my $p (@Page) { $p->set_revision("newest"); }
	    
	# 3) publish all pages 
	# autoflush stdout so we can monitor the progress...
	my $old_flush = $|;
	$| = 1;

	# first, libraries

	foreach my $p (@Page) {
            print $p->publish unless (! $p->is_library);
	}

	# next, templates

	foreach my $p (@Page) {
	    print $p->publish unless (! $p->is_template);
	}

	# next, plugin content
	foreach my $serv ($this->get_services()) {
	    my $mod = &ExSite::Module::get_module($serv->{name},1,$this);
	    if ($mod) { 
		my $publisher = $mod->ioctl("Publish");
		if (ref $publisher eq "CODE") {
		    $mod->info("Publishing Module $mod");
		    print $mod->$publisher || $@;
		}
	    }
	}

	# next, pages

	my $found_index;
	foreach my $p (@Page) {
	    next if (! $p->is_page);
	    $share{Page} = $p;
	    print $p->publish;
	    my $page = $p->get;
	    if ($page->{filename} =~ /^(default|index).(php|html?)$/) {
		$found_index = $p;
	    }
	}

	# lastly, reindex search
	if ($config{search}{reindex_on_publish} && 
	    $share{Search}) {
	    # update the search index for this page
	    print $share{Search}->index_plugins($this);
	}

	if (! $found_index) {
	    print $this->warn("Section::publish: no index page for section $section->{section_id}");
	    my $ml = new ExSite::ML;
	    print $ml->p("In most cases, it is recommended that you define an index page to serve 
as the front page of your site.  To set a particular page as the index page, 
edit that page's details, and set its filename to \"index.html\" (or 
equivalent, if your webserver is configured to allow other default pages).");
	}
	# reset flush settings
	$| = $old_flush;
    }
    return;
}

# remove the published section files from disk

sub unpublish {
    my ($this) = @_;
    my $section = $this->get;
    if (! $section) { 
	print $this->error("Section::unpublish: no section to remove");
	return;
    }
    elsif (! $section->{type} =~ /template/) { 
	print $this->error("Section::unpublish: cannot unpublish website templates");
	return;
    }
    else { # if native or standalone website
	my $diskpath = $this->diskpath;
	if (! $diskpath) { 
	    print $this->error("Section::unpublish: non-local unpublishing not supported");
	    return;
	}
	elsif (-e $diskpath) {
	    $this->info("Section::unpublish: removing files for section $section->{section_id}");

	    # autoflush stdout so we can monitor the progress...
	    my $old_flush = $|;
	    $| = 1;

	    # unpublish any child sections
	    my @csections = $share{DB}->fetch_match("section",{parent_id=>$this->id,
							       type=>"native website"});
	    foreach my $cs (@csections) {
		my $s = new ExSite::Section(section=>$cs);
		$s->unpublish;
		if ($s->errorcheck()) {
		    $this->error($s->fetch_diagnostics("error"));
		}
	    }
	
	    # remove libraries, templates, pages
	    my @Page = $this->get_all_pages();

	    foreach my $p (@Page) {
		print $p->unpublish;
		if ($p->errorcheck()) {
		    $this->error($p->fetch_diagnostics("error"));
		}
	    }

	    # remove library, template container dirs
	    my $rmdir =  $this->rmdir("$diskpath/_Template");
	    $rmdir .= $this->rmdir("$diskpath/_Library");

	    # remove the section directory if not a top-level site
	    if ($section->{type} eq "native website" &&
		$section->{publish_to}) {
		$rmdir .= $this->rmdir($diskpath);
		if (! $rmdir) {
		    print $this->show_diagnostics("error","html");
		    return;
		}
	    }
	    print $rmdir;

	    # reset flush settings
	    $| = $old_flush;
	}
    }
    return;
}

=pod

=head1 Creating/Modifying Sections

=head2 validate()

    my @errors = $section->validate($sectiondata);
    if (@errors == 0) { # section is OK }

Returns TRUE if the values in the the C<$sectiondata> hashref comprise
a valid and legal page description.  Returns an array of error
messages, which will be empty if the section data validates.

=head2 make()

=head2 make_section()

=head2 install()

   $section->make($sectiondata,$template,$approve_flag);

   $section->install($sectiondata,$template);

Creates a new section using the attributes in the C<$sectiondata> hashref.

The C<make()> and C<make_section()> versions perform validation and
permission checks on the section data first.  The C<install()> method 
does the actual installation.  C<$approve_flag> can be used to pre-authorize
the section creation;  use this if allowing unprivileged users to create
new sections on the system.

C<$template>, if defined, is a section ID that points to a website template.
This is used to initialize the new section with preconfigured pages.

=head2 delete()

Removes the current section from the system.  Records are moved to the
trash bin, and may be recovered from there.  (However, large sections'
trashballs may exceed the query size limit of the database.)  Attempts
to unpublish its files, as well.

=cut

sub validate {
    my ($this,$data) = @_;
    my ($section,$type,@err,@dupe);
    if (! $data) { $data = $this->get("section"); }

    # generic validation (using DBmap validation rules)
    $share{DB}->{action} = "insert";  # ignore missing primary keys
    push @err, $share{DB}->validate_record("section",$data);

    # special validation
    $type = $data->{type};
    if ($type =~ /^native website$/) {
	@dupe = $share{DB}->fetch_match("section",{type=>$type,
						   parent_id=>$data->{parent_id},
						   publish_to=>$data->{publish_to}});
	if ($data->{publish_to} =~ /\//) {
	    push @err, "invalid publication folder for native website";
	}
    }
    elsif ($type =~ /^stand-alone website$/) {
	@dupe = $share{DB}->fetch_match("section",{type=>$type,
						   publish_to=>$data->{publish_to}});
	if (! $data->{publish_to}) {
	    push @err, "stand-alone website has no publication directory";
	}
	elsif ($data->{publish_to} !~ /^\/.+/) {
	    push @err, "stand-alone website requires a complete publication path";
	}
	if (! $data->{url}) {
	    push @err, "stand-alone website is missing its URL";
	}
    }

    if (@dupe > 0) {
	# check for dupes
	if (@dupe > 1) {
	    push @err, "a section with that publication location already exists";
	}
	elsif (@dupe == 1) {
	    if ($data->{section_id} != $dupe[0]{section_id}) {
		push @err, "a section with that publication location already exists";
	    }
	}
    }

    return @err;
}

# make_section : insert a new section record
#
# can optionally pass website template ID to configure it with


# make : alias for make_section()

sub make {
    return &make_section(@_);
}

# make_section: validate data, then call install_section()
#
# If approve_flag is not true, then the operation only proceeds if the user
# has sufficient access to create website sections.  If allowing site creation 
# by unpriveleged users, you must set this flag to true.

sub make_section {
    my ($this,$section,$template,$approve_flag) = @_;
    my @err;
    if ($section->{section_id}) {
	push @err, "invalid attempt to insert an already-defined section record";
    }
    if (! $approve_flag && ! $this->allow_insert("section",$section)) {
	# permission denied
	push @err, "make_section: permission denied";
    }
    push @err, $this->validate($section);
    
    if (@err > 0) {
	$this->error(@err);
	return undef;
    }

    return $this->install($section,$template);

}

# install: actually creates the section (assumes the data is validated)

sub install {
    my ($this,$section,$template) = @_;
    
    my $type = $section->{type};
    my $db = $share{DB};

    # remove any residual timestamps
    delete $section->{mtime};
    delete $section->{ctime};
    $db->run_handler("pre_make_section",$section,$template);
    my $sid = $db->insert("section",$section);
    if (! $sid) {
	$this->error("failed to install new site \'$section->{filename}\'");
    }
    else {
	$this->info("installed new page $section->{filename}");
	if ($db->authorize >= 2) {
	    # add current user as an administrator
	    $db->insert("member_site_link",{member_id=>$db->my_uid,
					    section_id=>$sid,
					    priveleges=>"editor; designer; administrator"});
	}
	my $pid;
	if ($template) {
	    # copy pages
	    my %page_lookup;
	    my @pages = $db->fetch_match("page",{section_id=>$template},"parent_id asc, page_id asc");
	    # Use a page id lookup table to set parent_id for new pages
	    my %parent_lookup;
	    foreach my $page (@pages) {
		my %page = %$page;
		my %new = %page;
		delete $new{page_id};  # force an insert
		delete $new{ctime};
		delete $new{mtime};
		$new{template_id} = $section->{template_id};
		$new{section_id} = $sid;
		my $pid = $db->insert("page",\%new);
		# Build page_id lookup hash
		$parent_lookup{$page->{page_id}} = $pid || 0;
		# Set parent_id
		if ($page->{parent_id}){
			my $page_parent_id = 0;
			if (exists($parent_lookup{$page->{parent_id}})){
				$page_parent_id = $parent_lookup{$page->{parent_id}};
			}
			$db->update("page",{parent_id=>$page_parent_id},{page_id=>$pid});
		}
		# index the new pages by filename
		$page_lookup{$new{filename}} = $pid;
		if (! $pid) {
		    $this->error("failed to add page \"$page->{title}\"");
		}
		else {
		    # make copies of all content objects
		    my @content = $db->fetch_match("content",{page_id=>$page->{page_id}});
		    foreach my $cont (@content) {
			my $oldcid = $cont->{content_id};
			delete $cont->{content_id};  # force an insert
			$cont->{page_id} = $pid;
			my $cid = $db->insert("content",$cont);
			if (! $cid) {
			    $this->error("failed to add \"$cont->{name}\" content to page \"$page->{title}\"");
			}
			else {
			    # copy latest version
			    my @cdata = $db->fetch_match("content_data",
							 {content_id=>$oldcid},
							 "content_data_id");
			    if (@cdata > 0) {
				delete $cdata[-1]->{content_data_id};
				$cdata[-1]->{content_id} = $cid;
				my $cdid = $db->insert("content_data",$cdata[-1]);
				if (! $cdid) {
				    $this->error("failed to initialize \"$cont->{name}\" in page \"$page->{title}\"");
				}
			    }
			}
		    }
		}
	    }
	    # copy plug-in licenses
	    my @licenses = $db->fetch_match("service",{section_id=>$template});
	    # index the template pages by id
	    my $tpage_lookup = &keywise("page_id",\@pages);
	    foreach my $lic (@licenses) {
		delete $lic->{service_id};
		$lic->{section_id} = $sid;
		if ($lic->{page_id}) {
		    $lic->{page_id} = $page_lookup{$tpage_lookup->{$lic->{page_id}}->{filename}};
		}
		my $lid = $db->insert("service",$lic);
	    }
	}
    }
    $db->run_handler("post_make_section",$section,$template);
    return $sid;
}

# delete the section

sub delete {
    my $this = shift;
    my $id = $this->id;
    if ($id && $this->allow_delete("section",$this->{data})) {
	$this->unpublish();
	$share{DB}->trash_r("section",$id);
	return 1;
    }
    return 0;
}

1;


