package ExSite::Time;

#-----------------------------------------------------------------------
#
#   Copyright 2001-2006 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

=pod

=head1 ExSite::Time 

Utilities for handling dates and times in various formats.

We support the following time formats

    sql_date            yyyy-mm-dd
    iso_date            yyyy-mm-dd
    sql_datetime        yyyy-mm-dd hh:mm:ss
    time                hh:mm AM/PM
    time24              hh:mm (24-hour clock)
    hms                 hh:mm:ss (24-hour clock)
    sql_timestamp       yyyymmddhhmmss OR yyyy-mm-dd hh:mm:ss *
    unix_timestamp      nnnnnnnnnnn
    date                mmm dd, yyyy
    date_long           Month dd, yyyy
    datetime            wday Mon dd hh:mm yyyy
    datetime_long       Weekday Month dd hh:mm yyyy
    datetime2           hh:mm AM/PM, Month dd, yyyy
    arpa                wday, dd Mon yy hh:mm:ss ZONE
    cookie              Wdy, dd-mmm-yyyy hh:mm:ss GMT
    full                Weekday, Month dd, yyyy at hh:mm AM/PM
    dd-mmm-yy           dd-Mon-yy
    weekday             wday
    weekday_long        Weekday
    year                yyyy
    month               Mon
    month_long          Month
    raw                 [y,m,d,h,m,s]
    raw_year		yyyy
    raw_month		mm
    raw_day		dd
    raw_hour		hh
    raw_minute		mm
    raw_second		ss
    raw_weekday         nn

* depends on $config{timestamp_format}, which normally depends on which
which version of MySQL you are running.

Internally, we convert all dates/times to a 6-element time array: 
y, m, d, h, m, s.  The time array is initialized to the current datetime, 
so if you only set a partial datetime (eg. only the time), the remaining 
values are unchanged, and the full array still represents a complete 
datetime.

=head2 Usage

Make a time object (initialized to current time):

    my $t = new ExSite::Time;

Make a time object (initialized to some other time/format):

    my $t = new ExSite::Time("2001-12-25 09:00:00","sql_datetime");

Set the time to the current time:

    $t->set;

Set the time to some time/format:

    $t->set('Feb 1, 1968','date');
    $t->set('6:30 pm','time');
    $t->set('Fri Aug 26 12:14:43 2005','datetime');

=head3 Validation

Times can be correctly formatted, but still meaningless or even just
questionable.  For example:

    Feb 30, 2008       # obviously illegal day
    Feb 29, 2000       # not-so-obviously illegal day
    Feb 28, 1008       # strange, but legal, year

Time validation can catch these and other cases.  
To validate the date/time value:

    my $error = $t->validate();

The return value, C<$error>, will be set to a message explaining the 
formatting problems, if there are any.  Otherwise, it returns C<undef>.
To validate, date values must fall in the following ranges:

    year    within 100 years of current date (configurable)
    month   1 - 12
    day     1 - number of days in month
    hour    0 - 23
    min     0 - 59
    sec     0 - 59

=head2 Adjusting Times

To adjust the time by some amount:

    $t->add(30,'days');
    $t->add(-3,'months');

NB: some time units are not precisely defined, so we must approximate.
What we do is define 1 month as 1/12 of a year, and 1 year as 365
days.  These may not give expected results.  For example, adding 1
month will not leave you at the same time of day, nor on the same day
of the following month. (To understand why this is correct, consider
what the date should be if we start from Jan 31, and add 1 month.)

=head2 Displaying Times

Output the time in some format:

    print $t->write('date');
    print $t->write('unix_timestamp');
    print $t->write('weekday');

=head2 Comparing Times

Compare another time to the currently-set time
(returns 1 if the compared time is in the future, -1 if it is in 
the past, and 0 if it is the same as the current time):

    $t->compare('20050826091500','sql_timestamp');

If directly comparing two instances of this class, the format string
is not needed:

    $t->compare($t2);

Compare two Time instances using operator overloading:

    $t <=> $t2
    $t > $t2
    # etc...

Note that the <=> operator and all other comparison operators only compare
the date portion of an object.  To compare the entire date and time, use
the subtraction operator, eg:

    sort { $a - $b }   # time-sensitive sort
    sort { $a <=> $b } # time-insensitive sort

Convenience call to compare two times in same format:

    $t->compare2('Aug 26, 2005','Aug 28, 2005','datetime');

There are also a number of legacy calls to emulate an older Time package.

=head2 Internationalization

Month and weekday names are passed through the C<%msg> tool to allow for
translation in multilingual environments.  However, time formats do not 
change based on language preference.

=cut

use strict;
use ExSite::Config;
use ExSite::Base;
use POSIX qw(strftime);
use vars qw(@ISA);
@ISA = qw(ExSite::Base);

# IFP - allow direct comparisons of Time objects
use overload '<=>' => 'compare',
             '>'   => 'gt',
			 '<'   => 'lt',
			 '=='  => 'eq',
			 '>='  => 'ge',
			 '<='  => 'le',
			 '-'   => 'subtract',
			 'bool' => sub { my $this = shift; return defined $this->{time} };

use vars qw(@month_short @month @dayinmonth @weekday_short @weekday $weekday $timezone);

@month_short = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
@month = qw(January February March April May June July August September October November December);
@dayinmonth = qw(31 28 31 30 31 30 31 31 30 31 30 31);
@weekday_short = qw(Sun Mon Tue Wed Thu Fri Sat);
@weekday = qw(Sunday Monday Tuesday Wednesday Thursday Friday Saturday);

# new : create a Time object

sub new {
    my($this, $time, $format) = @_;

    # objectify object
    my $obj = {};
    my $class = ref($this) || $this;
    bless $obj, $class;

    # setup Base class features
    $obj->initialize_object;

    # load time
    $obj->set($time,$format);

    return $obj;
}

# set : set the current time, using a formatted time string and a format

sub set {
    my ($this,$value,$format) = @_;
    $weekday = undef;
    $timezone = undef;
    # eval the call to time() in case of a badly formatted value
    my $t = eval { $this->time($value,$format); };
    if ($@) {
	$this->error("ExSite::Time: bad time value ($format): $value [$@]");
	return $this->{time} = undef;
    }
    $this->{time} = $t;
    my @t = @$t;
    if (! $weekday && $this->now()->[0] - $t[0] < 50) {
	# timelocal weekday support is dodgy outside the current century
	use Time::Local;
	if ($t[1] > 0 && $t[1] <= 12 && $t[2] >= 1 && $t[2] <= 31) {
	    my $t = eval "timelocal($t[5],$t[4],$t[3],$t[2],$t[1]-1,$t[0]-1900);";
	    if (! $@) {
		(undef,undef,undef,undef,undef,undef,$weekday,undef,undef) =
		    localtime($t);
	    }
	}
    }
#    if (! $timezone) {
#	# ignore timezones for now
#    }
    $this->{wday} = $weekday;
    $this->{zone} = $timezone;
    return $this->{time};
}

# time : convert a formatted time string to a time array

sub time {
    my ($this,$value,$format) = @_;
    my $t = $this->now;
    my @time = @$t;

    if ($value) {
	$_ = $format;
	if (/^(sql|iso)_date$/) {
	    ($time[0],$time[1],$time[2]) = split /-/,$value;
	    $weekday = undef;
	}
	elsif (/^sql_datetime$/) {
	    my ($d,$t) = split / /,$value;
	    ($time[0],$time[1],$time[2]) = split /-/,$d;
	    ($time[3],$time[4],$time[5]) = split /:/,$t;
	    $weekday = undef;
	}
	elsif (/^sql_timestamp$/) {
	  if ($config{timestamp_format} eq "num") {
	    ($time[0],$time[1],$time[2],
	     $time[3],$time[4],$time[6]) = 
		 ($value =~ /(....)(..)(..)(..)(..)(..)/);
	    $weekday = undef;
	  }
	  else {
	    my ($d,$t) = split / /,$value;
	    ($time[0],$time[1],$time[2]) = split /-/,$d;
	    ($time[3],$time[4],$time[5]) = split /:/,$t;
	    $weekday = undef;
	  }
	}
	elsif (/^time$/) {
	    $value =~ /([012]?\d):(\d{2})(:(\d{2}))?/;
	    $time[3] = $1;
	    $time[4] = $2;
	    $time[5] = $4;
	    if ($value =~ /pm/i && $time[3] < 12) {
		$time[3] += 12;
	    }
	    elsif ($value =~ /am/i && $time[3] == 12) {
		$time[3] = 0;
	    }
	}
	elsif (/^time24|hms$/) {
	    $value =~ /([012]?\d):(\d{2})(:(\d{2}))?/;
	    $time[3] = $1;
	    $time[4] = $2;
	    $time[5] = $4;
	}
	elsif (/^unix_timestamp$/) {
	    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) =
		localtime($value);
	    @time = ($year+1900,$mon+1,$mday,$hour,$min,$sec);
	    $weekday = $wday;
	}
	elsif (/^date$/) {
	    my ($m,$d,$y) = ($value =~ /(\w+) (\d+), (\d{4})/);
	    $time[0] = $y;
	    $time[1] = $this->which($m,@month_short);
	    $time[2] = $d;
	    $weekday = undef;
	}
	elsif (/^date_long$/) {
	    my ($m,$d,$y) = ($value =~ /(\w+) (\d+), (\d{4})/);
	    $time[0] = $y;
	    $time[1] = $this->which($m,@month);
	    $time[2] = $d;
	    $weekday = undef;
	}
	elsif (/^datetime$/) {
	    #$value =~ /(\w+) (\w+) (\d+), ([012]?\d):(\d{2})(:(\d{2}))? (\d{4})/;
	    $value =~ /(\w+) (\w+) (\d+) ([012]?\d):(\d{2})(:(\d{2}))? (\d{4})/;
	    $time[0] = $8;
	    $time[1] = $this->which($2,@month_short);
	    $time[2] = $3;
	    $time[3] = $4;
	    $time[4] = $5;
	    $time[5] = $6;
	    $weekday = $this->which($1,@weekday_short) - 1;
	}
	elsif (/^datetime_long$/) {
	    #$value =~ /(\w+) (\w+) (\d+), ([012]?\d):(\d{2})(:(\d{2}))? (\d{4})/;
	    $value =~ /(\w+) (\w+) (\d+) ([012]?\d):(\d{2})(:(\d{2}))? (\d{4})/;
	    $time[0] = $8;
	    $time[1] = $this->which($2,@month);
	    $time[2] = $3;
	    $time[3] = $4;
	    $time[4] = $5;
	    $time[5] = $6;
	    $weekday = $this->which($1,@weekday) - 1;
	}
	elsif (/^full$/) {
	    # Weekday, Month dd, yyyy at hh:mm AM/PM
	    $value =~ /^(\w+), (\w+) (\d+), (\d{4}) at (\d{1,2})(:(\d{2})) ([AP]M)$/;
	    $time[0] = $4;
	    $time[1] = $this->which($2,@month);
	    $time[2] = $3;
	    $time[3] = $5;
	    $time[4] = $7;
	    $time[5] = 0;
	    $weekday = $this->which($1,@weekday) - 1;
	    if ($8 =~ /pm/i && $time[3] < 12) {
		$time[3] += 12;
	    }
	    elsif ($8 =~ /am/i && $time[3] == 12) {
		$time[3] = 0;
	    }
	}
	elsif (/^datetime2$/) {
	    $value =~ /(\d+):(\d+).*?, (\w+) (\d+), (\d+)/;
	    $time[0] = $5;
	    $time[1] = $this->which($3,@month);
	    $time[2] = $4;
	    $time[3] = $1;
	    $time[4] = $2;
	    if ($value =~ /pm/i && $time[3] < 12) {
		$time[3] += 12;
	    }
	    elsif ($value =~ /am/i && $time[3] == 12) {
		$time[3] = 0;
	    }
	}
	elsif (/^arpa$/) {
	    # wday, dd Mon yy hh:mm:ss ZONE
	    $value =~ /((\w+), )?(\d+) (\w+) (\d+) ([012]?\d):(\d{2})(:(\d{2}))? (\w+)/;
	    $time[0] = $5;
	    $time[1] = $this->which($4,@month_short);
	    $time[2] = $3;
	    $time[3] = $6;
	    $time[4] = $7;
	    $time[5] = $9;
	    $weekday = $this->which($2,@weekday_short);
	    $timezone = $10;
	}
	elsif (/^cookie$/) {
	    # Wdy, dd-mmm-yyyy hh:mm:ss GMT
	    $value =~ /((\w+), )?(\d+)-(\w+)-(\d+) ([012]?\d):(\d{2})(:(\d{2}))? (\w+)/;
	    $time[0] = $5;
	    $time[1] = $this->which($4,@month_short);
	    $time[2] = $3;
	    $time[3] = $6;
	    $time[4] = $7;
	    $time[5] = $9;
	    $weekday = $this->which($2,@weekday_short);
	    $timezone = $10;
	}
	elsif (/^dd-mmm-yy$/) {
	    my ($d,$m,$y) =~ /(\d+)-(\w+)-(\d+)/i;
	    $time[0] = $y > 50 ? $y+1900 : $y+2000;
	    $time[1] = $this->which($m,@month_short);
	    $time[2] = $d;
	}
	elsif (/^weekday$/) {
	    $weekday = $this->which($value,@weekday_short) - 1;
	}
	elsif (/^weekday_long$/) {
	    $weekday = $this->which($value,@weekday) - 1;
	}
	elsif (/^year$/) {
	    $time[0] = $value;
	    $weekday = undef;
	}
	elsif (/^month$/) {
	    $time[1] = $this->which($value,@month_short);
	    $weekday = undef;
	}
	elsif (/^month_long$/) {
	    $time[1] = $this->which($value,@month);
	    $weekday = undef;
	}
	elsif (/^raw$/) {
	    @time = @$value;
	}
	elsif (/^raw_year$/) {
	    $time[0] = $value;
	}
	elsif (/^raw_month$/) {
	    $time[1] = $value;
	}
	elsif (/^raw_day$/) {
	    $time[2] = $value;
	}
	elsif (/^raw_hour$/) {
	    $time[3] = $value;
	}
	elsif (/^raw_minute$/) {
	    $time[4] = $value;
	}
	elsif (/^raw_second$/) {
	    $time[5] = $value;
	}
	# IFP - $value can also be a Time object, in which case a format
	# string is not necessary.
	elsif ( ref($value) eq ( ref($this) || $this ) ) {
		@time = @{ $value->{time} };
	}
    }
    # Convert time values to ints, so they don't look like octal strings if
    # they have a leading zero.
    for (0..5) { $time[$_] += 0; }
    # untaint time values
    foreach my $t (@time) { 
	$t =~ /^(\d*)$/; 
	$t = $1; 
    }
    return \@time;
}

# write : output the time in an arbitrary format

sub write {
    my $this = shift;
    $_ = shift;     # format
    my $imonth = $this->{time}[1] - 1;
    if (/^(sql|iso)_date$/) {
	return sprintf "%04d-%02d-%02d", @{$this->{time}};
    }
    elsif (/^sql_datetime$/) {
	return sprintf "%04d-%02d-%02d %02d:%02d:%02d", @{$this->{time}};
    }
    elsif (/^sql_timestamp$/) {
      if ($config{timestamp_format} eq "num") {
	return sprintf "%04d%02d%02d%02d%02d%02d", @{$this->{time}};
      }
      else {
	return sprintf "%04d-%02d-%02d %02d:%02d:%02d", @{$this->{time}};
      }
    }
    elsif (/^time$/) {
	my $h = $this->{time}[3];
	my $m = $this->{time}[4];
	my $ampm = " AM";
	if ($h >= 24) {
	    $h -= 24;
	}
	elsif ($h > 11) {
	    if ($h > 12) { $h -= 12; }
	    $ampm = " PM";
	}
	if ($h == 0) { $h = 12; }
	return sprintf("%d:%02d", $h, $m).$ampm;
    }
    elsif (/^time24$/) {
	return sprintf "%02d:%02d", $this->{time}[3], $this->{time}[4];
    }
    elsif (/^hms$/) {
	return sprintf "%02d:%02d:%02d", $this->{time}[3], $this->{time}[4], $this->{time}[5];
    }
    elsif (/^unix_timestamp$/) { 
	use Time::Local;
	my @t = @{$this->{time}};
	if ($t[1] >= 1 && $t[1] <= 12 && $t[2] >= 1 && $t[2] <= 31) {
	    my $t;
	    if ($this->now()->[0] - $t[0] < 50) {
		$t = eval "timelocal($this->{time}[5],$this->{time}[4],$this->{time}[3],$this->{time}[2],$this->{time}[1]-1,$this->{time}[0]-1900);";
	    }
	    else {
		# YMMV - timelocal not well defined for dates outside the epoch
		$t = eval "timelocal($this->{time}[5],$this->{time}[4],$this->{time}[3],$this->{time}[2],$this->{time}[1]-1,$this->{time}[0]);";
	    }
	    return $t;
	}
	else {
	    $this->error("bad time value: $@");
	    return -1;
	}
    }
    elsif (/^date$/) {
	return $msg{$month_short[$imonth]}." $this->{time}[2], $this->{time}[0]";
    }
    elsif (/^date_long$/) {
	return $msg{$month[$imonth]}." $this->{time}[2], $this->{time}[0]";
    }
    elsif (/^datetime$/) {
	return $msg{$weekday_short[$this->{wday}]}." ".$msg{$month_short[$imonth]}." $this->{time}[2] ".$this->write('time24')." $this->{time}[0]";
    }
    elsif (/^datetime_long$/) {
	return $msg{$weekday[$this->{wday}]}." ".$msg{$month[$imonth]}." $this->{time}[2] ".$this->write('time24')." $this->{time}[0]";
    }
    elsif (/^full$/) {
	return $msg{$weekday[$this->{wday}]}.", ".$msg{$month[$imonth]}." $this->{time}[2], $this->{time}[0] $msg{at} ".$this->write('time');
    }
    elsif (/^datetime2$/) {
	return $this->write('time').", ".$msg{$month[$imonth]}." $this->{time}[2], $this->{time}[0]";
    }
    elsif (/^arpa$/) {
	# wday, dd Mon yyyy hh:mm:ss ZONE
	my $zone = $this->{zone};
 	if (! $zone) {
	    # guess an appropriate time zone from local system clock
	    $zone = strftime "%Z", localtime;
	}
	my $date = sprintf("%02d %s %d",$this->{time}[2],$month_short[$imonth],$this->{time}[0]);
	return "$weekday_short[$this->{wday}], $date ".$this->write('hms')." $zone";
    }
    elsif (/^cookie$/) {
	# wday, dd-Mon-yyyy hh:mm:ss GMT
	# FIXME:  we just write GMT into the string; we do not actually
	# convert the time zone or adjust the time.
	my $date = sprintf("%02d-%s-%d",$this->{time}[2],$month_short[$imonth],$this->{time}[0]);
	return "$weekday_short[$this->{wday}], $date ".$this->write('hms')." GMT";
    }
    elsif (/^dd-mmm-yy$/) {
	return sprintf("%d-%s-%02d",$this->{time}[2],$msg{$month_short[$imonth]},substr($this->{time}[0],2,2));
    }
    elsif (/^weekday$/) {
	return $msg{$weekday_short[$this->{wday}]};
    }
    elsif (/^weekday_long$/) {
	return $msg{$weekday[$this->{wday}]};
    }
    elsif (/^year$/) {
	return $this->{time}[0];
    }
    elsif (/^month$/) {
	return $msg{$month_short[$imonth]};
    }
    elsif (/^month_long$/) {
	return $msg{$month[$imonth]};
    }
    elsif (/^month_digit$/) {
	return $imonth; 
    }
    elsif (/^raw$/) {
	return $this->{time};
    }
    elsif (/^raw_year$/) {
    	return $this->{time}[0];
    }
    elsif (/^raw_month$/) {
    	return $this->{time}[1];
    }
    elsif (/^raw_day$/) {
    	return $this->{time}[2];
    }
    elsif (/^raw_hour$/) {
    	return $this->{time}[3];
    }
    elsif (/^raw_minute$/) {
    	return $this->{time}[4];
    }
    elsif (/^raw_second$/) {
    	return $this->{time}[5];
    }
}

# validate : see if the object represents a legal date/time

sub validate {
    my ($this) = @_;
    my @err;
    if (ref $this->{time} ne "ARRAY") {
	push @err, $msg{"date is not defined"};
	return @err;
    }
    my @date = @{$this->{time}};
    my $y = shift @date;
    my $m = shift @date;
    my $d = shift @date;
    my $now = $this->now();
    if ($y < 1) {
	push @err, $msg{"invalid year"};
    }
    if (abs($y - $now->[0]) > $config{form}{validate_year_range}) {
	# year must be within 100 years of today
	push @err, $msg{"year is out of range"};
    }
    if ($m < 1 || $m > 12) {
	push @err, $msg{"invalid month"};
    }
    if ($d < 1 || $d > 31) {
	push @err, $msg{"invalid day"};
    }
    else {
	# right number of days in month?
	my $dinm = $m > 0 ? $dayinmonth[$m-1] : 0;
	if ($m == 2 &&        # Feb
	    $y % 4 == 0 &&    # years divisible by 4 
	    $y % 100 != 0) {  # years not divisible by 100
	    # add a day to February for leap years
	    $dinm++;
	}
	if ($m > 0 && $d > $dinm) {
	    push @err, &substitute($msg{"too many days for [[month]]"},{month=>$month[$m-1]});
	}
    }

    my $h = shift @date;
    my $m = shift @date;
    my $s = shift @date;
    if ($h < 0 || $h > 23) {
	push @err, $msg{"invalid hour"};
    }
    if ($m < 0 || $m > 59) {
	push @err, $msg{"invalid minute"};
    }
    if ($s < 0 || $s > 59) {
	push @err, $msg{"invalid seconds"};
    }

    return @err == 0 ? undef : join "; ",@err;
}

# now : convert the current time to a time array

sub now {
    my $this = shift;
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = 
	localtime(time);
    $weekday = $wday;
    return [$year+1900,$mon+1,$mday,$hour,$min,$sec];
}

# add: add a number of time units (years, months, days, hours, minutes, 
# seconds) to the current time.

sub add {
    my ($this,$number,$what) = @_;
    my $t = $this->write("unix_timestamp");
    my $seconds = $number;
    if ($what =~ /year/i) {
	$seconds *= (365*24*60*60);
    }
    elsif ($what =~ /month/i) {    
	$seconds *= int(30.4375*24*60*60);   ### 1 month == 1/12 of a year
	#$seconds *= (30*24*60*60);          ### 1 month == 30 days
    }
    elsif ($what =~ /week/i) {
	$seconds *= (7*24*60*60);
    }
    elsif ($what =~ /day/i) {
	$seconds *= (24*60*60);
    }
    elsif ($what =~ /hour/i) {
	$seconds *= (60*60);
    }
    elsif ($what =~ /min/i) {
	$seconds *= 60;
    }
    $t += $seconds;
    return $this->set($t,"unix_timestamp");
}

# compare : return -1, 0, +1 if passed date is past, present, or future 
# with respect to the current date

sub compare {
    my ($this, $date, $fmt) = @_;
    my $now = $date ? $this->time($date,$fmt) : $this->now;
    my $then = $this->{time};
   
    return 1 if ($now->[0] > $then->[0]);
    return -1 if ($now->[0] < $then->[0]);
	       
    return 1 if ($now->[1] > $then->[1]);
    return -1 if ($now->[1] < $then->[1]);
	       
    return 1 if ($now->[2] > $then->[2]);
    return -1 if ($now->[2] < $then->[2]);

    return 0;
}

sub compare2 {
    my ($this, $date1, $date2, $fmt) = @_;
    $this->set($date2,$fmt);
    return $this->compare($date1,$fmt);
}

# IFP
# Wrapper methods around compare() for various types of comparisons. 
# These also implement the overloaded comparison operators, in which case:
#	a) both parameters must be instances of this class; and 
#	b) the $fmt parameter is undef.
#
# Note that the order of operands in compare() and diff() are the reverse
# of what you'd expect when using operators, so the comparisons in these
# methods are the reverse of what you'd expect, and the subtract()
# function is necessary to flip the operands around.
sub gt {
	my ($this, $that, $fmt) = @_;
	return ( $this->compare($that, $fmt) < 0 );
}
sub lt {
	my ($this, $that, $fmt) = @_;
	return ( $this->compare($that, $fmt) > 0 );
}
sub eq {
	my ($this, $that, $fmt) = @_;
	return ( $this->compare($that, $fmt) == 0 );
}
sub ge {
	my ($this, $that, $fmt) = @_;
	return ( $this->compare($that, $fmt) <= 0 );
}
sub le {
	my ($this, $that, $fmt) = @_;
	return ( $this->compare($that, $fmt) >= 0 );
}
sub subtract {
	my ($this, $that, $fmt) = @_;
	# Make sure that $that is a Time object.
	$that = new ExSite::Time($that, $fmt);  
	return $that->diff($this);
}

# diff: return the difference between the passed date and the internal date
# +ve differences indicate the passed date is in the future; -ve indicates
# the passed date is in the past.  The difference returned is in seconds.

sub diff {
    my ($this, $date, $fmt) = @_;
    my $s1 = $this->write("unix_timestamp");
    my $time = new ExSite::Time($date,$fmt);
    my $s2 = $time->write("unix_timestamp");
    return $s2 - $s1;
}

# which : finds a named time value (eg. month, weekday) in an array
# and returns the array location plus 1.

sub which {
    my ($this,$val,@list) = @_;
    my $i = 0;
    foreach my $item (@list) {
	$i++;
	return $i if ($val =~ /^$item$/i);
    }
    return 0;
}

###================= LEGACY FUNCTIONS ====================

#sub CompareToNow
#pre	sql date
#post	if date is in future, returns 1
#post	if date is in past, returns -1
#post	if date is now, returns 0

sub CompareToNow {
    my ($this,$t1,$t2) = @_;
    if ($t2) { 
	return $this->compare2($t1,$t2,"sql_date");
    }
    else {
	$this->set;
	return $this->compare($t1,"sql_date");
    }
}

#
# makePreviousDate
# accepts a number of days
# returns the sql date for the date in
# the past equalt to today minus the number
# of days specified
# ie, if $days = 7, we get the date for one week ago

sub MakePreviousDate {
    my ($this, $days) = @_;
    $this->add(-$days,"days");
    return $this->write("sql_date");
}
	
sub GetCurrentSqlDate{
    my ($this) = @_;
    $this->set;
    return $this->write('sql_date');
}

sub ConvertToDate{
    my ($this,$value) = @_;
    if ($value) { $this->set($value,"sql_date"); }
    return $this->write("date");
}

sub ConvertSqlToTime{
    my ($this,$value) = @_;
    if ($value) { $this->set($value,"time24"); }
    return $this->write("time");
}

sub ConvertFromTimeStamp{
    my ($this,$data) = @_;
    if (length $data == 14){
	$this->set($data,"sql_timestamp");
	return $this->write("sql_date");
    }
    else{
	$this->set($data,"unix_timestamp");
	return $this->write("datetime");
    }
}

sub GetWeekDay{
    my ($this,$date) = @_;
    $this->set($date,"sql_date");
    return $weekday[$this->{wday}];
}

sub ConvertToShortDate{
    my ($this,$value) = @_;
    my ($year,$month,$day);
    if ($value) { $this->set($value,'sql_date'); }
    return $this->write('dd-mmm-yy');
}

sub ConvertSqlToDateTime{
    my ($this,$data)= @_;	
    if ($data) { $this->set($data,'sql_datetime'); }
    return $this->write('datetime2');
}

1;
