#!/usr/bin/perl
#-----------------------------------------------------------------------
# cron.pl - scheduled task execution
#
# This program is designed to be executed by the cron process.  The 
# recommended usage is to execute once per hour, close to the hour.
# A crontab entry like the following should do this:
#
#    0 * * * *   cd /path/to/cgi-bin; ./cron.pl
#
# This says to change to the cgi directory and execute this program 
# once per hour on the hour.  You should always execute cron.pl from 
# the CGI directory to make sure you pick up local configurations.
#
# If you have an irregular environment, you may also want to set some
# environment variables in that cron command to spoof what the webserver
# would set, to ensure that autoconfiguration will be able find all of 
# ExSite's parts.  If not, consider manual configuration.
#-----------------------------------------------------------------------
#
#   Copyright 2001-2008 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

use ExSite::Config;
use ExSite::Form;
use ExSite::Module;
use ExSite::Misc;
use ExSite::Util;

if ($ENV{HTTP_USER_AGENT}) {
    # this appears to be an actual web request; abort
    # (in case .pl files are executable from URLs)
    &log("===== Aborting Cron; invalid environment"); 
    exit 1;
}

&exsite_init;
my $db = new ExSite::Form;
my $ncall = 0;   # number of cron commands issued
my $ndone = 0;   # number of commands actually acted on
my $logging = undef;

my @task = $db->fetch_all("crontask","crontask_id");

&log("===== Cron startup: scanning ".scalar @task." tasks.");

foreach my $t (@task) {
    if ($t->{status} eq "active") {

	# set UID
	if ($t->{member_id}) {
	    my $u = $db->fetch("member",$t->{member_id});
	    $u->{login} ? $db->do_login_temp($u->{login},1) : $db->logout;
	}
	else {
	    &log("Warning: Task $t->{crontask_id}: no UID to run task under");
	    $db->logout;
	}
	
	# do the task
	if ($t->{method} eq "preset") {
	    &do_task($t);
	}
	elsif ($t->{method} eq "hourly") {
	    &do_task($t);
	}
	else {
	    my @time = localtime(time);
	    my $hour = $time[2];
	    my $wday = $time[6];
	    if ($t->{method} eq "daily" && 
		$hour == $config{cron}{hour}) {
		&do_task($t);
	    }
	    elsif ($t->{method} eq "weekly" && 
		   $wday == $config{cron}{day} && 
		   $hour == $config{cron}{hour}) {
		&do_task($t);
	    }
	}
    }
}

&log("Cron finish: $ncall valid task(s) found; $ndone task(s) executed.");
&exsite_close;
close LOG;

sub do_task {
    my $task = shift;
    if ($task->{module}) {
	my $mod = &ExSite::Module::get_module($task->{module});
	if ($mod) {
	    my $cron = $mod->ioctl('Cron');
	    if (ref $cron eq "CODE") {
		# issue cron command to the module
		$ncall++;
		my $result;
		eval { $result = $mod->$cron($task->{command},
					     $task->{type},
					     $task->{id}); };
		if ($@) { 
		    # task crashed
		    $result .= " ERROR: $@"; 
		}
		else {
		    # mark the task as executed
		    my $time = new ExSite::Time;
		    $task->{exectime} = $time->write("sql_datetime");
		    if ($task->{method} eq "preset") {
			$task->{status} = "completed";
		    }
		    $db->update("crontask",$task);
		    $ndone++;
		}
		if ($result) { 
		    # log the result
		    &log("Task $task->{crontask_id}: $result"); 
		}
	    }
	    else {
		&log("!! Task $task->{crontask_id}: $task->{module} does not support task scheduling.");
	    }
	}
	else {
	    if ($@) {
		# module is broken
		&log("!! Task $task->{crontask_id}: $task->{module} failed to instantiate: $@");
	    }
	    else {
		# permission denied, probably
		&log("!! Task $task->{crontask_id}: no permission to run $task->{module}");
	    }
	}
    }
    else {
	&log("!! Task $task->{crontask_id} is misconfigured (no module).");
    }
    return;
}


sub log {
    my $message = shift;
    if (! defined $logging) {
	if ($config{log}{cron}) {
	    if (-w $config{log}{logfile}{cron}) {
		if (open LOG, ">>$config{log}{logfile}{cron}") {
		    $logging = 1;
		}
		else {
		    $db->error("failed to open cron logfile");
		}
	    }
	    else {
		$db->error("cron logfile is not writeable");
	    }
	}
	$logging = defined $logging ? $logging : 0;
    }
    if ($logging) {
	my $time = localtime;
	print LOG "$time\t$message\n";
    }
    # it's safe to leave this file open until the end of this pass through
    # the cron system, because there is no contention for it
}
