#!/usr/bin/perl
#-----------------------------------------------------------------------
# exsite-test.cgi - test an installation
#-----------------------------------------------------------------------
#
# THIS IS NOT A MULTI-PURPOSE TEST SCRIPT.
#
# This script only tests a default install of ExSite for basic 
# connectivity.  If you have customized your installation by installing
# files under different names or locations, modified the database,
# or other modifications that are not part of the default install, this
# script may report spurious errors.
#
#-----------------------------------------------------------------------

#-----------------------------------------------------------------------

use strict;
use CGI::Carp  qw(fatalsToBrowser);
use ExSite::Config;

BEGIN {
    # try to trap any missing prereqs before we start 
    require ExSite::Report;
    require ExSite::Misc;
}

use vars qw($db @dat $dat @required_mods @suggested_mods $nfail);

@required_mods = qw(
    CGI
    Digest::MD5
    Devel::Size
    MIME::Base64
    Net::SMTP
    Image::Info
    IO::String
    IPC::Open2
    IPC::Open3
    Time::Local
);

@suggested_mods = qw(
    MIME::Entity
    Storable
    GDBM_File
    Cwd
    Fcntl
    CGI::Carp
    LWP::Simple
    LWP::UserAgent
    Spreadsheet::WriteExcel
    XML::RSS
);

$| = 1;   # autoflush

print "Content-type: text/html\n\n";

print "<h1>testing ExSite install...</h1>\n";

print "<p>initializing...</p>\n";
&exsite_init;

print "<p>looking for static files...<br>\n";

if (! -e "$config{server}{HTMLroot}$config{server}{HTMLpath}/_ExSite/css") {
    print "FAILED: _ExSite files not found (did you copy all of the html files into your htdocs area?)</p>\n";
    $nfail++;
}
elsif (! -e "$config{server}{HTMLroot}$config{server}{HTMLpath}/_Modules/CMS/icon.gif") {
    print "FAILED: _Modules files not found (did you copy all of the html files into your htdocs area?)</p>\n";
    $nfail++;
}
else {
    print "PASSED.</p>\n";
}

print "<p>looking for ExSite plugin modules...<br>\n";

my $tmpfail = $nfail;
if (! -e "$config{server}{CGIroot}/Modules/BaseDCD.pm") {
    print "FAILED: BaseDCD.pm not found (did you copy all of the CGI files and folders into your cgi-bin area?)</p>\n";
    $nfail++;
}
if (opendir D, "$config{server}{CGIroot}/Modules") {
    my @plugins = readdir D;
    closedir D;
    my $nplug;
    foreach my $plug (@plugins) { 
	if ($plug =~ /\.pm$/) { $nplug++; }
    }
    if ($nplug < 2) {
	print "WARNING: No plug-ins were found (have you installed any plug-ins?)</p>\n";
	$nfail++;
    }
}
else {
    print "FAILED: plug-in area is missing (did you installed the cgi-bin area?)</p>\n";
    $nfail++;
}
if ($tmpfail == $nfail) {
    print "PASSED.</p>\n";
}

print "<p>connect to db...<br>\n";
eval '$db = new ExSite::Report';
if ($@ || (ref $db !~ /Report/)) {
    print "FAILED (reported error: <i>$@</i>)";
    print "<br><b>Double-check your database connection parameters</b>:
<pre>
server.db.name
server.db.type
server.db.user
server.db.pass
server.db.server
</pre>\n";
    $nfail++;
    exit 0;
}
else {
    print "PASSED.</p>\n";
}

print "<p>looking for tables...<br>\n";

eval '@dat = $db->get_tables()';
if ($@ || (scalar @dat <= 0)) {
    print "FAILED (reported error: <i>$@</i>)";
    print "<br><b>No database tables found.  Are you connecting to the right database?  Did you initialize the database?</b>\n";
    $nfail++;
    exit 0;
}
elsif (scalar grep /^member$/,@dat > 0) {
    print "FAILED";
    print "<br><b>No member table found - you will not be able to log in.  Are you connecting to the right database?  Did you initialize the database?</b>\n";
    $nfail++;
    exit 0;
}
elsif (scalar @dat < 6) {
    print "WARNING";
    print "<br><b>Insufficient number of tables found.  Are you connecting to the right database?  Did you initialize the database?</b>\n";
    $nfail++;
}
else {
    print "PASSED.</p>\n";
}

print "<p>looking for database map...<br>\n";

if (ref $db->{map} !~ /DBmap/) {
    print "FAILED";
    print "<br><b>No database map found. Double-check your database map parameters, eg.</b>:
<pre>
server.db.map = map
server.map.name = dbmap
server.map.type = text
</pre>\n";
    $nfail++;
    exit 0;
}
    
eval '$dat = $db->{map}->get_table_title("section",1)';
if ($dat !~ /Sites/) {
    print "FAILED";
    print "<br><b>Unexpected result from map query.  Is your <tt>dbmap</tt> 
directory installed under the cgi-bin area?</b>\n";
    $nfail++;
    exit 0;
}
eval '$dat = $db->{map}->get_ncol("member")';
if ($dat < 4) {  # should be at least 6 columns in this table
    print "FAILED";
    print "<br><b>Unexpected result from map query.  Is your <tt>dbmap</tt> 
directory installed under the cgi-bin area?</b>\n";
    $nfail++;
    exit 0;
}

print "PASSED.\n";

print "<p>looking for users...<br>\n";

eval '@dat = $db->fetch_all("member")';
if ($@ || (scalar @dat <= 0)) {
    print "FAILED (reported error: <i>$@</i>)";
    print "<br><b>No users found.  Did you initialize the database?  Did you create an admin user?</b>\n";
    $nfail++;
    exit 0;
}
else {
    eval '@dat = $db->fetch_match("member",{login=>"admin",password=>"password"})';
    if (@dat > 0) {
	print "WARNING: <b>you did not set a new login/password for the default user.  This installation is not secure!</b></p>\n";
    }
    else {
	print "PASSED.</p>\n";
    }
}

print "<p>checking publisher...<br>\n";

if (-e 'publish.cgi') {
    if (-u 'publish.cgi') {
	print "PASSED.</p>\n";
    }
    else {
	print "FAILED";
	print "<br><b>Publish program does not have permission to write files.  Did you set the permissions on publish.cgi?</b>\n";
	$nfail++;
    }
}
else {
    print "FAILED";
    print "<br><b>Publish program not found.  Did you compile publish.cgi?</b>\n";
    $nfail++;
}

# check Perl modules installation

my $nmodfail = 0;
print "<p>looking for required perl modules...<br>\n";
foreach my $mod (@required_mods) {
    print "-- $mod: ";
    if (eval "require $mod") {
	print "passed<br>\n";
    }
    else {
	print "FAILED: $@<br>\n";
	$nmodfail++;
    }
}
print "<p>looking for suggested perl modules...<br>\n";
foreach my $mod (@suggested_mods) {
    print "-- $mod: ";
    if (eval "require $mod") {
	print "passed<br>\n";
    }
    else {
	print "FAILED: $@<br>\n";
	$nmodfail++;
    }
}
if ($nmodfail > 0) {
    print "Failed modules may have to be (re-) installed to get full
functionality.</p>\n";
}
else {
    print "All modules found.</p>\n";
}
    
print "<h3>Exsite System Configuration</h3>
<ul>
<li> ExSite WebWare Version: ${ExSite::Config::VERSION}</li>
<li> Installation name: $config{site}{name}</li>
<li> Server: $config{server}{server}</li>
<li> Base HTdocs Directory: $config{server}{HTMLroot}</li>
<li> Base CGI Directory: $config{server}{CGIroot}</li>
<li> Base HTML URL: $config{server}{HTMLpath}</li>
<li> Base CGI URL: $config{server}{CGIpath}</li>
<li> Administrator: $config{site}{administrator}</li>
<li> Configuration: ".($config{server}{auto} ? "automatic" : "manual")."</li>
</ul>
";

my $status = ($nfail + $nmodfail) > 0 ? "performed" : "passed";
print "<h3>Done!</h3>\n<p>All basic connectivity tests $status.</p>\n";
if ($nfail > 0) {
    print "<p>Some tests failed.  Consult the messages above for tips
on solving the problems indicated.</p>\n";
}  
if ($nmodfail > 0) {
    print "<p>Some Perl modules are missing;  you can proceed, but 
certain features may not be available.</p>\n";
}  
print "<p>If there are no serious problems, you can now try to 
<a href=login.cgi>log in</a>.</p>\n";

### END

