#!/usr/bin/perl
#-----------------------------------------------------------------------
# login.cgi : authenticate a user and set their identity cookie
#
# This script accepts a "login" and "password" from an HTML form,
# and authenticates them against values looked up in a database.
# If a match is found, an identity cookie is set for future
# authentication.
#
# If called with no input, the script will throw up a simple login
# form. The site developer may code similar login forms into any HTML
# pages they choose.
#
# The script accepts input by POST or via the QUERY_STRING.  That is,
# even in the absence of a login form, login can still be done using
# an URL like /cgi/login.cgi?login=my_name&password=my_pwd
# (Obviously, you won't want to hard-code sensitive passwords into 
# publicly visible HTML.)
#
# Following a successful authentication, the user is redirected to
# a login page.  Default login targets are defined in 
# @ExSite::Config::login_destination, but these may be overridden by 
# supplying a "goto" argument in the login.cgi input data.
#-----------------------------------------------------------------------
#
#   Copyright 2001-2008 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

use strict;
use ExSite::Config;
use ExSite::Input;
use ExSite::Util;
use ExSite::ML;
use ExSite::DB;

use vars qw(%data $uid $access);

&exsite_init;
my $in = new ExSite::Input;
%data = %{$in->combine()}; #&DecodeInput;  # accepts input by post or query
my $ml = new ExSite::ML;

if (exists $data{login}) {
    my $db = new ExSite::DB;
    if ($data{login_action} eq "forgot") {
	&Page("Forgot Password",$db->login_reminder_form());
    }
    else {
	if ($config{throttle}{user}{rate} || $config{throttle}{system}{rate}) {
	    my $rate = &throttle();
	    if ($rate > 0) {
		# someone is hammering us; delay the login attempt
		sleep($rate * 10);
	    }
	}
	if (my $access = $db->login($data{login},$data{password},$data{goto},$data{remember})) {
	    my $destination = 
		$data{goto} || 
		$config{auth}{login_destination}[$access];
	    $ml->location($destination);
	}
	else {
	    &DenyLogin($db->show_diagnostics("warn","HTML"));
	}
    }
}
else {
    &PromptLogin;
}
&exsite_close;

sub PromptLogin {
    my $db = new ExSite::DB;
    my $access = $db->authorize();
    if ($access) {
	# user already logged in
	my $destination = $data{goto} || $config{auth}->{login_destination}->[$access];
	$ml->location($destination);
    }
    else {
	my $login_form = $db->login_form();
	&Page("$config{site}->{name} Login",
	      $config{site}{logo}.
	      $ml->h1($config{site}{name}).
	      $login_form);
    }
}

sub DenyLogin {
    my $contents = 
	$ml->h1("Login Denied").
	join(" ",@_).
	$ml->p(
	       $ml->a("Try again",{href=>"javascript:history.back()"})
	       );
    &Page("Login Denied",$contents);
}

sub Page {
    my ($title,$contents) = @_;
    $ml->Doctype($config{doctype});
    $ml->Doc($contents);
    $ml->__div({align=>"center"});
    $ml->__body();
    $ml->Prepend(
		 $ml->head(
			   $ml->title($title).
			   $ml->link(undef,{rel=>"shortcut icon",href=>"$config{server}{HTMLpath}/_ExSite/images/exsite.ico"}).
			   $ml->link(undef,{rel=>"stylesheet", type=>"text/css", href=>"$config{server}{HTMLpath}/_ExSite/css/ExSite.css"})
			   )
		 );
    $ml->__html();
    print "Content-type: text/html\n\n";
    $ml->Print();
}

### END

