#!/usr/bin/perl
#-----------------------------------------------------------------------
# publish.pl - write public content to disk
#
# Typically run via the publish.c[gi] setuid wrapper, to obtain the 
# necessary write permissions.
#
# USAGE:
#    publish.cgi?page=xxx - publishes page id XXX
#    publish.cgi?section=xxx - publishes section id XXX
#    publish.cgi?content=xxx - publishes content id XXX
#    publish.cgi?module=xxx - publishes plug-in module XXX
#    publish.cgi?-page=xxx - unpublishes page id XXX
#    publish.cgi?-section=xxx - unpublishes section id XXX
#    publish.cgi?-content=xxx - unpublishes content id XXX
#    publish.cgi?-module=xxx - unpublishes plug-in module XXX
#
#    The parameter slient=1 causes publish.pl to run silently.
#    This is useful when invoking publish.cgi indirectly to write
#    files to disk.  Otherwise it ends up writing to the top of
#    the current page.
#
# "publish" creates files
# "unpublish" removes files (presumably ones that were previously published)
#
#  For pages and sections, all publication data is obtained from the 
#  database.  For content, publication data can be obtained from the
#  database, or from the POST data (used when files are written 
#  directly to disk, instead of being stored in the database).  For
#  Modules, publication data is determined at the discretion of the
#  module.
#-----------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

# KLUDGE!  Include local dir in path to get ExSite libraries.
# This makes the path insecure for setuid execution of other programs.
BEGIN{ if (grep ( /^\.$/, @INC ) == 0) { push @INC,"."; } }

use strict;
use ExSite::Config;
use ExSite::Form;
use ExSite::Page;
use ExSite::Section;
use ExSite::Input;
use ExSite::Auth;
use ExSite::Misc;
use ExSite::ML;
use ExSite::Module qw(&get_module);
use ExSite::Search;

&exsite_init;
my $db = new ExSite::Form;
my $in = new ExSite::Input;
my $auth = new ExSite::Auth;
if ($config{search}{reindex_on_publish}) {
    $share{Search} = new ExSite::Search;
}
my $ml = new ExSite::ML;
my %query = %{$in->query};
my $delay = 3;                           # seconds to refresh
my $obj;                                 # cms object

if ($query{silent}) {
    # run silently, so we don't contaminate another page's output

    # save old streams
    open(OLDOUT, ">&STDOUT");
    open(OLDERR, ">&STDERR");
    # redirct streams to /dev/null
    open(STDOUT, '>', "/dev/null") || die "Can't redirect stdout";
    open(STDERR, ">&STDOUT")       || die "Can't dup stdout";
    # make unbuffered
    select(STDERR); $| = 1;
    select(STDOUT); $| = 1;
} 

# disable cookies in published pages
tied (%cookie)->publish_mode();

if ($db->authorize < 2) {
    &print_head;
    print "\n",$ml->p("Permission denied.");
}
elsif ($query{page}) {
    # publish one page
    $obj = new ExSite::Page(id=>$query{page});
    $obj->set_revision("newest");
    # redirect to published page when done
    my $url = $obj->get_url_canonical();
    if ($obj->is_page) {
	print "Refresh: ${delay}; URL=$url\n";
    }
    &print_head;
    if ($auth->is_admin() || $auth->owns($obj->get("page","section_id"))) {
	print $ml->h2("Publishing Page $query{page}");
	$share{Page} = $obj;
	$obj->info("Publishing Page $query{page}");
	print $obj->publish;
	if ($obj->is_page) {
	    print $ml->p("(Redirecting to $url in $delay seconds...  
Click 'Reload' if you still see the old page.)");
	}
    }
    else {
	print $ml->h2("Permission Denied"),
	  $ml->p("You do not have authorization to publish that page.");
    }
}
elsif ($query{'-page'}) {
    # unpublish one page
    &print_head;
    $obj = new ExSite::Page(id=>$query{'-page'});
    if ($auth->is_admin() || $auth->owns($obj->get("page","section_id"))) {
	print $ml->h2("Removing $obj->{data}{type} $query{-page}");
	$obj->info("Unpublishing Page $query{page}");
	print $obj->unpublish;
    }
    else {
	print $ml->h2("Permission Denied"),
	  $ml->p("You do not have authorization to remove that page.");
    }
}
elsif ($query{section}) {
    # publish whole section
    if ($auth->is_admin() || $auth->owns($query{section})) {
	$obj = new ExSite::Section(id=>$query{section});
	$obj->set_revision("newest");    # doesn't seem to do much
	# redirect to published home page
	my $url = $obj->get_url;
	print "Refresh: ${delay}; URL=$url\n";
	&print_head;
	print $ml->h2("Publishing Site $query{section}");
	# section publish prints directly for better flushing of output
	$obj->info("Publishing Section $query{section}");
	$obj->publish;
	print $ml->p("(Redirecting to $url in $delay seconds...)");
    }
    else {
	print $ml->h2("Permission Denied"),
	  $ml->p("You do not have authorization to publish that site.");
    }
}
elsif ($query{'-section'}) {
    # unpublish whole section
    &print_head;
    if ($auth->is_admin() || $auth->owns($query{-section})) {
	$obj = new ExSite::Section(id=>$query{'-section'});
	print $ml->h2("Removing Site $query{-section}");
	# section unpublish prints directly for better flushing of output
	$obj->info("Unpublishing Section $query{section}");
	$obj->unpublish;
    }
    else {
	print $ml->h2("Permission Denied"),
	  $ml->p("You do not have authorization to publish that site.");
    }
}
elsif ($query{content}) {
    # publish one content object
    $obj = new ExSite::Content(id=>$query{content});
    $obj->set_revision("newest");
    $obj->set_context;
    &print_head; 
    if ($auth->is_admin() || $auth->owns($obj->get("page","section_id"))) {
	print $ml->h2("Publishing Content $query{content}");
	$obj->info("Publishing Content $query{content}");
	print $obj->publish;
    }
    else {
	print $ml->h2("Permission Denied"),
	  $ml->p("You do not have authorization to publish that content.");
    }
}
elsif ($query{'-content'}) {
    # unpublish one content object
    &print_head; 
    $obj = new ExSite::Content(id=>$query{'-content'});
    $obj->set_context;
    print "\n";
    if ($auth->is_admin() || $auth->owns($obj->get("page","section_id"))) {
	print $ml->h2("Removing Content $query{-content}");
	$obj->info("Unpublishing Content $query{content}");
	print $obj->unpublish;
    }
    else {
	print $ml->h2("Permission Denied"),
	  $ml->p("You do not have authorization to remove that content.");
    }
}
elsif ($query{module}) {
    # publish plug-in
    &print_head; 
    $query{module} =~ /^(\w+)$/;
    my $mod = $1;

    $obj = &get_module($mod);
    if ($obj) { 
	my $publisher = $obj->ioctl("Publish");
	if (ref $publisher eq "CODE") {
	    $obj->info("Publishing Module $mod");
	    print $obj->$publisher || $@;
	}
	else {
	    print "$mod: this module does not publish";
	}
    }
    else {
	my $reason = $@ || "no license";
	print "$mod: permission denied ($reason)";
    }
    ### FIXME: reindex if $config{search}{reindex_on_publish}
}
elsif ($query{'-module'}) {
    # publish plug-in
    &print_head; 
    $query{module} =~ /^(\w+)$/;
    my $mod = $1;

    $obj = &get_module($mod);
    if ($obj) { 
	my $publisher = $obj->ioctl("Unpublish");
	if (ref $publisher eq "CODE") {
	    $obj->info("Unpublishing Module $mod");
	    print $obj->$publisher || $@;
	}
	else {
	    print "$mod: this module does not unpublish";
	}
    }
    else {
	my $reason = $@ || "no license";
	print "$mod: permission denied ($reason)";
    }
}
else {
    print $ml->h2("Error: nothing to publish! $ENV{QUERY_STRING}");
}

if ($config{search}{reindex_on_publish}) {
    $share{Search}->update($config{search}{threshold});
}

print $ml->div(
	       $ml->a("Close Window",{href=>"javascript:window.close()",class=>"button"}),
	       {align=>"center",style=>"padding:10px;"}
	       );

my @err = $db->fetch_diagnostics("error");
my @warn = $db->fetch_diagnostics("warn");
if (ref $obj) { 
    push @err, $obj->fetch_diagnostics("error") if ($obj->errorcheck("error"));
    push @warn, $obj->fetch_diagnostics("warn") if ($obj->errorcheck("warn"));
}
#print $db->show_diagnostics("info","html");

if (scalar @err + scalar @warn > 0) {
    print $ml->h3("Diagnostic Information:");
}
if (scalar @err > 0) {
    print $ml->p("Errors:"), $ml->ol(\@err);
}
if (scalar @warn > 0) {
    print $ml->p("Warnings:"), $ml->ol(\@warn);
}

&print_foot;
if ($query{silent}) {
    # restore normal I/O streams
    close(STDOUT);
    close(STDERR);
    open(STDOUT, ">&OLDOUT");
    open(STDERR, ">&OLDERR");
}
&exsite_close;


sub print_head {
    print "Content-type: text/html\n\n";

    # we print incomplete html and body tags because we want to ouput the page
    # in real time, and cannot close these tags until all output has bee flushed.
    print $config{doctype},"\n";
    print "<html>\n";
    print $ml->head(
		    $ml->title("$config{site}{name} Publisher").
		    $ml->link(undef,
			      {rel=>"stylesheet",
			       href=>"$config{server}{HTMLpath}/_ExSite/css/ExSite.css",
			       type=>"text/css"})
		    );
    print "<body>\n";
    print $ml->div(
		   $config{site}{logo}.
		   $ml->br.
		   $ml->h1("ExSite Publisher"),
		   {align=>"center"}
		   );
}

sub print_foot {
    print "</body>\n</html>\n";
}

